#Copyright 2003-2005 M. Wayne Davis
#All rights reserved

#Not for redistribution

#windows printing and metafile extensions
#Original Author: Michael I. Schwartz, mschwart@nyx.net

#set version number here
set version 1.08.1

set errorInfo ""
set errorCode ""
package require Tk

set time0 [clock clicks -milliseconds]
if {[winfo exists .]} {
  if {[tk windowingsystem] != "aqua"} {
    wm withdraw .
  } else {
    wm geometry . 1x1-2-2
  }
}

proc sputs {args} {
global tcl_platform
  if {($tcl_platform(platform) == "windows") || ([tk windowingsystem] == "aqua")} {
    puts $args
  }
}

if {[tk windowingsystem] == "aqua"} {
  proc ::tk::mac::OpenDocument {args} {
    foreach filename $args {
      open_file $filename
    }
  }
}

package require msgcat
##set ::msgcat::Loclist to get a list of loaded locales
set needs_translation [list]
proc ::msgcat::mcunknown {args} {
  if {[llength $args] > 2} {
    if {[string first % $args] == -1} {
      sputs "msgcat error (too many args): $args"
      set text [lrange $args 1 end]
    } else {
      set text [eval format [lrange $args 1 end]]
    }
  } else {
    set text [lindex $args 1]
  }
  set locale [lindex $args 0]
  global ::needs_translation
  if {[lsearch -exact $::needs_translation [lindex $args 1]] == -1} {
    lappend ::needs_translation [lindex $args 1]
  }
  return $text
}



############
## DNA edit window creator
############
proc create_window {{newdna ""} {filename ""} {filecomment ""} {circular linear} {methylated -1} {startindex 1} {title New_DNA} {tags ""} {genbank_header_list {}} {filetype "None"}} {
  global info
  global fea_info
  global winum
  global modifier
  global modstring
  global toolbar_images
  global empty_genbank_header

  incr winum
  set w .dna_window$winum

# variables refering to the DNA
  set info($w,startindex) $startindex
  if {$genbank_header_list != [list]} {
    set info($w,genbank_header_list) $genbank_header_list
  } else {
    set info($w,genbank_header_list) $empty_genbank_header
  }

  set info($w,circular) $circular
  if {$methylated > -1} {
    set info($w,Dam_Dcm_methylated) $methylated
  } else {
    set info($w,Dam_Dcm_methylated) $info(methylated_default)
  }
  set info($w,filename) $filename
  set info($w,filetype) $filetype
  set info($w,saved) 1



  if {[lsearch -exact $info(dnatitles_list) $title] > -1 } {
    set i 1
    while {[lsearch -exact $info(dnatitles_list) "$title\($i\)"] > -1 } {incr i}
    set title "$title\($i\)"
  }
set info($w,scanned) 0
set info($w,locked) [expr {[file isfile $info($w,filename)] && ![file writable $info($w,filename)]}]
toplevel $w 
wm title $w $title
set info($w,win_title) $title


lappend info(dnawindows_list) $w
lappend info(dnatitles_list) $title

wm protocol $w WM_DELETE_WINDOW "closewindow $w"

menu $w.menubar


#define file menu
menu  $w.menubar.filemenu 
$w.menubar add cascade -menu $w.menubar.filemenu -label [mc "File"] -underline 0
$w.menubar.filemenu add command -label [mc "New"] -accelerator "$modstring+N" -command "create_window" -underline 0
$w.menubar.filemenu add command -label [mc "Open..."] -accelerator "$modstring+O"  -command "open_file_dialog"
$w.menubar.filemenu add command -label [mc "Duplicate Selection"]  -accelerator "$modstring+D"  -command "duplicate_window $w selection" -underline 0
$w.menubar.filemenu add command -label [mc "Duplicate All"]  -accelerator "$modstring+Shift+D"  -command "duplicate_window $w all"
$w.menubar.filemenu add command -label [mc "Close"] -accelerator "$modstring+W" -command "closewindow $w"
$w.menubar.filemenu add separator
$w.menubar.filemenu add command -label [mc "Save"]  -accelerator "$modstring+S" -command "save_file $w Save"
$w.menubar.filemenu add command -label [mc "Save As..."] -command "save_file $w Save_as"
#exit is added below in the platform specific code

#define edit menu
menu  $w.menubar.edit 
$w.menubar add cascade -menu $w.menubar.edit -label [mc "Edit"]
$w.menubar.edit add command -label [mc "Can't Undo"] -state disabled -accelerator "$modstring+Z" -command "undo $w"
$w.menubar.edit add command -label [mc "Cut"]  -accelerator "$modstring+X" -command "clip_cut_copy $w cut"
$w.menubar.edit add command -label [mc "Copy"]  -accelerator "$modstring+C" -command "clip_cut_copy $w copy"
$w.menubar.edit add command -label [mc "Paste"]  -accelerator "$modstring+V" -command "clip_paste $w"
$w.menubar.edit add separator
$w.menubar.edit add command -label [mc "Cut Rev-Com"]  -accelerator "$modstring+Shift+X" -command "clip_cut_copy $w cut rev-com"
$w.menubar.edit add command -label [mc "Copy Rev-Com"]  -accelerator "$modstring+Shift+C" -command "clip_cut_copy $w copy rev-com"
$w.menubar.edit add command -label [mc "Paste Rev-Com"]  -accelerator "$modstring+Shift+V" -command "clip_paste $w rev-com"
$w.menubar.edit add separator
if {[tk windowingsystem] != "aqua"} {
  $w.menubar.edit add command -label [mc "Preferences..."]  -command "configure_preferences $w"
  $w.menubar.edit add separator
}

$w.menubar.edit add command -label [mc "Select All"]  -accelerator "$modstring+A" -command "select_all $w"
$w.menubar.edit add command -label [mc "Select From-To..."]  -accelerator "$modstring+Shift+A" -command "select_from_to $w"
$w.menubar.edit add command -label [mc "Jump To..."]  -accelerator "$modstring+J" -command "jump_to $w"
$w.menubar.edit add separator
$w.menubar.edit add command -label [mc "Find..."]  -accelerator "$modstring+F" -command "find_dialog $w"
$w.menubar.edit add command -label [mc "Find Again"]  -accelerator "$modstring+G" -command "if {\$info(find_find_again) != {}} {eval do_find $w \$info(find_find_again)} else {find_dialog $w}"
$w.menubar.edit add command -label [mc "Clear Find"]  -accelerator "" -command "$w.textarea tag delete foundf 1.0 end; $w.textarea tag delete foundr 1.0 end"
$w.menubar.edit add separator
$w.menubar.edit add command -label [mc "Convert to UPPERCASE"]  -accelerator "$modstring++" -command "change_case $w upper"
$w.menubar.edit add command -label [mc "Convert to lowercase"]  -accelerator "$modstring+-" -command "change_case $w lower"
$w.menubar.edit add command -label [mc "UPPER<->lower"] -accelerator "$modstring+=" -command "change_case $w exchange"
$w.menubar.edit add separator
$w.menubar.edit add command -label [mc "Reverse-Complement"]  -accelerator "$modstring+/" -command "rev_com_window $w"
$w.menubar.edit add separator
$w.menubar.edit add command -label [mc "Circular<->Linear"] -command "toggle_circ $w"
$w.menubar.edit add command -label [mc "Set Origin"]  -accelerator "$modstring+I" -command "set_origin $w"
$w.menubar.edit add command -label [mc "Linearize @ insert site"] -command "linearize $w"

#add enzymes menu
menu  $w.menubar.enzymes
$w.menubar add cascade -menu $w.menubar.enzymes -label [mc "Enzymes"]
$w.menubar.enzymes add command -label [mc "Enzyme Selector..."] -accelerator "$modstring+E" -command "enzyme_selection_dialog $w"
$w.menubar.enzymes add separator
if {[tk windowingsystem] != "aqua"} {
  $w.menubar.enzymes add checkbutton -label [mc "Selection Only"] -accelerator "$modstring+Shift-E" -onvalue "selection" -offvalue "all" -variable info(analyze_selection) -command "event generate . <<toggle_select>>"
} else {
  $w.menubar.enzymes add checkbutton -label [mc "Selection Only"]  -onvalue "selection" -offvalue "all" -variable info(analyze_selection) -command "event generate . <<toggle_select>>"
}
$w.menubar.enzymes add separator
$w.menubar.enzymes add command -label [mc "Graphic Map"] -accelerator "$modstring+M" -command "enz_graphic_map  $w \$info(enz_currently_selected) \$info(analyze_selection)"
$w.menubar.enzymes add command -label [mc "Graphic Map +U"] -accelerator "$modstring+Shift+M" -command "enz_graphic_map  $w \$info(enz_currently_selected) \$info(analyze_selection) {==1}"
$w.menubar.enzymes add separator
$w.menubar.enzymes add command -label [mc "Highlight"] -command "highlight_enzymes $w \$info(enz_currently_selected)";#uses selected
$w.menubar.enzymes add command -label [mc "Clear Highlighting"] -command "foreach enzyme \$enzymes {$w.textarea tag delete \$enzyme 1.0 end}"
$w.menubar.enzymes add separator
$w.menubar.enzymes add command -label [mc "Digestion"] -accelerator "$modstring+R" -command "enz_digest $w \$info(enz_currently_selected) \$info(analyze_selection)"
menu $w.menubar.enzymes.ladders
  foreach entry $info(ladderlist) {
    $w.menubar.enzymes.ladders add radiobutton -label $entry -value $entry -variable info(current_ladder)
  }
  $w.menubar.enzymes.ladders add separator
  $w.menubar.enzymes.ladders add command -label [mc "Edit..."] -accelerator "$modstring+Shift-R" -command "edit_ladders $w"
$w.menubar.enzymes add cascade -label [mc "Ladder"] -menu $w.menubar.enzymes.ladders
$w.menubar.enzymes add separator
$w.menubar.enzymes add command -label [mc "List Enzymes..."] -command "list_enzymes_dialog $w"
menu $w.menubar.enzymes.lists
  $w.menubar.enzymes.lists add command -label [mc "Site Usage"] -command "findenzymes $w; list_enzymes $w \$enzymes 1 \"Site Usage\" \$info(analyze_selection)"
  $w.menubar.enzymes.lists add command -label [mc "Unique Sites"] -command "findenzymes $w; list_enzymes $w \[enz_site_usage_filter $w \$enzymes \"==1\" \$info(analyze_selection)\] 0 \"Unique Sites\" \$info(analyze_selection)"
  $w.menubar.enzymes.lists add command -label [mc "Absent Sites"] -command "findenzymes $w; list_enzymes $w \[enz_site_usage_filter $w \$enzymes \"==0\" \$info(analyze_selection)\] 0 \"Absent Sites\" \$info(analyze_selection)"
$w.menubar.enzymes add cascade -label [mc "Quick Lists"] -menu $w.menubar.enzymes.lists
$w.menubar.enzymes add separator
$w.menubar.enzymes add command -label [mc "Silent Sites"]  -command "silent_sites $w \$info(enz_currently_selected)"
$w.menubar.enzymes add separator
$w.menubar.enzymes add command -label [mc "Text Map..."] -accelerator "$modstring+Shift+T" -command "enz_text_map $w "

#add ORFs menu
menu  $w.menubar.orf
$w.menubar add cascade -menu $w.menubar.orf -label [mc "ORFs"]
  $w.menubar.orf add command -label [mc "Find Next"] -accelerator "$modstring+>" -command "next_ORF $w forward "
  $w.menubar.orf add command -label [mc "Find Previous"] -accelerator "$modstring+<" -command "next_ORF $w reverse "

  $w.menubar.orf add separator
  $w.menubar.orf add command -label [mc "Translate..."] -accelerator "$modstring+T" -command "translation_window $w"
  $w.menubar.orf add separator
  $w.menubar.orf add command -label [mc "ORF map"] -command "orf_map $w \$info(analyze_selection)"
  $w.menubar.orf add separator
  #$w.menubar.orf add checkbutton -label "Selection Translate" -variable info(selection_shows_Trans) -onvalue 1 -offvalue 0 -command "selection_manager $w"
  $w.menubar.orf add checkbutton -label [mc "Selection Translate"] -variable info($w,selection_shows_Trans) -onvalue 1 -offvalue 0 -command "if {\$info($w,selection_shows_Trans)} {grid configure $w.infoframe.trans; selection_manager $w} else {grid remove $w.infoframe.trans}"
  set info($w,selection_shows_Trans) $info(selection_shows_Trans)
  #$w.menubar.orf add command -label "Codon Usage" -state disabled
  #$w.menubar.orf add command -label "Compare Codon Bias" -state disabled
  #$w.menubar.orf add command -label "Open Codon Bias Table" -state disabled
  #$w.menubar.orf add separator
  #$w.menubar.orf add radiobutton -label "Selection Translate" -variable info(selection_shows) -value "Translate" -command "selection_manager $w"
  #$w.menubar.orf add radiobutton -label "Selection None" -variable info(selection_shows) -value "None" -command "selection_manager $w"

#add oligos menu
#menu  $w.menubar.oligos
#  $w.menubar add cascade -menu $w.menubar.oligos -label "Oligos"
  #$w.menubar.oligos add command -label "Find Primers..." -command "primer_window $w";
  #$w.menubar.oligos add separator
  #$w.menubar.oligos add checkbutton -label "Selection Tm, %GC" -variable info(selection_shows_Tm) -onvalue 1 -offvalue 0 -command "selection_manager $w"
  #$w.menubar.oligos add radiobutton -label "Selection Tm, %GC" -variable info(selection_shows) -value "Tm" -command "selection_manager $w"
  #$w.menubar.oligos add radiobutton -label "Selection None" -variable info(selection_shows) -value "None" -command "selection_manager $w"

#add features menu
menu  $w.menubar.features
$w.menubar add cascade -menu $w.menubar.features -label [mc "Features"]
  $w.menubar.features add command -label [mc "New Feature..."] -state disabled
  $w.menubar.features add command -label [mc "List Features"] -command "list_features $w"
  $w.menubar.features add command -label [mc "Edit Features..."] -command "configure_features $w"
  $w.menubar.features add command -label [mc "Clear Features"] -accelerator "$modstring+Shift+H" -command "feature_clear $w"
  $w.menubar.features add separator
  $w.menubar.features add command -label [mc "Edit Feature Library..."] -command "edit_feature_library $w"
  $w.menubar.features add command -label [mc "New Feature in Library..."] -command "add_sel_feature_library $w"
  $w.menubar.features add command -label [mc "Open Feature Library..."] -command "read_featurelib"
  $w.menubar.features add command -label [mc "Annotate Features using Library"] -accelerator "$modstring+H" -command "feature_scan $w"

#add Tools menu
menu  $w.menubar.tools
$w.menubar add cascade -menu $w.menubar.tools -label [mc "Tools"]
  $w.menubar.tools add command -label [mc "Find Primers..."] -command "primer_window $w"
  $w.menubar.tools add separator
  $w.menubar.tools add command -label [mc "Align Sequences..."] -accelerator "$modstring+L" -command "align_dialog $w"
  $w.menubar.tools add separator
  $w.menubar.tools add command -label [mc "BLAST Sequences at NCBI..."] -accelerator "$modstring+B" -command "NCBI_Blast $w"
  $w.menubar.tools add command -label [mc "BLAST Sequences at Wormbase..."] -command "Wormbase_Blast $w"

#add Windows menu stub (menu items are maintained by update_windows_menu procedure)
menu $w.menubar.windows 
  $w.menubar add cascade -menu $w.menubar.windows -label [mc "Windows"]
update_windows_menu

#add Help menu
menu  $w.menubar.help 
$w.menubar add cascade -menu $w.menubar.help -label [mc "Help"]
  $w.menubar.help add command -label [mc "Standard Genetic Code"] -command "genetic_code_dialog $w"
  $w.menubar.help add command -label [mc "AA Info"] -command "aa_info_dialog $w"

  #add windows/Aqua specific items
  if {[tk windowingsystem] != "aqua"} {
    #Exit is automatically on the apple menu in Aqua
    $w.menubar.filemenu add separator
    $w.menubar.filemenu add command -label [mc "Exit"] -accelerator "$modstring+Q" -command close_all_windows

    #About... goes on the apple menu in aqua
    $w.menubar.help add separator
    $w.menubar.help add command -label [mc "About ApE"] -command "about_dialog $w"

    #add menu icons
    set iconlist [list $w.menubar.filemenu [mc New] new $w.menubar.filemenu [mc Open...] open $w.menubar.filemenu [mc Save] save $w.menubar.edit [mc "Cut"] cut $w.menubar.edit [mc "Copy"] copy $w.menubar.edit [mc "Paste"] paste $w.menubar.edit [mc "Cut Rev-Com"] cut $w.menubar.edit [mc "Copy Rev-Com"] copy $w.menubar.edit [mc "Paste Rev-Com"] paste $w.menubar.edit [mc "Select From-To..."] select $w.menubar.edit [mc "Jump To..."] jump $w.menubar.edit [mc "Find..."] find $w.menubar.edit [mc "Convert to UPPERCASE"] uppercase $w.menubar.edit [mc "Convert to lowercase"] lowercase $w.menubar.edit [mc "UPPER<->lower"] swap_case $w.menubar.edit [mc "Reverse-Complement"] reverse_com $w.menubar.edit [mc "Set Origin"] origin $w.menubar.enzymes [mc "Enzyme Selector..."] enzymes $w.menubar.enzymes [mc "Graphic Map"] graphic_map $w.menubar.enzymes [mc "Graphic Map +U"] graphic_map_u $w.menubar.enzymes [mc "Highlight"] highlight_enzymes $w.menubar.enzymes [mc "Clear Highlighting"] clear_highlighting $w.menubar.enzymes [mc "Digestion"] digest $w.menubar.enzymes [mc "Text Map..."] text_map $w.menubar.tools [mc "Find Primers..."] find_primers $w.menubar.tools [mc "Align Sequences..."] align]
    foreach {menu item icon} $iconlist {
      $menu entryconfigure $item  -compound right -image $toolbar_images($icon)
    }
  } else {
    #define the apple menu
    menu $w.menubar.apple 
    $w.menubar add cascade -menu $w.menubar.apple
    $w.menubar.apple add command -label [mc "About ApE..."] -command "about_dialog $w"
    $w.menubar.apple add separator
    $w.menubar.apple add command -label [mc "Preferences..."]  -command "configure_preferences $w"
    .menubar.filemenu entryconfigure New -state disabled
    .menubar.filemenu entryconfigure "Open..." -state disabled
    #bind . <Command-KeyPress-o> ""
    #bind . <Command-KeyPress-n> ""
    #bind . <Command-KeyPress-q> ""

    ##set the new window coordinates
    wm geometry $w "+$info(newaquax)+$info(newaquay)"
    incr info(newaquax) 20
    incr info(newaquay) 20
    if {$info(newaquax) > 100} {set info(newaquax) 20}
    if {$info(newaquay) > 100} {set info(newaquay) 20}
  }

#add menubar to the window
$w configure -menu $w.menubar

frame $w.toolbar -borderwidth 0 -height 1 -width 200 -relief flat

fill_toolbar $w

frame $w.infoframe -borderwidth 2 -relief ridge

  #checkbutton $w.infoframe.scanned -text "Scan" -variable info($w,scanned) -onvalue 1 -offvalue 0 -selectcolor white -activebackground gray75 -state disabled -disabledforeground black
  button $w.infoframe.locked -image $toolbar_images([expr {$info($w,locked)?"locked":"unlocked"}]) -borderwidth 0 -relief flat -command "toggle_locked $w"
  bind "$w.infoframe.locked" <Any-Enter> "tooltip_enter $w.infoframe.locked {[mc Lock/Unlock]\n[mc {File}]}"
  bind "$w.infoframe.locked" <Any-Leave> "tooltip_leave $w.infoframe.locked"
  checkbutton $w.infoframe.saved -text [mc "Saved"] -width 5 -variable info($w,saved) -onvalue 1 -offvalue 0 -selectcolor white -activebackground $info(bg_color) -disabledforeground $info(label_fg_color) -state disabled 
  bind "$w.infoframe.saved" <Any-Enter> "tooltip_enter $w.infoframe.saved {[mc {Click to Save File}]}"
  bind "$w.infoframe.saved" <Any-Leave> "tooltip_leave $w.infoframe.saved"
  bind $w.infoframe.saved <Button-1> "save_file $w Save"
  label $w.infoframe.totallenlabel -text [mc "Sequence"] -width 8 -height 1 
  label $w.infoframe.totallen -width 8 -height 1 
  label $w.infoframe.startlabel -text [mc "Start"] -width 6 -height 1 
  label $w.infoframe.start -width 8 -height 1  
  label $w.infoframe.sellenlabel -text [mc "Length"] -width 7 -height 1 
  label $w.infoframe.sellen -width 8 -height 1 
  label $w.infoframe.endlabel -text [mc "End"] -width 4 -height 1 
  label $w.infoframe.end -height 1 -width 8  
  label $w.infoframe.orflabel -text [mc "ORF"] -width 4 -height 1 
  label $w.infoframe.orf -height 1 -width 4 
  label $w.infoframe.tmlabel -text "Tm" -width 3 -height 1 
  label $w.infoframe.tm -height 1 -width 4 
  label $w.infoframe.gclabel -text "%GC" -width 4 -height 1 
  label $w.infoframe.gc -width 4 -height 1 
  label $w.infoframe.trans -height 1 -width 50  -justify center
  button $w.infoframe.circular -command "toggle_circ $w" -width 10 -activebackground $info(bg_color) -textvariable info($w,circular)
  $w.infoframe.circular configure -height 1  -pady 0 -border 2
  checkbutton $w.infoframe.methylation -text "Dam/Dcm" -onvalue 1 -offvalue 0  -width 10 -variable info($w,Dam_Dcm_methylated) -command "toggle_methylation $w" -selectcolor white -activebackground $info(bg_color)


  grid $w.infoframe.locked -row 0 -column 1
  grid $w.infoframe.saved -row 0 -column 2
  grid [label $w.infoframe.mouseindex -text {} -justify left] -row 1 -column 2  -sticky nwe
  grid $w.infoframe.totallenlabel -row 0 -column 4
  grid $w.infoframe.totallen  -row 1 -column 4
  grid $w.infoframe.startlabel -row 0 -column 5
  grid $w.infoframe.start -row 1  -column 5
  grid $w.infoframe.sellenlabel -row 0 -column 6
  grid $w.infoframe.sellen -row 1  -column 6
  grid $w.infoframe.endlabel  -row 0 -column 7
  grid $w.infoframe.end  -row 1 -column 7
  grid $w.infoframe.orflabel  -row 0 -column 8
  grid $w.infoframe.orf  -row 1 -column 8
  grid $w.infoframe.tmlabel  -row 0 -column 9
  grid $w.infoframe.tm  -row 1 -column 9
  grid $w.infoframe.gclabel  -row 0 -column 10
  grid $w.infoframe.gc  -row 1 -column 10
  grid $w.infoframe.circular -row 0 -column 11 -ipady 2
  grid $w.infoframe.methylation -row 1 -column 11
  grid $w.infoframe.trans -row 2 -column 4 -columnspan 6
  if {!$info($w,selection_shows_Trans)} {grid remove $w.infoframe.trans}

  label $w.infobox  -width 50

#define the index windows above, below and left of the sequence
text $w.horzindex -font dnafont -width 50 -height 1 -background $info(bg_color) -fg $info(label_fg_color) -wrap none -relief flat -cursor arrow
$w.horzindex insert 1.0 {1   *   10    *   20    *   30    *   40    *   50 }
bindtags $w.horzindex .

text $w.horzindex2 -font dnafont -width 50 -height 1 -background $info(bg_color) -fg $info(label_fg_color) -wrap none -relief flat -cursor arrow
$w.horzindex2 insert 1.0 {1   *   10    *   20    *   30    *   40    *   50 }
bindtags $w.horzindex2 .

text $w.vertindex -font dnafont -width 6 -background $info(bg_color) -fg $info(label_fg_color) -relief flat -cursor arrow
$w.vertindex insert 1.0 "1" allvertindex
$w.vertindex tag configure allvertindex -justify right
bindtags $w.vertindex .

#define textarea (sequence area) and bindings
text $w.textarea -font dnafont -width 60 -wrap char -borderwidth 1 -padx 0 -setgrid true 
bind $w.textarea <Any-Motion> "mouseover_info $w \[$w.textarea index @%x,%y\]"
bind $w.textarea <Any-Leave> "mouseover_info $w -1"
bind $w.textarea <KeyPress> "keyevent_manager $w None %K ; break"
bind $w.textarea <Alt-KeyPress> #
bind $w <$modifier-KeyPress> "keyevent_manager $w Control %K ; break"
bind $w.textarea <$modifier-KeyPress> "keyevent_manager $w Control %K ; break"
bind $w.textarea <Shift-$modifier-KeyPress> "keyevent_manager $w ShiftControl %K ; break"
bind $w.textarea <Key-Left> "textsetcursor $w insert-1c; break"
bind $w.textarea <Key-Right> "textsetcursor $w insert+1c; break"
bind $w.textarea <Key-Down> "textsetcursor $w insert+\$info($w,textwidth)c; break"
bind $w.textarea <Key-Up> "textsetcursor $w insert-\$info($w,textwidth)c; break"
bind $w.textarea <MouseWheel> "vertscrollmanager $w scroll \[expr {%D/-$info(mousewheelunits)}\] units; mouseover_info $w \[$w.textarea index @%x,%y\]; break"
if {[tk windowingsystem] != "aqua"} {
  bind $w.textarea <B1-MouseWheel> "vertscrollmanager $w scroll \[expr {%D/-$info(mousewheelunits)}\] units; text_select_xy $w %x %y"
  #bind $w.textarea <Button-2> "clip_paste $w; break"
} else {
  bind $w.textarea <B1-MouseWheel> "vertscrollmanager $w scroll \[expr {%D/-$info(mousewheelunits)}\] units; mouseover_info $w \[$w.textarea index @%x,%y\]; break"
  #bind $w.textarea <Button-3> "clip_paste $w; break"
}
bind $w.textarea <Shift-Key-Left> "textkeyselect $w \[$w.textarea index {insert - 1c}\]; break"
bind $w.textarea <Shift-Key-Right> "textkeyselect $w \[$w.textarea index {insert + 1c}\]; break"
bind $w.textarea <Shift-Key-Down> "textkeyselect $w \[$w.textarea index \"insert + \$info($w,textwidth)c\"\]; break"
bind $w.textarea <Shift-Key-Up> "textkeyselect $w \[$w.textarea index \"insert - \$info($w,textwidth)c\"\]; break"
bind $w.textarea <Button-1> "focus $w.textarea; textsetcursor $w \[text_closest_gap $w.textarea %x %y\]"
bind $w.textarea <Double-Button-1> "select_region $w 1.0 1.end"
bind $w.textarea <Shift-Button-1> "[bind Text <Shift-Button-1>]; selection_manager $w"
bind $w.textarea <Button1-Leave> "mouse_left_DNA $w %x %y"
bind $w.textarea <Button1-Enter> "cancel_scrolling_DNA"
bind $w.textarea <ButtonRelease-1> "cancel_scrolling_DNA"
bind $w.textarea <Button1-Motion> "text_select_xy $w %x %y"


bindtags $w.textarea [list $w.textarea . all]

scrollbar $w.vertscroll -command "vertscrollmanager $w" -relief flat -width 16

####bug in Aqua mouse scrolling
if {[tk windowingsystem] == "aqua"} {
  bind $w.vertscroll <Button-1> {set info(mousedown) %Y; set info(scrollmdown) [lindex [%W get] 0]}
}

grid columnconfigure $w 1 -weight 1
grid columnconfigure $w 2 -weight 0 -minsize 17

grid rowconfigure $w 0  -weight 0
grid rowconfigure $w 1  -weight 0
grid rowconfigure $w 2  -weight 0
grid rowconfigure $w 3  -weight 0
grid rowconfigure $w 4 -minsize 2c -weight 1
grid rowconfigure $w 5  -weight 0
grid rowconfigure $w 6  -weight 0


grid $w.toolbar -row 0 -column 0 -columnspan 3 -sticky new
grid $w.infoframe -row 1  -column 0 -columnspan 3 -sticky new
grid $w.infobox -row 2  -column 0  -columnspan 3  -sticky new
grid $w.horzindex -row 3 -column 1  -sticky nwe
grid  $w.vertindex -row 4 -column 0 -sticky nsw
grid $w.textarea -row 4 -column 1 -sticky nswe
grid $w.vertscroll -row 4 -column 2 -sticky nse
grid $w.horzindex2 -row 5 -column 1  -sticky swe

#grid [button $w.show_header_info -height 15 -width 15 -borderwidth 0 -image $toolbar_images(info_button) -command "toggle_comment $w"] -row 3 -column 0

grid [button $w.showcomment -height 15 -width 15 -borderwidth 0 -image $toolbar_images(comment_button) -command "toggle_comment $w"] -row 5 -column 2
grid [text $w.comment -height $info(comment_height) -width 1 -wrap char -yscrollcommand "optionscrollbar $w.cscroll" -font labelfont] -row 6 -column 0 -columnspan 2 -sticky nwe
bind "$w.showcomment" <Any-Enter> "tooltip_enter $w.showcomment {[mc Show/Hide]\n[mc {File Comment}]}"
bind "$w.showcomment" <Any-Leave> "tooltip_leave $w.showcomment"
$w.comment insert 1.0 $filecomment

bind $w.comment <$modifier-KeyPress> "comment_keyevent_manager $w Control %K ; break"
bind $w.comment <Shift-$modifier-KeyPress> "comment_keyevent_manager $w ShiftControl %K ; break"
bind $w.comment <Key> "if {\[string equal %A \"\"\] || \[file_lock_check $w\]} {sputs %A;break} else {set info($w,saved) 0}" 
bind $w.comment <Delete> "if {\[file_lock_check $w\]} {break} else {set info($w,saved) 0}"
bind $w.comment <Key-BackSpace> "if {\[file_lock_check $w\]} {break} else {set info($w,saved) 0}"
bind $w.comment <Key-Return> "if {\[file_lock_check $w\]} {break} else {set info($w,saved) 0}"
bind $w.comment <<Cut>> "if {\[file_lock_check $w\]} {break} else {set info($w,saved) 0}"
bind $w.comment <<Paste>> "if {\[file_lock_check $w\]} {break} else {set info($w,saved) 0}"
foreach event [list <Shift-Key-Down> <Shift-Key-Up> <Shift-Key-Right> <Shift-Key-Left> <Key-Down> <Key-Up> <Key-Right> <Key-Left>] {
  bind $w.comment $event #
}
bindtags $w.comment [list $w.comment Text all]
grid [scrollbar $w.cscroll -command "$w.comment yview"] -row 6 -column 2  -sticky nswe
if {!$info(show_comment)} {
  grid remove $w.comment
  grid remove $w.cscroll 
}

grid [label $w.filename -textvariable info($w,filename)  -justify right -width 60] -row 7 -column 1 -columnspan 1 -sticky nwe
grid [label $w.mouseindex -text {} -justify left] -row 7 -column 0  -sticky nwe


bind $w <FocusIn> "event generate . <<RaiseDialogs>>"
bind $w <Map> "event generate . <<RaiseDialogs>>"
bind $w <Activate> "event generate . <<RaiseDialogs>>"

wm deiconify $w
focus $w.textarea



$w.textarea insert 1.0 $newdna

set fea_info($w,count) 0
if {$tags != {}} {
  put_tags $w.textarea {1.0 1.end} $tags forward 1
}

$w.textarea mark set insert 1.0
set info($w,textwidth) 1
textarea_width_changed $w
if {![winfo exists $w]} {return}

set info($w,scanned) 0
set info($w,linked_events) [list]
$w.textarea tag configure sel -background $info(text_select_bg_color) -foreground $info(text_select_fg_color) -relief flat
$w.textarea tag configure tempsel -background $info(text_select2_bg_color) -foreground $info(text_select2_fg_color)
$w.textarea tag lower tempsel

global tcl_platform
if {($tcl_platform(platform) == "unix") && ([tk windowingsystem] != "aqua")} {
    selection handle -type UTF8_STRING -selection CUT_BUFFER0 $w unix_cubuffer_handler
    selection handle -type STRING -selection CUT_BUFFER0 $w unix_cubuffer_handler
    selection handle -type UTF8_STRING -selection PRIMARY $w unix_cubuffer_handler
    selection handle -type STRING -selection PRIMARY $w unix_cubuffer_handler
} elseif {$tcl_platform(platform) == "windows"} {
    wm deiconify $w
    if {([lsearch -glob [info loaded] "*libtkdnd.dll*"] > -1)} {
      dnd bindtarget $w Files <Drop> {HandleFileDrop  %D %T}
    }
}
if {($tcl_platform(platform) != "unix") || ([tk windowingsystem] == "aqua")} {
    bind $w <Shift-Button-3> {console show}
    bind $w.textarea <Shift-Button-3> {
      console show
      set w [lindex [winfo children .] 0]
    }
}
bind $w.textarea <Configure> "textarea_width_changed $w"

$w.textarea mark set insert 1.0
$w.textarea mark set anchor 1.0

update

window_on_screen $w

vertindexbalancer $w
selection_manager $w
focus -force $w.textarea
event generate $w.textarea <Button-1> -x 1 -y 1
event generate $w.textarea <ButtonRelease-1> -x 1 -y 1
return "ok"
}

##############
## read the toolbar button images
##############
proc read_toolbar_images {} {
  global toolbar_images
  #global argv0
  global info

  image create photo blank -height 16 -width 16
  blank blank

  set toolbar_images(separator) [image create photo -height 16 -width 3]
  $toolbar_images(separator) put {{gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30} {gray90 gray90 gray30}}
  set image_dir [file join $info(Accdir) "Icons and images"]
  if {![file isdirectory $image_dir]} {
    warning_dialog "Please put the toolbar button images in: $image_dir"
    set error 1
  } else {
    set error 0
  }

  foreach button [list new open duplicate save cut copy paste select jump find find_again clear_find uppercase lowercase swap_case reverse_com  origin enzymes enzymes_sonly graphic_map graphic_map_u highlight_enzymes clear_highlighting digest text_map find_primers align dna_win abi_win ana_win comment_button info_button error info question warning monkey_icon locked unlocked] {
    if {([file isfile "$image_dir/$button.gif"] == 0) || ([catch {set toolbar_images($button) [image create photo -file "$image_dir/$button.gif"]}] != 0)} {
      set toolbar_images($button) [image create photo -height 16 -width 16]
      $toolbar_images($button) copy blank
    } 
  }
  set toolbar_images(blank) [image create photo -height 16 -width 16]
  $toolbar_images(blank) copy blank
  image delete blank

  set toolbar_images(enzymes_all) [image create photo -height 16 -width 16]
  $toolbar_images(enzymes_all) copy $toolbar_images(enzymes)
##change the toolbar image when select only is toggled"
  bind . <<toggle_select>> {
    if {$info(analyze_selection)== "selection"} {
      $toolbar_images(enzymes) blank
      $toolbar_images(enzymes) copy $toolbar_images(enzymes_sonly)
    } else {
      $toolbar_images(enzymes) blank;
      $toolbar_images(enzymes) copy $toolbar_images(enzymes_all)
    }
  }
}

##############
## add buttons to the toolbar
##############
## toolbar list is a list of lists, each button element is a list of image name, tooltip, command
proc fill_toolbar w {
  global tcl_platform
  global toolbarlist
  global toolbar_images
  global info

  if {$toolbarlist == [list]} {
    $w.toolbar configure -height 1
    return
  }

  $w.toolbar configure -height 16
  set i 0
  foreach button $toolbarlist {
    set command2 ""
    switch $button {
      separator {set button_type separator}
      new {set button_type button; set tooltip "[mc {New File}] \n[mc Shift:][mc {Duplicate Selection}] "; set command "create_window"; set command2 "duplicate_window $w selection"}
      open {set button_type button; set tooltip [mc "Open File"]; set command "open_file_dialog"}
      duplicate {set button_type button; set tooltip "[mc "Duplicate Selection"]\n[mc "Duplicate All"]"; set command "duplicate_window $w selection"; set command2 "duplicate_window $w all"}
      save {set button_type button; set tooltip "[mc {Save File}] \n[mc Shift:][mc {Save As}] "; set command "save_file $w Save"; set command2 "save_file $w Save_as"}
      cut {set button_type button; set tooltip "[mc {Cut to Clipboard}] \n[mc Shift:][mc {Cut Rev-Com}]"; set command "clip_cut_copy $w cut"; set command2 "clip_cut_copy $w cut rev-com"}
      copy {set button_type button; set tooltip "[mc {Copy to Clipboard}] \n[mc Shift:][mc {Copy Rev-Com}]"; set command "clip_cut_copy $w copy"; set command2 "clip_cut_copy $w copy rev-com"}
      paste {set button_type button; set tooltip "[mc {Paste from Clipboard}] \n[mc Shift:][mc {Paste Rev-Com}]"; set command "clip_paste $w"; set command2 "clip_paste $w rev-com"}
      cut_rc {set button_type button; set tooltip [mc "Cut(RC) to Clipboard"]; set command "clip_cut_copy $w cut rev-com"}
      copy_rc {set button_type button; set tooltip [mc "Copy(RC) to Clipboard"]; set command "clip_cut_copy $w copy rev-com"}
      paste_rc {set button_type button; set tooltip [mc "Paste(RC) from Clipboard"]; set command "clip_paste $w rev-com"}
      select {set button_type button; set tooltip "[mc {Select From-To}] \n[mc Shift:][mc {Jump to}]"; set command "select_from_to $w"; set command2 "jump_to $w"}
      jump {set button_type button; set tooltip [mc "Jump To"]; set command "jump_to $w"}
      find {set button_type button; set tooltip "[mc {Find...}] \n[mc Shift:][mc {Clear Find}]"; set command "find_dialog $w"; set command2 "$w.textarea tag delete foundf 1.0 end; $w.textarea tag delete foundr 1.0 end"}
      find_again {set button_type button; set tooltip [mc "Find Again"]; set command "if {\$info(find_find_again) != {}} {eval do_find $w \$info(find_find_again)} else {find_dialog $w}"}
      clear_find {set button_type button; set tooltip [mc "Remove Find Highlighting"]; set command "$w.textarea tag delete foundf 1.0 end; $w.textarea tag delete foundr 1.0 end"}
      uppercase {set button_type button; set tooltip "[mc {Change to Uppercase}] \n[mc Shift:][mc {Change to Lowercase}]"; set command "change_case $w upper"; set command2 "change_case $w lower"}
      lowercase {set button_type button; set tooltip [mc "Change to Lowercase"]; set command "change_case $w lower"}
      swap_case {set button_type button; set tooltip [mc "UPPER<->lower Exchange"]; set command "change_case $w exchange"}
      reverse_com {set button_type button; set tooltip [mc "Reverse Complement"]; set command "rev_com_window $w"}
      origin {set button_type button; set tooltip [mc "Set Origin"]; set command "set_origin $w"}
      enzymes {set button_type button; set tooltip "[mc {Enzyme Selection Dialog}]\n[mc Shift:][mc {Toggle Selected Only}]"; set command "enzyme_selection_dialog $w"; set command2 "if {\$info(analyze_selection)== \"all\"} {set info(analyze_selection) selection} else {set info(analyze_selection) all}; event generate . <<toggle_select>>"}
      graphic_map {set button_type button; set tooltip "[mc {Graphic Map}] \n[mc Shift:][mc {+Unique}]"; set command "enz_graphic_map $w \$info(enz_currently_selected) \$info(analyze_selection)"; set command2 "enz_graphic_map  $w \$info(enz_currently_selected) \$info(analyze_selection) {==1}"}
      graphic_map_u {set button_type button; set tooltip [mc "Graphic Map +Unique Enzymes"]; set command "enz_graphic_map  $w \$info(enz_currently_selected) \$info(analyze_selection) {==1}"}
      highlight_enzymes {set button_type button; set tooltip "[mc {Highlight Selected Enzymes}] \n[mc Shift:][mc {Clear Highlight}]"; set command "highlight_enzymes $w \$info(enz_currently_selected)"; set command2 "foreach enzyme \$enzymes {$w.textarea tag delete \$enzyme 1.0 end}"}
      clear_highlighting {set button_type button; set tooltip [mc "Clear Enzyme Highlighting"]; set command "foreach enzyme \$enzymes {$w.textarea tag delete \$enzyme 1.0 end}"}
      digest {set button_type button; set tooltip [mc "Digestion Window"]; set command "enz_digest $w \$info(enz_currently_selected) \$info(analyze_selection)"}
      text_map {set button_type button; set tooltip [mc "Text Map"]; set command "enz_text_map $w"}
      find_primers {set button_type button; set tooltip [mc "Find Primers"]; set command "primer_window $w"}
      align {set button_type button; set tooltip [mc "Align Sequences"]; set command "align_dialog $w"}
      temp {set tooltip ""; set command ""}
      default {set button_type button; set tootip "Unknown button type $button"; set command ""}
    }
    switch $button_type {
      button {
        pack [button "$w.toolbar.$button" -image $toolbar_images($button) -relief flat -highlightthickness 0 -width 16 -height 16 -command $command] -side left -padx 0
        if {1 || ([tk windowingsystem] != "aqua")} {
          bind "$w.toolbar.$button" <Any-Enter> "$w.toolbar.$button configure -relief raised; tooltip_enter $w.toolbar.$button \{$tooltip\}"
          bind "$w.toolbar.$button" <Any-Leave> "$w.toolbar.$button configure -relief flat; tooltip_leave $w.toolbar.$button"
        }
        bind "$w.toolbar.$button" <Button-1> "$w.toolbar.$button configure -relief sunken"
        bind "$w.toolbar.$button" <ButtonRelease-1> "if \{\[$w.toolbar.$button cget -relief\] == \"sunken\"\} \{$w.toolbar.$button configure -relief raised; $command\}"
        if {$command2 != ""} {
          bind "$w.toolbar.$button" <Shift-ButtonRelease-1> "if \{\[$w.toolbar.$button cget -relief\] == \"sunken\"\} \{$command2\}; $w.toolbar.$button configure -relief raised"
          if {[tk windowingsystem] != "aqua"} {
            set secbut 3
          } else {
            set secbut 2
          }
          bind "$w.toolbar.$button" <Button-$secbut> "$w.toolbar.$button configure -relief sunken"
          bind "$w.toolbar.$button" <ButtonRelease-$secbut> "if \{\[$w.toolbar.$button cget -relief\] == \"sunken\"\} \{$w.toolbar.$button configure -relief raised; $command2\}"
        }

        bindtags "$w.toolbar.$button" "$w.toolbar.$button"
      }
      separator {
        if {[tk windowingsystem] == "aqua"} {
          pack [frame "$w.toolbar.separator$i" -width 5 -height 16 -relief flat] -side left -padx 0
          incr i
        } else {
          pack [button "$w.toolbar.separator$i" -image $toolbar_images(separator) -relief flat -border 0 -width 3 -height 16] -side left -padx 0
          bindtags "$w.toolbar.separator$i" "$w.toolbar.separator$i"
          incr i
        }
      }
    }
  }
}

##############
## procedure called to display a tooltip
##############
proc tooltip_show {window tip} {
  global tooltip_after
  global tcl_platform

  #if {([tk windowingsystem] == "aqua")} {
  #  return
  #}
  set tooltip_after showing
  if {[winfo exists .tooltip]} {destroy .tooltip}
  toplevel .tooltip -borderwidth 1 -background black

  if {([tk windowingsystem] == "aqua")} {
    tk::unsupported::MacWindowStyle style .tooltip help {noActivates noUpdates}
  } else {
    wm overrideredirect .tooltip 1
  }
  wm geometry .tooltip "+[expr {[winfo rootx $window]+[winfo height $window]/2}]+[expr {[winfo rooty $window]+[winfo height $window]}]"
  pack [label .tooltip.label1 -text $tip -background #ffffc0 -foreground black -border 0] -padx 0 -pady 0 -ipadx 0 -ipady 0
}

##############
## procedure called on leaving object with a tooltip
##############
proc tooltip_leave {window} {
  global tooltip_after


  if {[winfo exists .tooltip]} {destroy .tooltip}
  if {[string equal -length 5 $tooltip_after "after"] == 1} {
    after cancel $tooltip_after
  } else {
    set tooltip_after "showing_[after 300 {set tooltip_after "not_showing"}]"
  }  
}

##############
## procedure called on entering object with a tooltip
##############
proc tooltip_enter {window tip} {
  global tooltip_after

  if {[winfo exists .tooltip]} {destroy .tooltip}

  if {[string equal -length 7 $tooltip_after "showing"] == 1} {
    tooltip_show $window $tip
    after cancel {set tooltip_after "not_showing"}
  } else {
    set tooltip_after [after 1200 [list tooltip_show $window $tip]]
  }
}

##############
## open a file
##############
#recognizes fasta, plain and strider files

proc open_file_dialog {{filename ""}} {
  global info dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1
  while {($info(default_dnadir) != "") && (![file isdirectory $info(default_dnadir)])} {
    set info(default_dnadir) [file dirname $info(default_dnadir)]
    if {[file dirname $info(default_dnadir)] == $info(default_dnadir)} {
      set info(default_dnadir) ""
    }
  }
  if {$filename == ""} {
    set focus_win [focus]
    if {$focus_win == ""} {set focus_win .}
    set filename [tk_getOpenFile -initialdir $info(default_dnadir) -title "open DNA file" -parent $focus_win -multiple 1]
    if {([winfo exists $focus_win]) && ($focus_win != ".")} {
      wm deiconify [winfo toplevel $focus_win]
      focus $focus_win
    }
    update
  } else {
    regsub -all {\\} $filename {/} filename
  }

  foreach file $filename {
    open_file $file
  }
  set dialogblock 0
}

##############
## open a file
##############
#recognizes fasta, plain and strider files
proc open_file {filename} {
  global info empty_genbank_header

  if {$filename != ""} {
    regsub -all {\\} $filename {/} filename
    foreach win $info(dnawindows_list) {
      if {$info($win,filename) == $filename} {
        wm deiconify $win
        raise $win
        focus $win
        update
        return
      }
    }
    if {([file exists $filename]) && ([file readable $filename])} {
      set fileid [open $filename r]
      #read in the 1st header line and determine filetype (storing info from the line)
      fconfigure $fileid -translation binary
      set striderheader [read $fileid 112]
      fconfigure $fileid -translation auto
      seek $fileid 0
      set a [gets $fileid]
      if {[eof $fileid] != 1} {
        set b [gets $fileid]
      } else {
        set b ""
      }
      seek $fileid 0
      set a [gets $fileid]

      #if the string is long enough to be a strider file, extract the strider header info
        set strversion -1; set strtype -1
        binary scan $striderheader cccx14cx10IIx60I strversion strtype strtopology strmet strlength strminus strcomlength
 
      #check for header types
      if {$striderheader == ""} {
        set filetype "empty"
      } elseif {[string index $a 0] == ">"} {
        set filetype "fasta"
        set filecomment [string range $a 1 end]
      } elseif {[string range $a 0 4] == "LOCUS"} {
         set filetype "genbank"
      } elseif {[string range $a 0 1] == "ID"} {
         set filetype "embl"
      } elseif {([string first "Created:" $a] > -1) && ($b == "^^")} {  
         set filetype "editseq"
      } elseif {($strversion == 0) && (($strtype == 1) || ($strtype == 2))} {
         set filetype "strider"
      } elseif {([string length $a] > 10) && ([regexp -nocase {[^ACGTN*BDHKMRSVWY 0-9]} $a] == 0)} {
         set filetype "plain"
      } elseif {[string range $a 0 3] == "ABIF"} {
         set filetype "abi"
      } else {
          set filetype "unknown"
          tk_messageBox -message [mc "Not fasta, plain, genbank, EMBL or strider file"] -type ok -icon warning -default ok
          close $fileid
          set dialogblock 0
          return error
      }

      #read in DNA and comment for each file type
      set genbank_features [list]
      set genbank_header_info [list]

      switch $filetype {
        empty {
          set newdna ""
          set filecomment ""
          set startindex 1
          set circular "linear"
          set methylated $info(methylated_default)
          set new_tags [list]
          set new_title [mc "New_DNA"]
          set filetype "None"
        }
        fasta {
          sputs "fasta: $filename"
          set newdna ""
          set filecomment $a
          while {[eof $fileid] != 1} {append newdna [gets $fileid]}
          if {[set gt [string first ">" $newdna]] > -1} {
            set newdna [string range $newdna 0 $gt]
          }
          set startindex 1
          set circular "linear"
          set methylated $info(methylated_default)
          set new_tags [list]
          set new_title [file tail $filename]
          set filetype "Fasta"
        }
        plain {
          sputs "plain: $filename"
          set newdna $a
          while {[eof $fileid] != 1} {append newdna [gets $fileid]}
          set filecomment ""
          set startindex 1
          set circular "linear"
          set methylated $info(methylated_default)
          set new_tags [list]
          set new_title [file tail $filename]
          set filetype "Plain Text"
        }
        strider {
          sputs "strider: $filename"
          seek $fileid 112
          set newdna [read $fileid $strlength]
          #regsub -all {\*} $newdna "N" newdna
          set filecomment [read $fileid $strcomlength]
          set startindex [expr {1+$strminus}]
          set circular [expr ($strtopology?"circular":"linear")]
          if {($strmet != 0) && ($strmet != 1)} {set strmet 0}
          set methylated $strmet
          set new_tags [list]
          set new_title [file tail $filename]
          set filetype "DNA Strider"
        } 
        abi {
          sputs "abi: $filename"
          close $fileid
          open_abi $filename
          if {$info(dnadir_follows_open)} {
            set info(default_dnadir) [file dirname $filename]
          }
          set dialogblock 0
          return ok
        }
        genbank {
          sputs "genbank: $filename"
          set filecomment ""
          set methylated $info(methylated_default)

          if {![regexp {LOCUS       (.*) ([0-9]*) bp (   |ss-|ds-|ms-)(NA    |DNA   |RNA   |tRNA  |rRNA  |mRNA  |uRNA  |snRNA |snoRNA)[ ]*(linear  |circular|        )[ ]*([ A-Z]{3})[ ]*(..-...-....)} $a x name size stranded type circular div date] || $circular== "        "} {
            set circular "linear"
          }        
          set genbank_type_array(0) "LOCUS       "
          set genbank_data_array(0) [string range $a 12 end]
          set i 0
          set text ""
          while {([gets $fileid nextline] != -1) && ([string range $nextline 0 7] != "FEATURES") && ([string range $nextline 0 5] != "ORIGIN")} {
            set lastlen [string length $text]
            set text [string range $nextline 12 end]
            set type [string range $nextline 0 11]

            if {$type == "            "} {
              if {($lastlen + [string first " " "$text "]) > 66 } {
                append genbank_data_array($i) " " $text
              } else {
                append genbank_data_array($i) \n $text
              }
            } else {
              incr i
              set genbank_type_array($i) $type
              set genbank_data_array($i) $text
            }
          }
          set genbank_header_info [list]
          set vnti 0
          for {set j 0} {$j <= $i} {incr j} {
            if {$genbank_type_array($j) == "COMMENT     "} {
              if {[string first "This file is created by Vector NTI" $genbank_data_array($j)] == 0} {
                set vnti 1
              }
              if {$vnti} continue
              if {[regexp {ApEinfo:methylated:(1|0)} $genbank_data_array($j) temp methylated]} {
                unset temp
                continue
              }
              append filecomment $genbank_data_array($j)
            }
            lappend genbank_header_info $genbank_type_array($j) $genbank_data_array($j)
          }
          unset i j
          array unset genbank_type_array
          array unset genbank_data_array

          if {[string range $nextline 0 7] == "FEATURES"} {
            while {([gets $fileid nextline] != -1) && ([string range $nextline 0 5] != "ORIGIN")} {
              lappend genbank_features $nextline
            }
          }
          set newdna ""
          while {([gets $fileid nextline] != -1) && ($nextline != "//")} {
            set newdna "$newdna$nextline"
          }

          set new_tags [feature_table_to_taglist $genbank_features]
          
          set startindex 1

          set new_title [file tail $filename]
          set filetype "Genbank"
        }
        embl {
          sputs "embl: $filename"
          if {([string match "*;*circular*;*" $a])} { 
            set circular "circular"
          } else {
            set circular "linear"
          }
          set filecomment $a
          while {([gets $fileid nextline] != -1) && ([regexp {([A-Z]{2})\ *(.*)} $nextline nextline type text])} {
            if {$type == "FT"} {
              lappend genbank_features "  [string range $nextline 2 end]"
            } elseif {$type == "SQ"} {
              break
            }  else {
              set filecomment "$filecomment\n$nextline" 
            }
          }
          set newdna ""
          while {([gets $fileid nextline] != -1) && ($nextline != "//")} {
            set newdna "$newdna$nextline"
          }
          set new_tags [feature_table_to_taglist $genbank_features]
          set startindex 1
          set methylated $info(methylated_default)
          set new_title [file tail $filename]
          set filetype "EMBL"
        }
        editseq {
          sputs "editseq: $filename"
          set newdna ""
          set filecomment $a
          gets $fileid
          while {[eof $fileid] != 1} {append newdna [gets $fileid]}
          if {[regexp -nocase {[^ACGTN*BDHKMRSVWY 0-9]} $newdna] == 0} {
            set startindex 1
            set circular "linear"
            set methylated $info(methylated_default)
            set new_tags [list]
            set new_title [file tail $filename]
            set filetype "Editseq"
          } else {
            close $fileid
            tk_messageBox -message [mc "This editseq file contains protein sequence. I can't open it. Sorry."] -title [mc "Can't open protein sequences"] -icon info -type ok
            set dialogblock 0
            return error
          }
        }
        default {
          sputs "unknown filetype: $filename"
          set dialogblock 0
          close $fileid
          return error
        }  
      }
      close $fileid
      regsub -all {[^abcdghkmnrstvwyABCDGHKMNRSTVWY*]} $newdna "" newdna
      if {$info(dnadir_follows_open)} {
        set info(default_dnadir) [file dirname $filename]
      }
      create_window $newdna $filename $filecomment $circular $methylated $startindex $new_title $new_tags $genbank_header_info $filetype
      return ok
    } else {
      sputs  open error:$filename [file exists $filename] [file readable $filename]
      return error
    }
  } else {
    sputs filename == ""
    return error
  }
}

###########
## Duplicate a window
###########
proc duplicate_window {w} {
  global info
  set dna [$w.textarea get 1.0 1.end]
  set tags [get_tags $w.textarea {1.0 1.end} forward]
  create_window $dna "" [$w.comment get 1.0 end-1char] $info($w,circular) $info($w,Dam_Dcm_methylated) $info($w,startindex) New_DNA $tags $info($w,genbank_header_list)
} 

###########
## Make a new window from selection (duplicate all if no selection)
###########
proc duplicate_window {w {region selection}} {
  global info
  if {([llength [$w.textarea tag ranges sel]] > 0) && ($region == "selection")} {
    set start [$w.textarea index sel.first]
    set end [$w.textarea index sel.last]
    set circ linear
  } else {
    set start 1.0
    set end [$w.textarea index "1.end"]
    set circ $info($w,circular)
  }
  set dna [$w.textarea get $start $end]
  set tags [get_tags $w.textarea [list $start $end] forward]
  # {newdna ""} {filename ""} {filecomment ""} {circular linear} {methylated 1} {startindex 1} {title New_DNA} {tags ""}
  create_window $dna "" "$info($w,filename) from [expr {[string range $start 2 end]+1}] to [string range $end 2 end]\n[$w.comment get 1.0 end-1char]" $circ $info($w,Dam_Dcm_methylated) 0 New_DNA $tags $info($w,genbank_header_list)
}

###########
## Get a list of fixed fonts on the system (in the background)
###########
proc get_fixed_fonts {{i 0}} {
  global info temp_info

  if {$i == 0} {
    set info(fixedfontlist) [list courier]
  }
  if {$i < [llength [font families]]} {
    set font [lindex [font families] $i]
    if {![regexp {[^a-z,0-9,A-Z,\ ,\.\}\{]} \{$font\}]} {
      if [font metrics \{$font\} -fixed] {
        lappend info(fixedfontlist) $font
      }
      after 1 "get_fixed_fonts [expr {$i+1}]"
    }
  } else {
    if {[winfo exists .dialog.generaloptions.f1.menubutton1.menu]} {
      .dialog.generaloptions.f1.menubutton1.menu delete 0
      foreach fontfam $info(fixedfontlist) {
        .dialog.generaloptions.f1.menubutton1.menu add radiobutton -label $fontfam -indicatoron 1 -value $fontfam -variable temp_info(textfontfamily) -font [list $fontfam 10]
      }
    }
  }
}


###########
## Replacement for tk_messageBox
###########
rename tk_messageBox old_tk_messageBox


if {[tk windowingsystem] != "aqua"} {
###########
## On non-Aqua, it just needs to reset the focus
###########
proc tk_messageBox {args} {
  set focus_store [focus]
  set ret [eval old_tk_messageBox $args]
  focus $focus_store
  return $ret
}
} else {
###########
## Replacement for buggy and unattractive Aqua tk_messageBox
###########
proc tk_messageBox {args} {
  global info
  global messageboxok toolbar_images

  set default ""
  set icon "info"
  set message ""
  set parent ""
  set title ""
  set type ok
  foreach {name value} $args {
    if {[regsub -- - $name "" name] > 0} {
      set $name $value
    }
  }

  if {[winfo exists .t]} {destroy .t}
  toplevel .t -background $info(bg_color)

  wm overrideredirect .t 1
  wm title .t ""
  set icon monkey_icon
  set font boldlabelfont

  if {[tk windowingsystem] != "aqua"} {
    wm withdraw .t
  }

  set iconimage $toolbar_images($icon)

  #add in the image and message
  grid [canvas .t.c -height [image height $iconimage] -width [image width $iconimage] -bg $info(bg_color) -borderwidth 0] -row 0 -column 0 -rowspan 2 -sticky nwse
  .t.c create image 2 2 -image $iconimage -anchor nw



  #add in the buttons

    switch $type {
      ok {
        set buttonlist [list [mc OK]]
        if {$default == ""} {set default ok}
      }
      abortretryignore {
        set buttonlist [list [mc Abort] [mc Retry] [mc Ignore]]
      } 
      okcancel  {
        set buttonlist [list [mc Cancel] [mc OK]]
       }
      retrycancel {
        set buttonlist [list [mc Cancel] [mc Retry]]
      } 
      yesno {
        set buttonlist [list [mc No] [mc Yes]]
      }
      yesnocancel {
        set buttonlist [list [mc No] [mc Cancel] [mc Yes]]
      }
    }
    set wd 8

  grid [label .t.l -text $message -font $font -justify left] -row 0 -column 1 -columnspan [llength $buttonlist]  -sticky nwse -padx 10

  set i 1
  foreach button $buttonlist {
    grid [button .t.[string tolower $button] -command "set messageboxok $button" -height 1 -width $wd -text $button] -row 1 -column $i -sticky nw -padx 15
    incr i
  }




  if {($default != "") && ([winfo exists .t.$default])} {
    .t.$default configure -default active
    bind .t <Key-Return> ".t.$default invoke"
  }


  #move it to +3+80 relative to parent
  if {$parent != ""} {
    regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $parent] wgeom winx winy
    wm geometry .t "+[expr {(($winx<0)?0:$winx)+3}]+[expr {(($winy<0)?0:$winy)+80}]"
  } else {
    wm geometry .t "+200+200"
  }


  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .t] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 20}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 20}
  wm geometry .t "+$winx+$winy"
  update

  if {[tk windowingsystem]=="aqua"} {
    foreach button $buttonlist {
      grid remove ".t.[string tolower $button]"
      grid configure ".t.[string tolower $button]"
    }
  }

  wm deiconify .t

  focus .t
  grab .t
  vwait messageboxok
  grab release .t
  destroy .t
  return [string tolower $messageboxok]
}
}

###########
## creates a color picker button
###########
proc color_button {name variable {color red} {size 12}} {
  global [regsub {\(.*\)} $variable ""] info
  if {$color == ""} {set color $info(transp_feature_color)}
  set $variable $color
  set rgb_color [eval format "\#%04X%04X%04X" [winfo rgb . $color]] 
  set new_image [image create photo]
  $new_image put [string repeat "{[string repeat "$rgb_color " $size] } " $size]
  button $name -image $new_image -width $size -height $size
  $name configure -command "
    set button_priv \[tk_chooseColor -initialcolor \$$variable\];
    if \{\$button_priv != \"\"\} \{
      set $variable \$button_priv
    \}"

  bind $name <<updatevar>> "$new_image put  \[string repeat \"{\[string repeat \"\[eval format #%%04X%%04X%%04X \[winfo rgb . \$$variable\]\]  \" $size] } \" $size\]"
  trace add variable $variable write "event generate $name <<updatevar>>; #"
  bind $name <Destroy> "image delete $new_image"
  return $name
}

###########
## Program Preferences dialog
###########
proc configure_preferences {w} {
  global info temp_info abi_info temp_abi
  global keysubst
  global allowedkeys 

  global ok
  global selectedframe dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  toplevel .dialog
  wm title .dialog [mc "Preferences..."]
  wm protocol .dialog WM_DELETE_WINDOW "set ok -1"
  bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
  if {$w != "."} {
    regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
    wm geometry .dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  }
  if {[tk windowingsystem] != "aqua"} {
    wm transient .dialog $w
  } else {
  ::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
  }
  set ok 0



  ##frame selector frame
  grid [frame .dialog.frameselect -relief ridge -borderwidth 2] -row 0 -column 0 -sticky ns
  grid [listbox .dialog.frameselect.list -background white -foreground black -width 9 -exportselection 0] -row 0 -column 0 -sticky ns
  grid rowconfigure .dialog.frameselect 0 -weight 1
  .dialog.frameselect.list insert end [mc General]
  .dialog.frameselect.list insert end [mc Color]
  .dialog.frameselect.list insert end [mc ABI]
  .dialog.frameselect.list insert end [mc Files]
  if {[tk windowingsystem] == "win32"} {
    .dialog.frameselect.list insert end [mc Filetypes]
  }
  .dialog.frameselect.list selection set 0
  bind .dialog.frameselect.list <Button-1> {
    grid remove $selectedframe
    switch [.dialog.frameselect.list get [.dialog.frameselect.list index @%x,%y]] "
      [mc General] {
        set selectedframe .dialog.generaloptions
      }
      [mc Color] {
        set selectedframe .dialog.coloroptions
      }
      [mc ABI] {
        set selectedframe .dialog.abioptions
      }
      [mc Files] {
        set selectedframe .dialog.files
      }
      [mc Filetypes] {
        set selectedframe .dialog.filetypes
      }
    "
    grid configure $selectedframe
  }

  ###general options frame
  array set temp $info(keysubstmap)
  set keysubst(A) ""
  set keysubst(C) ""
  set keysubst(G) ""
  set keysubst(T) ""
  foreach sub [array names temp] {
    if {([string is upper $sub]) || (![string is alpha $sub])} {
      append keysubst($temp($sub)) $sub
    }
  }
  set allowedkeys [string map {"ACGTacgt" "ACGT" "ACGTNacgtn*" "ACGTN" "ABCDGHKMNRSTVWYabcdghkmnrstvwy*" "degenerate code"} $info(allowedkeys)]

  foreach element [list print_font_size dnafontsize textfontfamily graphicfontsize graphicfontfamily locale find_orf_starts draw_graphic_connectors show_comment comment_height selection_shows_Trans methylated_default] {
    set temp_info($element) $info($element)
  }


  grid [frame .dialog.generaloptions  -relief ridge -borderwidth 2] -row 0 -column 1 -sticky nswe
  grid [frame .dialog.generaloptions.f1 -relief ridge -borderwidth 3] -row 1 -column 0 -columnspan 2 -sticky nwe
  grid [label .dialog.generaloptions.f1.label1 -text [mc "Text font:"]] -row 0 -column 0 -sticky w

  grid [menubutton .dialog.generaloptions.f1.menubutton -textvariable temp_info(dnafontsize) -menu .dialog.generaloptions.f1.menubutton.menu -indicatoron 1 -relief raised] -row 0 -column 1 -sticky w
  menu .dialog.generaloptions.f1.menubutton.menu 
    foreach fontsize [list 8 10 12 14 18] {
      .dialog.generaloptions.f1.menubutton.menu add radiobutton -label $fontsize -indicatoron 1 -value $fontsize -variable temp_info(dnafontsize)
    }
  bind .dialog.generaloptions.f1.menubutton.menu <<MenuSelect>> "focus .dialog"
  grid [label .dialog.generaloptions.f1.label2 -text "pt"] -row 0 -column 2 -sticky w
  grid [menubutton .dialog.generaloptions.f1.menubutton1 -textvariable temp_info(textfontfamily) -menu .dialog.generaloptions.f1.menubutton1.menu -indicatoron 1 -relief raised] -row 0 -column 3 -sticky w
  menu .dialog.generaloptions.f1.menubutton1.menu 
  bind .dialog.generaloptions.f1.menubutton1.menu <<MenuSelect>> "focus .dialog"
  if {[llength $info(fixedfontlist)] == 0} {
    .dialog.generaloptions.f1.menubutton1.menu add command -label [mc "Searching for fixed fonts on the system, please wait..."]
    get_fixed_fonts
  } else {
    foreach fontfam $info(fixedfontlist) {
      .dialog.generaloptions.f1.menubutton1.menu add radiobutton -label $fontfam -indicatoron 1 -value $fontfam -variable temp_info(textfontfamily) -font [list $fontfam 10]
    }
  }
  grid [label .dialog.generaloptions.f1.label3 -text [mc "Graphic font:"]] -row 1 -column 0 -sticky w
  grid [menubutton .dialog.generaloptions.f1.menubutton2 -textvariable temp_info(graphicfontsize) -menu .dialog.generaloptions.f1.menubutton2.menu -indicatoron 1 -relief raised] -row 1 -column 1 -sticky w
  menu .dialog.generaloptions.f1.menubutton2.menu 
    foreach fontsize [list 8 10 12 14 18] {
      .dialog.generaloptions.f1.menubutton2.menu add radiobutton -label $fontsize -indicatoron 1 -value $fontsize -variable temp_info(graphicfontsize)
    }
  bind .dialog.generaloptions.f1.menubutton2.menu <<MenuSelect>> "focus .dialog"
  grid [label .dialog.generaloptions.f1.label4 -text "pt"] -row 1 -column 2 -sticky w
  grid [menubutton .dialog.generaloptions.f1.menubutton3 -textvariable temp_info(graphicfontfamily) -menu .dialog.generaloptions.f1.menubutton3.menu -indicatoron 1 -relief raised] -row 1 -column 3 -sticky w
  menu .dialog.generaloptions.f1.menubutton3.menu 
    foreach fontfam [list helvetica times courier] {
      .dialog.generaloptions.f1.menubutton3.menu add radiobutton -label $fontfam -indicatoron 1 -value $fontfam -variable temp_info(graphicfontfamily) -font [list $fontfam 10]
    }
  bind .dialog.generaloptions.f1.menubutton3.menu <<MenuSelect>> "focus .dialog"
  grid [label .dialog.generaloptions.f1.label5 -text [mc "Printer font:"]] -row 2 -column 0 -sticky w
  grid [menubutton .dialog.generaloptions.f1.menubutton4 -textvariable temp_info(print_font_size) -menu .dialog.generaloptions.f1.menubutton4.menu -indicatoron 1 -relief raised] -row 2 -column 1 -sticky w
  menu .dialog.generaloptions.f1.menubutton4.menu 
    foreach fontsize [list 6 8 10 12 14 18] {
      .dialog.generaloptions.f1.menubutton4.menu add radiobutton -label $fontsize -indicatoron 1 -value $fontsize -variable temp_info(print_font_size)
    }
  bind .dialog.generaloptions.f1.menubutton4.menu <<MenuSelect>> "focus .dialog"
  grid [label .dialog.generaloptions.f1.label6 -text "pt"] -row 2 -column 2 -sticky w
  grid [label .dialog.generaloptions.f1.label7 -text "[font actual courier -family]"] -row 2 -column 3 -sticky w
  grid [label .dialog.generaloptions.f1.label8 -text [mc "Locale:"]] -row 3 -column 0 -sticky w
  grid [menubutton .dialog.generaloptions.f1.menubutton5 -textvariable temp_info(locale) -menu .dialog.generaloptions.f1.menubutton5.menu -indicatoron 1 -relief raised] -row 3 -column 1 -sticky w

  menu .dialog.generaloptions.f1.menubutton5.menu 
    foreach localefile [glob -directory [file join $info(Accdir) Msgs] *.msg] {
      set localefile [string tolower [file rootname [file tail $localefile]]]
      .dialog.generaloptions.f1.menubutton5.menu add radiobutton -label $localefile -indicatoron 1 -value $localefile -variable temp_info(locale)
    }
  bind .dialog.generaloptions.f1.menubutton5.menu <<MenuSelect>> "focus .dialog"
   

  grid columnconfigure .dialog.generaloptions.f1 3 -weight 1 -pad 0

  grid [frame .dialog.generaloptions.f2 -relief ridge -borderwidth 3] -row 2 -column 0 -columnspan 2 -sticky nwe
  grid [label .dialog.generaloptions.f2.label1 -text [mc "Allowed keys:"]] -row 0 -column 0  -sticky w
  grid [menubutton .dialog.generaloptions.f2.menubutton -textvariable allowedkeys -menu .dialog.generaloptions.f2.menubutton.menu -indicatoron 1 -relief raised] -row 0 -column 1  -sticky w
  menu .dialog.generaloptions.f2.menubutton.menu 
    .dialog.generaloptions.f2.menubutton.menu add radiobutton -label "ACGT" -value "ACGT" -variable allowedkeys -state active
    .dialog.generaloptions.f2.menubutton.menu add radiobutton -label "ACGTN" -value "ACGTN" -variable allowedkeys
    .dialog.generaloptions.f2.menubutton.menu add radiobutton -label "degenerate code" -value "degenerate code" -variable allowedkeys
  bind .dialog.generaloptions.f2.menubutton.menu <<MenuSelect>> "focus .dialog"
  grid [frame .dialog.generaloptions.f2.f1 -relief flat -borderwidth 3] -row 2 -column 0  -columnspan 2 -sticky nwe
  grid [label .dialog.generaloptions.f2.f1.label2 -text [mc "Key substitutions:"]] -row 1 -column 0  -sticky w
  grid [entry .dialog.generaloptions.f2.f1.entryA -textvariable keysubst(A) -width 3 -justify right] -row 1 -column 1 -sticky w
  grid [label .dialog.generaloptions.f2.f1.label3 -text "=A,  "] -row 1 -column 2  -sticky w
  grid [entry .dialog.generaloptions.f2.f1.entryC -textvariable keysubst(C) -width 3 -justify right] -row 1 -column 3 -sticky w
  grid [label .dialog.generaloptions.f2.f1.label4 -text "=C,  "] -row 1 -column 4  -sticky w
  grid [entry .dialog.generaloptions.f2.f1.entryG -textvariable keysubst(G) -width 3 -justify right] -row 1 -column 5 -sticky w
  grid [label .dialog.generaloptions.f2.f1.label5 -text "=G,  "] -row 1 -column 6  -sticky w
  grid [entry .dialog.generaloptions.f2.f1.entryT -textvariable keysubst(T) -width 3 -justify right] -row 1 -column 7 -sticky w
  grid [label .dialog.generaloptions.f2.f1.label6 -text "=T.  "] -row 1 -column 8  -sticky w
  grid columnconfigure .dialog.generaloptions.f2 1 -weight 1

  grid [frame .dialog.generaloptions.f4 -relief ridge -borderwidth 3] -row 3 -column 0 -columnspan 2 -sticky nwe
  grid [label .dialog.generaloptions.f4.label1 -text [mc "Find ORFs starting with:"]] -row 0 -column 0  -sticky w
  grid [menubutton .dialog.generaloptions.f4.menubutton -textvariable temp_info(find_orf_starts) -menu .dialog.generaloptions.f4.menubutton.menu -indicatoron 1 -relief raised] -row 0 -column 1
    menu .dialog.generaloptions.f4.menubutton.menu
    .dialog.generaloptions.f4.menubutton.menu add radiobutton -label "Met" -value "Met" -variable temp_info(find_orf_starts)
    .dialog.generaloptions.f4.menubutton.menu add radiobutton -label "Stop+1" -value "Stop+1" -variable temp_info(find_orf_starts)
  bind .dialog.generaloptions.f4.menubutton.menu <<MenuSelect>> "focus .dialog"
  grid columnconfigure .dialog.generaloptions.f4 2 -weight 1 -pad 0

  grid [frame .dialog.generaloptions.f5 -relief ridge -borderwidth 3] -row 4 -column 0 -columnspan 2 -sticky nwe
  grid [checkbutton .dialog.generaloptions.f5.connectors -text [mc "Draw Connectors in Graphic Map"] -selectcolor white -onvalue 1 -offvalue 0 -variable temp_info(draw_graphic_connectors)] -row 0 -column 0 -columnspan 3 -sticky nw
  grid [checkbutton .dialog.generaloptions.f5.meth -text [mc "Dam/Dcm default"] -selectcolor white -onvalue 1 -offvalue 0 -variable temp_info(methylated_default)] -row 1 -column 0 -columnspan 3 -sticky nw
  grid [checkbutton .dialog.generaloptions.f5.comment -text [mc "Show File Comment"] -selectcolor white -onvalue 1 -offvalue 0 -variable temp_info(show_comment)] -row 2 -column 0 -sticky nw
  grid [label .dialog.generaloptions.f5.label1 -text [mc "Comment Height"]] -row 2 -column 1 -sticky ne
  grid [entry .dialog.generaloptions.f5.commheight -textvariable temp_info(comment_height) -width 3 -justify right ] -row 2 -column 2 -sticky nw
  grid [checkbutton .dialog.generaloptions.f5.transel -text [mc "Translate Selection"] -selectcolor white -onvalue 1 -offvalue 0 -variable temp_info(selection_shows_Trans)] -row 3 -column 0 -columnspan 3 -sticky nw

  grid rowconfigure .dialog.generaloptions 0 -weight 1
  grid rowconfigure .dialog.generaloptions 1 -weight 1
  grid rowconfigure .dialog.generaloptions 2 -weight 1
  set selectedframe .dialog.generaloptions

  ###color options frame
  grid [frame .dialog.coloroptions  -relief ridge -borderwidth 2] -row 0 -column 1 -sticky nswe
  grid [label .dialog.coloroptions.l1  -text [mc "Background"]] -row 0 -column 1 -sticky n
  grid [label .dialog.coloroptions.l2  -text [mc "Foreground"]] -row 0 -column 2 -sticky n
  grid [label .dialog.coloroptions.l3  -text [mc "Program"]] -row 1 -column 0 -sticky nw
  grid [color_button .dialog.coloroptions.b1  temp_info(bg_color) $info(bg_color)] -row 1 -column 1 -sticky n
  grid [label .dialog.coloroptions.l4  -text [mc "Menu"]] -row 2 -column 0 -sticky nw
  grid [color_button .dialog.coloroptions.b2  temp_info(menu_bg_color) $info(menu_bg_color)] -row 2 -column 1 -sticky n
  grid [color_button .dialog.coloroptions.b3  temp_info(menu_fg_color) $info(menu_fg_color)] -row 2 -column 2 -sticky n
  grid [label .dialog.coloroptions.l5  -text [mc "Menu (selected)"]] -row 3 -column 0 -sticky nw
  grid [color_button .dialog.coloroptions.b4  temp_info(menu_select_bg_color) $info(menu_select_bg_color)] -row 3 -column 1 -sticky n
  grid [color_button .dialog.coloroptions.b5  temp_info(menu_select_fg_color) $info(menu_select_fg_color)] -row 3 -column 2 -sticky n
  grid [label .dialog.coloroptions.l6  -text [mc "Text"]] -row 4 -column 0 -sticky nw
  grid [color_button .dialog.coloroptions.b6  temp_info(text_bg_color) $info(text_bg_color)] -row 4 -column 1 -sticky n
  grid [color_button .dialog.coloroptions.b7  temp_info(text_fg_color) $info(text_fg_color)] -row 4 -column 2 -sticky n
  grid [label .dialog.coloroptions.l7  -text [mc "Text (selected)"]] -row 5 -column 0 -sticky nw
  grid [color_button .dialog.coloroptions.b8  temp_info(text_select_bg_color) $info(text_select_bg_color)] -row 5 -column 1 -sticky n
  grid [color_button .dialog.coloroptions.b9  temp_info(text_select_fg_color) $info(text_select_fg_color)] -row 5 -column 2 -sticky n
  grid [label .dialog.coloroptions.l8  -text [mc "Text (background)"]] -row 6 -column 0 -sticky nw
  grid [color_button .dialog.coloroptions.b10  temp_info(text_select2_bg_color) $info(text_select2_bg_color)] -row 6 -column 1 -sticky n
  grid [color_button .dialog.coloroptions.b11  temp_info(text_select2_fg_color) $info(text_select2_fg_color)] -row 6 -column 2 -sticky n

  grid [label .dialog.coloroptions.l9  -text [mc "Find(forward)"]] -row 7 -column 0 -sticky nw
  grid [color_button .dialog.coloroptions.b12  temp_info(find_forward_highlight) $info(find_forward_highlight)] -row 7 -column 1 -sticky n
  grid [label .dialog.coloroptions.l10  -text [mc "Find(reverse)"]] -row 8 -column 0 -sticky nw
  grid [color_button .dialog.coloroptions.b13  temp_info(find_reverse_highlight) $info(find_reverse_highlight)] -row 8 -column 1 -sticky n
  grid [label .dialog.coloroptions.l11  -text [mc "Enzyme highlight"]] -row 9 -column 0 -sticky nw
  grid [color_button .dialog.coloroptions.b14  temp_info(enzyme_highlight_color) $info(enzyme_highlight_color)] -row 9 -column 1 -sticky n

  if {[tk windowingsystem] == "win32"} {
    set defaultcolors [list bg_color SystemButtonFace menu_bg_color SystemMenu menu_fg_color SystemMenuText menu_select_bg_color SystemHighlight menu_select_fg_color SystemHighlightText text_bg_color SystemWindow text_fg_color SystemWindowText text_select_bg_color SystemHighlight text_select_fg_color SystemHighlightText]
  } else {
    set defaultcolors [list bg_color gray85 menu_bg_color gray85 menu_fg_color black menu_select_bg_color #008 menu_select_fg_color white text_bg_color white text_fg_color black text_select_bg_color #008 text_select_fg_color white text_select2_bg_color #6060d0 text_select2_fg_color white]
  }
  grid [button .dialog.coloroptions.b15 -text [mc "Restore Program Defaults"] -command "
    foreach {var color} {$defaultcolors} {
      set temp_info(\$var) \$color
    }
  "] -row 10 -column 0 -columnspan 2

  grid remove .dialog.coloroptions

  ###ABI options frame
  grid [frame .dialog.abioptions -relief ridge -borderwidth 2] -row 0 -column 1 -sticky nswe
  foreach base [list A C G T N] row [list 1 2 3 4 5] {
    set temp_abi($base) $abi_info(colors,$base)
    grid [label .dialog.abioptions.label$base -justify left -text "$base\:" ] -row $row -column 0 -sticky nw
    grid [color_button .dialog.abioptions.button$base  temp_abi($base) $temp_abi($base)] -row $row -column 1 -sticky nw
  }
  set temp_abi(non) $abi_info(colors,B)
  grid [label .dialog.abioptions.labelnon -justify left -text "BDHKMRSVWY\:" ] -row 6 -column 0 -sticky nw
  grid [color_button .dialog.abioptions.buttonnon  temp_abi(non) $temp_abi(non)] -row 6 -column 1 -sticky nw
  grid remove .dialog.abioptions

  ###Files options frame
  foreach var [list default_dnadir dnadir_follows_open default_featuredir featuredir_follows_open default_enzymedir default_file_format save_file_format_warning default_file_extension] {
    set temp_info($var) $info($var)
  }
  grid [frame .dialog.files -relief ridge -borderwidth 2] -row 0 -column 1 -sticky nswe
  grid [frame .dialog.files.f1 -relief ridge -borderwidth 2] -row 0 -column 0 -sticky nswe
  grid [label .dialog.files.f1.dnalabel -text [mc "Default DNA Directory:"]] -row 0 -column 0
  grid [label .dialog.files.f1.dnadir -textvariable temp_info(default_dnadir)] -row 1 -column 0 -columnspan 2 -sticky nswe
  grid [button .dialog.files.f1.changednadir -text [mc "Change Directory"] -command {set temp_info(default_dnadir) [tk_chooseDirectory -initialdir $temp_info(default_dnadir) -mustexist 1 -title "Chose new DNA Directory..." -parent .dialog]}] -row 2 -column 0
  grid [checkbutton .dialog.files.f1.follows -text [mc "Follows Last Opened File"] -selectcolor white -onvalue 1 -offvalue 0 -variable temp_info(dnadir_follows_open)] -row 2 -column 1
  grid [frame .dialog.files.f2 -relief ridge -borderwidth 2] -row 1 -column 0 -sticky nswe
  grid [label .dialog.files.f2.dnalabel -text [mc "Default Feature Directory:"]] -row 0 -column 0
  grid [label .dialog.files.f2.dnadir -textvariable temp_info(default_featuredir)] -row 1 -column 0 -columnspan 2 -sticky nswe
  grid [button .dialog.files.f2.changednadir -text [mc "Change Directory"] -command {set temp_info(default_featuredir) [tk_chooseDirectory -initialdir $temp_info(default_featuredir) -mustexist 1 -title "Chose New Feature Directory..." -parent .dialog]}] -row 2 -column 0
  if {[tk windowingsystem] == "aqua"} {
    grid [button .dialog.files.f2.exportfeatures -text [mc "Create Default Feature File"] -command {
      if {(![file exists [file join $temp_info(default_featuredir) "Default_Features.txt"]]) || ([tk_messageBox -message [mc "Replace Default_Features.txt in directory %1$s?" $temp_info(default_featuredir)] -title [mc "Default Features file exists"] -icon question -type yesno -default no] == "yes")} {
          file copy [file join $info(Accdir) "Features" "Default_Features.txt"] [file join $temp_info(default_featuredir) "Default_Features.txt"]
      }
    }] -row 3 -column 0
  }
  grid [checkbutton .dialog.files.f2.follows -text [mc "Follows Last Opened File"] -selectcolor white -onvalue 1 -offvalue 0 -variable temp_info(featuredir_follows_open)] -row 2 -column 1
  grid [frame .dialog.files.f3 -relief ridge -borderwidth 2] -row 2 -column 0 -sticky nswe
  grid [label .dialog.files.f3.dnalabel -text [mc "Default Enzyme Directory:"]] -row 0 -column 0
  grid [label .dialog.files.f3.dnadir -textvariable temp_info(default_enzymedir)] -row 1 -column 0 -columnspan 2 -sticky nswe
  grid [button .dialog.files.f3.changednadir -text [mc "Change Directory"] -command {set temp_info(default_enzymeedir) [tk_chooseDirectory -initialdir $temp_info(default_enzymedir) -mustexist 1 -title "Chose New Enzyme Directory..." -parent .dialog]}] -row 2 -column 0

  grid [frame .dialog.files.f4 -relief ridge -borderwidth 2] -row 3 -column 0 -sticky nswe
  grid [label .dialog.files.f4.formatlabel -text [mc "Save Files in Format:"]] -row 0 -column 0 -sticky wn
  tk_optionMenu .dialog.files.f4.format temp_info(default_file_format) Genbank {DNA Strider}
  grid .dialog.files.f4.format -row 0 -column 1 -sticky wn
  grid [checkbutton .dialog.files.f4.formatwarning -text [mc "Warn if Changing File Formats"] -selectcolor white -onvalue 1 -offvalue 0 -variable temp_info(save_file_format_warning)] -row 1 -column 0 -sticky nw -columnspan 2
  grid [label .dialog.files.f4.extensionlabel -text [mc "Default File Extension"]] -row 2 -column 0 -sticky wn
  tk_optionMenu .dialog.files.f4.extension temp_info(default_file_extension) .ape .seq .dna .str .gb
  grid .dialog.files.f4.extension -row 2 -column 1 -sticky wn
  grid columnconfigure .dialog.files.f4 1 -weight 1 -uniform 1
  grid remove .dialog.files

  ###Filetypes options frame
  if {[tk windowingsystem] == "win32"} {
    global possible_extensions
    grid [frame .dialog.filetypes  -relief ridge -borderwidth 2] -row 0 -column 1 -sticky nswe
    set registered [get_registered_types]
    set ape_registered [lindex $registered 0]
    set other_registered [lindex $registered 1]
    set other_appl [lindex $registered 2]
    set r 0
    foreach extension $possible_extensions {
      if {[lsearch -exact $ape_registered $extension] > -1} {
        set temp_info($extension,selected) 2
        grid [checkbutton .dialog.filetypes.button$r -text "$extension"  -selectcolor white -variable temp_info($extension,selected) -onvalue 2 -offvalue 0] -row $r -column 0 -sticky w
      } elseif {[set pos [lsearch $other_registered $extension]] > -1} {
        set temp_info($extension,selected) -3
        grid [label .dialog.filetypes.button$r -text [format [mc {%1$s is already: %2$s}] $extension [lindex $other_appl $pos]] -fg red] -row $r -column 0 -sticky w
      } else {
        set temp_info($extension,selected) -1
        grid [checkbutton .dialog.filetypes.button$r -text "$extension" -selectcolor white -variable temp_info($extension,selected) -onvalue 1 -offvalue -1] -row $r -column 0 -sticky w
      }
      incr r
    }
    grid remove .dialog.filetypes
  }

  ###ok buttons frame
  grid [frame .dialog.f3 -relief flat] -row 1 -column 0 -columnspan 2
  grid [button .dialog.f3.ok -command "set ok 1" -text [mc "OK"] -default active] -row 0 -column 1 -padx 10 
  grid [button .dialog.f3.cancel -command "set ok -1" -text [mc "Cancel"]] -row 0 -column 0 -padx 10
  grid columnconfigure .dialog.f3 0 -weight 1 -uniform 1
  grid columnconfigure .dialog.f3 1 -weight 1 -uniform 1


  grid columnconfigure .dialog 1 -weight 1
  grid rowconfigure .dialog 0 -weight 1
  #wm resizable .dialog 0 0
  update idletasks

  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .dialog] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry .dialog "+$winx+$winy"
  update

  focus .dialog
  #grab set .dialog

  vwait ok
  destroy .dialog
  set dialogblock 0
  bind . <<RaiseDialogs>> ""

  ###execute changes
  if {$ok == 1} {
    #text changes
    if {$temp_info(dnafontsize) != $info(dnafontsize)} {
      font configure dnafont  -family $temp_info(textfontfamily) -size $temp_info(dnafontsize)
      font configure labelfont -size $temp_info(dnafontsize)
      font configure boldlabelfont -size $temp_info(dnafontsize)
      font configure entryfont -size $temp_info(dnafontsize)
    } elseif {$temp_info(textfontfamily) != $info(textfontfamily)} {
      font configure dnafont  -family $temp_info(textfontfamily) -size $temp_info(dnafontsize)
    }
    if {($temp_info(graphicfontfamily) != $info(graphicfontfamily)) || ($temp_info(graphicfontsize) != $info(graphicfontsize))} {
      font configure graphicfont -family $temp_info(graphicfontfamily) -size $temp_info(graphicfontsize)
    }

    if {($temp_info(locale) != $info(locale))} {
      mclocale $temp_info(locale)
      mcload [file join $info(Accdir) Msgs]
    }

    if {(![string is integer $temp_info(comment_height)]) || ($temp_info(comment_height) > 10) ||($temp_info(comment_height) <0)} {
      set temp_info(comment_height) $info(comment_height)
    }

    foreach element [array names temp_info] {
      set info($element) $temp_info($element)
    }

    set info(allowedkeys) [string map {"ACGTN" "ACGTNacgtn*" "ACGT" "ACGTacgt" "degenerate code" "ABCDGHKMNRSTVWYabcdghkmnrstvwy*"} $allowedkeys] 
    set info(keysubstmap) [list]
    foreach sub [array names keysubst] {
      foreach char [split $keysubst($sub)] {
        if {[string is alpha $char]} {
          lappend info(keysubstmap) [string toupper $char] $sub [string tolower $char] [string tolower $sub]
        } else {
          lappend info(keysubstmap) $char $sub
        }
      }
    }

    #color changes
    tk_setPalette background $info(bg_color) activeBackground $info(bg_color) activeForeground black selectForeground $info(text_select_fg_color) selectBackground $info(text_select_bg_color);#selectColor white
    option add *Text.background $info(text_bg_color)
    option add *Text.foreground $info(text_fg_color)
    option add *Entry.background $info(text_bg_color)
    option add *Entry.foreground $info(text_fg_color)

    option add *Menu*foreground $info(menu_fg_color)
    option add *Menu*background $info(menu_bg_color)
    option add *Menu*activeBackground $info(menu_select_bg_color)
    option add *Menu*activeForeground $info(menu_select_fg_color)

    option add *Checkbutton.activeBackground $info(bg_color)
    option add *Button.selectColor white
    option add *Radiobutton.activeBackground $info(bg_color)

    label .temp
    set info(label_fg_color) [.temp cget -foreground]
    destroy .temp

    foreach w [winfo children .] {
      if {[winfo exists $w.menubar]} {
        recolor_menu $w.menubar
      }
      if {[winfo exists $w.textframe.text]} {
        $w.textframe.text configure -background $info(text_bg_color) -foreground $info(text_fg_color)
        $w.textframe.text tag configure sel -background $info(text_select_bg_color) -foreground $info(text_select_fg_color) -relief flat
      }
      if {[winfo exists $w.text]} {
        $w.text configure -background $info(text_bg_color) -foreground $info(text_fg_color)
        $w.text tag configure sel -background $info(text_select_bg_color) -foreground $info(text_select_fg_color) -relief flat
      }
    }
    foreach w [regexp -all -inline ".dna_window\[0-9\]* " "[winfo children .] "] {
      set w [string range $w 0 end-1]
      if {[winfo exists $w.textarea]} {
        $w.textarea configure -background $info(text_bg_color) -foreground $info(text_fg_color)
        $w.textarea tag configure sel -background $info(text_select_bg_color) -foreground $info(text_select_fg_color) -relief flat
        $w.textarea tag configure tempsel -background $info(text_select2_bg_color) -foreground $info(text_select2_fg_color)
        $w.textarea tag configure foundf -background $info(find_forward_highlight)
        $w.textarea tag configure foundr -background $info(find_reverse_highlight)
        $w.horzindex configure -background $info(bg_color) -fg $info(label_fg_color)
        $w.horzindex2 configure -background $info(bg_color) -fg $info(label_fg_color)
        $w.vertindex configure -background $info(bg_color) -fg $info(label_fg_color)
        $w.infoframe.saved configure -disabledforeground $info(label_fg_color)
        $w.comment configure -background $info(text_bg_color) -foreground $info(text_fg_color)
      }
    }

    #ABI changes
    foreach base [list A C G T N] {
      set abi_info(colors,$base) $temp_abi($base)
    }
    foreach base [list B D H K M R S V W Y] {
      set abi_info(colors,$base) $temp_abi(non)
    }
    foreach window [winfo children .] { 
      if {([string match ".abi_window*" $window]) && (![string match "*analysis*" $window])} {
        foreach base [list A C G T N B D H K M R S V W Y] {
          $window.c itemconfigure $base -fill $abi_info(colors,$base)
        }
      }
    }
    #filetype changes
    if {[tk windowingsystem] == "win32"} {
      set on_list [list]
      set off_list [list]
      set tot_on 0
      foreach extension $possible_extensions {
        if {$temp_info($extension,selected) > 0} {incr tot_on}
        if {$temp_info($extension,selected) == 0} {lappend off_list $extension}
        if {$temp_info($extension,selected) == 1} {lappend on_list $extension}
      }
      file_typer $on_list $off_list [expr {$tot_on==0}]
    }

    save_defaults
  }

  array unset keysubst
  array unset temp_info
  array unset temp_abi
  unset allowedkeys

  unset ok
  unset selectedframe
}

###########
## recolor menus when color schemes change
###########
proc recolor_menu {m} {
  global info
  $m configure -background $info(menu_bg_color) -foreground $info(menu_fg_color) -activebackground $info(menu_select_bg_color) -activeforeground $info(menu_select_fg_color)
  foreach m2 [winfo children $m] {
    recolor_menu $m2
  }
}

###########
## Save program defaults
###########
proc save_defaults {} {
  global info

  set required_items [list text_select2_fg_color graphicfontsize locale menu_select_bg_color menu_bg_color menu_select_fg_color textfontfamily default_featuredir graphicfontfamily find_orf_starts menu_fg_color default_file_extension methylated_default selection_shows_Trans default_dnadir default_file_format save_file_format_warning dnadir_follows_open draw_graphic_connectors featuredir_follows_open text_select_bg_color find_reverse_highlight text_bg_color show_comment comment_height find_forward_highlight text_select2_bg_color enzyme_highlight_color dnafontsize text_select_fg_color text_fg_color bg_color print_font_size]
  set oldfilename [file join $info(user_defaults_dir) "ApE_Defaults.txt"]

  if {[file exists $oldfilename] && [file writable $oldfilename]} {
    #make a unique temporary filename
    set base [file join $info(Accdir) "O"]
    set i 0
    while {([file exists [set newfilename "$base$i"]])} {incr i}

    #open temp file for writing, old prefs file for reading
    set newfile [open $newfilename w]
    set oldfile [open $oldfilename r]

    set infolist [array names info]
    set saved_items [list]

    #read the prefs file and update values for all items that exist in the info array
    while {[gets $oldfile nextline] != -1} {
      if {([string index $nextline 0] == "#") || ($nextline =="") || ([string first \t $nextline] == -1)} {
        puts $newfile $nextline
      } else {
        set l [split $nextline \t]
        if {[lsearch $infolist [lindex $l 0]] > -1} {
          puts $newfile "[lindex $l 0]\t$info([lindex $l 0])"
          lappend saved_items [lindex $l 0]
        } else {
          puts $newfile $nextline
        }
      }
    }
    close $oldfile

    #add items that should exist but don't
    foreach item $required_items {
      if {[lsearch $saved_items $item] == -1} {
        puts $newfile "$item\t$info($item)"
      }
    } 
    close $newfile
    file delete $oldfilename
    file rename $newfilename $oldfilename

  } else {
    warning_dialog "Can't find the \"ApE_Defaults.txt\" file, or file is locked."
  }
}

###########
## returns text with all non-ATCG characters removed
###########
proc atgcfilter {s {allowed_keys "acgtACGT"} {message 0}} {

  regsub -all "\[^$allowed_keys\]" $s {} result
  regsub -all "\[$allowed_keys\]" $s {} disallowed
  regsub -all "\n" $disallowed "" disallowed
  regsub -all "\t" $disallowed "" disallowed
  regsub -all "\f" $disallowed "" disallowed
  regsub -all "\ " $disallowed "" disallowed
  regsub -all "\r" $disallowed "" disallowed
  set disallowed [join [lsort -unique [split $disallowed ""]] ""]
  if {$message && ($disallowed != [list])} {
    if {[string length $disallowed] > 100} {
      set disallowed "[string range $disallowed 0 100]..."
    }
    if {[string first N $allowed_keys] > -1} {
      set answer [tk_messageBox -icon warning -title [mc "Disallowed characters being inserted..."] -message [mc "Remove disallowed characters: %1\$s ? \n\n You may remove them (Yes), convert them to N (No), or cancel the insertion (Cancel)." $disallowed] -type yesnocancel -default cancel]
    } else {
      set answer [tk_messageBox -icon warning -title [mc "Disallowed characters being inserted..."] -message [mc "Remove disallowed characters: %1\$s ? \n\n You may remove them (OK) or cancel the insertion (Cancel)." $disallowed] -type okcancel -default cancel]
    }
sputs $answer
    if {($answer == "yes") || ($answer == "ok")} {
      return $result
    } elseif {($answer == "no") && ([string first N $allowed_keys] > -1)} {
      regsub -all -nocase "\[^$allowed_keys\]" $s "N" s
      return $s
    } else {
      return ""
    }
  }
  return $result
}


#########
## close window- keep track of open windows, so you can exit if all are closed
#########
proc closewindow {w} {
  #in aqua, there is a menubar child of .
  if {$w == ".menubar"} return
  global abi_info
  global info

  set proceede 1
  if {[string match "*_analysis*" $w]} {
    #####close an analysis window
     #look at all of the events linked to windows linked to this analysis window
    foreach window $info($w,linked_windows) {
      set templist [list]
      foreach event $info($window,linked_events) {
        #if the event doesn't refer to this analysis window, keep it in the list
        if {[winfo toplevel [lindex $event 0]] != $w} { 
          lappend templist $event
        }
      }
      set info($window,linked_events) $templist
    }
    array unset info "$w,*"

    if {[winfo exists $w]} {destroy $w}
  } elseif {[string match ".dna_window*" $w]} {
    ####close a dna window
    if {$info($w,saved) == 0} {
      wm deiconify $w
      raise $w
      set answer [tk_messageBox -icon warning -title [mc "ApE: File not saved"] -message "[mc {Do you want to save the changes you made to}]\n[expr {($info($w,filename)=="")?[wm title $w]:$info($w,filename)}]?" -type yesnocancel -default yes -parent $w]
      if {$answer == "yes"} {
       save_file $w Save
      } elseif {$answer == "cancel"} {
        set proceede 0
      }
    }
    if {$proceede == 1} {
      unlink_analysis_window $w
      if {[winfo exists $w]} {destroy $w}
      array unset info "$w,*"
      if {[llength [array names info $w]] > 0} {warning_dialog "info array still has elements: [array names info $w]"}
      set templist [list]
      set temptitleslist [list]
      foreach element $info(dnawindows_list) {
        if {$element != $w} {lappend templist $element; lappend temptitleslist $info($element,win_title)}
      }
      set info(dnawindows_list) $templist
      set info(dnatitles_list) $temptitleslist
    }
  } elseif {[string match ".abi_window*" $w]} {
    #####close an abi window
    unlink_analysis_window $w
    if {[winfo exists $w]} {destroy $w}
    array unset abi_info "$w,*"
  } 

  update_windows_menu
  if {[tk windowingsystem] != "aqua"} {
    if {([llength $info(dnawindows_list)] == 0)} {
      if {([llength [winfo children .]] > 0) && ([tk_messageBox -message [mc "Do you want to exit?"] -title [mc "Last DNA window"] -icon question -type yesno -default no] == "no")} {
        create_window
      } else {
        final_exit
      }
    }
  } else {
    if {([llength $info(dnawindows_list)] == 0)} {
      bind . <Command-KeyPress-o> open_file_dialog
      bind . <Command-KeyPress-n> create_window
      bind . <Command-KeyPress-q> close_all_windows
      .menubar.filemenu entryconfigure New -state normal
      .menubar.filemenu entryconfigure "Open..." -state normal
    }
  }
  return $proceede
}

########
## Rename exit to final_exit, so exit commands from the outside can be re-directed
########
rename exit realy_final_exit

########
## Exiting, no more dialogs
########
proc final_exit {} {
  global info fea_info

  if {[tk windowingsystem] == "win32"} {
    dde servername " "
  }
  if {$info(enzymes_saved) != 1} {
    if {[tk_messageBox -message "Do you want to save changes to the enzyme set?" -type yesno] == "yes"} {
      save_enzymes "" 0
    }
  }
  if {($fea_info(library_file) == "") && ([llength $fea_info(library)] > 0)} {
    if {[tk_messageBox -message "Do you want to save changes to the feature file?" -type yesno] == "yes"} {
      save_feature_library .
    }
  }

  save_defaults

  realy_final_exit
}

########
## Now, direct exit commands to close all windows
########
proc exit {} {
  close_all_windows
}

########
## break the event links when a dna window is modified
########
proc unlink_analysis_window {w} {
  global info
#unbind all the bound events pointing to $w
  if {[lsearch [array names info] "$w,linked_events"] == -1} {return}
  foreach event $info($w,linked_events) {
    if {[llength $event] == 3} {
      [lindex $event 0] configure -background gray95
      if {[winfo class [lindex $event 0]] == "Text"} {  
        [lindex $event 0] tag bind [lindex $event 1] [lindex $event 2] "
          tk_messageBox -message {The window this analysis refers to has been modified.} -icon info -type ok -default ok
          [lindex $event 0] tag bind [lindex $event 1] [lindex $event 2] {}
        "
      } else {
        [lindex $event 0] bind [lindex $event 1] [lindex $event 2] "
          tk_messageBox -message {The window this analysis refers to has been modified.} -icon info -type ok -default ok
          [lindex $event 0] bind [lindex $event 1] [lindex $event 2] {}
        "
      }
    } else {
      bind [lindex $event 0] [lindex $event 1] "
        tk_messageBox -message {The window this analysis refers to has been modified.} -icon info -type ok -default ok
        bind [lindex $event 0] [lindex $event 1] {}
      " 
    }
  }
#take $w out of all the analysis,linked_windows lists
  foreach a [array names info "*,linked_windows"] {
    if {[set i [lsearch -exact $info($a) $w]] != -1 } {
      set info($a) [lreplace $info($a) $i $i]
    }
  }
  set info($w,linked_events) [list]
}

########
## Make sure the window is on the screen
########
proc window_on_screen {w} {
  update
  if {![winfo exists $w]} {return}
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winw winh winx winy
  if {[expr $winx + $winw] > [winfo screenwidth $w]} {set winx [expr {[winfo screenwidth $w]-$winw}]}
  if {[expr $winy + $winh] > [winfo screenheight $w]} {set winy [expr {[winfo screenheight $w]-$winh}]}
  if {$winx < 20} {set winx 20}
  if {($winy < 20)} {set winy 20}
  wm geometry $w +$winx+$winy
}

########
## Close all windows for exit
########
proc close_all_windows {} {
  ##close analysis windows first
  foreach window [winfo children .] {
    if {[string first "analysis" $window] > -1} {
      closewindow $window
    }
  }

  ##then close the dna windows, asking for saves
  set really 1
  foreach window [winfo children .] {
    if {[closewindow $window] == 0} {
      set really 0
      break
    }
  }

  if {([tk windowingsystem] == "aqua") && ($really == 1)} {
    final_exit
  }
}

########
## set the selection in the textbox (Shift-arrow)
########
proc textkeyselect {w new} {
  if {[string equal [$w.textarea tag nextrange sel 1.0 end] ""]} {
    if {[$w.textarea compare $new < insert]} {
      $w.textarea tag add sel $new insert
    } else {
      $w.textarea tag add sel insert $new
    }
    $w.textarea mark set anchor insert
  } else {
    if {[$w.textarea compare $new < anchor]} {
      set first $new
      set last anchor
    } else {
      set first anchor
      set last $new
    }
    $w.textarea tag remove sel 1.0 $first
    $w.textarea tag add sel $first $last
    $w.textarea tag remove sel $last end
  }
  $w.textarea mark set insert $new
  $w.textarea see insert

  vertindexbalancer $w
  selection_manager $w
  update idletasks
}

########
## set the cursor in the textbox (arrow movements)
########
proc textsetcursor {w pos} {
    if {[$w.textarea compare $pos == end]} {
	set pos {end - 1 chars}
    }
    $w.textarea mark set insert $pos
    $w.textarea mark set anchor insert
    $w.textarea tag remove sel 1.0 end
    $w.textarea tag remove tempsel 1.0 end
    $w.textarea see insert

    vertindexbalancer $w
    selection_manager $w
}

########
## return the index of the closes gap to a mouse press
########
proc text_closest_gap {w x y} {
    if {$y < 1} {
      set y 2
    }
    if {$y >= [winfo height $w]} {
      set y [expr {[winfo height $w]-5}]
    }
    set pos [$w index @$x,$y]
    set bbox [$w bbox $pos]
    if {[string equal $bbox ""]} {
	return $pos
    }
    if {($x - [lindex $bbox 0]) < ([lindex $bbox 2]/2)} {
	return $pos
    }
    $w index "$pos + 1 char"
  
}

########
## configure the mouseindex and info box for textarea motion events
########
proc mouseover_info {w index} {
  if {$index > 0} {
    $w.mouseindex configure -text [expr {1+[lindex [regexp -inline 1.(.*) $index] 1]}]
    $w.infoframe.mouseindex configure -text [expr {1+[lindex [regexp -inline 1.(.*) $index] 1]}]
    set info [list]
    foreach tag [$w.textarea tag names $index] {
      if {[lindex [$w.textarea tag bind $tag <<Metadata>>] 2] == "enzyme"} {
        #set name [lindex [$w.textarea tag bind $tag <<Metadata>>] 0]
        #lappend info "$name\([enz_count $w $name all]):[lindex [$w.textarea tag bind $tag <<Metadata>>] 4]"
        lappend info "[lindex [$w.textarea tag bind $tag <<Metadata>>] 0]:[lindex [$w.textarea tag bind $tag <<Metadata>>] 4]"
      } else {
        if {[set name [lindex [$w.textarea tag bind $tag <<Metadata>>] 0]] != ""} {
          lappend info $name
        }
      }
    }
    $w.infobox configure -text [join $info ","]
  } else {
    $w.mouseindex configure -text ""
    $w.infoframe.mouseindex configure -text ""
    $w.infobox configure -text ""
  }
}

#######
## delete text from index to del_index, then add text with tags (allows undo)
#######
proc register_undo {w type {args [list]} {label ""} {direction Undo}} {
  global info

  if {$type != "cant"} {
    set info($w,undo) [list $type $args $direction]
    $w.menubar.edit entryconfigure 0 -label "[mc $direction] [mc $label]" -state normal
  } else {
    set info($w,undo) [list]
    $w.menubar.edit entryconfigure 0 -label "[mc {Can't Undo}] [mc $label]" -state disabled
  }
}

#######
## uses info($w,undo) to undo the last edit
#######
proc undo {w} {
  global info

  if {$info($w,undo) != {}} {
  set type [lindex $info($w,undo) 0]
  set args [lindex $info($w,undo) 1]
  set direction [lindex $info($w,undo) 2]
  switch $type {
    rev_com {
      #format: first, last, cleanup
      set first [lindex $args 0]
      set last [lindex $args 1]
      set tag_cleanup [lindex $args 2]

      if {($first != 1.0) || ($last != [$w.textarea index 1.end])} {
        select_region $w $first $last
      }
      rev_com_window $w [expr {($direction == "Undo")?"Redo":"Undo"}]
      $w.textarea see insert
      #tag_cleanup_undo $w $tag_cleanup
    }
    change_case {
      #format: first, last, action
      set first [lindex $args 0]
      set last [lindex $args 1]
      set action [lindex $args 2]

      if {($first != 1.0) || ($last != [$w.textarea index 1.end])} {
        select_region $w $first $last
      }
      if {$action == "upper"} {
        set action lower
      } elseif {$action == "lower"} {
        set action upper
      }
      if {($direction == "Redo") || ($action == "exchange")} {
        change_case $w $action [expr {($direction == "Undo")?"Redo":"Undo"}]
      } else {
        change_case $w $action Redo [lindex $args 3]
      }
    }
    typing -
    delete -
    cut -
    paste {
      set first [lindex $args 0]
      set last [lindex $args 1]
      set label [lindex $args 2]
      set undo_text [lindex $args 3]
      set undo_tags [lindex $args 4]
      set tag_cleanup [lindex $args 5]
      set text_was_selected [lindex $args 6]
      set insert_at_end [lindex $args 7]
      text_ins_del $w $type $first $last $label $undo_text $undo_tags $tag_cleanup $text_was_selected $insert_at_end [expr {($direction == "Undo")?"Redo":"Undo"}]
    }
    set_origin {
      set index [lindex $args 0]
      select_region $w $index $index
      set_origin $w [expr {($direction == "Undo")?"Redo":"Undo"}]
    }
    toggle_circ {
      set tag_cleanup [lindex $args 0]
      toggle_circ $w [expr {($direction == "Undo")?"Redo":"Undo"}]
      #tag_cleanup_undo $w $tag_cleanup
    }

    default {

    }
  }
  }
}

#######
## delete text from forst to last then insert undo_text with undo_tags at that location
#######
proc text_ins_del {w type first last label undo_text undo_tags tag_cleanup text_was_selected insert_at_end direction} {

  set new_undo_text [$w.textarea get $first $last]
  set new_undo_tags [get_tags $w.textarea [list $first $last]]
  $w.textarea delete $first $last
  $w.textarea insert $first $undo_text
  set ins_last [$w.textarea index $first+[string length $undo_text]chars]
  put_tags $w.textarea [list $first $ins_last] $undo_tags forward 0
  if {$text_was_selected} {
    if {$insert_at_end} {
      select_region $w $first $ins_last
    } else {
      select_region $w $ins_last $first
    }
  } else {
    if {$insert_at_end} {
      $w.textarea mark set anchor $ins_last
      $w.textarea mark set insert $ins_last
    } else {
      $w.textarea mark set insert $first
      $w.textarea mark set anchor $first
    }
  }
  #tag_cleanup_undo $w $tag_cleanup
  register_undo $w $type [list $first $ins_last $label $new_undo_text $new_undo_tags [feature_cleanup $w.textarea] $text_was_selected $insert_at_end] $label $direction
}

#######
## sends the scrollbar commands to both the index box and the text box
#######
proc vertscrollmanager {w args} {
  global info
  if {[$w.textarea index insert] == 2.0} {
    $w.textarea mark set insert 1.end
  }
  if {[lindex $args 0] == "moveto"} {
    set numlines [expr {([lindex [split [$w.textarea index 1.end] .] 1]-1)/$info($w,textwidth)+1.0}]
    set toline [expr {int([lindex $args 1]*$numlines)}]
    $w.textarea yview "1.[expr {$toline*$info($w,textwidth)+1}]"
    $w.vertindex yview "[expr {$toline+1}].0" 
  } else {
    eval [concat $w.textarea yview $args]
    eval [concat $w.vertindex yview $args]
  }
  vertindexbalancer $w
}

#######
## fix the aqua scrolling bug (again)
#######
if {[tk windowingsystem] == "aqua"} {
proc vertscrollmanager {w args} {
  global info
  if {[$w.textarea index insert] == 2.0} {
    $w.textarea mark set insert 1.end
  }
  if {[lindex $args 0] == "moveto"} {
    #set moved_fraction [expr {1.0*([winfo pointery .]-$info(mousedown))/([winfo height $w.vertscroll]-40)}] 
    #set moveto [expr {$moved_fraction+$info(scrollmdown)}]
    set moveto [expr {$info(scrollmdown)+(1.0*[winfo pointery .]-$info(mousedown))/([winfo height $w.vertscroll]-40)}]

    set numlines [expr {([lindex [split [$w.textarea index 1.end] .] 1]-1)/$info($w,textwidth)+1.0}]
    set toline [expr {int($moveto*$numlines)}]

    $w.textarea yview "1.[expr {$toline*$info($w,textwidth)+1}]"
    $w.vertindex yview "[expr {$toline+1}].0" 
  } else {
    eval [concat $w.textarea yview $args]
    eval [concat $w.vertindex yview $args]
  }
  vertindexbalancer $w
}
}


#######
##  makes the index numbers along the left edge and scrollbar match the text box
#######
proc vertindexbalancer {w } {
  global info
  set textend [expr round (([lindex [split [$w.textarea index end-1chars] .] 1]-1) / $info($w,textwidth) )+1]
  set vertend [lindex [split [$w.vertindex index end-1chars] .] 0]
  #$w.vertindex configure -state normal
  if {$vertend < $textend} {
    while {$vertend < $textend} {
      $w.vertindex insert end "\n [expr $vertend*$info($w,textwidth)+1]" allvertindex
      incr vertend
    }
  }
  if {($vertend > $textend) && ($vertend > 1)} {
    $w.vertindex delete [expr ($textend > 1 ? $textend : 1)+1].0 end
  }
  #$w.vertindex tag configure allvertindex -justify right
  #$w.vertindex configure -state disabled

  $w.vertindex yview moveto [lindex [$w.textarea yview] 0]
#make the scrollbar match the text view
  set numlines [expr {([lindex [split [$w.textarea index 1.end] .] 1]-1)/$info($w,textwidth)+1}]
  set firstline [expr {([lindex [split [$w.textarea index @0,0] .] 1])/$info($w,textwidth)}]
  #set lastline [expr {([lindex [split [$w.textarea index "@0,[winfo height $w.textarea]"] .] 1])/$info($w,textwidth)+1}]
  set lastline $numlines

  #set mod [expr {[winfo height $w.textarea] % [font metrics dnafont -linespace]}]
  set mod [expr {[winfo height $w.textarea] % [font metrics dnafont -linespace]}]

  set wheight [expr {([winfo height $w.textarea]/ [font metrics dnafont -linespace])}]

  if {($lastline > $wheight)} {
    set lastline [expr {$wheight+$firstline}]
  }
  #if {$mod < 4} {incr numlines}

  if {($numlines == 0.0) || (($firstline == 0) && ($lastline == $numlines))} {
    if {[grid info $w.vertscroll] != {}} {
      grid remove $w.vertscroll
    }
  } else {
    if {[grid info $w.vertscroll] == ""} {
      grid configure $w.vertscroll
      update idletasks
    } 
    eval [concat $w.vertscroll set [expr {1.0*$firstline/$numlines}] [expr {1.0*$lastline/$numlines}]]
   #$w.infobox configure -text "$mod $lastline $numlines [expr {1.0*$firstline/$numlines}] [expr {1.0*$lastline/$numlines}]"
  }
}

##########
## change the vertical balance if width changes
##########
proc textarea_width_changed {w} {
  global info

  update
  if {![winfo exists $w]} {return}
  if {[winfo width $w.textarea] == 0} {
    after 200
  }
  
  if {[tk windowingsystem] == "aqua"} {
    set width [expr {int(([winfo width $w.textarea]-6-2.0*[$w.textarea cget -borderwidth])/[lindex [wm grid $w] 2])}]
  } else {
    set width [expr {int(([winfo width $w.textarea]-2.0*[$w.textarea cget -borderwidth])/[lindex [wm grid $w] 2])}]
  }
  if {$width != $info($w,textwidth)} {
    set info($w,textwidth) $width


    ####redo the top and bottom numbers
    set iline ""
    set count 10
    while {[string length $iline] < $width} {
      set iline "$iline    *[string repeat " " [expr {5-[string length $count]}]]$count"
      incr count 10
    }

    $w.horzindex delete 1.0 end
    $w.horzindex insert 1.0 $iline

    $w.horzindex2 delete 1.0 end
    $w.horzindex2 insert 1.0 $iline


    ####redo the left side numbers

    $w.vertindex delete 2.0 end 
    set vertend 1
    set textend [expr {([lindex [split [$w.textarea index 1.end] .] 1]-1)/$info($w,textwidth)+1.0}]
    while {$vertend < $textend} {
      $w.vertindex insert end "\n [expr $vertend*$info($w,textwidth)+1]" allvertindex
      incr vertend
    }

    $w.vertindex yview moveto [lindex [$w.textarea yview] 0] 
  }
  $w.textarea see insert
update
  vertindexbalancer $w
}

##########
## sets the selection location info for the textbox (start and end of selection)
##########
proc selection_manager {w} {
  global info
  $w.textarea tag remove tempsel 1.0 end

  if {[$w.textarea index insert] == 2.0} {
    $w.textarea mark set insert 1.end
  }
  if {[$w.textarea tag names 2.0] == "sel"} {
    $w.textarea tag remove sel 2.0 
  }
  $w.infoframe.totallen configure -text [lindex [split [$w.textarea index end-1chars] .] 1]
  if {[$w.textarea tag ranges sel] == ""} {
    $w.infoframe.startlabel configure -text [mc "insert@"]
    set first [expr [lindex [split [$w.textarea index insert] .] 1]+1]
    $w.infoframe.start configure -text "$first\<[expr {($first-1) % 3}]>"
    $w.infoframe.sellenlabel configure -text ""
    $w.infoframe.sellen configure -text ""
    $w.infoframe.endlabel configure -text ""
    $w.infoframe.end configure -text ""
    $w.infoframe.orf configure -text ""
    $w.infoframe.orflabel configure -text ""
    $w.infoframe.tm configure -text ""
    $w.infoframe.tmlabel configure -text ""
    $w.infoframe.gc configure -text ""
    $w.infoframe.gclabel configure -text ""
    $w.infoframe.trans configure -text ""
  } else {
    if {[$w.textarea index sel.last] == 2.0} {
      $w.textarea tag remove sel 1.end 2.0
    }
    $w.textarea tag add tempsel sel.first sel.last
    $w.textarea tag lower tempsel sel
    $w.infoframe.startlabel configure -text [mc "Start"]
    set first [expr {1+[lindex [split [$w.textarea index sel.first] .] 1]}]
    set last [expr {[lindex [split [$w.textarea index sel.last] .] 1]}]
    $w.infoframe.start configure -text "$first\<[expr {($first-1) % 3}]>"
    $w.infoframe.sellenlabel configure -text [mc "Length"]
    $w.infoframe.endlabel configure -text [mc "End"]
    $w.infoframe.end configure -text "$last\<[expr {($last-1) % 3}]>"
    $w.infoframe.sellen configure -text "[expr {$last-$first+1}]\<[expr {($last-$first+1) % 3}]>"
    $w.infoframe.orflabel configure -text [mc "ORF"]
    set selection [$w.textarea get sel.first sel.last]
    if {[regexp -nocase "^(atg)" $selection] == 1} {
      set orftemp "M"
    } else {
      set orftemp ""
    }
    if {[regexp -nocase -indices "^(...)*?(taa|tag|tga)" $selection found] == 1} {
      if {[expr [string length $selection]- [lindex $found 1]] < 4 } {
        append orftemp " *"
      } else {
        append orftemp "---"
      }
    }
    $w.infoframe.orf configure -text $orftemp
    $w.infoframe.tmlabel configure -text "Tm"
    if {[string is double [set tm [Tm $selection $info(primer_dna) $info(primer_salt)]]]} {
      $w.infoframe.tm configure -text "[format %2.0f $tm]C"
    } else {
      $w.infoframe.tm configure -text "$tm"
    }
    $w.infoframe.gclabel configure -text "%GC"
    $w.infoframe.gc configure -text "[format %2.0f [GC% $selection]]%"

    if {[string length $selection] > 144} {
      $w.infoframe.trans configure -text "[translate [string range $selection 0 59]]...[translate [string range $selection end-[expr {([string length $selection] % 3) + 59}] end-[expr {[string length $selection] % 3}]]]"
    } else {
      $w.infoframe.trans configure -text [translate $selection]
    }
  }
  mouseover_info $w [$w.textarea index insert]
}

#########
## manages all key press events in the textbox
#########
proc keyevent_manager {w modifier k} { 
  global info
  set k [string map {asterisk *} $k]
  #$w.infobox configure -text "$modifier $k" ;#this let's you see the keypress keysyms
  switch $modifier {
    None {
      if {([string length $k] == 1)} {set k [string map $info(keysubstmap) $k]}
      if {([string length $k] == 1) && ([regexp  "\[$info(allowedkeys)\]" $k] == 1) && (![file_lock_check $w])}  {
        if {[llength [$w.textarea tag ranges sel]] > 0} {
          set undo_text [$w.textarea get sel.first sel.last]
          set undo_tags [get_tags $w.textarea [list [$w.textarea index sel.first] [$w.textarea index sel.last]]]
          set insert_at_end [expr {[$w.textarea index sel.last] == [$w.textarea index insert]}]
          $w.textarea delete sel.first sel.last
          set text_was_selected 1
        } else {
          set undo_text ""
          set undo_tags {}
          set insert_at_end 1
          set text_was_selected 0
        }
        if {[$w.textarea index insert] > 1.0} {
          $w.textarea insert insert $k {}
          $w.textarea mark set anchor insert
        } else {
          $w.textarea insert 1.0 $k {}
          $w.textarea mark set anchor 1.1
          $w.textarea mark set insert 1.1
        }
        $w.textarea see insert
        register_undo $w typing [list [$w.textarea index insert-1chars] [$w.textarea index insert] "Typing" $undo_text $undo_tags [feature_cleanup $w.textarea] $text_was_selected $insert_at_end] "Typing" Undo
        set info($w,scanned) 0
        set info($w,saved) 0
        unlink_analysis_window $w
      } else {

        switch $k {
          "BackSpace" {
            if {![file_lock_check $w]} {
              if {[llength [$w.textarea tag ranges sel]] > 0} {
                set undo_text [$w.textarea get sel.first sel.last]
                set undo_tags [get_tags $w.textarea [list [$w.textarea index sel.first] [$w.textarea index sel.last]]]
                set insert_at_end [expr {[$w.textarea index sel.last] == [$w.textarea index insert]}]
                $w.textarea delete sel.first sel.last
                register_undo $w delete [list [$w.textarea index insert] [$w.textarea index insert] "Delete" $undo_text $undo_tags [feature_cleanup $w.textarea] 1 $insert_at_end] "Delete" Undo
                set info($w,scanned) 0
                set info($w,saved) 0
                unlink_analysis_window $w
              } elseif {[$w.textarea index insert] != 1.0} {
                set undo_text [$w.textarea get [$w.textarea index insert-1chars] [$w.textarea index insert]]
                set undo_tags [get_tags $w.textarea [list [$w.textarea index insert-1chars] [$w.textarea index insert]]]
                $w.textarea delete insert-1chars
                $w.textarea mark set anchor insert
                $w.textarea see insert
                register_undo $w delete [list [$w.textarea index insert] [$w.textarea index insert] "Delete" $undo_text $undo_tags [feature_cleanup $w.textarea] 0 1] "Delete" Undo
                set info($w,scanned) 0
                set info($w,saved) 0
                unlink_analysis_window $w
              }
            }
          } 
          "Delete" {
            if {![file_lock_check $w]} {
              if {[llength [$w.textarea tag ranges sel]] > 0} {
                set undo_text [$w.textarea get sel.first sel.last]
                set undo_tags [get_tags $w.textarea [list [$w.textarea index sel.first] [$w.textarea index sel.last]]]
                set insert_at_end [expr {[$w.textarea index sel.last] == [$w.textarea index insert]}]
                $w.textarea delete sel.first sel.last
                register_undo $w delete [list [$w.textarea index insert] [$w.textarea index insert] "Delete" $undo_text $undo_tags [feature_cleanup $w.textarea] 1 $insert_at_end] "Delete" Undo
                set info($w,scanned) 0
                set info($w,saved) 0
                unlink_analysis_window $w
              } elseif {[$w.textarea index insert] != [$w.textarea index "1.0 lineend"]} {
                set undo_text [$w.textarea get [$w.textarea index insert] [$w.textarea index insert+1chars]]
                set undo_tags [get_tags $w.textarea [list [$w.textarea index insert] [$w.textarea index insert+1chars]]]
                $w.textarea delete insert
                $w.textarea mark set anchor insert+1chars
                $w.textarea see insert
                register_undo $w delete [list [$w.textarea index insert] [$w.textarea index insert] "Delete" $undo_text $undo_tags [feature_cleanup $w.textarea] 0 0] "Delete" Undo
                set info($w,scanned) 0
                set info($w,saved) 0
                unlink_analysis_window $w
              }
            }
          }
          "space" {
          } 
          default {
            #sputs "unknown key:$k"
            return
	    }
        } 
      } 
    } 
    Control {
      switch -glob -- $k {
        Control_L {return}
        Control_R {return}
        [Zz] {undo $w}
        [xX] {clip_cut_copy $w cut}
        [cC] {clip_cut_copy $w copy}
        [vV] {clip_paste $w}
        [aA] {select_all $w}
        [hH] {feature_scan $w}
        [jJ] {jump_to $w}
        [nN] {create_window; return}
        [oO] {open_file_dialog; return}
        [dD] {duplicate_window $w selection; return}
        [sS] {save_file $w "Save"; return}
        [wW] {closewindow $w; return}
        [iI] {if {$info($w,circular)== "circular"} {set_origin $w} else {bell}}
        [uU] {#toggle_circ $w}
        minus {change_case $w lower}
        plus {change_case $w upper}
        greater {next_ORF $w}
        less {next_ORF $w reverse}
        [fF] {find_dialog $w}
        [gG] {if {$info(find_find_again) != {}} {eval do_find $w $info(find_find_again)} else {find_dialog $w}}
        [rR] {enz_digest $w $info(enz_currently_selected) $info(analyze_selection)}
        [eE] {enzyme_selection_dialog $w}
        [mM] {enz_graphic_map  $w $info(enz_currently_selected) $info(analyze_selection)}
        [tT] {translation_window $w}
        [lL] {align_dialog $w}
        [bB] {NCBI_Blast $w}
        [qQ] {close_all_windows; return}
        [0]  {if {[info commands console] != ""} {
               console show
               uplevel #0 {set w [lsearch -regexp -inline [winfo children .] {.dna_window[0-9]+}]
                 sputs "w= $w"
               }
              }
             }
        slash {rev_com_window $w}
        equal {change_case $w exchange}
	}
    }
    ShiftControl {
      switch -glob -- $k {
        Control_L {return}
        Shift_L {return}
        Control_R {return}
        Shift_R {return}
        [xX] {clip_cut_copy $w cut rev-com}
        [cC] {clip_cut_copy $w copy rev-com}
        [Vv] {clip_paste $w rev-com}
        minus {change_case $w lower}
        plus {change_case $w upper}
        greater {next_ORF $w}
        less {next_ORF $w reverse}
        [aA] {select_from_to $w}
        [dD] {duplicate_window $w all; return}
        [hH] {feature_clear $w}
        [rR] {edit_ladders $w}
        [tT] {enz_text_map $w}
        [mM] {enz_graphic_map  $w $info(enz_currently_selected) $info(analyze_selection) {==1}}
        [eE] {set info(analyze_selection) [expr {($info(analyze_selection)=="all")?"selection":"all"}]; event generate . <<toggle_select>>}
	}

    }
  }  
  vertindexbalancer $w
  selection_manager $w
  #$w.infobox configure -text "$modifier $k" ;#this let's you see the keypress keysyms
}


###########
## limited control keys for the comment window
###########
proc comment_keyevent_manager {w modifier k} {
  global info
  switch $modifier {
    Control {
      switch -glob $k {
        Control_L {return}
        Control_R {return}
        [cC] {event generate $w.comment <<Copy>>}
        [xX] {event generate $w.comment <<Cut>>}
        [vV] {event generate $w.comment <<Paste>>}
        [aA] {select_all $w.comment}
        [nN] {create_window; return}
        [oO] {open_file_dialog; return}
        [dD] {duplicate_window $w selection; return}
        [sS] {save_file $w "Save"; return}
        [wW] {closewindow $w; return}
        [fF] {find_dialog $w}
        [gG] {if {$info(find_find_again) != {}} {eval do_find $w $info(find_find_again)} else {find_dialog $w}}
        [qQ] {close_all_windows; return}
        [0]  {if {[info commands console] != ""} {
               console show
               uplevel #0 {set w [lsearch -regexp -inline [winfo children .] {.dna_window[0-9]+}]
                 sputs "w= $w"
               }
              }
             }
	}
    }
    ShiftControl {
      switch -glob $k {
        Control_L {return}
        Shift_L {return}
        Control_R {return}
        Shift_R {return}
        [dD] {duplicate_window $w all; return}
	}

    }
  }  
}

###########
## return the locked state of a file (when attempting to change contents)
###########
proc file_lock_check {w} {
  global info toolbar_images

  set info($w,locked) [expr {[file isfile $info($w,filename)] && ![file writable $info($w,filename)]}]
  $w.infoframe.locked configure -image $toolbar_images([expr {$info($w,locked)?"locked":"unlocked"}])
  return $info($w,locked)
}

###########
## return the locked state of a file
###########
proc file_islocked {w} {
  global info

  set info($w,locked) [expr {[file isfile $info($w,filename)] && ![file writable $info($w,filename)]}]
  return $info($w,locked)
}

###########
## toggle the locked state of a file
###########
proc toggle_locked {w} {
  global info tcl_platform toolbar_images

  if {[file isfile $info($w,filename)]} {
    if {$tcl_platform(platform) == "windows"} {
      file attributes $info($w,filename) -readonly [expr {![file attributes $info($w,filename) -readonly]}]

    } elseif {$tcl_platform(platform) == "unix"} {
      if {[file owned $info($w,filename)]} {
        file attributes $info($w,filename) -permissions [expr {([file attributes $info($w,filename) -permissions] ^ 000200)}]
      }
    }
    set info($w,locked) [expr {![file writable $info($w,filename)]}]
  } else {
    if {$info($w,filename) == ""} {
      if {[tk_messageBox -title [mc "File is not saved."] -message "[mc {You can't lock a file until it is saved. Do you want to save this file now?}]" -type "yesno" -icon question -default no] == "no"} {
        set info($w,locked) 0
      } else {
        save_file $w "Save_as"
      }
    } else {
      if {[tk_messageBox -title [mc "File doesn't exist."] -message "[mc {The file %1$s no longer exists on the disk. Do you want to save this file now?} $info($w,filename)]" -type "yesno" -icon question -default no] == "no"} {
        set info($w,locked) 0
      } else {
        save_file $w "Save_as"
      }
    }
  }
  $w.infoframe.locked configure -image $toolbar_images([expr {$info($w,locked)?"locked":"unlocked"}])
}

###########
## save a file
###########
proc save_file {w mode {type ""}} {
global info toolbar_images
global tcl_platform

  set focus_win [focus]
  if {$focus_win == ""} {set focus_win .}

  while {(![file isdirectory $info(default_dnadir)]) && ([llength [file split $info(default_dnadir)]] > 1)} {
    set info(default_dnadir) [file dirname $info(default_dnadir)]
    if {[file dirname $info(default_dnadir)] == $info(default_dnadir)} {
      set info(default_dnadir) ""
    }
  }
  if {(![file isdirectory $info(default_dnadir)]) && ([llength [file split $info(default_dnadir)]] == 1)} {
    set info(default_dnadir) ""
  }

  if {$type == ""} {
    set type $info(default_file_format)
  }

  if {($mode != "Save_as") && ($info($w,filetype) != "None") && ($info($w,filetype) != $type) && $info(save_file_format_warning)} {
    set response [tk_messageBox -title [mc "Changing File Format"] -message "$info($w,filename) \n[mc {File is currently of filetype:%1$s. It will be saved as type: %2$s. Do you want to change the file name?} $info($w,filetype) $type]" -type "yesnocancel"  -icon warning -default no]
    if {$response == "yes"} { 
      set mode "Save_as"
    }
  }

  if {($mode == "Save_as") || ($info($w,filename) == "")} {
    set filename [tk_getSaveFile -title [mc "Save File As..."] -initialfile [file tail $info($w,filename)] -initialdir $info(default_dnadir) -parent $focus_win -defaultextension $info(default_file_extension) -filetypes { {"All files" {"*.*"}} {"ApE" {.str .ape .dna .seq .gb}}}]
    focus $focus_win
  } else {
    set filename $info($w,filename)
  }
  if {$filename == ""} {return}
  if {[file extension $filename] == ""} {set filename "$filename$info(default_file_extension)"}

  while {([file isfile $filename] && ![file writable $filename]) && ($filename != "") } {
    set response [tk_messageBox -title [mc "File is not writable."] -message "$filename \n[mc {File is locked. Do you want to unlock it (Yes), save it with a different name (No) or cancel save (Cancel?)}]" -type "yesnocancel"  -icon warning -default no]
    if {$response == "yes"} {
      set info($w,filename) $filename
      toggle_locked $w
      if {![file writable $filename]} {
        set filename [tk_getSaveFile -title [mc "Sorry, couldn't unlock. Save File As..."] -initialdir $info(default_dnadir) -parent $focus_win -defaultextension $info(default_file_extension) -filetypes { {"All files" {"*.*"}} {"strider" {.str .test}} {"genbank" {.gb}} {"fasta" {.fasta}}}]
        set info($w,filename) $filename
      }
    } elseif {$response == "no"} {
      set filename [tk_getSaveFile -title [mc "Save File As..."] -initialdir $info(default_dnadir) -parent $focus_win -defaultextension $info(default_file_extension) -filetypes { {"All files" {"*.*"}} {"strider" {.str .test}} {"genbank" {.gb}} {"fasta" {.fasta}}}]
    } else {
      set filename ""
    }
  }
  if {$filename == ""} {return}

  if {$type == "DNA Strider"} {
    ######write strider format
    if {$info($w,circular) == "circular"} {set topology 1} else {set topology 0}
    regexp "1.(.*)" [$w.textarea index end-1chars] fill dnalength
    set fileid [open $filename w]
    fconfigure $fileid -translation binary
    puts -nonewline $fileid [binary format "cccx14cx10IIx60Ix12" 0 1 $topology $info($w,Dam_Dcm_methylated) $dnalength [expr {$info($w,startindex)-1}] [string length [$w.comment get 1.0 end-1char]]]
    set text [$w.textarea get 1.0 end-1chars]
    regsub -all -nocase "n" $text "*" text
    puts -nonewline $fileid $text
    puts -nonewline $fileid [$w.comment get 1.0 end-1char]
    close $fileid
    set info($w,filetype) "DNA Strider"

  } elseif {$type == "Genbank"} {
    #####write genbank format
    set fileid [open $filename w]
    set text [$w.textarea get 1.0 end-1chars]
    set locus_line [lindex $info($w,genbank_header_list) 1]
    if {![regexp {(.{16})[ ]*([0-9]*) bp (   |ss-|ds-|ms-)(NA    |DNA   |RNA   |tRNA  |rRNA  |mRNA  |uRNA  |snRNA |snoRNA)[ ]*(linear  |circular|        )[ ]*([ A-Z]{3})[ ]*(..-...-....)} $locus_line x genbank_locus_name size stranded type circular genbank_division_code date] || $circular== "        "} {
      set genbank_locus_name ""
      set genbank_division_code ""
    }
    puts $fileid [format "LOCUS       %s %+11s bp ds-DNA     %-8s %-3s %-11s" $genbank_locus_name [string length $text] $info($w,circular) $genbank_division_code [string toupper [clock format [clock seconds] -format "%d-%h-%Y"]]]

    if {[lsearch $info($w,genbank_header_list) "COMMENT     "] == -1} {
      lappend info($w,genbank_header_list) "COMMENT     " ""
    }
    set comment_done 0
    foreach {type data} [lrange $info($w,genbank_header_list) 2 end] {
      if {$type == "COMMENT     "} {
        if {!$comment_done} {
          set data [$w.comment get 1.0 end-1char]
        } else {
          continue
        }
      }
      while {([string length $data] > 67) || ([string first \n $data] > -1)} {
        if {([set split [string first \n $data]] == -1) || ($split > 67)} {
          set split [string last " " $data 67]
        }
        if {$split < 0} {set split 67}
        set firstline [regsub -all {[ ]*$} [string range $data 0 [expr {$split-1}]] ""]
        set data [regsub {^[ ]*[\n]?} [string range $data $split end] ""]
        puts $fileid $type$firstline
        set type "            "
      } 
      puts $fileid $type$data

    }

    puts $fileid "COMMENT     ApEinfo:methylated:$info($w,Dam_Dcm_methylated)"

    ##ADD FEATURES HERE
    set temp_fea_list [list]
    foreach tag [$w.textarea tag names] {
      if {[regexp {f[0-9]+#} $tag]} {
        set x1 [lindex [$w.textarea tag ranges $tag] 0]
        regsub "1." $x1 "" x1
        if {$x1 != "" } {
          lappend temp_fea_list [list $x1 $tag]
        } else {
          sputs "feature_list- empty feature $x1 $tag"
        }
      }
    }
    #set temp_fea_list [lsort -index 0 -integer $temp_fea_list]
    if {[llength $temp_fea_list] > 0} {
      puts $fileid "FEATURES             Location/Qualifiers"
    }
    foreach fea $temp_fea_list {
      set tag [lindex $fea 1]
      set metadata [$w.textarea tag bind $tag <<Metadata>>]

      set loc [list]
      foreach {x1 x2} [$w.textarea tag ranges $tag] {
        regsub "1." $x1 "" x1
        incr x1
        regsub "1." $x2 "" x2
        lappend loc "$x1..$x2"
      }
      if {[llength $loc] > 1} {
        set loc "join([join $loc ,])"
      }
      if {[lindex $metadata 3]} {
        set loc "complement($loc)"
      }
      ##put feature definition line
      ##first the type and location indexes
      if {[string length $loc] <= 58} {
        puts $fileid [format "     %-16s%s" [lindex $metadata 2] $loc]
      } else {
        puts $fileid [format "     %-16s%s" [lindex $metadata 2] [string range $loc 0 [string last "," $loc 58]]]
        set loc [string range $loc [expr {[string last "," $loc 58]+1}] end]
        while {[string length $loc] > 58} {  
          set split [string last "," $loc 58]
          if {$split < 0} {set split 58}
          set firstline [string range $loc 0 $split]
          set loc [string range $loc [expr {$split+1}] end]
          puts $fileid "                     $firstline"
        }
        puts $fileid "                     $loc"
      }


      ##put sub-data lines
      set feature_comment [lindex $metadata 4]
        ##if feature comment isn't in a paired format, make it a note
      if {([llength $feature_comment] == 1) || (([llength $feature_comment] % 2) != 0)} {
        set feature_comment [list note $feature_comment]
      }
        ##add label and colors data
      lappend feature_comment ApEinfo_label [lindex $metadata 0] ApEinfo_fwdcolor [lindex [$w.textarea tag bind $tag <<Revcolors>>] 0] ApEinfo_revcolor [lindex [$w.textarea tag bind $tag <<Revcolors>>] 1]
      foreach {type data} $feature_comment {
        set data "/$type=$data"
        while {[string length $data] > 58} {
          if {([set split [string first \n $data]] == -1) || ($split > 58)} {
            set split [string last " " $data 58]
          }
          if {$split < 0} {set split 58}
          set firstline [regsub -all {[ ]*$} [string range $data 0 [expr {$split-1}]] ""]
          set data [regsub {^[ ]*[\n]?} [string range $data $split end] ""]
          puts $fileid "                     $firstline"
        }
        puts $fileid "                     $data"
      }
    }

    ##base count line
    if {[regexp {[^atcgATCG]} $text]} {
      puts $fileid [format "BASE COUNT%+9s a %+9s c %+9s g %+9s t %+9s other" [string length [regsub -all {[^aA]} $text ""]] [string length [regsub -all {[^cC]} $text ""]] [string length [regsub -all {[^gG]} $text ""]] [string length [regsub -all {[^tT]} $text ""]] [string length [regsub -all {[acgtACGT]} $text ""]]]
    } else {
      puts $fileid [format "BASE COUNT%+9s a %+9s c %+9s g %+9s t" [string length [regsub -all {[^aA]} $text ""]] [string length [regsub -all {[^cC]} $text ""]] [string length [regsub -all {[^gG]} $text ""]] [string length [regsub -all {[^tT]} $text ""]]]
    }
    puts $fileid "ORIGIN"

    ##sequence output
    set i 0
    while {$i < [string length $text]} {
      puts $fileid [format "%+9s %s %s %s %s %s %s" [expr {$i+1}] [string range $text $i [expr {$i+9}]] [string range $text [expr {$i+10}] [expr {$i+19}]] [string range $text [expr {$i+20}] [expr {$i+29}]] [string range $text [expr {$i+30}] [expr {$i+39}]] [string range $text [expr {$i+40}] [expr {$i+49}]] [string range $text [expr {$i+50}] [expr {$i+59}]]]
      incr i 60
    }
    puts $fileid "//"

    close $fileid
    set info($w,filetype) "Genbank"
  } else {
    tk_messageBox -title [mc "Filetype not savable"] -message "I don't know how to save files in format:$type"
  }


  set info($w,saved) 1 

  set info($w,filename) $filename
  set info($w,win_title) [file tail $filename]
  set info($w,locked) [expr {[file isfile $info($w,filename)] && ![file writable $info($w,filename)]}]
  $w.infoframe.locked configure -image $toolbar_images([expr {$info($w,locked)?"locked":"unlocked"}])
  if {$info(dnadir_follows_open)} {
    set info(default_dnadir) [file dirname $filename]
  }
  wm title $w $info($w,win_title)
  set info(dnatitles_list) [list]
  foreach element $info(dnawindows_list) {
    lappend info(dnatitles_list) $info($element,win_title)
  }
  update_windows_menu
}

###########
## cut and copy  function
###########
proc clip_cut_copy {w {function copy} {direction normal}} {
  global info
  global tcl_platform

  if {($function == "cut") && [file_lock_check $w]} {return}

  if {[llength [$w.textarea tag ranges sel]] > 0} {
    clipboard clear 
    if {$direction == "rev-com"} {
      clipboard append -displayof $w [revcom [$w.textarea get sel.first sel.last]]
      set info(clipboard_text) [revcom [$w.textarea get sel.first sel.last]]
      set info(clipboard_tags) [get_tags $w.textarea [list [$w.textarea index sel.first] [$w.textarea index sel.last]] reverse]
    } else {
      clipboard append -displayof $w [$w.textarea get sel.first sel.last]
      set info(clipboard_text) [$w.textarea get sel.first sel.last]
      set info(clipboard_tags) [get_tags $w.textarea [list [$w.textarea index sel.first] [$w.textarea index sel.last]]]
    }
    if {$function == "cut"} {
      set undo_text [$w.textarea get sel.first sel.last]
      set undo_tags [get_tags $w.textarea [list [$w.textarea index sel.first] [$w.textarea index sel.last]]]
      set insert_at_end [expr {[$w.textarea index sel.last] == [$w.textarea index insert]}]
      $w.textarea delete sel.first sel.last
      register_undo $w cut [list [$w.textarea index insert] [$w.textarea index insert] "Cut" $undo_text $undo_tags [feature_cleanup $w.textarea] 1 $insert_at_end] "Cut" Undo
      set info($w,scanned) 0
      set info($w,saved) 0
      unlink_analysis_window $w
      vertindexbalancer $w
      selection_manager $w
    }
    if {($tcl_platform(platform) == "unix") && ([tk windowingsystem] != "aqua")} {
      if {$function == "copy"} {
        $w.textarea tag add tempsel sel.first sel.last
      }
      selection own -selection CUT_BUFFER0 $w
      selection own -selection PRIMARY $w
      if {$function == "copy"} {
        $w.textarea tag add sel tempsel.first tempsel.last
      }
      #warning_dialog "cutbuffer0 owned by:[selection own -displayof $w -selection CUT_BUFFER0]; primary oned by [selection own -displayof $w -selection PRIMARY]"
    }
  }
}

###########
## paste  function
###########
proc clip_paste {w {direction normal}} {
  global info
  global tcl_platform

  if {[file_lock_check $w]} {return}

  if {($tcl_platform(platform) == "unix") && ([tk windowingsystem] != "aqua")} {
    #sputs "getting selection from [selection own -displayof . -selection PRIMARY]...."
     #warning_dialog "cutbuffer0 owned by:[selection own -displayof $w -selection CUT_BUFFER0]; primary oned by [selection own -displayof $w -selection PRIMARY]"
    if {[catch {set text [selection get -displayof $w -selection PRIMARY]}] != 0} {return}
  } else {
    if {[catch { set text [selection get -displayof $w -selection CLIPBOARD]}] != 0} {return}
  }
  set text [atgcfilter $text $info(allowedkeys) 1]
  if {$direction == "rev-com"} {set text [revcom $text]}
  if {$text == ""} return

  if {[llength [$w.textarea tag ranges sel]] > 0} {
    set undo_text [$w.textarea get sel.first sel.last]
    set undo_tags [get_tags $w.textarea [list [$w.textarea index sel.first] [$w.textarea index sel.last]]]
    set insert_at_end [expr {[$w.textarea index sel.last] == [$w.textarea index insert]}]
    set text_was_selected 1
    $w.textarea delete sel.first sel.last
  } else {
    set undo_text ""
    set undo_tags {}
    set insert_at_end 1
    set text_was_selected 0
  }
###some sort of bug (maybe only on windows) when inserting at insert mark, if insert is at 1.0
  if {[$w.textarea compare insert == 1.0]} {
    $w.textarea mark set insert 1.end
    $w.textarea mark set anchor 1.end
    $w.textarea insert 1.0 $text {}
    $w.textarea mark set insert 1.[string length $text] 
    $w.textarea mark set anchor 1.0
  } else {
    $w.textarea insert insert $text {}
  }

  if {($direction == "rev-com") && ($text == [revcom $info(clipboard_text)])} {
    put_tags $w.textarea [list [$w.textarea index insert-[string length $text]chars] [$w.textarea index insert]] $info(clipboard_tags) reverse
  } elseif {($direction != "rev-com") && ($text == $info(clipboard_text))} {
    put_tags $w.textarea [list [$w.textarea index insert-[string length $text]chars] [$w.textarea index insert]] $info(clipboard_tags)
  } 

  register_undo $w paste [list [$w.textarea index insert-[string length $text]chars] [$w.textarea index insert] "Paste" $undo_text $undo_tags [feature_cleanup $w.textarea] $text_was_selected $insert_at_end] "Paste" Undo
  set info($w,scanned) 0
  set info($w,saved) 0
  unlink_analysis_window $w
 
  select_region $w insert-[string length $text]chars insert
  vertindexbalancer $w
}

###########
## send out the cut buffer for unix pastes
###########
proc unix_cubuffer_handler {{offset 0} {maxchars 1000000}} {
  global info
  #sputs "$offset $maxchars"
  #sputs [string range $info(clipboard_text) $offset [expr {$offset+$maxchars-1}]]
  return [string range $info(clipboard_text) $offset [expr {$offset+$maxchars-1}]]
}

###########
## get tags from a selection for storage
###########
#(list of tag data pairs siutable for array get; data = list of indices, list of properties, list of bindings, list of actions bound)
proc get_tags {text range {direction forward}} {
  set first [lindex $range 0]
  set last [lindex $range 1]

  set firstchar [lindex [split $first .] 1]
  set lastchar [lindex [split $last .] 1]

  foreach tagname [$text tag names $first] {
    set tags($tagname) $first
  }


  foreach {action tagname index} [$text dump -tag $first+1char $last+1chars] {
    lappend tags($tagname) $index
  }

  foreach tagname [$text tag names $last] {
    if {[lindex $tags($tagname) 0] != $last} {
      lappend tags($tagname) $last
    } else {
      array unset tags $tagname
    }
  }

  set result [list]

  foreach tagname [$text tag names] {
    if {[lsearch -exact [array names tags] $tagname] == -1} {continue}
    regsub -all {[0-9]*?\.} $tags($tagname) "" tags($tagname)
    set indexlist [list]
    foreach index $tags($tagname) {
      if {$direction == "forward"} {
        lappend indexlist [expr {$index - $firstchar}]
      } else {
        set indexlist [linsert $indexlist 0 [expr {$lastchar - $index }]]
      }
    }
    set formatlist [list [$text tag cget $tagname -background] [$text tag cget $tagname -foreground] [$text tag cget $tagname -font]]
    set bindlist [$text tag bind $tagname]
    set actionlist [list]
    foreach binding $bindlist {
      lappend actionlist [$text tag bind $tagname $binding]
    }

    ###change the metadata and colors if the tags are getting reversed
    if {$direction == "reverse"} {
      #change the direction data in the Features metadata
      if {[set pos [lsearch $bindlist <<Metadata>>]] != -1} {
        set metadata [lindex $actionlist $pos]
        #set dir [lindex $metadata 3]
        set metadata [lreplace $metadata 3 3 [expr {![lindex $metadata 3]}]]
        set actionlist [lreplace $actionlist $pos $pos $metadata]
      }
      #change the colors if a reverse color is provided
      if {[set pos [lsearch $bindlist <<Revcolors>>]] != -1} {
        set colorlist [lindex $actionlist $pos]
        if {[lindex $formatlist 0] != [lindex $colorlist 0]} {
          set formatlist [lreplace $formatlist 0 0 [lindex $colorlist 0]]
        } else {
          set formatlist [lreplace $formatlist 0 0 [lindex $colorlist 1]]
        }
      }
    }
    lappend result $tagname [list $indexlist $formatlist $bindlist $actionlist]
    #set tags($tagname) [list $indexlist $formatlist $bindlist $actionlist]
    #sputs "$tagname: $tags($tagname)"
  }
  array unset tags sel
  array unset tags tempsel
  #return [array get tags]
  return $result
}

###########
## put tags back onto a selection (see get_tags)
###########
proc put_tags {text range taglist {direction forward} {rename 1}} {
  global fea_info

  set first [lindex $range 0]
  set last [lindex $range 1]


  set name_store [list]
  foreach {tagname tagdata} $taglist {
    set formatlist [lindex $tagdata 1]
    set bindlist [lindex $tagdata 2]
    set actionlist [lindex $tagdata 3]

    if {([regexp {f[0-9]+#} $tagname]) && ($rename == 1)} {
      set newtagname "f$fea_info([winfo parent $text],count)#"
      incr fea_info([winfo parent $text],count)
      lappend name_store $tagname $newtagname
    } else {
      set newtagname $tagname
    }


    ###change the metadata and colors if the tags are getting reversed
    if {$direction == "reverse"} {
      #change the direction data in the Features metadata
      if {[set pos [lsearch $bindlist <<Metadata>>]] != -1} {
        set metadata [lindex $actionlist $pos]
        set metadata [lreplace $metadata 3 3 [expr {![lindex $metadata 3]}]]
        set actionlist [lreplace $actionlist $pos $pos $metadata]
      }
      #change the colors if a reverse color is provided
      if {[set pos [lsearch $bindlist <<Revcolors>>]] != -1} {
        set colorlist [lindex $actionlist $pos]
        if {[lindex $formatlist 0] != ""} {
          if {[lindex $formatlist 0] != [lindex $colorlist 0]} {
            set formatlist [lreplace $formatlist 0 0 [lindex $colorlist 0]]
          } else {
            set formatlist [lreplace $formatlist 0 0 [lindex $colorlist 1]]
          }
        }
      }
      if {$newtagname == "foundf"} {set newtagname "foundr"} elseif {$newtagname == "foundr"} {set newtagname "foundf"}
    }

    if {($rename == 0) || (![regexp {fn[0-9]+#} $tagname])} {
      $text tag configure $newtagname -background [lindex $formatlist 0] -foreground [lindex $formatlist 1] -font [lindex $formatlist 2]
      foreach {start end} [lindex $tagdata 0] {
        if {$direction == "forward"} {
          $text tag add $newtagname "$first+$start chars" "$first+$end chars"
        } else {
          $text tag add $newtagname "$last-$end chars" "$last-$start chars"
        }
      }
      foreach binding $bindlist action $actionlist {
        $text tag bind $newtagname $binding $action
      }
    }
  }

 
  ##make new fn# tags for split features that just got added in rename mode
  foreach {old_tag new_tag} $name_store {
    if {[string match {f[0-9]*#} $old_tag]} {
      set old_fn_tag "fn[regexp -inline {[0-9]+} $old_tag]#"
      if {([lsearch -exact $taglist $old_fn_tag] > -1) && ([llength [lindex $taglist [expr {[lsearch -exact $taglist $old_tag]+1}] 0]] > 2)} {
        set new_fn_tag "fn[regexp -inline {[0-9]+} $new_tag]#"
        lappend name_store $old_fn_tag $new_fn_tag
        foreach {start end} [lindex $taglist [expr {[lsearch -exact $taglist $old_fn_tag]+1}] 0] {
          if {$direction == "forward"} {
            $text tag add $new_fn_tag "$first+$start chars" "$first+$end chars"
          } else {
            $text tag add $new_fn_tag "$last-$end chars" "$last-$start chars"
          }
        }
      }
    }
  }
  return $name_store
}

#############
## copy tags into a destination analysis window
#############
proc copy_tags {text range dest dest_index {direction forward} {tagappend ""}} {
  set first [lindex $range 0]
  set last [lindex $range 1]

  if {[$text compare $first > $last]} {
    set temp $first
    set first $last
    set last $temp
    unset temp
  }
  set firstchar [lindex [split $first .] 1]
  set lastchar [lindex [split $last .] 1]

  foreach tagname [$text tag names $first] {
    set tags($tagname) $first
  }
  foreach {type tag index} [$text dump -tag $first+1char $last+1chars] {
    lappend tags($tag) $index
  }
  foreach tagname [$text tag names $last] {
    lappend tags($tagname) [$text index $last+1char]
  }
  set dest_line [$dest get $dest_index "$dest_index lineend"]

  array unset tags sel
  array unset tags tempsel
  foreach tag [array names tags] {
    if {($tag == "sel") ||($tag == "tempsel") || ([regexp {fn[0-9]+#} $tag])} continue

    set tagname "$tag$tagappend"
    regsub -all {[0-9]*?\.} $tags($tag) "" tags($tag)

    foreach {index1 index2} $tags($tag) {
      if {$direction == "forward"} {
        set tag_start [expr {$index1 - $firstchar}]
        set tag_end [expr {$index2 - $firstchar}]
      } else {
        set tag_start [expr {$lastchar - $index2+1}]
        set tag_end [expr {$lastchar - $index1+1}]
      }
      if {[regexp -indices "(?:\[\\\ -\]*\[^-\\\ \]){$tag_start}((?:\[\\\ -\]*\[^-\\\ \]){[expr {$tag_end-$tag_start}]})" $dest_line a b]} {
        $dest tag add $tagname [$dest index $dest_index+[lindex $b 0]char] [$dest index $dest_index+[expr {1+[lindex $b 1]}]char]
      }
      #$dest tag add $tag [$dest index $dest_index+[string length [regexp -inline "(?:\[\\\ -\]*\[^-\\\ \]){$tag_start}" $dest_line]]char] [$dest index $dest_index+[string length [regexp -inline "(?:\[\\\ -\]*\[^-\\\ \]){$tag_end}" $dest_line]]char]
    }
    $dest tag configure $tagname -background [$text tag cget $tag -background]
    $dest tag bind $tagname <<Metadata>> [$text tag bind $tag <<Metadata>>]
  }
}

###########
## select all function
###########
proc select_all {w} {
  $w.textarea tag add sel 1.0 end-1chars
  $w.textarea mark set insert 1.0
  $w.textarea mark set anchor insert
  selection_manager $w
}

###########
## select from_to window
###########
proc select_from_to {w} {
  global ok
  global info dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  if {[winfo exists .select_to]} {destroy .select_to}
  set s [toplevel .select_to -background $info(bg_color)]
  if {[tk windowingsystem] != "aqua"} {
    wm transient $s $w
  } else {
    ::tk::unsupported::MacWindowStyle style $s document {closeBox}
  }
  wm title $s [mc "Select from to"]
  wm protocol $s WM_DELETE_WINDOW "set ok -1"
  raise $s
  bind . <<RaiseDialogs>> "wm deiconify $s; raise $s"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry $s "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"

  set wgeomlist [split [wm geometry $w] "x+"]
  set sgeom "+[expr [lindex $wgeomlist 2]+3]+[expr [lindex $wgeomlist 3]+150]"
  wm geometry $s $sgeom
  set from 0
  set to 0
  set ok 0
  frame $s.fromtoframe  
  pack [label $s.fromtoframe.fromlabel -text [mc "Select from"]] -expand false -side left

  pack [entry $s.fromtoframe.from -width 10 -validate key -vcmd "check_char %S" -selectforeground $info(text_select_fg_color) -selectbackground $info(text_select_bg_color) -foreground $info(text_fg_color) -background $info(text_bg_color)] -expand true -fill x -side left
  pack [label $s.fromtoframe.tolabel -text [mc "to"]] -expand false -side left
  pack [entry $s.fromtoframe.to -width 10 -validate key -vcmd "check_char %S" -selectforeground $info(text_select_fg_color) -selectbackground $info(text_select_bg_color) -foreground $info(text_fg_color) -background $info(text_bg_color)] -expand true -fill x -side left
  button $s.ok -text [mc "OK"] -command "set ok 1" -default active 
  button $s.cancel -text [mc "Cancel"] -command "set ok -1"
  grid rowconfigure $s 0 -weight 0
  grid rowconfigure $s 1 -weight 0
  grid columnconfigure $s 0 -weight 1
  grid columnconfigure $s 1 -weight 1 
  grid configure $s.fromtoframe -row 0 -column 0 -columnspan 2 -sticky we
  grid configure $s.ok -row 1 -column 0
  grid configure $s.cancel -row 1 -column 1 
  
  if {[llength [$w.textarea tag ranges tempsel]] > 0} {
    $s.fromtoframe.from insert 0 [lindex [split [$w.textarea index sel.first] .] 1]
    $s.fromtoframe.to insert 0 [lindex [split [$w.textarea index sel.last] .] 1]
  } else {
    $s.fromtoframe.from insert 0 [lindex [split [$w.textarea index insert] .] 1]
  }

  $s.fromtoframe.from selection range 0 end

  update idletasks
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $s] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry $s "+$winx+$winy"

  #grab set $s
  bind $s <Key-Return> "set ok 1"
  focus $s
  focus $s.fromtoframe.from

  vwait ok
  if {$ok == 1} {
    set from [string trimleft [$s.fromtoframe.from get] 0]
    set to [string trimleft [$s.fromtoframe.to get] 0]

    if {($from > $to) && ($from != "") && ($to != "") } {set temp $from; set from $to; set to $temp}
    if {$from == ""} {set from 1.0} else {set from "1.$from"}
    if {$to == ""} {set to "end-1char"} else {set to "1.$to"}
    select_region $w "$from-1chars" $to
  }
  #grab release $s
  destroy $s
  set dialogblock 0
  bind . <<RaiseDialogs>> ""
  focus $w
  focus $w.textarea
}

###########
## jump to window
###########
proc jump_to {w} {
  global ok
  global info dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  if {[winfo exists .dialog]} {destroy .dialog}
  set s [toplevel .dialog -background $info(bg_color)]
  if {[tk windowingsystem] != "aqua"} {
    wm transient $s $w
  } else {
    ::tk::unsupported::MacWindowStyle style $s document {closeBox}
  }
  wm title $s [mc "Jump to"]
  wm protocol $s WM_DELETE_WINDOW "set ok -1"
  raise $s
  bind . <<RaiseDialogs>> "wm deiconify $s; raise $s"


  set wgeomlist [split [wm geometry $w] "x+"]
  set sgeom "+[expr [lindex $wgeomlist 2]+3]+[expr [lindex $wgeomlist 3]+150]"
  wm geometry $s $sgeom

  grid [label $s.tolabel -text [mc "Jump to:"]] -row 0 -column 0 -sticky nwe
  grid [entry $s.to -width 10 -validate key -vcmd "check_char %S" -selectforeground $info(text_select_fg_color) -selectbackground $info(text_select_bg_color) -foreground $info(text_fg_color) -background $info(text_bg_color)] -row 0 -column 1 -sticky nwe
  grid [button $s.ok -text [mc "OK"] -command "set ok 1" -default active] -row 1 -column 0 -sticky n
  grid [button $s.cancel -text [mc "Cancel"] -command "set ok -1"] -row 1 -column 1 -sticky n
  grid rowconfigure $s 0 -weight 0
  grid rowconfigure $s 1 -weight 0
  grid columnconfigure $s 0 -weight 1
  grid columnconfigure $s 1 -weight 1 
  $s.to insert 0 [lindex [split [$w.textarea index insert] .] 1]
  $s.to selection range 0 end

  update idletasks
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $s] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry $s "+$winx+$winy"

  #grab set $s
  bind $s <Key-Return> "$s.ok invoke"
  focus $s
  focus $s.to
  vwait ok
  set dialogblock 0
  if {$ok == 1} {
    $w.textarea tag remove sel 1.0 end
    set to [string trimleft [$s.to get] 0]
    if {$to == ""} {set to 0}
    if {$to > 0} {incr to -1}
    select_region $w 1.$to 1.$to
  }
  #grab release $s
  destroy $s
  bind . <<RaiseDialogs>> ""
  focus $w
  focus $w.textarea
  unset ok
}

###########
## checks for numeric character in validation of numeric entries
###########
proc check_char {c} {
  if {[string equal $c " "]} {return 0}
  if {![string is integer $c]} {return 0}
  if {[regexp {[^0-9]} $c]} {return 0}
  return  1
}

proc entry_integer {c w mode min max} {
  if {$mode != "key"} {
    if {(![string is integer -strict $c]) || ($c< $min)} {
      $w delete 0 end
      $w insert end $min
      after idle "if {\[winfo exists $w\]} {$w config -validate all}"
    } elseif {$c > $max} {
      $w delete 0 end
      $w insert end $max
      after idle "if {\[winfo exists $w\]} {$w config -validate all}"
    }
  return 1
  } else {
    if {[string equal $c " "]} {return 0}
    if {[regexp {[^0-9]} $c]} {return 0}
    return  1
  }
}

###########
## find dialog
###########
proc find_dialog {w} {
  global info
  global ok dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1


  if {[winfo exists .dialog]} {destroy .dialog}
  set s [toplevel .find_dialog -background $info(bg_color)]
  if {[tk windowingsystem] != "aqua"} {
    wm transient $s $w
  } else {
    ::tk::unsupported::MacWindowStyle style $s document {closeBox}
  }
  wm title $s [mc "Find..."]
  wm protocol $s WM_DELETE_WINDOW "set ok -1"
  raise $s
  bind . <<RaiseDialogs>> "wm deiconify $s; raise $s"

  set wgeomlist [split [wm geometry $w] "x+"]
  set sgeom "+[expr [lindex $wgeomlist 2]+3]+[expr [lindex $wgeomlist 3]+150]"
  wm geometry $s $sgeom

  set ok 0
  label $s.parent_info -text "[mc {Find in}] $info($w,win_title):"
  entry $s.find -width 50 -selectforeground $info(text_select_fg_color) -selectbackground $info(text_select_bg_color) -foreground $info(text_fg_color) -background $info(text_bg_color)
  $s.find insert 0 [lindex $info(find_find_again) 0]
  frame $s.mod_frame 
    pack [checkbutton  $s.mod_frame.casens -text [mc "case sensitive"] -onvalue 1 -offvalue 0 -variable info(find_casens) -selectcolor white] -side left
    pack [checkbutton  $s.mod_frame.revcom -text [mc "also find rev-com of string"] -onvalue 1 -offvalue 0 -variable info(find_revcom) -selectcolor white] -side left
    pack [checkbutton  $s.mod_frame.aa -text [mc "as amino acid"] -onvalue 1 -offvalue 0 -variable info(find_as_aa) -selectcolor white] -side left
    pack [checkbutton  $s.mod_frame.exact -text [mc "as literal"] -onvalue 1 -offvalue 0 -variable info(find_as_literal) -selectcolor white] -side left 
       
  frame $s.action_frame
    pack [button $s.action_frame.next -text [mc "Find Next"] -command "set ok next" -default active] -side left 
    pack [button $s.action_frame.prev -text [mc "Find Prev"] -command "set ok prev"] -side left 
    pack [button $s.action_frame.all -text [mc "Highlight All"] -command "set ok highlight"] -side left
    pack [button $s.action_frame.cancel -text [mc "Cancel"] -command "set ok -1"] -side left 

  pack $s.parent_info -side top -fill x -expand true
  pack $s.find -side top -fill x -expand true
  pack $s.mod_frame -side top -fill x -expand true
  pack $s.action_frame -side top -fill x -expand true
 
  update idletasks
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $s] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry $s "+$winx+$winy"

  #grab set $s
  bind $s <Key-Return> "$s.action_frame.next invoke"
  focus $s
  focus $s.find

  vwait ok
  set dialogblock 0
  if {$ok != -1} {
    do_find $w [$s.find get] $info(find_casens) $info(find_revcom) $info(find_as_aa) $info(find_as_literal) $ok
  }
  #grab release $s 
  destroy $s
  bind . <<RaiseDialogs>> ""
}

###########
## perform find function
###########
proc do_find {w s casens revcom as_aa as_literal action} {
  global info

  if {$as_aa} {
    set pat [aa_regexp $s]
  } elseif {!$as_literal} {
    regsub -all -nocase {[^abcdghkmnrstvwy\|\;\,]} $s "" s
    set pat [regsubpattern $s] 
    while {[regsub -all {\|\|} $pat "|" pat] > 0} {}
    regsub {^\|} $pat "" pat
    regsub {\|$} $pat "" pat
  } else {
    set pat $s 
  }

  if {$pat == ""} {
    set info(find_find_again) [list]
    return
  } else {
    set info(find_find_again) [list $s $casens $revcom $as_aa $as_literal $action]
  }

  if {$revcom} {
    set rpat "($pat)|([revcom_regexp $pat])"
  } else {
    set rpat $pat
  }

  if {!$casens || $as_aa} {set switch2 "-nocase"} else {set switch2 "--"}

  switch $action {
    next {
      catch {$w.textarea mark set insert sel.first+1chars}
      set first [$w.textarea search -regexp -count last -forwards $switch2 $rpat insert]
      if {$first != ""} {
        select_region $w $first "$first+$last chars"
      } else {
        bell
      }
    }
    prev {
      catch {$w.textarea mark set insert sel.first}
      set first [$w.textarea search -regexp -count last -backwards $switch2 $rpat insert]
      if {$first != ""} {
        select_region $w $first "$first+$last chars"
      } else {bell}
    }
    highlight {
      set cur 1.0
      $w.textarea tag remove foundf 1.0 end
      $w.textarea tag remove foundr 1.0 end
      $w.textarea tag configure foundf -background $info(find_forward_highlight)
      $w.textarea tag configure foundr -background $info(find_reverse_highlight)
      if {!$casens || $as_aa} {
        set switches "-nocase"
      } else {
        set switches "--"
      }
      while {[set first [$w.textarea search -regexp -count last -forwards $switches $pat $cur end]] != ""} {
        set cur [$w.textarea index "$first+1chars"]
        $w.textarea tag add foundf $first "$first+$last chars"
        $w.textarea tag bind foundf <<Revcolors>> [list gray70 gray50]
      }
      $w.textarea tag raise foundf
      if {$revcom} {
        set cur 1.0
        set pat [revcom_regexp $pat]
        while {[set first [$w.textarea search -regexp -count last -forwards $switches $pat $cur end]] != ""} {
          set cur [$w.textarea index "$first+1chars"]
          $w.textarea tag add foundr $first "$first+$last chars"
          $w.textarea tag bind foundr <<Revcolors>> [list gray50 gray70]
        }
        $w.textarea tag raise foundr
      }
    }
  }

  focus $w
  focus $w.textarea
}


###########
## returns regexp pattern for aa input 
###########
proc aa_regexp {pattern} {
  set temp [list]
  set lookup [list A (GC.) C (TG\[CT\]) D (GA\[CT\]) E (GA\[AG\]) F (TT\[CT\]) G (GG.) H (CA\[CT\]) I (AT\[ACT\]) K (AA\[AG\]) L ((CT.)|(TT\[AG\])) M (ATG) N (AA\[CT\]) P (CC.) Q (CA\[AG\]) R ((CG.)|(AG\[AG\])) S ((TC.)|(AG\[CT\])) T (AC.) V (GT.) W (TGG) X (...) Y (TA\[CT\]) * ((TA\[AG\])|(TGA))]  

  foreach seq [split $pattern " ;,"] {
    lappend temp [string map -nocase $lookup $seq]
  }
  return [join $temp |]
}

###########
## returns revrse complement of simple regexp patterns
###########
proc revcom_regexp {pattern} {
  set temp ""
  foreach char [split $pattern ""] {
      set temp "[string map {A T C G G C T A a t c g g c t a \[ \] \] \[ ) ( ( )} $char]$temp"
  }
  return $temp
}

##########
## returns the reverse complement of the input
##########
proc revcom {pattern} {
  set temp ""
  foreach char [split $pattern ""] {
    set temp "[string map {A T B V C G D H G C H D K M M K N N * * R Y S S T A V B W W Y R a t b v c g d h g c h d k m m k n n r y s s t a v b w w y r} $char]$temp"
  }
  return $temp
}

##########
## returns the reverse complement of the input
##########
proc rev {pattern} {
  set temp ""
  foreach char [split $pattern ""] {
    set temp "$char$temp"
  }
  return $temp
}

################
## Translate dna, 1 or 3 characters, direction:1 forward, 0 reverse, spacing between characters
################
proc translate {dna {chars 1} {direction 1} {spacing 0}} {

  set i [string range $dna 0 [expr [string length $dna]/3*3-1]]

  while {[regsub -nocase "((...)*)((.\[bdhkmnrsvwy*\].)|(..\[bdhkmnrsvwy*\])|(\[bdhkmnrsvwy*\]..))(.*)" $i {\1xxx\7} i]} {}
  set i [regexp -inline -all {...} $i]

  if {$chars == 1} {
    if {$direction == 1} {
      set result [string map -nocase {xxx X ttt F ttc F tta L ttg L tct S tcc S tca S tcg S tat Y tac Y taa * tag * tgt C tgc C tga * tgg W ctt L ctc L cta L ctg L cct P ccc P cca P ccg P cat H cac H caa Q cag Q cgt R cgc R cga R cgg R att I atc I ata I atg M act T acc T aca T acg T aat N aac N aaa K aag K agt S agc S aga R agg R gtt V gtc V gta V gtg V gct A gcc A gca A gcg A gat D gac D gaa E gag E ggt G ggc G gga G ggg G} $i]
    } else {
      set result [string map -nocase {xxx X aaa F gaa F taa L caa L aga S gga S tga S cga S ata Y gta Y tta * cta * aca C gca C tca * cca W aag L gag L tag L cag L agg P ggg P tgg P cgg P atg H gtg H ttg Q ctg Q acg R gcg R tcg R ccg R aat I gat I tat I cat M agt T ggt T tgt T cgt T att N gtt N ttt K ctt K act S gct S tct R cct R aac V gac V tac V cac V agc A ggc A tgc A cgc A atc D gtc D ttc E ctc E acc G gcc G tcc G ccc G} $i]
    }
  } else {
    if {$direction == 1} {
      set result [string map -nocase {xxx Xxx ttt Phe ttc Phe tta Leu ttg Leu tct Ser tcc Ser tca Ser tcg Ser tat Tyr tac Tyr taa End tag End tgt Cys tgc Cys tga End tgg Trp ctt Leu ctc Leu cta Leu ctg Leu cct Pro ccc Pro cca Pro ccg Pro cat His cac His caa Gln cag Gln cgt Arg cgc Arg cga Arg cgg Arg att Ile atc Ile ata Ile atg Met act Thr acc Thr aca Thr acg Thr aat Asn aac Asn aaa Lys aag Lys agt Ser agc Ser aga Arg agg Arg gtt Val gtc Val gta Val gtg Val gct Ala gcc Ala gca Ala gcg Ala gat Asp gac Asp gaa Glu gag Glu ggt Gly ggc Gly gga Gly ggg Gly} $i]
    } else {
      set result [string map -nocase {xxx Xxx aaa Phe gaa Phe taa Leu caa Leu aga Ser gga Ser tga Ser cga Ser ata Tyr gta Tyr tta End cta End aca Cys gca Cys tca End cca Trp aag Leu gag Leu tag Leu cag Leu agg Pro ggg Pro tgg Pro cgg Pro atg His gtg His ttg Gln ctg Gln acg Arg gcg Arg tcg Arg ccg Arg aat Ile gat Ile tat Ile cat Met agt Thr ggt Thr tgt Thr cgt Thr att Asn gtt Asn ttt Lys ctt Lys act Ser gct Ser tct Arg cct Arg aac Val gac Val tac Val cac Val agc Ala ggc Ala tgc Ala cgc Ala atc Asp gtc Asp ttc Glu ctc Glu acc Gly gcc Gly tcc Gly ccc Gly} $i]
    }
  }
  return [join $result [string repeat " " $spacing]]
}

###############
##finds the melting temperature of a sequence, [dna] in nM, [Na] in mM
###############
proc Tm {seq {dna_conc 250} {salt_conc 50}} {
  if {[string length $seq] > 100} {return "---"}
  regsub -nocase -all {[^acgt]} $seq "t" seq
  set seq [string tolower $seq]
  array set s {aa 222 tt 222 at 204 ta 213 ca 227 tg 227 gt 224 ac 224 ct 210 ag 210 ga 222 tc 222 cg 272 gc 244 gg 199 cc 199}
  array set h {aa 79 tt 79 at 72 ta 72 ca 85 tg 85 gt 84 ac 84 ct 78 ag 78 ga 82 tc 82 cg 106 gc 98 gg 80 cc 80}
  
  if {[string first [string index $seq 0] "at"] > -1} {
    set htot -23
    set stot -41
  } else {
    set htot -1
    set stot 28
  }
  if {[string first [string index $seq end] "at"] > -1} {
    incr htot -23
    incr stot -41
  } else {
    incr htot -1
    incr stot 28
  }
  set l [expr [string length $seq]-1]
  for {set i 0} {$i < $l} {incr i} {
    incr htot $h([string range $seq $i [expr $i+1]])
    incr stot $s([string range $seq $i [expr $i+1]])
  }
  set log_dna [expr {1.987 * log( $dna_conc / 1000000000.0)}]
  set log_salt [expr {12 * log10($salt_conc / 1000.0)}]
  set result [expr {$htot * 100.0 / ($stot * 0.1 - $log_dna) - 273.15 + $log_salt}]
  if {$result > 0} {return $result} else {return <0}
}


##########
## finds the %GC of a sequence
##########
proc GC% {seq} {
set gc [string map -nocase {a "" t "" n "" b "" d "" h "" k "" m "" r "" v "" w "" y "" * ""} $seq]
return [expr {100.0*[string length $gc]/[string length $seq]}] 
}

##########
## find ORF- finds the next ATG-Stop after the cursor
##########
proc next_ORF {w {direction forward}} {
  global info
  set c [list]
  if {$direction == "forward"} {
    if {[$w.textarea tag ranges sel] != {}} {
      set text [$w.textarea get sel.first+1chars 1.end]
      set i [$w.textarea index sel.first+1chars]
    } else {
      set text [$w.textarea get insert 1.end]
      set i [$w.textarea index insert]
    }

    if {$info(find_orf_starts) == "Met"} {
      set c [lindex [lindex [regexp -nocase -indices -inline {ATG} $text] 0] 0]
    } elseif {[$w.textarea compare 1.3 > $i]} {
      set c 0
    } else {
      set c [lindex [lindex [regexp -nocase -indices -inline "(?:taa|tag|tga|tar)" $text] 0] 0]
      if {$c != [list]} {
        incr c 3
      }
    }
    if {$c != [list]} {
      set text [string range $text $c end]
      set d [lindex [lindex [regexp -nocase -indices -inline "^(?:...)*?(?:taa|tag|tga|tar)" $text] 0] 1]
      if {$d == [list]} {set d [expr {3*([string length $text]/3)}]}
      lappend c [expr {$c+$d}]
    }
  } else {
#find previous
    if {[$w.textarea tag ranges sel] != {}} {
      set text [$w.textarea get 1.0 sel.last-1chars]
    } else {
      set text [$w.textarea get 1.0 insert]
    }
    set stoplist [lsort -decreasing -integer -index 0 [regexp -inline -all -nocase -indices {(?:taa|tag|tga|tar)} $text]]
    set stoplist [concat $stoplist [list {2 4} {1 3} {0 2}]]
    if {((([$w.textarea tag ranges sel] != {}) && ([$w.textarea compare 1.end-3chars < sel.last])) || ([$w.textarea compare 1.end-3chars < insert])) && ([string length $text] > 2)} {
      set stoplist [concat [list [list [expr {[string length $text]-3}] [expr {[string length $text]-1}]]] $stoplist]
    }
    set startlist [lsort -decreasing -integer -index 0 [regexp -inline -all -nocase -indices {(?:atg)} $text]]
    set c [list]
    foreach stop2 $stoplist {
      foreach stop1 $stoplist {
        if {([lindex $stop1 0] < [lindex $stop2 0]) &&([expr {fmod([lindex $stop2 0] - [lindex $stop1 0], 3)}]==0.0)} {
          if {$info(find_orf_starts) == "Met"} {
             if {[regexp -nocase -indices {(?:atg)(?:...)*$} [string range $text [lindex $stop1 0] [lindex $stop2 1]] start] == 1} {
               set c [list [expr {[lindex $start 0]+[lindex $stop1 0]}] [lindex $stop2 1]]
             }
          } else {
            set c [list [expr {3+[lindex $stop1 0]}] [lindex $stop2 1]]
          }
          break
        }
      }
      if {$c != [list]} break
    }

    set i 1.0

  }
  if {$c == [list]} {
    bell
    return
  } else {
    $w.textarea tag remove sel 1.0 end
    select_region $w [$w.textarea index $i+[lindex $c 0]chars] [$w.textarea index $i+[expr {1+[lindex $c 1]}]chars]
  }
}

##########
## changes case of the selection (or window if no selection)
##########
proc change_case {w {action "upper"} {undo Undo} {undo_text ""}} { 
  global info

  if {[file_lock_check $w]} {return}

  if {[$w.textarea tag ranges sel] == ""} {
    set first 1.0
    set last [$w.textarea index end-1chars]

    set reset 0
  } else {
    set first [$w.textarea index sel.first]
    set last [$w.textarea index sel.last]
    catch {set anchorstore [$w.textarea index anchor]}
    set reset 1
  }
  set insertstore [$w.textarea index insert]
  set scrollstore [lindex [$w.textarea yview] 0]
  set textstore [$w.textarea get $first $last]
  set tagstore [get_tags $w.textarea [list $first $last]]
  $w.textarea delete $first $last
  if {($undo == "Undo") || ($action == "exchange")} {
    switch $action {
      upper {
        $w.textarea insert $first [string toupper $textstore] {}
        register_undo $w change_case [list $first $last $action $textstore] "Change Case" Undo
      }
      lower {
        $w.textarea insert $first [string tolower $textstore] {}
        register_undo $w change_case [list $first $last $action $textstore] "Change Case" Undo
      }
      exchange {
        $w.textarea insert $first [swap $textstore] {}
        register_undo $w change_case [list $first $last $action] "Change Case" Undo
      }
      default {
        $w.textarea insert $first $textstore {}
      }
    }
  } else {
    $w.textarea insert $first $undo_text {}
    register_undo $w change_case [list $first $last $action] "Change Case" Redo
  }
  $w.textarea yview moveto $scrollstore
  put_tags $w.textarea [list $first $last] $tagstore forward 0
  if {$reset} {
    $w.textarea tag add sel $first $last
    catch {$w.textarea mark set anchor $anchorstore}
  } 
  $w.textarea mark set insert $insertstore



  set info($w,saved) 0
}

##########
## returns the case swap of the input
##########
proc swap {pattern} {
  return [string map {A a B b C c D d G g H h K k M m N n R r S s T t V v W w Y y a A b B c C d D g G h H k K m M n N r R s S t T v V w W y Y} $pattern]
}

##########
## rotates the origin of a circular sequence to the insertion point-needs to move the tags too
##########
proc set_origin {w {undo Undo}} {
  global info
  if {([$w.textarea index insert] == 1.0) ||[file_lock_check $w]} {return}
  set before [$w.textarea get 1.0 insert]
  set tagstore [get_tags $w.textarea [list 1.0 [$w.textarea index insert]]]
  $w.textarea delete 1.0 insert
  $w.textarea insert end-1chars $before {}
  put_tags $w.textarea [list 1.end-[string length $before]chars 1.end] $tagstore forward 0
  selection_manager $w

  register_undo $w set_origin [list [$w.textarea index 1.end-[string length $before]chars]] "Set Origin" $undo
  set info($w,scanned) 0
  set info($w,saved) 0
  unlink_analysis_window $w
  vertindexbalancer $w
}

##########
## toggle circular to linear
##########
proc toggle_circ {w {undo Undo}} {
  global info

  if {[file_lock_check $w]} {return}

  if {($info($w,circular) == "linear") } {
    set info($w,circular) circular
    $w.menubar.edit entryconfigure [mc "Set Origin"] -state normal
    $w.menubar.edit entryconfigure [mc "Linearize @ insert site"] -state normal
  } else {
    set info($w,circular) linear
    $w.menubar.edit entryconfigure [mc "Set Origin"] -state disabled
    $w.menubar.edit entryconfigure [mc "Linearize @ insert site"] -state disabled
  }
  register_undo $w toggle_circ [list [feature_cleanup $w.textarea]] "Circular<->Linear" $undo
  set info($w,scanned) 0
  set info($w,saved) 0
  unlink_analysis_window $w
}

##########
## linearize at insert site
##########
proc linearize {w {undo Undo}} {
  set_origin $w
  toggle_circ $w
  register_undo $w cant [list] "Linearize" $undo
}

##########
## reverse complements the selection if there is one, or the whole dna if not -needs to deal with tags
##########
proc rev_com_window {w {undo Undo}} {
  global info 

  if {[file_lock_check $w]} {return}
  if {[$w.textarea tag ranges sel] == ""} {
    set first 1.0
    set last [$w.textarea index end-1chars]
    set reset 0
  } else {
    set first [$w.textarea index sel.first]
    set last [$w.textarea index sel.last]
    set anchorstore $first
    catch {set anchorstore [$w.textarea index anchor]}
    set reset 1
  }
  set insertstore [$w.textarea index insert]
  set selected [$w.textarea get $first $last]
  set tagstore [get_tags $w.textarea [list $first $last]]
  $w.textarea delete $first $last
  $w.textarea insert $first [revcom $selected] {}

  #if rev-com the first time, rename the tags and store the name equivalences
  #if undoing ($undo=="Redo"), then don't change the names so the equivalences can be re-exchanged
  set put_tag_store [put_tags $w.textarea [list $first $last] $tagstore reverse [expr {$undo=="Undo"}]]

  if {$reset} {
    $w.textarea tag add sel $first $last
    $w.textarea mark set insert $insertstore
    $w.textarea mark set anchor $anchorstore
  } else {
    regexp "1.(.*)" $insertstore fill insertstore
    regexp "1.(.*)" [$w.textarea index 1.end] fill dnaend
    $w.textarea mark set insert "1.[expr {int($dnaend-$insertstore)}]"
    $w.textarea see insert
  }

  register_undo $w rev_com [list $first $last $put_tag_store] "Reverse Complement" $undo

  vertindexbalancer $w
  selection_manager $w
  set info($w,scanned) 0 
  set info($w,saved) 0
  unlink_analysis_window $w
}
##############
## enzyme selection dialog
##############
proc enzyme_selection_dialog {w} {
  global info
  global enzinfo
  global enzymes
  global enz_select_window
  global enz_select_groups
  global enz_select_meth select_numsites_lab select_numsites_comp
  global ok dialogblock modstring modifier
  if {$dialogblock == 1} {return}
  set dialogblock 1

  set s ".enz_selector"

  foreach window $info(dnawindows_list) {
    findenzymes $window
  }
  if {[winfo exists $s]} {destroy $s}
  toplevel $s -background $info(bg_color)

  bind . <<RaiseDialogs>> "wm deiconify $s; raise $s"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry $s "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient $s $w
  } else {
    ::tk::unsupported::MacWindowStyle style $s document {closeBox resizable}
  }
  wm title $s [mc "Enzyme Selection..."]
  wm protocol $s WM_DELETE_WINDOW "$s.actionframe.cancel invoke"
  set ok 0
  menu $s.menubar

  menu $s.menubar.filemenu 
  $s.menubar add cascade -menu $s.menubar.filemenu -label [mc "File"]
  $s.menubar.filemenu add command -label [mc "Open new enzymes file"] -command "reload_enzymes"
  $s.menubar.filemenu add command -label [mc "Save enzymes file"] -command {save_enzymes "" 0; wm deiconify .enz_selector}
  $s.menubar.filemenu add separator
  $s.menubar.filemenu add command -label [mc "Save current enzymes as Default"] -command {save_enzymes "" 1; wm deiconify .enz_selector}
  $s.menubar.filemenu add separator
  $s.menubar.filemenu add command -label [mc "Import Enzymes From DNA Strider..."] -command "import_strider_enzymes .enz_selector"
  $s.menubar.filemenu add separator
  $s.menubar.filemenu add command -label [mc "Close Selection Dialog"] -accelerator "$modstring+W" -command "$s.actionframe.cancel invoke"

  menu $s.menubar.enzmenu 
  $s.menubar add cascade -menu $s.menubar.enzmenu -label [mc "Enzymes"]
  $s.menubar.enzmenu add command -label [mc "New Enzyme..."] -command "add_enzyme_dialog $s"
  $s.menubar.enzmenu add command -label [mc "Delete Selected Enzymes"] -command {delete_selected $info(enz_currently_selected)}
  $s.menubar.enzmenu add separator
  $s.menubar.enzmenu add command -label [mc "New Group..."] -command "add_enz_group $s \$info(enz_currently_selected)"
  $s.menubar.enzmenu add command -label [mc "Delete Groups..."] -command "delete_enz_group $s"
  $s.menubar.enzmenu add command -label [mc "Read Groups File..."] -command "read_enz_groups $s"


  menu $s.menubar.listmenu
  $s.menubar add cascade -menu $s.menubar.listmenu -label [mc "List"]
  $s.menubar.listmenu add command -label [mc "List enzymes"] -command {set ok List_enzymes0}
  $s.menubar.listmenu add command -label [mc "List enzymes + #sites"] -command {set ok List_enzymes1}
  $s.menubar.listmenu add command -label [mc "List enzymes + site positions"] -command {set ok List_enzymes2}
  $s.menubar.listmenu add command -label [mc "List enzyme + recognition sequences"] -command {set ok List_enzymes3}

  if {[tk windowingsystem] == "aqua"} {
    menu $s.menubar.apple 
    $s.menubar.apple add command -label [mc "About ApE..."] -command "about_dialog ." -state disabled
    $s.menubar.apple add separator
    $s.menubar.apple add command -label [mc "Preferences..."]  -command "configure_preferences ." -state disabled
    $s.menubar add cascade -menu $s.menubar.apple
  }

  $s configure -menu $s.menubar

  bind $s <$modifier-KeyPress-w> "$s.actionframe.cancel invoke"
  bind $s <$modifier-KeyPress-m> "$s.actionframe.graphic invoke"
  bind $s <$modifier-KeyPress-M> "$s.actionframe.graphicU invoke" 
  bind $s <$modifier-KeyPress-g> "$s.actionframe.digest invoke" 
  bind $s <$modifier-KeyPress-T> "$s.actionframe.textmap invoke"

##Region of interest frame###
  grid [frame $s.location -relief ridge -borderwidth 3] -row 0 -column 0 -sticky nwe
  grid [label $s.location.label -text [mc "Window"] -font boldlabelfont] -row 0 -column 0 -sticky wns
  menubutton $s.location.win_button
    menu $s.location.win_button.menu 
    foreach window $info(dnawindows_list) {
      $s.location.win_button.menu add radiobutton -command "fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment $window; $s.location.win_button configure -textvariable info($window,win_title); set enz_select_meth \$info($window,Dam_Dcm_methylated)" -label $info($window,win_title) -variable enz_select_window -value $window 
    }
    $s.location.win_button configure -textvariable info($w,win_title) -menu $s.location.win_button.menu  -indicatoron 1 -relief raised -width 50 
    bind $s.location.win_button <<MenuSelect>> "focus $s"

  set enz_select_window $w
  set enz_select_meth $info($w,Dam_Dcm_methylated)
  grid $s.location.win_button -row 0 -column 1 -sticky wens

  grid [checkbutton $s.location.local -height 1 -width 10 -variable info(analyze_selection) -textvariable info(analyze_selection) -onvalue "selection" -offvalue "all" -indicator off -selectcolor $info(bg_color) -activebackground $info(bg_color) -command "fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment \$enz_select_window; event generate . <<toggle_select>>"] -row 0 -column 2 -sticky wns 
  grid [checkbutton $s.location.methylation -text Dam/Dcm -variable enz_select_meth -onvalue 1 -offvalue 0 -command {set info($enz_select_window,Dam_Dcm_methylated) $enz_select_meth;toggle_methylation $enz_select_window;fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment $enz_select_window} -selectcolor white] -row 0 -column 3 -sticky wns
  grid columnconfigure $s.location 1 -weight 1

###Enzyme list frame ######
  grid [frame $s.textframe] -row 1 -column 0 -sticky nswe
  grid [scrollbar $s.textframe.scroll -command "$s.textframe.textbox yview"] -row 0 -column 1 -sticky nsw
  grid [text $s.textframe.textbox -yscrollcommand "optionscrollbar $s.textframe.scroll" -height 20 -background $info(text_bg_color) -foreground $info(text_fg_color) -font labelfont -wrap none] -row 0 -column 0 -sticky nswe
  set tabwidth [font measure [$s.textframe.textbox cget -font] "HpyCH4III(200) "]
  $s.textframe.textbox configure -tabs "$tabwidth left" -width [expr {0+$tabwidth*8/[font measure [$s.textframe.textbox cget -font] "0"]}]
  #grid [scrollbar $s.textframe.scrollx -command "$s.textframe.textbox xview" -orient horizontal] -row 1 -column 0 -sticky nwe

  grid columnconfigure $s.textframe 0 -weight 1
  grid rowconfigure $s.textframe 0 -weight 1

##Enzyme Comment box
  grid [label $s.enz_comment -justify center] -row 2 -column 0 -sticky nswe

###Select_enzymes frame 
  grid [frame $s.groupsframe -relief ridge -borderwidth 3] -row 3 -column 0 -sticky nwe
  grid [label $s.groupsframe.label -text [mc "Select Enzymes"] -font boldlabelfont] -row 0 -column 0 -sticky w

  #number of sites menu button
  menubutton $s.groupsframe.numsites
    menu $s.groupsframe.numsites.menu 
    foreach label [list [mc "unique (1)"] [mc any] [mc "absent (0)"] [mc "present (>0)"] [mc "not unique (>1)"] <2 [mc "double (2)"] >2 <3 [mc "triple (3)"] >3 <4 >4] comparison [list "==1" ">-1" "==0" ">0" ">1" "<2" "==2" ">2" "<3" "==3" ">3" "<4" ">4"] {
      $s.groupsframe.numsites.menu add radiobutton -label $label -value $comparison -command "set select_numsites_lab {$label}" -variable select_numsites_comp
    }
    $s.groupsframe.numsites configure -indicatoron 1 -relief raised -width 12 -textvariable select_numsites_lab -menu $s.groupsframe.numsites.menu
    $s.groupsframe.numsites.menu invoke 0
    bind $s.groupsframe.numsites.menu <<MenuSelect>> "focus $s"
  grid $s.groupsframe.numsites -row 0 -column 1 -sticky w 

  #enzyme groups menu button
  menubutton $s.groupsframe.groups_button
    menu $s.groupsframe.groups_button.menu 
    foreach element $enzinfo(enz_groups_list) {
      $s.groupsframe.groups_button.menu add radiobutton -label $element -variable enz_select_groups -value $element 
    }
    $s.groupsframe.groups_button configure -indicatoron 1 -relief raised -width 10 -textvariable enz_select_groups -menu $s.groupsframe.groups_button.menu
    $s.groupsframe.groups_button.menu invoke 0
    bind $s.groupsframe.groups_button.menu <<MenuSelect>> "focus $s"

  grid $s.groupsframe.groups_button -row 0 -column 2 -sticky w

  #select functions button set
  grid [button $s.groupsframe.plus -text "Select" -command {tagmanage .enz_selector.textframe.textbox add [enz_site_usage_filter $enz_select_window $enzinfo(enz_groups,$enz_select_groups) $select_numsites_comp $info(analyze_selection)]}] -row 0 -column 3 -sticky w
  grid [button $s.groupsframe.minus -text "De-select" -command {tagmanage .enz_selector.textframe.textbox remove [enz_site_usage_filter $enz_select_window $enzinfo(enz_groups,$enz_select_groups) $select_numsites_comp $info(analyze_selection)]}] -row 0 -column 4 -sticky w
  grid [button $s.groupsframe.and -text "AND" -command {tagmanage .enz_selector.textframe.textbox booland [enz_site_usage_filter $enz_select_window $enzinfo(enz_groups,$enz_select_groups) $select_numsites_comp $info(analyze_selection)]}] -row 0 -column 5 -sticky w
  grid [button $s.groupsframe.clearall -text [mc "clear all"] -command {tagmanage .enz_selector.textframe.textbox booland {}}] -row 0 -column 6 -sticky w
  grid [button $s.groupsframe.memplus -text "Sel to Mem" -command {set enzinfo(enz_groups,Mem\ Recall) $info(enz_currently_selected)}] -row 0 -column 7 -padx {10 0} -sticky w
  grid columnconfigure $s.groupsframe 7 -weight 1

##Perfom Action frame######
  grid [frame $s.actionframe -relief ridge  -borderwidth 3] -row 4 -column 0 -sticky nsew
  grid [label $s.actionframe.label -text [mc "Perform Action"] -font boldlabelfont] -row 0 -column 0 -sticky w
  grid [button $s.actionframe.graphic -text [mc "Graphic map"] -command {set ok Graphic_map}] -row 0 -column 1
  grid [button $s.actionframe.graphicU -text [mc "Graphic map +U"] -command {set ok Graphic_mapU}] -row 0 -column 2
  grid [button $s.actionframe.digest -text [mc "Digest"] -command {if {$info(enz_currently_selected) != {}} {set ok Digest}}] -row 0 -column 3
  grid [button $s.actionframe.hilight -text [mc "Highlight"] -command {set ok Highlight}]  -row 0 -column 4
  grid [button $s.actionframe.textmap -text [mc "Text"] -command {set ok Textmap}]  -row 0 -column 5
  grid [button $s.actionframe.cancel -text [mc "Close"] -command "set ok Cancel"]  -row 0 -column 6
  grid columnconfigure $s.actionframe 1 -weight 1
  grid columnconfigure $s.actionframe 2 -weight 1
  grid columnconfigure $s.actionframe 3 -weight 1
  grid columnconfigure $s.actionframe 4 -weight 1
  grid columnconfigure $s.actionframe 5 -weight 1

  fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment $enz_select_window

  grid columnconfigure $s 0 -weight 1
  grid rowconfigure $s 0 -weight 0
  grid rowconfigure $s 1 -weight 1
  grid rowconfigure $s 2 -weight 0
  grid rowconfigure $s 3 -weight 0
  grid rowconfigure $s 4 -weight 0

  update idletasks
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $s] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry $s "+$winx+$winy"

  tagmanage .enz_selector.textframe.textbox add $info(enz_currently_selected)
  $s.textframe.textbox configure -state disabled -cursor arrow
  bind $s <Any-MouseWheel> {[winfo toplevel %W].textframe.textbox yview scroll [expr {%D/-$info(mousewheelunits)}] units}
  bind $s <Any-Button-1> {}


  bindtags $s.textframe.textbox {$s.textframe.textbox .enz_selector}
  #grab set $s
  bind $s <Key-Return> "$s.actionframe.cancel invoke"
  focus $s

  vwait ok

  set dialogblock 0
  bind . <<RaiseDialogs>> ""

  if {$info(enzymes_saved) == 0} {
    if {[tk_messageBox -message "Do you want to save changes to the current enzyme set?" -type yesno] == "yes"} {
      save_enzymes "" 0
    } else {
      set info(enzymes_saved) -1
    }
  }

  switch $ok {
    Graphic_map {
      enz_graphic_map $enz_select_window $info(enz_currently_selected) $info(analyze_selection)
    }
    Graphic_mapU {
      enz_graphic_map $enz_select_window $info(enz_currently_selected) $info(analyze_selection) {==1}
    }
    Digest {
      enz_digest $enz_select_window $info(enz_currently_selected) $info(analyze_selection)
    }
    Highlight {
      highlight_enzymes $enz_select_window $info(enz_currently_selected)
    }
    Textmap {
      enz_text_map $enz_select_window 0
    }
    List_enzymes0 {
      list_enzymes $enz_select_window $info(enz_currently_selected) 0 "Selected Enzymes" $info(analyze_selection)
    }
    List_enzymes1 {
      list_enzymes $enz_select_window $info(enz_currently_selected) 1 "Selected Enzymes" $info(analyze_selection)
    }
    List_enzymes2 {
      list_enzymes $enz_select_window $info(enz_currently_selected) 2 "Selected Enzymes" $info(analyze_selection)
    }
    List_enzymes3 {
      list_enzymes $enz_select_window $info(enz_currently_selected) 3 "Selected Enzymes" $info(analyze_selection)
    }
  }

  #grab release .enz_selector
  destroy .enz_selector
  unset select_numsites_lab select_numsites_comp enz_select_meth ok 
  set enz_select_groups "All"
}

#################
## insert enzymes into the enzyme selector text box
#################
proc fill_enzymes_box {textbox commentbox w } {
  set columns 8
  global enzymes
  global enzinfo
  global info

  set enzymelist $enzymes
  while {[expr {[llength $enzymelist] % $columns}] > 0} {
    lappend enzymelist ""
  }
  set rows [expr int([llength $enzymelist]/$columns)]

  $textbox configure -state normal
  $textbox delete 1.0 end
  for {set i 0} {$i < $rows} {incr i} {
    set prepend " "
    for {set j 0} {$j < $columns} {incr j} {
      set beginenz [$textbox index "end-1 chars"]
      set enzyme [lindex $enzymelist [expr $j*$rows+$i]]
      if {$enzyme != ""} {
        $textbox insert end "$prepend$enzyme ([enz_count $w $enzyme $info(analyze_selection)]) "
        $textbox tag add $enzyme "$beginenz+1 chars" "end-1 chars"
        if {$enzinfo(compat,$enzyme) != ""} {
          if {[llength $enzinfo(compat,$enzyme)] < 11} {
            $textbox tag bind $enzyme <Any-Enter> "$commentbox configure -text {$enzyme: $enzinfo(enz_comment,$enzyme) (compat w/: $enzinfo(compat,$enzyme))}"
          } else {
            $textbox tag bind $enzyme <Any-Enter> "$commentbox configure -text {$enzyme: $enzinfo(enz_comment,$enzyme) (compat w/: [lrange $enzinfo(compat,$enzyme) 0 9]+[expr {[llength $enzinfo(compat,$enzyme)]-10}] others)}"
          }
        } else {
          $textbox tag bind $enzyme <Any-Enter> "$commentbox configure -text {$enzyme: $enzinfo(enz_comment,$enzyme)}"
        }
        $textbox tag bind $enzyme <Any-Leave> "$commentbox configure -text {}"
        $textbox tag bind $enzyme <Button-1> "tagmanage $textbox click $enzyme"
        $textbox tag bind $enzyme <Double-Button-1> "tagmanage $textbox click $enzyme; edit_enzyme $enzyme .enz_selector"
        $textbox tag bind $enzyme <Shift-Button-1> "tagmanage $textbox booland {}; tagmanage $textbox click $enzyme"
        if {[tk windowingsystem] != "aqua"} {
          $textbox tag bind $enzyme <Button-3> "tagmanage $textbox booland {}; tagmanage $textbox click $enzyme"
        } else {
          $textbox tag bind $enzyme <Button-2> "tagmanage $textbox booland {}; tagmanage $textbox click $enzyme"
        }
        set prepend "\t"
      } else {
        $textbox insert end "$prepend$enzyme "
      }
    }
    $textbox insert end "\n"
  }
  $textbox delete end-1char end
  $textbox configure -state disabled
  unset enzymelist
}


#################
## filter an incoming group for site number and return remaining elements
#################
proc enz_site_usage_filter {w {enzlist [list]} {comparison "==1"} {region all}} {
  global info enz_select_groups

  set templist [list]
  if {$enzlist != "Difference..."} {
    foreach enzyme $enzlist {
      if "[enz_count $w $enzyme $region] $comparison" {lappend templist $enzyme}
    }
  } else {
    foreach enzyme [enzyme_diffs $w] {
      if "[enz_count $w $enzyme $region] $comparison" {lappend templist $enzyme}
    }
  }
  return $templist
}

#################
## return the number of sites for enzyme in window
#################
proc enz_count {w enzyme region} {
  global info

  if {[$w.textarea tag ranges tempsel] == ""} {set region "all"}
  if {$region == "all"} {
    return [llength $info($w,$enzyme)]
  } else {
    regexp "\\.(.*)" [$w.textarea index tempsel.first] fill start 
    regexp "\\.(.*)" [$w.textarea index tempsel.last] fill end
    set i 0
    foreach site $info($w,$enzyme) {
      if {$site > $end} {break}
      if {$site >= $start} {incr i}
    }
    return $i
  }
}

#################
## retrun enzymes that have diff # cuts from w
#################
proc enzyme_diffs {w} {
  global ok2 wins wins_title analyze_selection info enzymes

  set a [toplevel .dialog -background $info(bg_color)]
  wm title $a [mc "Enzyme Difference..."]
  set ok 0
  wm protocol $a WM_DELETE_WINDOW "set ok2 -1"
  set bindstore [bind . <<RaiseDialogs>>]
  bind . <<RaiseDialogs>> "wm deiconify $a; raise $a"
  if {[winfo exists .enz_selector]} {
    regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .enz_selector] wgeom winx winy
    wm geometry $a "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
    if {[tk windowingsystem] != "aqua"} {
      wm transient $a .enz_selector
    } else {
      ::tk::unsupported::MacWindowStyle style $a document {closeBox}
    }
  }

  grid [label $a.l1 -text "Window 1: $info($w,win_title), $info(analyze_selection)" -font labelfont] -row 1 -column 0 -columnspan 3 -sticky w

  grid [label $a.l2 -text "Window 2:" -font labelfont] -row 2 -column 0 -sticky w
  set win_menu [tk_optionMenu $a.win wins_title temp]
  $win_menu configure -background $info(menu_bg_color) -foreground $info(menu_fg_color) -activebackground $info(menu_select_bg_color) -activeforeground $info(menu_select_fg_color)
  foreach window $info(dnawindows_list) {
      $win_menu add radiobutton -label $info($window,win_title) -variable wins -value $window -command "set wins_title \{$info($window,win_title)\}"
  }
  $win_menu delete 0
  set wins $w
  set wins_title "$info($w,win_title)"
  grid $a.win -row 2 -column 1 -sticky we

  grid [checkbutton $a.local2 -height 1 -width 10 -variable analyze_selection -textvariable analyze_selection -onvalue "selection" -offvalue "all" -indicator off -selectcolor $info(bg_color) -activebackground $info(bg_color)] -row 2 -column 2 -sticky wns 
  set analyze_selection all

  grid [button $a.ok -text "OK" -command "set ok2 1" -padx 8 -pady 3 -default active] -row 3 -column 1
  grid columnconfigure $a 1 -weight 1

  bind $a <Key-Return> "$a.ok invoke"
  wm deiconify $a; raise $a; focus $a
  vwait ok2

  set templist [list]
  if {$ok2 == 1} {
    foreach enzyme $enzymes {
      if {[enz_count $w $enzyme $info(analyze_selection)] != [enz_count $wins $enzyme $analyze_selection]} {lappend templist $enzyme}
    }
  }
  bind . <<RaiseDialogs>> $bindstore
  destroy $a
  unset ok2 wins wins_title analyze_selection
  set templist
}

#################
## manage the enzyme selector dialog enzyme textbox tags
#################
proc tagmanage {textbox action {taglist [list]}} {
  global enzymes
  global info

  switch $action {
    click {
       foreach etag $taglist {
          if {[$textbox tag cget $etag -background] == "red"} {
            $textbox tag configure $etag -background white
          } else {
            $textbox tag configure $etag -background red
          }
       }
    }
    add {
       foreach etag $taglist {
          $textbox tag configure $etag -background red
       }
    }
    remove {
       foreach etag $taglist {
          $textbox tag configure $etag -background white
       }
    }
    booland {
       foreach etag $enzymes {
          if {[lsearch -exact $taglist $etag] == -1} {
             $textbox tag configure $etag -background white  
          }
       }
    }
    diff {

    }
  }
  set info(enz_currently_selected) [list]
  foreach etag $enzymes {
    if {[$textbox tag cget $etag -background] == "red"} {
      lappend info(enz_currently_selected) $etag
    }
  }
}

##############
## Delete selected enzymes from the enzyme library
##############
proc delete_selected {enzyme_list} {
  global info
  global enz_select_window

  if {([llength $enzyme_list] == 0) || ([tk_messageBox -message "Are you sure you want to delete these [llength $enzyme_list] enzymes?" -type okcancel -default cancel  -icon question] == "cancel")} {
    return
  }
  wm deiconify .enz_selector
  .enz_selector configure -cursor watch
  .enz_selector.textframe.textbox configure -cursor watch
  update

  foreach enzyme $enzyme_list {
    delete_enzyme $enzyme
  }
  set info(enzymes_saved) 0
  foreach window $info(dnawindows_list) {
    set info($window,scanned) 0
    findenzymes $window
  }

  fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment $enz_select_window
  .enz_selector configure -cursor arrow
  .enz_selector.textframe.textbox configure -cursor arrow
  .enz_selector.textframe.textbox configure -cursor arrow
  update
}

##############
## List enzymes Dialog
##############
proc list_enzymes_dialog {w} {
  global info
  global ok
  global analyze_selection
  global enzymes dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

##open configuration dialog
  toplevel .dialog -background $info(bg_color)
  wm title .dialog [mc "List Enzymes..."]
  set ok 0
  wm protocol .dialog WM_DELETE_WINDOW "set ok -1"
  bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry .dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .dialog $w
  } else {
    ::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
  }

  set modelist [list [mc "Name"] [mc "Name, #Sites"] [mc "Name, Site, #Sites, Positions"] [mc "Site, Name"]]
  grid [frame .dialog.f1 -relief ridge -borderwidth 3] -row 0 -column 0  -sticky nwe
  grid [label .dialog.f1.label1 -text [mc "Format:"] ] -row 0 -column 0 -sticky w
  grid [menubutton .dialog.f1.formatmbutton -direction below -indicatoron 1 -textvariable info(enz_list_mode) -menu .dialog.f1.formatmbutton.menu] -row 0 -column 1 -sticky w
  menu .dialog.f1.formatmbutton.menu 
  foreach mode $modelist {
    .dialog.f1.formatmbutton.menu add radiobutton -label $mode -variable info(enz_list_mode)
  }
  bind .dialog.f1.formatmbutton.menu <<MenuSelect>> {focus .dialog}
  grid [label .dialog.f1.label2 -text [mc "Columns:"] ] -row 0 -column 2 -sticky w
  tk_optionMenu .dialog.f1.colentry info(enz_list_columns) 1 2 3 4 5 6 7 8 9 10
  grid .dialog.f1.colentry -row 0 -column 3 -sticky w

  grid [frame .dialog.f3 -relief ridge -borderwidth 3] -row 1 -column 0  -sticky nwe
  grid [label .dialog.f3.label1 -text [mc "Enzymes:"] ] -row 0 -column 0 -sticky w
  grid [radiobutton .dialog.f3.allenz -text [mc "All"] -value "all" -variable info(enz_list_enz) -selectcolor white] -row 0 -column 1 -sticky w
  grid [radiobutton .dialog.f3.selenz -text [mc "Selected Enzymes Only"] -value "selected" -variable info(enz_list_enz) -selectcolor white] -row 0 -column 2 -sticky w
  grid [radiobutton .dialog.f3.absenz -text [mc "Absent Enzymes"] -value "absent" -variable info(enz_list_enz) -selectcolor white] -row 0 -column 3 -sticky w
  grid [radiobutton .dialog.f3.unienz -text [mc "Unique Enzymes"] -value "unique" -variable info(enz_list_enz) -selectcolor white] -row 0 -column 4 -sticky w

  set analyze_selection $info(analyze_selection)
  grid [frame .dialog.f4 -relief ridge -borderwidth 3] -row 2 -column 0  -sticky nwe
  grid [label .dialog.f4.label1 -text [mc "DNA:"]] -row 0 -column 0 -sticky w
  grid [radiobutton .dialog.f4.selall -text [mc "All"] -value "all" -variable analyze_selection -selectcolor white] -row 0 -column 1 -sticky w
  grid [radiobutton .dialog.f4.selonly -text [mc "Selected DNA only"] -value "selection" -variable analyze_selection -selectcolor white] -row 0 -column 2 -sticky w

  grid [button .dialog.ok -command {set ok 1} -text [mc "OK"]  -default active] -row 3 -column 0 -sticky n

  grid columnconfigure .dialog 0 -weight 1

  update idletasks
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .dialog] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry .dialog "+$winx+$winy"
  wm resizable .dialog 0 0
  focus .dialog
  #grab .dialog
  vwait ok
  set dialogblock 0

  if {$ok == 1} {
    findenzymes $w
    set mode [lsearch -exact $modelist $info(enz_list_mode)]
    switch $info(enz_list_enz) {
      "all" {
        list_enzymes $w $enzymes $mode "All Enzymes" $analyze_selection
      }
      "selected" {
        list_enzymes $w $info(enz_currently_selected) $mode "Selected Sites" $analyze_selection
      }
      "absent" {
        list_enzymes $w [enz_site_usage_filter $w $enzymes "==0" $analyze_selection] $mode "Absent Sites" $analyze_selection
      }
      "unique" {
        list_enzymes $w [enz_site_usage_filter $w $enzymes "==1" $analyze_selection] $mode "Unique Sites" $analyze_selection
      }
    }
  }
  destroy .dialog
  bind . <<RaiseDialogs>> ""
  unset ok 

}

##############
## List enzymes
##############
#mode: (sorted by *)
#0:*name    
#1:*name, #_of_sites       
#2:name, comment, *#_of_sites, position  
#3:*comment, name


proc list_enzymes {w enz_list {mode 0} {title "Enzyme List"} {region "all"}} {
  global info
  global enzinfo
  global modifier
  global modstring

  #set region $info(analyze_selection)
  if {([$w.textarea tag ranges tempsel] == "") && ($region == "selection")} {set region "all"}
  if {$region == "all"} {
    set start 0
    regexp "\\.(.*)" [$w.textarea index end-1chars] fill end
  } else {
    regexp "\\.(.*)" [$w.textarea index tempsel.first] fill start 
    regexp "\\.(.*)" [$w.textarea index tempsel.last] fill end
  }

  findenzymes $w

  set templist [list]
  foreach enzyme $enz_list {
    set sitelist [list]
    foreach site $info($w,$enzyme) {
      if {$site >= $end} {break}
      if {$site >= $start} {lappend sitelist $site}
    }
    lappend templist [list $enzyme [llength $sitelist] $sitelist $enzinfo(enz_pattern,$enzyme) [lindex $sitelist 0]]
  }

  set a [new_analysis_window $w $title]
  create_textframe $a 120
    
  $a.textframe.text insert end "[clock format [clock seconds] -format $info(clock_format)]\n"
  $a.textframe.text insert end "$info($w,win_title) \n"
  $a.textframe.text insert end "[mc From] $start [mc to] $end\. \n\n"
  $a.textframe.text insert end "$title\n\n"

  switch $mode {
    0 {
      $a.textframe.text configure -tabs "2c" -width 75
      set templist [lsort -index 0 $templist]
      set rows [expr {ceil(1.0*[llength $templist]/$info(enz_list_columns))}]
      set columns [expr {$info(enz_list_columns) - 1 }]
      for {set i 0} {$i < $rows} {incr i} {
        set spacer "\t"
        for {set j 0} {$j <= $columns} {incr j} {
          if {$j == $columns} {set spacer "\n"}
          if {[set enzyme [lindex $templist [expr {$i*$info(enz_list_columns)+$j}]]] != ""} {
            $a.textframe.text insert end "[lindex $enzyme 0]$spacer"
          }
        }
      } 
    }
    1 {
      $a.textframe.text configure -tabs "2c 3c 5c 6c 8c 9c 11c 12c 14c 15c 16c 18c" -width 100
      set templist [lsort -index 0 $templist]
      set rows [expr {ceil(1.0*[llength $templist]/$info(enz_list_columns))}]
      set columns [expr {$info(enz_list_columns) - 1 }]
      for {set i 0} {$i < $rows} {incr i} {
        set spacer "\t"
        for {set j 0} {$j <= $columns} {incr j} {
          if {$j == $columns} {set spacer "\n"}
          if {[set enzyme [lindex $templist [expr {$i*$info(enz_list_columns)+$j}]]] != ""} {
            $a.textframe.text insert end "[lindex $enzyme 0]\t[lindex $enzyme 1]$spacer"
          }
        }
      }
    }
    2 {
      $a.textframe.text configure -tabs "2c 5c 5.5c" -width 100
      set templist [lsort -integer -index 1 $templist]
      foreach enzyme $templist {
        $a.textframe.text insert end "[lindex $enzyme 0]\t[lindex $enzyme 3]\t[lindex $enzyme 1]\t[lindex $enzyme 2]\n"
      }
    }
    3 {
      $a.textframe.text configure -tabs "6c " -width 50
      set templist [lsort -index 3 $templist]
      foreach enzyme $templist {
        $a.textframe.text insert end "[lindex $enzyme 3]\t[lindex $enzyme 0]\n"
      }
    }
  }
  $a.textframe.text configure -state disabled

  set outputmenu [add_output_menu $a.textframe.text "Text" text $w]
  bind $a <Button-1> "focus $a.textframe.text" 
  bindtags $a.textframe.text [list $a.textframe.text AnalysisText . all]
  wm deiconify $a
  window_on_screen $a
  focus $a.textframe.text
}

##############
## Highlights the enzymes in list enz_list in window w
##############
proc highlight_enzymes {w enz_list} {
  global enzymes
  global enzinfo
  global info

  set bg_color $info(enzyme_highlight_color)
  findenzymes $w

  foreach enzyme $enzymes {
    $w.textarea tag delete $enzyme 1.0 end 
  }

  foreach enzyme $enz_list { 
    set length [string length $enzinfo(flatpat,$enzyme)]
    set comment $enzinfo(enz_comment,$enzyme)
    foreach site $info($w,$enzyme) {
      $w.textarea tag add $enzyme "1.[expr ($site -1)]" "1.[expr ($site + $length -1)]"
    }
    $w.textarea tag configure $enzyme -background $bg_color
    #metadat format: (0)name, (1)special_info, (2)type, (3)fwd/rev, (4)annotations_string
    $w.textarea tag bind $enzyme <<Metadata>> [list $enzyme {} enzyme 0 $comment]
    #$w.textarea tag bind $enzyme <Any-Enter> "\[winfo parent %W\].infobox configure -text \"$enzyme: $comment\""
    #$w.textarea tag bind $enzyme <Any-Leave> "\[winfo parent %W\].infobox configure -text {}"
  }
}

##############
## Make a new window to contain an analysis result
##############
proc new_analysis_window {winlist title} {
  global info
  global modifier

  set w [lindex $winlist 0]
  set childlist [winfo children .]
  set i 0
  while {[lsearch -exact $childlist "$w\_analysis$i"] > -1 } {incr i}
  set a [toplevel "$w\_analysis$i"]
  bind $a <FocusIn> "event generate . <<RaiseDialogs>>"
  bind $a <Map> "event generate . <<RaiseDialogs>>"
  bind $a <Activate> "event generate . <<RaiseDialogs>>"
  bind $a <$modifier-Key-n> "create_window"
  bind $a <$modifier-Key-q> "close_all_windows"
  wm protocol $a WM_DELETE_WINDOW "closewindow $a"
  wm title $a "[wm title $w] [mc $title]"
  set info($a,linked_windows) $winlist
  update_windows_menu
  return $a
}


##############
## Add a double scrolled text ananysis frame
##############
proc create_textframe {a width {row 0} {column 0}} {
  set screen_fraction 0.5
  set text_height [expr {int([winfo screenheight .]*$screen_fraction/[font metrics dnafont -linespace])}]
  grid [frame $a.textframe] -row $row -column $column -sticky nswe
  grid columnconfigure $a.textframe 0 -weight 1
  grid rowconfigure $a.textframe 0 -weight 1
  grid [text $a.textframe.text -yscrollcommand "optionscrollbar $a.textframe.scrolly" -xscrollcommand "optionscrollbar $a.textframe.scrollx" -wrap none -width $width -height $text_height -font dnafont] -row 0 -column 0 -rowspan 2 -sticky nswe
  #grid [text $a.textframe.text -yscrollcommand "optionyscrollbar $a.textframe.scrolly" -xscrollcommand "optionxscrollbar $a.textframe.scrollx" -wrap none -width $width -font dnafont] -row 0 -column 0 -rowspan 2 -sticky nswe
  #grid [text $a.textframe.text -yscrollcommand "optionyscrollbar $a.textframe.scrolly" -xscrollcommand "$a.textframe.scrollx set" -wrap none -width $width -font dnafont] -row 0 -column 0 -rowspan 2 -sticky nswe

  grid [frame $a.textframe.spacer -width 2 -height 17] -row 2 -column 0 -sticky nswe
  grid [scrollbar $a.textframe.scrollx -orient horizontal -command "$a.textframe.text xview"] -row 2 -column 0 -sticky swe
  grid [scrollbar $a.textframe.scrolly -orient vertical -command "$a.textframe.text yview"] -row 0 -column 1 -sticky nse
  if {([tk windowingsystem] == "aqua")} {
    grid [frame $a.textframe.sp -height 10] -row 1 -column 1 -sticky swe
    bind $a.textframe.scrolly <Button-1> {set mousedown %Y; set scrollmdown [lindex [%W get] 0]}
    $a.textframe.scrolly configure -command "vertscroll_bug $a.textframe text scrolly"
  }
  grid columnconfigure $a.textframe 0 -weight 1
  grid rowconfigure $a.textframe 0 -weight 1

  grid columnconfigure $a 0 -weight 1
  grid rowconfigure $a 0 -weight 1
}

##############
## Fix the Aqua text scrolling bug for text windows
##############
if {[tk windowingsystem] == "aqua"} {
proc vertscroll_bug {w txt sc args} {
  global mousedown scrollmdown
  if {[lindex $args 0] == "moveto"} {
    #catch the moveto- this is mouse movement
    #I assume the arrows are at the bottom and ~40pixels high
    #the moveto value is the fraction the scrollbar was at when
    #the mouse was first clicked plus the fraction the mouse has
    # moved

    #the fraction the mouse has moved is the difference between
    #where the mouse was first clicked and where it is now, divided
    #by the length of the scrollbar in pixels.

    set moveto [expr {$scrollmdown+(1.0*[winfo pointery .]-$mousedown)/([winfo height $w.$sc]-40)}]
    $w.$txt yview moveto $moveto
  } else {
    #all of the other commands are fine
    eval [concat $w.$txt yview $args]
  }
}
}

##############
## Add the output menu to an analysis window
##############
proc add_output_menu {w label type return_window} {
  global info modifier modstring

  set top [winfo toplevel $w]
  if {[tk windowingsystem] != "aqua"} {
    set outputmenu [menu $w.outputmenu]
    bind $w <Button-3> "tk_popup $outputmenu %X %Y"
    #bind $top <Button-1> "if \{\[winfo ismapped $outputmenu\]\} \{$outputmenu unpost\}"
  } else {
    menu $top.menubar
    set outputmenu [menu $top.menubar.outputmenu]
    $top.menubar add cascade -menu $outputmenu -label [mc $label]

    menu $top.menubar.apple
    $top.menubar.apple add command -label [mc "About ApE..."] -command "about_dialog $top"
    $top.menubar.apple add separator
    $top.menubar.apple add command -label [mc "Preferences..."]  -command "configure_preferences $top"
    $top.menubar add cascade -menu $top.menubar.apple

    $top.menubar add cascade -menu [menu $top.menubar.windows] -label [mc "Windows"]
    update_windows_menu
    menu $top.menubar.help
    $top.menubar add cascade -menu $top.menubar.help -label [mc "Help"]
    $top.menubar.help add command -label [mc "Standard Genetic Code"] -command "genetic_code_dialog $top"
    $top.menubar.help add command -label [mc "AA Info"] -command "aa_info_dialog $top"
    $top configure -menu $top.menubar
    bind $w <Button-2> "tk_popup $outputmenu %X %Y"
  }

  switch $type {
    text {
      $outputmenu add command -label [mc "Copy"] -accelerator "$modstring+C" -command "event generate $w <<Copy>>"
      $outputmenu add command -label [mc "Save"] -accelerator "$modstring+S" -command "event generate $w <<Save>>"
      $outputmenu add command -label [mc "Print"] -accelerator "$modstring+P" -command "event generate $w <<Print>>"
      if {([lsearch -glob [info loaded] "*MacCarbonPrint.dylib*"] > -1) ||  ([lsearch -glob [info loaded] "*wmf.dll*"] > -1)} {
        $outputmenu add command -label [mc "Page Setup"] -accelerator "$modstring+Shift+P" -command "page_setup_dialog $top"
        bind $w <$modifier-KeyPress-P> "event generate $w <<Print-Setup>>"
        bind $w <<Print-Setup>> "page_setup_dialog $top"
      }
      $outputmenu add command -label [mc "Raise Parent"] -accelerator "$modstring+Space" -command "event generate $w <<Return>>"
      $outputmenu add command -label [mc "Close"] -accelerator "$modstring+W" -command "closewindow $top"
      bind $w <$modifier-KeyPress-c> "event generate $w <<Copy>>"
      bind $w <$modifier-KeyPress-s> "event generate $w <<Save>>"
      bind $w <$modifier-KeyPress-p> "event generate $w <<Print>>"
      bind $w <$modifier-KeyPress-w> "closewindow $top"
      bind $w <$modifier-KeyPress-space> "event generate $w <<Return>>"
      bind $w <Double-Button-1> "event generate $w <<Return>>; break"
      bind $w <<Save>> "text_save $w"
      bind $w <<Print>> "print_text $w"
      bind $w <<Return>> "wm deiconify $return_window; raise $return_window; focus $return_window.textarea"
      lappend info($return_window,linked_events) [list $w <<Return>>]
    }
    image {
      set c $w
      if {([lsearch -glob [info loaded] "*wmf.dll*"] > -1)} {
        $outputmenu add command -label [mc "Copy Metafile"] -accelerator "$modstring+C" -command "event generate $c <<CopyEMF>>"
        bind $c <$modifier-KeyPress-c> "event generate $c <<CopyEMF>>"
        bind $c <<CopyEMF>> "canvas_to_metafile copy $c both"
      }
      $outputmenu add command -label [mc "Save"] -accelerator "$modstring+S" -command "event generate $c <<Save>>"
      $outputmenu add command -label [mc "Print"] -accelerator "$modstring+P" -command "event generate $c <<Print>>"
      if {([lsearch -glob [info loaded] "*MacCarbonPrint.dylib*"] > -1) ||  ([lsearch -glob [info loaded] "*wmf.dll*"] > -1)} {
        $outputmenu add command -label [mc "Page Setup"] -accelerator "$modstring+Shift+P" -command "page_setup_dialog $top"
        bind $c <$modifier-KeyPress-P> "event generate $c <<Print-Setup>>"
        bind $c <<Print-Setup>> "page_setup_dialog $top"
      }
      $outputmenu add command -label [mc "Raise Parent"] -accelerator "$modstring+Space" -command "event generate $c <<Return>>"
      $outputmenu add command -label [mc "Close"] -accelerator "$modstring+W" -command "closewindow $top"
      bind $c <$modifier-KeyPress-s> "event generate $c <<Save>>"
      bind $c <$modifier-KeyPress-p> "event generate $c <<Print>>"
      bind $c <$modifier-KeyPress-space> "event generate $c <<Return>>"
      bind $c <$modifier-KeyPress-w> "closewindow $top"
      bind $c <Double-Button-1> "event generate $c <<Return>>; break"
      bind $c <<Save>> "canvas_save $c"
      bind $c <<Print>> "print_canvas $c $c"
      bind $c <<Return>> "wm deiconify $return_window; raise $return_window; focus $return_window.textarea; break"
      lappend info($return_window,linked_events) [list $c <<Return>>]
    }
    none {
    }
  }
  return $outputmenu
}


##############
## Make a digest window
##############
proc enz_digest {w enz_list region} {
  global enzymes
  global info
  global modifier
  global modstring

  if {$enz_list == [list]} {
    enzyme_selection_dialog $w
    return 
  }

  if {[$w.textarea tag ranges tempsel] == ""} {set region "all"}
  if {$region == "all"} {
    set start 1
    regexp "\\.(.*)" [$w.textarea index end-1chars] fill total_length
    set end $total_length
  } else {
    regexp "\\.(.*)" [$w.textarea index tempsel.first] fill start
    incr start 
    regexp "\\.(.*)" [$w.textarea index tempsel.last] fill end
    set total_length [expr {$end - $start+1}]
  }
  if {$total_length == 0} return
  incr end
  findenzymes $w

  set templist [list]
  foreach enzyme $enz_list {
    foreach site $info($w,$enzyme) {
      if {$site >= $end} {break}
      if {$site >= $start} {lappend templist [list $site $enzyme]}
    }
  }

  if {($region != "all") || ($info($w,circular) != "circular")} {
    lappend templist [list $start "Start"]
    lappend templist [list $end "End"]
  }
  set templist [lsort -integer -index 0 $templist]
  set fraglist [list]
  set i 0
  set lastsite [lindex [lindex $templist 0] 0]
  set lastenz [lindex [lindex $templist 0] 1]
  foreach pair [lrange $templist 1 end] {
    if {[expr {[lindex $pair 0]- $lastsite}] > 0} {
      lappend fraglist [list [expr {[lindex $pair 0]- $lastsite}] $lastenz $lastsite [lindex $pair 1] [lindex $pair 0] $i [expr {round(100.0*([lindex $pair 0]- $lastsite)/$total_length)}]]
      set lastsite [lindex $pair 0]
      set lastenz [lindex $pair 1]
      incr i
    }
  }
  if {($region == "all") && ($info($w,circular) == "circular")} {
    if {$templist != [list]} {
      lappend fraglist [list [expr {[lindex [lindex $templist 0] 0]+ $total_length - $lastsite}] $lastenz $lastsite [lindex [lindex $templist 0] 1] [lindex [lindex $templist 0] 0] $i [expr {round(100.0*([lindex [lindex $templist 0] 0]+ $total_length - $lastsite)/$total_length)}]]
    } else {    
      tk_messageBox -type ok -message "No Enzymes Cut -- Circular product." -default ok  -icon info
      return 
    }
  }
  set fraglist [lsort -integer -index 0 -decreasing $fraglist]

  set a [new_analysis_window $w "Digest"]  

  grid [canvas $a.t -yscrollcommand "optionscrollbar $a.s" -width 500 -height 250 -background white] -row 0 -column 0 -sticky nswe
  grid [scrollbar $a.s -orient vertical -command "$a.t yview"] -row 0 -column 1 -sticky nsw
  grid rowconfigure $a 0 -weight 1
  grid columnconfigure $a 0 -weight 1

  #draw file info
  $a.t create text 10 10 -text "[clock format [clock seconds] -format $info(clock_format)]" -tags [list label table] -anchor nw -font [list $info(graphicfontfamily) $info(graphicfontsize)]
  set textheight [expr {[lindex [$a.t bbox label] 3] - [lindex [$a.t bbox label] 1]}]
  $a.t create text 10 [expr {10+$textheight}] -text "[wm title $w]" -tags [list label table] -anchor nw -font [list $info(graphicfontfamily) $info(graphicfontsize)]

  #draw table header
  $a.t create text 10 [expr {10+2*$textheight}] -text "Size\t site1\t\t site2\t\t Mass %" -tags [list title table] -anchor nw -font [list $info(graphicfontfamily) $info(graphicfontsize)]
  set textright [lindex [$a.t bbox title] 2]
  $a.t create line 10 [lindex [$a.t bbox title] 3] [lindex [$a.t bbox title] 2] [lindex [$a.t bbox title] 3]
  set i 1
  if {$info(ladderlist) != [list]} {
    set dig_ladder $info(ladder,$info(current_ladder))
  } else {
    set dig_ladder [list]
  }

  if {($dig_ladder != [list]) && ([lindex [lindex $fraglist end] 0] > [lindex $dig_ladder end])} {
    set shortfrag [lindex $dig_ladder end]
  } else {
    set shortfrag [lindex [lindex $fraglist end] 0]
  }
  if {$shortfrag < 10} {
    set shortfrag 10
  }

  if {($dig_ladder != [list]) && ([lindex [lindex $fraglist 0] 0] < [lindex $dig_ladder 0])} {
    set longfrag [lindex $dig_ladder 0]
  } else {
    set longfrag [lindex [lindex $fraglist 0] 0]
  }

  if {[lindex [lindex $fraglist 0] 0] > $shortfrag} {
    set gelscale [expr {210.0/(log10($longfrag)- log10($shortfrag))}]
  } else {
    set gelscale 20
  }
  set gelorigin [expr {30.0 + $gelscale * log10($longfrag) }]

  #draw ladder
  if {$dig_ladder != [list]} {
    set lastladderband 0
    foreach band $dig_ladder {
      if {$lastladderband != $band} {
        if {[lindex $band 0] > 0} {
          $a.t create line [expr {$textright + 5}] [expr {$gelorigin - $gelscale * log10($band)}] [expr {$textright + 55}] [expr {$gelorigin - $gelscale *log10($band)}] -width 3 -capstyle round -tags [list ladderband$band scalable]
        }
        $a.t bind ladderband$band <Any-Enter> "$a.t itemconfigure laddertext -text [string range $band 0 end-3][expr {([string length $band]>3)?",":{}}][string range $band end-2 end]"
        $a.t bind ladderband$band <Any-Leave> "$a.t itemconfigure laddertext -text {}"
        set lastladderband $band
      } else {
        $a.t itemconfigure ladderband$band -width [expr {[$a.t itemcget ladderband$band -width]+1}]
     # warning_dialog ladderband$band [$a.t itemcget ladderband$band -width]
      }
    }
    $a.t create text [expr {$textright + 8}] 10  -anchor nw -justify left -font [list $info(graphicfontfamily) $info(graphicfontsize)] -tags laddertext
  }
  foreach fragment $fraglist {
    #draw table entry for the band
    $a.t create text 10 [expr {10+($i+2)*$textheight}] -tags [list table frag$i] -font [list $info(graphicfontfamily) $info(graphicfontsize)] -anchor nw -text "[lindex $fragment 0]\t[lindex $fragment 1]\t[lindex $fragment 2]\t[lindex $fragment 3]\t[lindex $fragment 4]\t[lindex $fragment 6]"
    #draw gel bands
    if {[lindex $fragment 0] > 0} {
      $a.t create line [expr {$textright + 60}] [expr {$gelorigin - $gelscale * log10([lindex $fragment 0])}] [expr {$textright + 110}] [expr {$gelorigin - $gelscale *log10([lindex $fragment 0])}] -width 3 -capstyle round -tags [list band$i scalable]
    }
    $a.t bind frag$i <Any-Enter> "$a.t itemconfigure frag$i -fill red; $a.t itemconfigure band$i -fill red -width 5; $a.t raise band$i "
    $a.t bind frag$i <Any-Leave> "$a.t itemconfigure frag$i -fill black; $a.t itemconfigure band$i -fill black -width 3"
    $a.t bind band$i <Any-Enter> "$a.t itemconfigure frag$i -fill red; $a.t itemconfigure band$i -fill red -width 5; $a.t raise band$i "
    $a.t bind band$i <Any-Leave> "$a.t itemconfigure frag$i -fill black; $a.t itemconfigure band$i -fill black -width 3"
    #draw map area band highlight
    if {[lindex $fragment 2] < [lindex $fragment 4]} {
      $a.t create line [expr {$textright + 120}] [expr {30+ ([lindex $fragment 2]-$start)* 210 / ($end - $start)}] [expr {$textright + 120}] [expr {30+ ([lindex $fragment 4]-$start)* 210 / ($end - $start)}] -width 3 -tags [list band$i scalable]
      $a.t bind frag$i <Button-1> "select_region $w [$w.textarea index 1.[expr {[lindex $fragment 2]-1}]] [$w.textarea index 1.[expr {[lindex $fragment 4]-1}]]" 
      $a.t bind band$i <Button-1> "select_region $w [$w.textarea index 1.[expr {[lindex $fragment 2]-1}]] [$w.textarea index 1.[expr {[lindex $fragment 4]-1}]]"  
      lappend info($w,linked_events) [list $a.t "frag$i" "<Button-1>"] [list $a.t "band$i" "<Button-1>"]
    } else {
#needs to deal with band button presses for circular fragments
#may need to give the two halves unique tags, then rebind the enter and leave for frag, as well as for band
      $a.t create line [expr {$textright + 120}] 30 [expr {$textright + 120}] [expr {30+ ([lindex $fragment 4]-$start)* 210 / ($end - $start)}] -width 3 -tags [list band$i scalable]
      $a.t create line [expr {$textright + 120}] [expr {30+ ([lindex $fragment 2]-$start)* 210 / ($end - $start)}] [expr {$textright + 120}] 240 -width 3 -tags [list band2$i scalable]
      $a.t bind frag$i <Button-1> "select_region $w 1.0 [$w.textarea index 1.[expr {[lindex $fragment 4]-1}]]" 
      $a.t bind frag$i <Shift-Button-1> "select_region $w [$w.textarea index 1.[expr {[lindex $fragment 2]-1}]] [$w.textarea index 1.[expr {$total_length}]]" 
      $a.t bind band$i <Button-1> "select_region $w 1.0 [$w.textarea index 1.[expr {[lindex $fragment 4]-1}]]" 
      $a.t bind band$i <Shift-Button-1> "select_region $w [$w.textarea index 1.[expr {[lindex $fragment 2]-1}]] [$w.textarea index 1.[expr {$total_length}]]" 
      $a.t bind band2$i <Button-1> "select_region $w [$w.textarea index 1.[expr {[lindex $fragment 2]-1}]] [$w.textarea index 1.end]" 
      $a.t bind band2$i <Shift-Button-1> "select_region $w 1.0 [$w.textarea index 1.[expr {[lindex $fragment 4]-1}]]" 
      $a.t bind frag$i <Any-Enter> "+$a.t itemconfigure band2$i -fill red -width 5; $a.t raise band2$i "
      $a.t bind frag$i <Any-Leave> "+$a.t itemconfigure band2$i -fill black -width 3"
      $a.t bind band$i <Any-Enter> "+$a.t itemconfigure band2$i -fill red -width 5; $a.t raise band2$i "
      $a.t bind band$i <Any-Leave> "+$a.t itemconfigure band2$i -fill black -width 3"
      $a.t bind band2$i <Any-Enter> "$a.t itemconfigure frag$i -fill red; $a.t itemconfigure band$i -fill red -width 5; $a.t raise band$i; $a.t itemconfigure band2$i -fill red -width 5; $a.t raise band2$i  "
      $a.t bind band2$i <Any-Leave> "$a.t itemconfigure frag$i -fill black; $a.t itemconfigure band$i -fill black -width 3; $a.t itemconfigure band2$i -fill black -width 3"
      lappend info($w,linked_events) [list $a.t "frag$i" "<Button-1>"] [list $a.t "frag$i" "<Shift-Button-1>"] [list $a.t "band$i" "<Button-1>"] [list $a.t "band$i" "<Shift-Button-1>"] [list $a.t "band2$i" "<Button-1>"] [list $a.t "band2$i" "<Shift-Button-1>"]
    }
    incr i
  }
 
  #draw map area tick marks
  foreach site $templist {
    $a.t create line [expr {$textright + 120}] [expr {30+ ([lindex $site 0]-$start)* 210 / ($end - $start)}] [expr {$textright + 130}] [expr {30+ ([lindex $site 0]-$start)* 210 / ($end - $start)}] -tags  scalable
    $a.t create text [expr {$textright + 132}] [expr {30+ ([lindex $site 0]-$start)* 210 / ($end - $start)}] -anchor w -text "[lindex $site 0] [lindex $site 1]" -tags  scalable
  }
  $a.t configure -scrollregion [list 10 10 [lindex [$a.t bbox table] 2] [lindex [$a.t bbox table] 3]]
  #$a.t configure -scrollregion [$a.t bbox all]

  bind $a.t <Configure> "if {\[winfo height $a.t\] > 80} {$a.t scale scalable \[lindex \[$a.t bbox scalable\] 0\] \[lindex \[$a.t bbox scalable\] 1\] 1 \[expr \{(\[winfo height $a.t\]-60.0)/(\[lindex \[$a.t bbox scalable\] 3\]-\[lindex \[$a.t bbox scalable\] 1\])\}\]}"

  set outputmenu [add_output_menu $a.t "Image" image $w]
  $outputmenu insert 0 command -label "Copy Text" -accelerator "$modstring+Shift+C" -command "event generate $a.t <<CopyText>>"
  bind $a.t <$modifier-KeyPress-C> "event generate $a.t <<CopyText>>"
  bind $a.t <<CopyText>> "\
    clipboard clear;\
    foreach row \[$a.t find withtag table\] \{\
      clipboard append \"\[$a.t itemcget \$row -text\]\\n\"\
    \}"
  bindtags $a.t [list $a.t $a all]
  bind $a <Button-1> "focus $a.t" 
  wm deiconify $a
  window_on_screen $a
  focus $a.t
}

##############
## draws a text map of enzymes
##############
proc enz_text_map {w {dialog 1}} {
  global info
  global enzymes
  global ok
  global modifier
  global modstring
  global analyze_selection dialogblock
  if {$dialogblock == 1} {return}


##open configuration dialog
  if {$dialog} {
    set dialogblock 1
    toplevel .dialog
    wm title .dialog [mc "Configure Text Map..."]
    set ok 0
    wm protocol .dialog WM_DELETE_WINDOW "set ok -1"
    bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
    regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
    wm geometry .dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
    if {[tk windowingsystem] != "aqua"} {
      wm transient .dialog $w
    } else {
      ::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
    }
    grid [frame .dialog.f3 -relief ridge -borderwidth 3] -row 1 -column 0  -sticky nwe
    grid [label .dialog.f3.label1 -text [mc "Enzymes:"]] -row 0 -column 0 -sticky w
    grid [radiobutton .dialog.f3.allenz -text [mc "All"] -value "all" -variable info(enz_text_enz) -selectcolor white -activebackground $info(bg_color)] -row 0 -column 1 -sticky w
    grid [radiobutton .dialog.f3.selenz -text [mc "Selected Enzymes Only"] -value "selected" -variable info(enz_text_enz) -selectcolor white -activebackground $info(bg_color)] -row 0 -column 2 -sticky w

    set analyze_selection $info(analyze_selection)
    grid [frame .dialog.f4 -relief ridge -borderwidth 3] -row 2 -column 0  -sticky nwe
    grid [label .dialog.f4.label1 -text [mc "DNA:"]] -row 0 -column 0 -sticky w
    grid [radiobutton .dialog.f4.selall -text [mc "All"] -value "all" -variable analyze_selection -selectcolor white -activebackground $info(bg_color)] -row 0 -column 1 -sticky w
    grid [radiobutton .dialog.f4.selonly -text [mc "Selected DNA only"] -value "selection" -variable analyze_selection -selectcolor white -activebackground $info(bg_color)] -row 0 -column 2 -sticky w

    grid [frame .dialog.f1 -relief ridge -borderwidth 3] -row 3 -column 0 -sticky nwe
    grid [checkbutton .dialog.f1.check1 -text [mc "Show 2nd Strand"] -onvalue 1 -offvalue 0 -variable info(enz_text_2nd) -selectcolor white -activebackground $info(bg_color)] -row 0 -column 0 -sticky w
    grid [checkbutton .dialog.f1.copyhighlight -text [mc "Copy Highlighting"] -onvalue 1 -offvalue 0 -variable info(enz_text_copy_highlight) -selectcolor white -activebackground $info(bg_color)] -row 0 -column 1 -sticky w
    grid [label .dialog.f1.widthlabel -text [mc "Characters/line:"]] -row 0 -column 2 -sticky w
    grid [entry .dialog.f1.width -width 4  -textvariable info(enz_text_width) -validate key -vcmd "check_char %S"] -row 0 -column 3 -sticky w
    grid columnconfigure .dialog.f1 0 -weight 1 -pad 0

    grid [frame .dialog.f2 -relief ridge -borderwidth 3] -row 4 -column 0  -sticky nwe
    grid [label .dialog.f2.label1 -text [mc "Translation:"]] -row 0 -column 0  -sticky w
    grid [radiobutton .dialog.f2.tnone -text [mc "No Translation"] -value "off" -variable info(enz_text_trans) -selectcolor white -activebackground $info(bg_color)] -row 0 -column 1 -sticky w
    grid [radiobutton .dialog.f2.11 -text [mc "1 frame 1 letter"] -value "1_1" -variable info(enz_text_trans) -selectcolor white -activebackground $info(bg_color)] -row 1 -column 1 -sticky w
    grid [radiobutton .dialog.f2.31 -text [mc "3 frame 1 letter"] -value "3_1" -variable info(enz_text_trans) -selectcolor white -activebackground $info(bg_color)] -row 1 -column 2 -sticky w
    grid [radiobutton .dialog.f2.61 -text [mc "6 frame 1 letter"] -value "6_1" -variable info(enz_text_trans) -selectcolor white -activebackground $info(bg_color)] -row 1 -column 3 -sticky w
    grid [radiobutton .dialog.f2.13 -text [mc "1 frame 3 letter"] -value "1_3" -variable info(enz_text_trans) -selectcolor white -activebackground $info(bg_color)] -row 2 -column 1 -sticky w
    grid [radiobutton .dialog.f2.33 -text [mc "3 frame 3 letter"] -value "3_3" -variable info(enz_text_trans) -selectcolor white -activebackground $info(bg_color)] -row 2 -column 2 -sticky w
    grid [radiobutton .dialog.f2.63 -text [mc "6 frame 3 letter"] -value "6_3" -variable info(enz_text_trans) -selectcolor white -activebackground $info(bg_color)] -row 2 -column 3 -sticky w
    grid [checkbutton .dialog.f2.check1 -text [mc "Genes"] -onvalue 1 -offvalue 0 -variable info(enz_text_genes) -selectcolor white -activebackground $info(bg_color)] -row 3 -column 0 -sticky w


    grid [button .dialog.ok -command {set ok 1} -text [mc "OK"]  -default active] -row 5 -column 0 -sticky n

    grid columnconfigure .dialog 0 -weight 1
    wm resizable .dialog 0 0

    update idletasks
    regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .dialog] wgeom winw winh winx winy
    if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
    if {$winx < 0} {set winx 0}
    if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
    if {$winy < 0} {set winy 0}
    wm geometry .dialog "+$winx+$winy"

    focus .dialog
    vwait ok
    set dialogblock 0
    destroy .dialog
    bind . <<RaiseDialogs>> ""
    if {$ok != 1} {return}
    unset ok 
    if {$info(enz_text_enz) == "selected"} {
      set enzlist $info(enz_currently_selected)
    } else {
      set enzlist $enzymes
    }
  } else {
    ###No dialog
    set analyze_selection $info(analyze_selection)
     if {$info(enz_currently_selected) != {}} {
      set enzlist $info(enz_currently_selected)
    } else {
      set enzlist $enzymes
    }
  }

  if {$info(enz_text_width) == ""} {set info(enz_text_width) 100}
##now make the text map
  set region $analyze_selection
  if {[$w.textarea tag ranges tempsel] == ""} {set region "all"}
  if {$region == "all"} {
    set start 0
    regexp "\\.(.*)" [$w.textarea index end-1chars] fill total_length
    set end $total_length
  } else {
    regexp "\\.(.*)" [$w.textarea index tempsel.first] fill start 
    regexp "\\.(.*)" [$w.textarea index tempsel.last] fill end
    set total_length [expr {$end - $start}]
  }

  if {$total_length < 1} return
  findenzymes $w

  #set templist [linear_enz_sort $w $enzlist $start $end]

  set a [new_analysis_window $w "Text Map"]

  create_textframe $a [expr {$info(enz_text_width)+21}]

  set text [$w.textarea get "1.$start" "1.$end"]

#proc translate {dna {chars 1} {direction 1} {spacing 0}} {}
  switch $info(enz_text_trans) {
    off {set trans [list]}
    "1_1" {
      set trans [list [translate $text 1 1 2]]  
    }
    "3_1" {
      set trans [list "[translate $text 1 1 2]" " [translate [string range $text 1 end] 1 1 2]" "  [translate [string range $text 2 end] 1 1 2]"]
    }
    "6_1" {
      set trans [list "[translate $text 1 1 2]" " [translate [string range $text 1 end] 1 1 2]" "  [translate [string range $text 2 end] 1 1 2]" "[translate [string range $text 0 end] 1 -1 2]" " [translate [string range $text 1 end] 1 -1 2]" "  [translate [string range $text 2 end] 1 -1 2]" ]
    }
    "1_3" {
      set trans [list [translate $text 3 1 0]]
    }
    "3_3" {
      set trans [list "[translate $text 3 1 0]" " [translate [string range $text 1 end] 3 1 0]" "  [translate [string range $text 2 end] 3 1 0]"]
    }
      "6_3" {
     set trans [list [translate $text 3 1 0] [translate [string range $text 1 end] 3 1 0] [translate [string range $text 2 end] 3 1 0] [translate [string range $text 0 end-[expr {[string length $text] % 3}]] 3 -1 0] [translate [string range $text 0 end-[expr {1+([string length $text] % 3)}]] 3 -1 0] [translate [string range $text 0 end-[expr {2+([string length $text] % 3)}]] 3 -1 0]]
    }
  }

  if {$info(enz_text_genes) == 1} {
    set geneslist [uppercase_cdna $text]
    set genesline [translate [lindex $geneslist 0] 1 1 2]
    foreach intron [lsort -decreasing -integer -index 0 [lindex $geneslist 1]] {
      set genesline "[string range $genesline 0 [expr {[lindex $intron 0]-1}]][string repeat " " [lindex $intron 1]][string range $genesline [lindex $intron 0] end]"
    }
  }

  ##start output
  $a.textframe.text insert end "[clock format [clock seconds] -format $info(clock_format)]\n"
  $a.textframe.text insert end "$info($w,win_title)\nText Map\n"
#put the output out 1 line at a time ($info(enz_text_width) bp per line)
  for {set x 0} {$x < [expr {[string length $text]}]} {incr x $info(enz_text_width)} {
    set linelength [expr {[string length $text]- $x}]
    if {$linelength > $info(enz_text_width)} {
      set linelength $info(enz_text_width)
    }
    set templist [linear_enz_sort $w $enzlist [expr {$start+$x+1}] [expr {$start+$x+$linelength}]]
    set tickline [string repeat "         *" [expr {$linelength /10}]]
    set i 0; set maxi 0; set enzline(0) ""
    foreach pair $templist {
      set i 0
      while {($i <= $maxi) && ([string length $enzline($i)] >= [expr {[lindex $pair 0]-$start - $x -1}])} {
        incr i
      }
      if {$i > $maxi} {
        set maxi $i
        set enzline($i) ""
      }
      set tickline [string replace $tickline [expr {[lindex $pair 0] -$start - $x -1}] [expr {[lindex $pair 0] - $start - $x -1}] "|"]
      set enzline($i) "$enzline($i)[string repeat " " [expr {[lindex $pair 0] - $start - $x -1 -[string length $enzline($i)]}]][lindex $pair 0] [lindex $pair 1]"
    }
    for {set i $maxi} {$i >= 0} {incr i -1} {
      $a.textframe.text insert end "[string repeat " " 7]$enzline($i)\n"
    }
    $a.textframe.text insert end "[string repeat " " 7]$tickline\n"
    set textline [string range $text $x [expr {$x+$linelength-1}]]
    $a.textframe.text insert end "[string repeat " " [expr {6-[string length [expr {$x+$start+1}]]}]][expr {$x+$start+1}] " {} $textline dnatag " [expr {$x+$start+$linelength}]\n" {}

    if {$info(enz_text_copy_highlight)} {
      copy_tags $w.textarea [list 1.[expr {$x+$start}] 1.[expr {$x+$start+$linelength-1}]] $a.textframe.text [$a.textframe.text index "end-2line linestart +7char"] "forward" ""
    }

    if {$info(enz_text_2nd) == 1} {
      $a.textframe.text insert end "       [string map {A T B V C G D H G C H D K M M K N N R Y S S T A V B W W Y R a t b v c g d h g c h d k m m k n n r y s s t a v b w w y r} $textline]\n"
    }

    if {$info(enz_text_genes) == 1} {
      $a.textframe.text insert end "[string repeat " " 7][string range $genesline $x [expr {$x+$linelength-1}]]\n"
    }

    foreach transline $trans {
      $a.textframe.text insert end "[string repeat " " 7][string range $transline $x [expr {$x+$linelength-1}]]\n"
    }


    $a.textframe.text insert end "\n"
  }

  foreach tag [$w.textarea tag names] {
     if {[lsearch -exact [$a.textframe.text tag names] $tag] > -1} {
       $a.textframe.text tag raise $tag
     }
  }

  $a.textframe.text tag bind dnatag <Double-Button> "alignlink $a.textframe.text $w \[$a.textframe.text index @%x,%y\] 1 goto 1; break"
  $a.textframe.text tag bind dnatag <$modifier-Button> "alignlink $a.textframe.text $w \[$a.textframe.text index @%x,%y\] 1 goto 0; break"
  $a.textframe.text tag bind dnatag <Shift-$modifier-Button> "alignlink $a.textframe.text $w \[$a.textframe.text index @%x,%y\] 1 select 0; break"
  $a.textframe.text tag bind dnatag <Shift-Double-Button> "alignlink $a.textframe.text $w \[$a.textframe.text index @%x,%y\] 1 select 1; break"
  lappend info($w,linked_events) [list $a.textframe.text dnatag <Double-Button>]
  lappend info($w,linked_events) [list $a.textframe.text dnatag <$modifier-Button>]
  lappend info($w,linked_events) [list $a.textframe.text dnatag <Shift-$modifier-Button>]
  lappend info($w,linked_events) [list $a.textframe.text dnatag <Shift-Double-Button>]
  $a.textframe.text tag bind dnatag <Any-Leave> {%W tag remove highlite 1.0 end}
  $a.textframe.text tag bind dnatag <Any-Motion> {%W tag remove highlite 1.0 end; %W tag add highlite current}
  $a.textframe.text tag bind dnatag <Any-Enter> {%W tag remove highlite 1.0 end; %W tag add highlite current}
  $a.textframe.text tag configure highlite -background red
#####
  set outputmenu [add_output_menu $a.textframe.text "Text" text $w]
  bind $a <Button-1> "focus $a.textframe.text" 
  bindtags $a.textframe.text [list $a.textframe.text AnalysisText . all]
  wm deiconify $a
  window_on_screen $a
  focus $a.textframe.text
}

##############
## returns cDNA extracted as uppercase exons, and location and size of introns
##############
proc uppercase_cdna {dna} {
  set templist [list]
  #find leading region
  set templist [concat $templist [regexp -inline -indices {([a-y]+)[A-Y]} $dna]]
  #find introns- lowercase gt..ag between uppercase exons
  set templist [concat $templist [regexp -all -inline -indices {[A-Y](g[tc][a-y]*?ag)[A-Y]} $dna]]
  #find trailing region- all lowercase after stop (stop not necessarily in frame with cDNA)
  set templist [concat $templist [regexp -all -inline -indices {(?:TGA|TAA|TAG)([a-y]+)} $dna]]

  set intronlist [list]
  foreach {i j} $templist {
    lappend intronlist $j
  }

  set intronlist [lsort -unique -integer -index 1 $intronlist]
  set sizelist [list]
  set sum 0
  foreach intron $intronlist { 
    set dna [string replace $dna [expr {[lindex $intron 0]-$sum}] [expr {[lindex $intron 1]-$sum}] ""]
    lappend sizelist [list [expr {[lindex $intron 0]-$sum}] [expr {[lindex $intron 1]-[lindex $intron 0]+1}]]
    incr sum [expr {[lindex $intron 1]-[lindex $intron 0]+1}]
  }

  return [list $dna $sizelist]
}

##############
## draws a graphic map of enzymes
##############
proc enz_graphic_map {w enz_list region {numsites "none"}} {
  global enzymes
  global info
  global enzinfo
  global modifier
  global modstring

  if {[$w.textarea tag ranges tempsel] == ""} {set region "all"}

  if {$region == "all"} {
    set start 0
    regexp "\\.(.*)" [$w.textarea index end-1chars] fill total_length
    set end $total_length
  } else {
    regexp "\\.(.*)" [$w.textarea index tempsel.first] fill start 
    regexp "\\.(.*)" [$w.textarea index tempsel.last] fill end
    set total_length [expr {$end - $start}]
  }

  findenzymes $w

  if {$numsites != "none"} {
    set extra_enz [enz_site_usage_filter $w $enzymes $numsites all]
  } else {
    set extra_enz [list]
  }

  set a [new_analysis_window $w "Map"]

#make the canvas

  set c [canvas $a.canvas -width 500 -height 500 -background white]

####check for circular or linear map
  if {($region == "all") && ($info($w,circular) == "circular")} {
    ########draw circular map
    $c create oval 100 100 300 300 
    set pi 3.141593629
    set templist [circ_enz_sort $w [lsort -unique [concat $enz_list $extra_enz]] $total_length]

    #draw enzymes
    set maxx 350; set maxy 350; set minx 50; set miny 50

    set charheight [expr [font metrics graphicfont -linespace]]
    set anchor "sw"; set bump [expr -$charheight]; set nexty 200; set comparison ">"
    set lastloc -1
    foreach pair $templist  {
      set angle [expr {2*$pi*[lindex $pair 0]/$total_length}]
      if {($anchor == "sw") && ($angle > [expr {$pi/2}])} {set anchor "nw"; set nexty 200; set comparison "<"; set bump $charheight}
      if {($anchor == "nw") && ($angle > $pi)} {set anchor "ne"; set nexty 200}
      if {($anchor == "ne") && ($angle > [expr {$pi*1.5}])} {set anchor "se"; set nexty 200; set comparison ">"; set bump [expr -$charheight]}
      if {$lastloc != [lindex $pair 0]} {
        set lineid [$c create line [expr {200.5 + 100*sin($angle)}] [expr {200.5 - 100*cos($angle)}] [expr {200.5 + 110*sin($angle)}] [expr {200.5 - 110*cos($angle)}] -tags [lindex $pair 1]]
      } else {
        $c itemconfigure $lineid -tags [concat [$c itemcget $lineid -tags] [lindex $pair 1]]
      }
      set lastloc [lindex $pair 0]
      set y [expr {200 - 115*cos($angle)}]
      if "$y $comparison $nexty" {set y $nexty}
      set nexty [expr {$y + $bump}]

      set id [$c create text [expr {200 + 115*sin($angle)}] $y -text "[lindex $pair 2]" -anchor $anchor -font [list $info(graphicfontfamily) $info(graphicfontsize)] -tags [lindex $pair 1]] 

      if {$info(draw_graphic_connectors)} {
        $c create line [expr {200.5 + 110*sin($angle)}] [expr {200.5 - 110*cos($angle)}] [expr {200 + 115*sin($angle)}] $y  -tags [lindex $pair 1]
      }

      if {[lsearch -exact $enz_list [lindex $pair 1]] > -1} {
        $c bind $id <Any-Enter> "$c raise [lindex $pair 1]; $c itemconfigure [lindex $pair 1] -fill red"
        $c bind $id <Any-Leave> "$c itemconfigure [lindex $pair 1] -fill black"
      } else {
        $c itemconfigure [lindex $pair 1] -fill gray50
        $c bind $id <Any-Enter> "$c raise [lindex $pair 1]; $c itemconfigure [lindex $pair 1] -fill blue"
        $c bind $id <Any-Leave> "$c itemconfigure [lindex $pair 1] -fill gray50"
      }
    
      if {[lsearch -exact $enzymes [lindex $pair 1]] > -1} {
        if {($anchor == "sw") ||($anchor == "nw")} {
          $c bind $id <Any-Enter> "+$c itemconfigure seqtext -text $enzinfo(enz_pattern,[lindex $pair 1])"
          $c bind $id <Any-Leave> "+$c itemconfigure seqtext -text {}"
        } else {
          $c bind $id <Any-Enter> "+$c itemconfigure seqtext -text $enzinfo(enz_pattern,[lindex $pair 1])"
          $c bind $id <Any-Leave> "+$c itemconfigure seqtext -text {}"
        }
      } else {
        $c itemconfigure $id -font [list $info(graphicfontfamily) $info(graphicfontsize) bold] -fill black
        $c bind $id <Any-Enter> "$c raise [lindex $pair 1]; $c itemconfigure [lindex $pair 1] -fill red"
        $c bind $id <Any-Leave> "$c itemconfigure [lindex $pair 1] -fill black"
      }

      $c bind $id <Button-1> "$w.textarea tag remove tempsel 1.0 end; textsetcursor $w 1.[expr [lindex $pair 0]-1]" 
      $c bind $id <Shift-Button-1> "select_region $w \[$w.textarea index insert\] {1.[expr [lindex $pair 0]-1]}" 
      $c bind $id <Shift-Double-Button-1> "select_region $w \[$w.textarea index insert\] {1.[expr [lindex $pair 0]-1]}; event generate $c <<Return>>;break"
      lappend info($w,linked_events) [list $c $id <Button-1>] [list $c $id <Shift-Button-1>] [list $c $id <Shift-Double-Button-1>]
      
      scan [$c bbox $id] "%u %u %u %u" x1 y1 x2 y2
      if {$x1 < $minx} {set minx $x1}
      if {$y1 < $miny} {set miny $y1}
      if {$x2 > $maxx} {set maxx $x2}
      if {$y2 > $maxy} {set maxy $y2}
    }


    #draw features
    foreach {name tag} [feature_find $w] {
      if {[llength [set ranges [$w.textarea tag ranges $tag]]] > 2} {
        set start_index 2
      } else {
        set start_index 0
      }

      if {[string match fn* $tag]} {
        ##draw split tag
        regsub "fn" $tag "f" tag2
        set tag_color [$w.textarea tag cget $tag2 -background]
        set direction [lindex [$w.textarea tag bind $tag2 <<Metadata>>] 3]
        set start_angle [expr {90.0-(360.0*[string range [lindex $ranges 1] 2 end]/$total_length)}]
        set extent_angle [expr {360.0*([string range [lindex $ranges 1] 2 end]-[string range [lindex $ranges $start_index] 2 end])/$total_length}]
        if {$extent_angle < 0.9} {set extent_angle 0.9}
        if {$start_index == 2} {set extent_angle [expr {$extent_angle+360.0}]}
        $c create arc 110 110 290 290 -style arc -start $start_angle -extent $extent_angle -outline black
        if {$direction == 0} {
          set arrow [string range [lindex $ranges 1] 2 end]
          set angle1 [expr {2*$pi*$arrow/$total_length+.05}]
          set angle2 [expr {$angle1-0.14}]
        } else {
          set arrow [string range [lindex $ranges $start_index] 2 end]
          set angle1 [expr {2*$pi*$arrow/$total_length-.05}]
          set angle2 [expr {$angle1+0.14}]
        }
        set ranges [$w.textarea tag ranges $tag2]
      } else {
        #draw non-split tag
        set direction [lindex [$w.textarea tag bind $tag <<Metadata>>] 3]
        set tag_color [$w.textarea tag cget $tag -background]
        if {$direction == 0} {
          set arrow [string range [lindex $ranges 1] 2 end]
          set angle1 [expr {2*$pi*$arrow/$total_length+.05}]
          set angle2 [expr {$angle1-0.14}]
        } else {
          set arrow [string range [lindex $ranges $start_index] 2 end]
          set angle1 [expr {2*$pi*$arrow/$total_length-.05}]
          set angle2 [expr {$angle1+0.14}]
        }
      }
      foreach {start stop} $ranges {
        set start_angle [expr {90.0-(360.0*[string range $stop 2 end]/$total_length)}]
        set extent_angle [expr {360.0*([string range $stop 2 end]-[string range $start 2 end])/$total_length}]
        #bug in arc drawing for extents <0.45, on windows
        if {$extent_angle < 0.9} {set extent_angle 0.9}
        $c create arc 110 110 290 290 -style arc -width 5 -start $start_angle -extent $extent_angle -outline $tag_color
      }
      $c create polygon [expr {200.5 + 90*sin($angle1)}] [expr {200.5 - 90*cos($angle1)}] [expr {200.5 + 82*sin($angle2)}] [expr {200.5 - 82*cos($angle2)}] [expr {200.5 + 98*sin($angle2)}] [expr {200.5 - 98*cos($angle2)}] -fill $tag_color -width 1 -outline $tag_color
    }

    #make the enzyme sequence text box
    $c create text 200 190 -tags seqtext -justify center -anchor s -font "Helvetica $info(graphicfontsize)"

    #draw center info
    $c create text 200 200  -text "[file tail $info($w,filename)]" -justify center -font "Helvetica $info(graphicfontsize)"
    $c create text 200 [expr {205+$info(graphicfontsize)}]  -text "[expr $total_length] bp" -justify center -font "Helvetica $info(graphicfontsize)"
  } else {

 #######draw linear map
    $c create line 10 350 390 350
    set templist [linear_enz_sort $w [lsort -unique [concat $enz_list $extra_enz]] $start $end]

    #draw enzymes
    set maxx 400; set maxy 370; set minx 0; set miny 340

    set bump [expr -[font metrics graphicfont -linespace]]
    set xright 0
    set lastloc -1
    foreach pair $templist  {
      set x [expr {10 + 380 * ([lindex $pair 0] - $start) / $total_length}]
      if {$lastloc != [lindex $pair 0]} {
        set lineid [$c create line $x 340 $x 350 -tags [lindex $pair 1]]
      } else {
        $c itemconfigure $lineid -tags [concat [$c itemcget $lineid -tags] [lindex $pair 1]]
      }
      set lastloc [lindex $pair 0]
      
      if {$x > $xright} {
        set y 335
        set nexty [expr 335+$bump]
      } else {
        set y $nexty
        set nexty [expr $nexty+$bump]
      }
      while {[$c find overlapping $x [expr {$y-1}] [expr {$x+1}] [expr {$y-2}]] != ""} {
        set y $nexty
        set nexty [expr $nexty+$bump]      
      }
      set id [$c create text $x $y -text "[lindex $pair 0] [lindex $pair 1] \([llength $info($w,[lindex $pair 1])]\)" -anchor "sw" -font [list $info(graphicfontfamily) $info(graphicfontsize)] -tags [lindex $pair 1]] 
      if {[lsearch -exact $enz_list [lindex $pair 1]] > -1} {
        $c bind $id <Any-Enter> "$c raise [lindex $pair 1]; $c itemconfigure [lindex $pair 1] -fill red"
        $c bind $id <Any-Leave> "$c itemconfigure [lindex $pair 1] -fill black"
      } else {
        $c itemconfigure [lindex $pair 1] -fill gray50
        $c bind $id <Any-Enter> "$c raise [lindex $pair 1]; $c itemconfigure [lindex $pair 1] -fill blue"
        $c bind $id <Any-Leave> "$c itemconfigure [lindex $pair 1] -fill gray50"
      }

      $c bind $id <Any-Enter> "+$c itemconfigure seqtext -text $enzinfo(enz_pattern,[lindex $pair 1])"
      $c bind $id <Any-Leave> "+$c itemconfigure seqtext -text {}"
      $c bind $id <Button-1> "select_region $w 1.[expr [lindex $pair 0]-1] 1.[expr [lindex $pair 0]-1]" 
      $c bind $id <Shift-Button-1> "select_region $w \[$w.textarea index insert\] {1.[expr [lindex $pair 0]-1]}" 
      $c bind $id <Shift-Double-Button-1> "select_region $w \[$w.textarea index insert\] {1.[expr [lindex $pair 0]-1]}; event generate $c <<Return>>; break" 
      lappend info($w,linked_events) [list $c $id <Button-1>] [list $c $id <Shift-Button-1>] [list $c $id <Shift-Double-Button-1>]

      scan [$c bbox $id] "%u %u %u %u" x1 y1 x2 y2
      if {$y1 < $miny} {set miny $y1}
      if {$x2 > $maxx} {set maxx $x2}
      if {$y == 335} {set xright [expr {23 + $x2}]}
    }
    #draw file identifiers
    set y [expr {350+5}]
    set ystep [font metrics [list $info(graphicfontfamily) $info(graphicfontsize)] -displayof . -linespace]
    incr y $ystep
    $c create text 10 $y -text "[file tail $info($w,filename)] from [expr {1+$start}] to $end" -anchor "sw" -font [list $info(graphicfontfamily) $info(graphicfontsize)]
    incr y $ystep
    $c create text 10 $y  -tags seqtext -justify left -anchor sw -font [list $info(graphicfontfamily) $info(graphicfontsize)]
    incr y 2
    #draw features
    foreach {name tag} [feature_find $w $start $end] {
      ##need to check for splitting around circles- draw two features
      ## make drawing a proc and call it twice
      regsub "fn" $tag "f" tag
      linear_feature_draw $c $y $w $start $end $total_length $tag $tag
    }
  }

  $c move all [expr {-[lindex [$c bbox all] 0]}] [expr {-[lindex [$c bbox all] 1]}]
  $c configure -scrollregion [$c bbox all] -xscrollcommand "optionscrollbar $a.xscroll " -yscrollcommand "optionscrollbar $a.yscroll " -confine True
  #if {[set dist [expr $maxx-$minx]] < 500} {$c configure -width $dist}
  #if {[set dist [expr $maxy-$miny]] < 500} {$c configure -height $dist}
  bind $c <Configure> {
    set templist [%W bbox all]
    set newx [expr {([winfo width %W]-[lindex $templist 2]+[lindex $templist 0])/2}]
    set newy [expr {([winfo height %W]-[lindex $templist 3]+[lindex $templist 1])/2}]
    if {$newx < 0} {set newx 0}
    if {$newy < 0} {set newy 0}
    %W move all [expr {$newx-[lindex $templist 0]}] [expr {$newy-[lindex $templist 1]}]
    %W configure -scrollregion [%W bbox all]
  }
  grid rowconfigure $a 0 -weight 1
  grid columnconfigure  $a 0 -weight 1
  grid configure $c -row 0 -column 0 -sticky nsew
  grid configure [scrollbar $a.yscroll -command "$c yview " ] -row 0 -column 1 -sticky nes 
  grid configure  [scrollbar $a.xscroll -orient horizontal -command "$c xview"] -row 1 -column 0 -sticky wse


  set outputmenu [add_output_menu $c "Image" image $w]

  bind $c <MouseWheel> {%W yview scroll [expr {%D/-$info(mousewheelunits)}] units}
  bind $c <Shift-MouseWheel> {%W xview scroll [expr {%D/-$info(mousewheelunits)}] units}
  bind $a <MouseWheel> {[winfo toplevel %W].canvas yview scroll [expr {%D/-$info(mousewheelunits)}] units}
  bind $a <Shift-MouseWheel> {[winfo toplevel %W].canvas xview scroll [expr {%D/-$info(mousewheelunits)}] units}
  bind $a <Button-1> "focus $c" 
  wm deiconify $a
  window_on_screen $a
  focus $c
}

################
## Return a list of {site name} list pairs for a set of enzyme lists,with pairs sorted into quadrants, 2nd and 4th quadrants reversed in order
################
proc circ_enz_sort {w enz_list len} {
  global info

  set templist [list]
  set a1 -1
  set a2 -1
  set a3 -1
  foreach enzyme $enz_list {
    foreach site $info($w,$enzyme) {
      lappend templist [list $site $enzyme "$site $enzyme \([llength $info($w,$enzyme)]\)"]
    }
  }

  foreach {name text_tag} [feature_find $w] {
    if {[llength [$w.textarea tag ranges $text_tag]] == 2} {
      set tx1 [string range [lindex [$w.textarea tag ranges $text_tag] 0] 2 end]
    } else {
      set tx1 [string range [lindex [$w.textarea tag ranges $text_tag] 2] 2 end]
    }
    set tx2 [string range [lindex [$w.textarea tag ranges $text_tag] 1] 2 end]
    set ttg $text_tag
    regsub fn $text_tag f ttg
    if {[lindex [$w.textarea tag bind $ttg <<Metadata>>] 3]} {
      lappend templist [list [expr {$tx2}] $text_tag "$name [expr {$tx2}]...[expr {$tx1+1}]"]
    } else {
      lappend templist [list [expr {$tx1+1}] $text_tag "$name [expr {$tx1+1}]...[expr {$tx2}]"]
    }
  }

  set templist [lsort -integer -index 0 $templist]
  set i 0
  foreach element $templist {
    set index [lindex $element 0]
    if {$index < [expr .25 *$len]} {set a1 $i}
    if {$index < [expr .5 *$len]} {set a2 $i}
    if {$index < [expr .75 *$len]} {set a3 $i}
    incr i
  }
  return [concat [lsort -integer -index 0 -decreasing [lrange $templist -1 $a1]] [lrange $templist [expr $a1+1] $a2] [lsort -integer -index 0 -decreasing [lrange $templist [expr $a2+1] $a3]] [lrange $templist [expr $a3+1] $i]]
}

################
## Return a list of {site name} list pairs for a set of enzyme lists
################
proc linear_enz_sort {w enz_list start end} {
  global info
  set templist [list]
  foreach enzyme $enz_list {
    foreach site $info($w,$enzyme) {
      if {$site > $end} {break}
      if {$site >= $start} {lappend templist [list $site $enzyme]}
    }
  }
  set templist [lsort -integer -index 0 $templist]
}

################
## draw arrows on the linear graphic map
################
proc linear_feature_draw {c y w start end total_length tag {tags [list]}} {
  global info
  lappend tags temptag
  set ranges [feature_coords $w $tag $start $end]
  set tag_color [$w.textarea tag cget $tag -background]
  set direction [lindex [$w.textarea tag bind $tag <<Metadata>>] 3]
  set label [lindex [$w.textarea tag bind $tag <<Metadata>>] 0]
  if {$direction == 0} {
    #draw right pointing
    set x2 -1
    foreach range [lrange $ranges 0 end-1]  {
      set x1 [expr {10 + 380 * ([lindex $range 0] - $start) / $total_length}]
      if {($x2 > -1)} {
         #draw intron
        if {($x1 >= $x2)} {
          $c create line [expr {$x2+0}] 395 [expr {($x1+$x2)/2.0}] 405 $x1 395 -tag [concat $tags intron] -fill $tag_color -width 2
        } else {
          #split around the circle
          $c create line $x2 395 [expr {$x2+1}] 395 -tag $tags -fill $tag_color -width 4
          $c create line [expr {$x2+5}] 395 [expr {$x2+6}] 395 -tag $tags -fill $tag_color -width 4
          $c create line [expr {$x2+10}] 395 [expr {$x2+11}] 395 -tag $tags -fill $tag_color -width 4
          $c create line $x1 395 [expr {$x1-1}] 395 -tag $tags -fill $tag_color -width 4
          $c create line [expr {$x1-5}] 395 [expr {$x1-6}] 395 -tag $tags -fill $tag_color -width 4
          $c create line [expr {$x1-10}] 395 [expr {$x1-11}] 395 -tag $tags -fill $tag_color -width 4
        }
      }
      set x2 [expr {10 + 380 * ([lindex $range 1] - $start) / $total_length}]
      #draw exon
      $c create rectangle $x1 385 $x2 405 -tag [concat $tags exon] -fill $tag_color -outline black -width 1
    }
    set x1 [expr {10 + 380 * ([lindex [lindex $ranges end] 0] - $start) / $total_length}]
    #draw final intron
    if {($x2 > -1)} {
      if {($x1 >= $x2)} {
        $c create line [expr {$x2+0}] 395 [expr {($x1+$x2)/2.0}] 405 $x1 395 -tag [concat $tags intron] -fill $tag_color -width 2
      } else {
        #split around the circle
        $c create line $x2 395 [expr {$x2+1}] 395 -tag $tags -fill $tag_color -width 4
        $c create line [expr {$x2+5}] 395 [expr {$x2+6}] 395 -tag $tags -fill $tag_color -width 4
        $c create line [expr {$x2+10}] 395 [expr {$x2+11}] 395 -tag $tags -fill $tag_color -width 4
        $c create line $x1 395 [expr {$x1-1}] 395 -tag $tags -fill $tag_color -width 4
        $c create line [expr {$x1-5}] 395 [expr {$x1-6}] 395 -tag $tags -fill $tag_color -width 4
        $c create line [expr {$x1-10}] 395 [expr {$x1-11}] 395 -tag $tags -fill $tag_color -width 4
      }
    }
    #draw final exon
    set x2 [expr {10 + 380 * ([lindex [lindex $ranges end] 1] - $start) / $total_length}]
    set x3 [expr {$x2-15}]
    if {$x3 < $x1} {set x3 $x1}
    if {[expr {$x2-$x1}] < 4} {set x2 [expr {$x1+4}]}
    $c create polygon $x1 385 $x3 385 $x3 375 $x2 395 $x3 415 $x3 405 $x1 405 -tag [concat $tags exon] -fill $tag_color -outline black -width 1
    set x1 [expr {10 + 380 * ([lindex [lindex $ranges 0] 0] - $start) / $total_length}]
    #draw label
    $c create text $x1 405 -text "$label" -anchor nw -font [list $info(graphicfontfamily) $info(graphicfontsize) bold] -tag [concat $tags "text$tag"]
    $c create text $x1 [expr {410+$info(graphicfontsize)}] -text "([expr {1+[lindex [lindex $ranges 0] 0]}]...[expr {1+[lindex [lindex $ranges end] end]}])" -anchor nw -font [list $info(graphicfontfamily) $info(graphicfontsize) bold] -tag [concat $tags "text$tag"]
    $c bind $tag <Any-Enter> "$c itemconfigure text$tag -fill red;$c itemconfigure seqtext -text {[lindex [$w.textarea tag bind $tag <<Metadata>>] 0]}"
    $c bind $tag <Any-Leave> "$c itemconfigure text$tag -fill black;$c itemconfigure seqtext -text {}"
    $c bind $tag <Button-1> "select_region $w 1.[lindex [lindex $ranges 0] 0] 1.[lindex [lindex $ranges 0] 0]" 
    $c bind $tag <Shift-Button-1> "select_region $w \[$w.textarea index insert\] {1.[lindex [lindex $ranges 0] 0]}" 
    $c bind $tag <Shift-Double-Button-1> "select_region $w \[$w.textarea index insert\] {1.[lindex [lindex $ranges 0] 0]}; event generate $c <<Return>>; break" 
    lappend info($w,linked_events) [list $c $tag <Button-1>] [list $c $tag <Shift-Button-1>] [list $c $tag <Shift-Double-Button-1>]

    $c lower intron exon
  } else {
   #draw left pointing
    set x1 [expr {10 + 380 * ([lindex [lindex $ranges 0] 0] - $start) / $total_length}]
    set x2 [expr {10 + 380 * ([lindex [lindex $ranges 0] 1] - $start) / $total_length}]
    set x3 [expr {$x1+15}]
    if {$x3 > $x2} {set x3 $x2}
    if {[expr {$x2-$x1}] < 4} {set x1 [expr {$x2-4}]}
    $c create polygon $x2 385 $x3 385 $x3 375 $x1 395 $x3 415 $x3 405 $x2 405 -tag [concat $tags exon] -fill $tag_color -outline black -width 1
    foreach range [lrange $ranges 1 end]  {
      set x1 [expr {10 + 380 * ([lindex $range 0] - $start) / $total_length}]
      if {$x1 >= $x2} {
        $c create line [expr {$x2+1}] 395 [expr {($x1+$x2)/2.0}] 405 $x1 395 -tag [concat $tags intron] -fill $tag_color -width 2
      } else {
        #split around the circle
        $c create line $x2 395 [expr {$x2+1}] 395 -tag $tags -fill $tag_color -width 4
        $c create line [expr {$x2+5}] 395 [expr {$x2+6}] 395 -tag $tags -fill $tag_color -width 4
        $c create line [expr {$x2+10}] 395 [expr {$x2+11}] 395 -tag $tags -fill $tag_color -width 4
        $c create line $x1 395 [expr {$x1-1}] 395 -tag $tags -fill $tag_color -width 4
        $c create line [expr {$x1-5}] 395 [expr {$x1-6}] 395 -tag $tags -fill $tag_color -width 4
        $c create line [expr {$x1-10}] 395 [expr {$x1-11}] 395 -tag $tags -fill $tag_color -width 4
      }
      set x2 [expr {10 + 380 * ([lindex $range 1] - $start) / $total_length}]
      #draw exon
      $c create rectangle $x1 385 $x2 405 -tag [concat $tags exon] -fill $tag_color -outline black -width 1
    }
    #draw label
    $c create text $x2 405 -text "$label" -anchor ne -justify right -font [list $info(graphicfontfamily) $info(graphicfontsize) bold] -tag [concat $tags "text$tag"]
    $c create text $x2 [expr {410+$info(graphicfontsize)}] -text "([expr {1+[lindex [lindex $ranges end] end]}]...[expr {1+[lindex [lindex $ranges 0] 0]}])" -anchor ne -justify right -font [list $info(graphicfontfamily) $info(graphicfontsize) bold] -tag [concat $tags "text$tag"]
    $c bind $tag <Any-Enter> "$c itemconfigure text$tag -fill red;$c itemconfigure seqtext -text {[lindex [$w.textarea tag bind $tag <<Metadata>>] 0]}"
    $c bind $tag <Any-Leave> "$c itemconfigure text$tag -fill black;$c itemconfigure seqtext -text {}"
    $c bind $tag <Button-1> "select_region $w 1.[lindex [lindex $ranges end] end] 1.[lindex [lindex $ranges end] end]" 
    $c bind $tag <Shift-Button-1> "select_region $w \[$w.textarea index insert\] {1.[lindex [lindex $ranges end] end]}" 
    $c bind $tag <Shift-Double-Button-1> "select_region $w \[$w.textarea index insert\] {1.[lindex [lindex $ranges end] end]}; event generate $c <<Return>>; break" 
    lappend info($w,linked_events) [list $c $tag <Button-1>] [list $c $tag <Shift-Button-1>] [list $c $tag <Shift-Double-Button-1>]

    $c lower intron exon
  }

  set bump [expr {45+2*$info(graphicfontsize)}]
  set i 0
  while {([llength [eval [concat $c find overlapping [$c bbox temptag]]]] != [llength [$c find withtag temptag]]) && ($i < 9)} {
    $c move temptag 0 $bump
    incr i
  }
  $c dtag temptag temptag
}

################
## Select a region of the DNA textarea
################
proc select_region {w from_index to_index} {
  if {$from_index == 2.0} {set from_index 1.end}
  if {$to_index == 2.0} {set to_index 1.end}
  $w.textarea tag remove tempsel 1.0 end
  #$w.textarea tag raise tempsel
  $w.textarea tag remove sel 1.0 end
  #$w.textarea tag raise sel
  $w.textarea mark set insert $to_index
  $w.textarea mark set anchor $from_index
  if {[$w.textarea compare insert > anchor]} {
    $w.textarea tag add tempsel anchor insert
    $w.textarea tag add sel anchor insert
  } else {
    $w.textarea tag add tempsel insert anchor 
    $w.textarea tag add sel insert anchor 
  }
  $w.textarea see insert

  vertindexbalancer $w
  selection_manager $w
}

################
## Select from anchor to new x y position (mouse movement selections
################
proc text_select_xy {w x y} {
  global info

  #if {$y < 0} {
  #  vertscrollmanager $w scroll -2 units
  #} elseif {$y >= [winfo height $w.textarea]} {
  #  vertscrollmanager $w scroll 2 units
  #}

  set cur [text_closest_gap $w.textarea $x $y]
  if {$cur == 2.0} {
    set cur [$w.textarea index 1.end]
  }
  if {[catch {$w.textarea index anchor}]} {
    $w.textarea mark set anchor $cur
  }
  set anchor [$w.textarea index anchor]

  if {[$w.textarea compare $cur < anchor]} {
    set first $cur
    set last anchor
  } else {
    set first anchor
    set last $cur
  }

  $w.textarea tag remove sel 1.0 end
  $w.textarea tag remove tempsel 1.0 end
  $w.textarea mark set insert $cur
  $w.textarea tag add tempsel $first $last
  $w.textarea tag add sel $first $last
  selection_manager $w
  update idletasks
}

################
## scroll and select if mouse leaves DNA with button down
################
proc mouse_left_DNA {w x y} {
  global info

  #$w.infobox configure -text ""
  if {[winfo exists $w]} {
    set x [expr {[winfo pointerx $w]-[winfo rootx $w.textarea]}]
    set y [expr {[winfo pointery $w]-[winfo rooty $w.textarea]}]
    if {$y < 0} {
      vertscrollmanager $w scroll -2 units
      text_select_xy $w $x 3
    } elseif {$y >= [winfo height $w.textarea]} {
      vertscrollmanager $w scroll 2 units
      text_select_xy $w $x [expr {[winfo height $w.textarea]-5}]
    } else {
     # return
   }
    set info(cancel_scroll) [after 70 [list mouse_left_DNA $w $x $y]]
  } else {
    cancel_scrolling_DNA
  }
}

################
## Stop scrolling when mouse re-enters
################
proc cancel_scrolling_DNA {} {
  global info
  if {$info(cancel_scroll) != "" } {
    after cancel $info(cancel_scroll)
    set info(cancel_scroll) ""
  }
}

########
## deletes unneeded scrollbars from the grid manager, and sends scrollbar coordinates to the scrollbar if needed
########
proc optionscrollbar {s args} {
  if {$args == {0 0}} {return}
  if {$args == {0 1}} {
    if {[grid info $s] != {}} {
      grid remove $s
    }
  } else {
    if {[grid info $s] == ""} {
      grid configure $s
      update idletasks
    } 
  }
  eval [concat $s set $args]
}

########
## deletes unneeded scrollbars from the grid manager, and sends scrollbar coordinates to the scrollbar if needed
########
proc optionxscrollbar {s args} {
  if {$args == {0 0}} {return}
  regexp {([0-9]*)x([0-9]*)\+} [wm geometry [winfo toplevel $s]] g x y
  if {$args == {0 1}} {
    if {[grid info $s] != {}} {
      grid remove $s
      wm geometry [winfo toplevel $s] "$x\x[expr {$y-[$s cget -width]}]"
    }
  } else {
    if {[grid info $s] == ""} {
      wm geometry [winfo toplevel $s] "$x\x[expr {$y+[$s cget -width]}]"
      grid configure $s
      update idletasks
    } 
  }
  eval [concat $s set $args]
}

########
## deletes unneeded scrollbars from the grid manager, and sends scrollbar coordinates to the scrollbar if needed
########
proc optionyscrollbar {s args} {
  if {$args == {0 0}} {return}
  regexp {([0-9]*)x([0-9]*)\+} [wm geometry [winfo toplevel $s]] g x y
  if {$args == {0 1}} {
    if {[grid info $s] != {}} {
      grid remove $s
      wm geometry [winfo toplevel $s] "[expr {$x-[$s cget -width]-1}]x$y"
    }
  } else {
    if {[grid info $s] == ""} {
      wm geometry [winfo toplevel $s] "[expr {$x+[$s cget -width]+1}]x$y"
      grid configure $s
      update idletasks
    } 
  }
  eval [concat $s set $args]
}

##############
## toggle the comment textbox on and off
##############
proc toggle_comment {w} {
    if {[grid info $w.comment] == ""} {
      grid configure $w.comment
      if {[$w.cscroll get] != {0.0 1.0}} {
        grid configure $w.cscroll
      }
      update idletasks
    } else {
      grid remove $w.comment
      grid remove $w.cscroll
    }
}

##############
## open an abi file and create a new window
##############
proc open_abi {filename} {
  global abi_info abiwinum

  set childlist [winfo children .]
  set w ".abi_window$abiwinum"
  incr abiwinum
  set error_status [read_abi_info $filename $w]
  if {$error_status =="ok"} {
    new_abi_window $w
  } else {
    array unset abi_info $w*
    tk_messageBox -message "Error opening file:$error_status" -type ok -title "Error Opening File" -icon warning -default ok
  }
}

##############
## read the header info for an abi file
##############
proc read_abi_info {filename w} {
  global abi_info

  set error_status ok
  set record_list [list]
  set fid [open $filename r]
  fconfigure $fid -translation binary
  seek $fid 0
  set magnum [read $fid 4]
  if {$magnum != "ABIF"} {set error_status "Wrong Filetype- ABIF not found"}
  if {$error_status == "ok"} {
    seek $fid 18
    set bindata [read $fid 4]
    binary scan $bindata I numrecords
    seek $fid 26
    set bindata [read $fid 4]
    binary scan $bindata I records_offset

    seek $fid 0 end
    set file_size [tell $fid]
    if {[expr {$records_offset+28*$numrecords}]<= $file_size} {
      seek $fid $records_offset
      for {set i 0} {$i<$numrecords} {incr i} {
        binary scan [read $fid 28] a4ISSIIII tag_name tag_number data_type ele_len num_ele rec_len rec_data blank
        #set tag_name [string map {_ 0} $tag_name]
        lappend record_list [list $tag_name $tag_number $num_ele $rec_data]
        set abi_info($w,filename) $filename
      }
    } else {
      set error_status "Wrong Filetype- Data Past End of File"
    }
  }
  #for data of type 10:
  #binary scan [binary format I $rec_data] Scc year month day

  #for data of type 11:
  #binary scan [binary format I $rec_data] ccc hour min sec

  set a [clock clicks -milliseconds]
  foreach record $record_list {
    switch -glob -- [lindex $record 0] {
      DATA {#raw and processed data tracks
        if {[lindex $record 1] < 0 } {#set to <5 to read in raw data
          #raw data traces- not read at this time, see line above
          set track [expr [lindex $record 1]-1]
          seek $fid [lindex $record 3]
          for {set i 0} {$i<[lindex $record 2]} {incr i} {
            binary scan [read $fid 2] S point
            lappend abi_info($w,raw_data,$track) $point
          }
          #sputs "read raw data track #$track from [lindex $record 6] for $i points"
        } elseif {[lindex $record 1] < 9 } {
           #runtime data
        } else {
          #processed data
          set track [expr {[lindex $record 1]-9}]
          seek $fid [expr [lindex $record 3]]
          for {set i 0} {$i<[lindex $record 2]} {incr i} {
            binary scan [read $fid 2] S point
            lappend abi_info($w,processed_data,$track) $point
          }
         # sputs "read processed data track #$track from [lindex $record 6] for $i points"
        } ;#if/else
      } 
      FWO_ {#filter wheel order
        set abi_info($w,filter_wheel) [binary format I [lindex $record 3]]
      } 
      PBAS {#basecalls
        if {[lindex $record 1] == 1 } {
          set type edited
        } else {
          set type original
        }
        seek $fid [expr [lindex $record 3]]
        set abi_info($w,basecall,$type) [read $fid [lindex $record 2]]
      }
      PLOC {#basecalls positions
        if {[lindex $record 1] == 1 } {
          set type edited
        } else {
          set type original
        }
        seek $fid [lindex $record 3]
        for {set i 0} {$i<[lindex $record 2]} {incr i} {
          binary scan [read $fid 2] S point
          lappend abi_info($w,basecall_loc,$type) $point
        }
      }
      RUND {
        binary scan [binary format I [lindex $record 3]] Scc year month day
        set abi_info($w,run_date,[lindex $record 1]) "$month/$day/$year"
      }
      RUNT {
        binary scan [binary format I [lindex $record 3]]  ccc hour min sec
        set abi_info($w,run_time,[lindex $record 1]) "$hour:$min:$sec"
      }
      SMPL {
        seek $fid [expr [lindex $record 3]+1]
        set abi_info($w,sample_name) [read $fid [expr {[lindex $record 2]-1}]]
      }
      S/N% {
        seek $fid [expr [lindex $record 3]]
        binary scan [read $fid 8] SSSS abi_info($w,scale_track,0) abi_info($w,scale_track,1) abi_info($w,scale_track,2) abi_info($w,scale_track,3)
      }
      default {
        #sputs "untrapped data:$record [string match "FWO?" [lindex $record 0]]"
      }
    } ;#switch
  } ;#foreach
  if {$record_list == [list]} {set error_status "Wrong Filetype- No Header Read"}
  close $fid
  return $error_status
}

##############
## Make a new ABI trace window
##############
proc new_abi_window {w} {
  global abi_info
  global modstring modifier
  global info abiwinum


  set abi_info($w,vert_scale) 1
  set abi_info($w,horiz_scale) 1
  set abi_info($w,sequence) $abi_info($w,basecall,edited)
  set abi_info($w,rev_com) 0

  toplevel $w
  wm withdraw $w
  bind $w <FocusIn> "event generate . <<RaiseDialogs>>"
  bind $w <Map> "event generate . <<RaiseDialogs>>"
  bind $w <Activate> "event generate . <<RaiseDialogs>>"
  bind $w <$modifier-Key-n> "create_window"
  bind $w <$modifier-Key-q> "close_all_windows"
  wm protocol $w WM_DELETE_WINDOW "closewindow $w"
  wm title $w [file tail $abi_info($w,filename)]


  grid [canvas $w.c  -xscrollcommand "$w.xs set" -bg white -height 400] -row 0 -rowspan 2 -column 0 -sticky nswe
  #grid [scrollbar $w.ys -command "$w.c yview"] -row 0 -column 1 -sticky nswe
  grid [scale $w.ys -from 5 -to 1 -sliderrelief raised -showvalue 0 -resolution .1 -troughcolor white -background $info(bg_color) -width 10 -command "abi_scale $w"] -row 1 -column 1 -sticky nswe
  grid [scale $w.xs2 -from 5 -to 0.5 -sliderrelief raised -showvalue 0 -resolution .1 -troughcolor white -background $info(bg_color) -width 10 -command "abi_x_scale $w"] -row 1 -column 2 -sticky nswe
  $w.xs2 set $abi_info($w,horiz_scale)
  grid [label $w.yslabel -text y] -row 0 -column 1 -sticky nswe
  grid [label $w.xslabel -text x] -row 0 -column 2 -sticky nswe
  grid [scrollbar $w.xs -command "$w.c xview" -orient horizontal] -row 2 -column 0 -sticky nswe
  grid [label $w.label -text $abi_info($w,filename)] -row 3 -column 0 -sticky wes
  grid rowconfigure $w 1 -weight 1
  grid columnconfigure $w 0 -weight 1


  set b [menu $w.menubar]

  ##file menu
  $b add cascade -menu [menu $b.filemenu] -label "File"
  $b.filemenu add command -label [mc "Open"] -accelerator "$modstring+O" -command "open_file_dialog"
  bind $w <$modifier-KeyPress-o> "open_file_dialog"
  $b.filemenu add command -label [mc "Save Window"] -accelerator "$modstring+S" -command "abi_save $w.c"
  bind $w <$modifier-KeyPress-s> "abi_save $w.c"
  if {([lsearch -glob [info loaded] "*MacCarbonPrint.dylib*"] > -1) ||  ([lsearch -glob [info loaded] "*wmf.dll*"] > -1)} {
    $b.filemenu add command -label [mc "Page Setup"] -accelerator "$modstring+Shift+P" -command "page_setup_dialog $w"
    bind $w <$modifier-KeyPress-P> "page_setup_dialog $w"
  }
  $b.filemenu add command -label [mc "Print Window"] -accelerator "$modstring+P" -command "abi_print $w.c"
  bind $w <$modifier-KeyPress-p> "abi_print $w.c"
  $b.filemenu add command -label [mc "New DNA from basecalls"] -accelerator "$modstring+Shift+N" -command "create_window \$abi_info($w,sequence) \"\" \"\[file tail \$abi_info($w,filename)\] basecalls\" linear 0 1 \"\[file tail \$abi_info($w,filename)\] basecalls\""
  bind $w <Shift-$modifier-KeyPress-n> "create_window \$abi_info($w,sequence) \"\" \"\[file tail \$abi_info($w,filename)\] basecalls\" linear 0 1 \"\[file tail \$abi_info($w,filename)\] basecalls\""
  $b.filemenu add separator
  $b.filemenu add command -label [mc "File Info"] -command "abi_file_info $w"
  $b.filemenu add separator
  $b.filemenu add command -label [mc "Close"] -accelerator "$modstring+W" -command "closewindow $w"
  bind $w <$modifier-KeyPress-w> "closewindow $w"

  ##edit menu
  $b add cascade -menu [menu $b.editmenu] -label "Edit"
  if {([lsearch -glob [info loaded] "*wmf.dll*"] > -1)} {
    $b.editmenu add command -label [mc "Copy as Metafile"] -accelerator "$modstring+C" -command "event generate $w.c <<CopyEMF>>"
  }

  if {[tk windowingsystem] != "aqua"} {
    $b.editmenu add command -label [mc "Preferences..."] -command "configure_preferences $w" -accelerator "$modstring+D"
    bind $w <$modifier-KeyPress-d> "configure_preferences $w"
  }
  #$b.editmenu add command -label [mc "Find..."] -command "abi_find_dialog $w" -accelerator "$modstring+F"
  #bind $w <$modifier-KeyPress-f> "abi_find_dialog $w"
  if {[tk windowingsystem] != "aqua"} {
    $b.editmenu add checkbutton -label [mc "Reverse-Complement"] -command "abi_revcom $w" -variable abi_info($w,rev_com) -onvalue 1 -offvalue 0 -accelerator "$modstring+/"
  } else {
    $b.editmenu add checkbutton -label [mc "Reverse-Complement"] -command "abi_revcom $w" -variable abi_info($w,rev_com) -onvalue 1 -offvalue 0 
  }
  bind $w <$modifier-KeyPress-/> "set abi_info($w,rev_com) \[expr {!\$abi_info($w,rev_com)}\];abi_revcom $w"
  $b.editmenu add separator
  $b.editmenu add command -label [mc "Align..."] -command "align_dialog $w" -accelerator "$modstring+L"
  bind $w <$modifier-KeyPress-l> "align_dialog $w"
  
  ##windows menu
  $b add cascade -menu [menu $b.windows] -label "Windows"

  ##help menu
  $b add cascade -menu [menu $b.help] -label [mc "Help"]
  $b.help add command -label [mc "Standard Genetic Code"] -command "genetic_code_dialog $w"
  $b.help add command -label [mc "AA Info"] -command "aa_info_dialog $w"

  ##platform specific menu items
  if {[tk windowingsystem] != "aqua"} {
    #Exit is automatically on the apple menu in Aqua
    $b.filemenu add separator
    $b.filemenu add command -label [mc "Exit"] -accelerator "$modstring+Q" -command close_all_windows

    #About... goes on the apple menu in aqua
    $b.help add separator
    $b.help add command -label [mc "About ApE"] -command "about_dialog $w"
  } else {
    #define the apple menu
    menu $b.apple 
    $b add cascade -menu $b.apple
    $b.apple add command -label [mc "About ApE..."] -command "about_dialog $w"
    $b.apple add separator
    $b.apple add command -label [mc "Preferences..."]  -command "configure_preferences $w"
    .menubar.filemenu entryconfigure New -state disabled
    .menubar.filemenu entryconfigure "Open..." -state disabled


    ##set the new window coordinates
    wm geometry $w "+$info(newaquax)+$info(newaquay)"
    incr info(newaquax) 20
    incr info(newaquay) 20
    if {$info(newaquax) > 100} {set info(newaquax) 20}
    if {$info(newaquay) > 100} {set info(newaquay) 20}
  }
  bind $w <$modifier-KeyPress-q> "close_all_windows"

  $w configure -menu $b

  set outputmenu [menu $w.c.outputmenu]
  $outputmenu add command -label [mc "Save Window"] -accelerator "$modstring+S" -command "abi_save $w.c" 
  if {([lsearch -glob [info loaded] "*wmf.dll*"] > -1)} {
    $outputmenu add command -label [mc "Copy as Metafile"] -accelerator "$modstring+C" -command "event generate $w.c <<CopyEMF>>"
    bind $w.c <$modifier-KeyPress-c> "event generate $w.c <<CopyEMF>>"
    bind $w.c <<CopyEMF>> "abi_copy_metafile $w.c both"
  }
  $outputmenu add command -label [mc "Print Window"] -accelerator "$modstring+P" -command "abi_print $w.c"
  $outputmenu add command -label [mc "New DNA from basecalls"] -accelerator "$modstring+Shift+N" -command "create_window \$abi_info($w,sequence) \"\" \"\[file tail \$abi_info($w,filename)\] basecalls\" linear 0 1 \"\[file tail \$abi_info($w,filename)\] basecalls\""

  $outputmenu add command -label [mc "Align..."] -command "align_dialog $w" -accelerator "$modstring+L"
  if {[tk windowingsystem] != "aqua"} {
    $outputmenu add checkbutton -label [mc "Reverse-Complement"] -command "abi_revcom $w" -variable abi_info($w,rev_com) -onvalue 1 -offvalue 0 -accelerator "$modstring+/"
  } else {
    $outputmenu add checkbutton -label [mc "Reverse-Complement"] -command "abi_revcom $w" -variable abi_info($w,rev_com) -onvalue 1 -offvalue 0 
  }



  if {[tk windowingsystem] != "aqua"} {
    bind $w.c <Button-3> "tk_popup $outputmenu %X %Y"
  } else {
    bind $w.c <Button-2> "tk_popup $outputmenu %X %Y"
  }


  for {set track 0} {$track < 4} {incr track} {
    set i 0
    set trace [list]
    foreach point $abi_info($w,processed_data,$track) {
      lappend trace $i [expr {300-$point/8}]
      incr i
    } 
    $w.c create line $trace -fill $abi_info(colors,[string index $abi_info($w,filter_wheel) $track]) -tags [list scalable trace [string index $abi_info($w,filter_wheel) $track]]
  }
  set i 0
  foreach call [split $abi_info($w,basecall,edited) ""] x $abi_info($w,basecall_loc,edited) {
    $w.c create text $x 15 -text $call -anchor n -font "Courier 10" -width 0 -fill $abi_info(colors,$call) -tags [list basecall $call]
    incr i
    if {[expr {$i % 10}] == 0} {
      $w.c create text $x 0 -text $i -anchor n -font "Courier 10" -width 0 -fill black
    }
  }
  set trace_top [lindex [$w.c bbox trace] 1]
  set trace_bottom [lindex [$w.c bbox trace] 3]
  $w.c move trace 0 [expr {35-$trace_top}] 
  $w.c scale trace  0 35 1 [expr {355.0/($trace_bottom-$trace_top)}]
  $w.c configure -scrollregion [list 0 0 [lindex [$w.c bbox all] 2] 400]
  wm deiconify $w
  raise $w
  focus $w
  focus $w.c

  bind $w.c <Configure> "if {\[winfo height $w.c\] > 80} {
    $w.c scale scalable  0 35 1 \[expr {(\[winfo height $w.c\]-45.0)/(\[lindex \[$w.c bbox trace\] 3\]-\[lindex \[$w.c bbox trace\] 1\])*\$abi_info($w,vert_scale)}\] 
    $w.c scale selection 0 -2 1 \[expr {\[winfo height $w.c\]/\[lindex \[$w.c coords selection\] 3\]}\]
  }"
  bind $w <Any-MouseWheel> {[winfo toplevel %W].c xview scroll [expr {%D/-$info(mousewheelunits)}] units}
  bind $w.c <Shift-Button-3> {console show}
  $w.c create rectangle -2 -2 -2 [winfo height $w.c] -tags [list selection] -fill gray90
  $w.c lower selection
  bind $w.c <Button-1> "abi_click_select $w.c %x"
  bind $w.c <Shift-Button-1> "abi_shift_select $w.c %x"
 
  if {([tk windowingsystem] == "win32") && ([lsearch -glob [info loaded] "*libtkdnd.dll*"] > -1)} {
    dnd bindtarget $w Files <Drop> {HandleFileDrop  %D %T}
  }

  update_windows_menu
}

###########
## change the selection with button1
###########
proc abi_click_select {c mouse_x} {
  set x [$c canvasx $mouse_x]
  $c coords selection [$c canvasx $mouse_x] -2 [$c canvasx $mouse_x] [winfo height $c]
}

###########
## change the selection with shift-button1
###########
proc abi_shift_select {c mouse_x} {
  set x [$c canvasx $mouse_x]
  if {$x < [expr {([lindex [$c coords selection] 0]+[lindex [$c coords selection] 2])/2}]} {
    $c coords selection $x -2 [lindex [$c coords selection] 2] [winfo height $c]
  } else {
    $c coords selection [lindex [$c coords selection] 0] -2 $x [winfo height $c]
  }
}


###########
## change the selection with shift-button1
###########
proc abi_select_region {w from_c to_c} {
  global abi_info
  foreach {x0 i x1 j} [$w.c bbox all] {break}
  foreach {s0 i s1 j} [$w.c coords selection] {break}

  if {[expr {$from_c+1}] > [llength $abi_info($w,basecall_loc,edited)]} {
    set from_c [expr {[llength $abi_info($w,basecall_loc,edited)]-1}]
  }
  if {[expr {$to_c+1}] > [llength $abi_info($w,basecall_loc,edited)]} {
    set to_c [expr {[llength $abi_info($w,basecall_loc,edited)]-1}]
  }

  set a [lindex $abi_info($w,basecall_loc,edited) [expr {$from_c-1}]]
  set b [lindex $abi_info($w,basecall_loc,edited) $from_c]
  set c [lindex $abi_info($w,basecall_loc,edited) [expr {$from_c+1}]]
  
  if {$from_c == 0} {
    set loc_list [concat 0 [expr {($b+$c)/2}]]
  } elseif {[expr {$from_c+1}] == [llength $abi_info($w,basecall_loc,edited)]} {
    set loc_list [concat [expr {($a+$b)/2}] $x1]
  } else {
    set loc_list [list [expr {($a+$b)/2}] [expr {($b+$c)/2}]]
  }

  if {$to_c eq "select"} {
    set from_mid [expr {([lindex $loc_list 0]+[lindex $loc_list 1]) / 2.0}]
    set from_mid [expr {$abi_info($w,rev_com)?($x1-$abi_info($w,horiz_scale)*($from_mid+3)):($abi_info($w,horiz_scale)*$from_mid)}]
    set sel_mid [expr {($s0 + $s1) / 2.0}]
    if {$from_mid < $sel_mid} {
      $w.c coords selection [expr {$abi_info($w,rev_com)?($x1-$abi_info($w,horiz_scale)*([lindex $loc_list 1]+3)):($abi_info($w,horiz_scale)*[lindex $loc_list 0])}] -2 $s1 [winfo height $w.c]
    } else {
      $w.c coords selection $s0 -2 [expr {$abi_info($w,rev_com)?($x1-$abi_info($w,horiz_scale)*([lindex $loc_list 0]+3)):($abi_info($w,horiz_scale)*[lindex $loc_list 1])}] [winfo height $w.c]
    }
    $w.c xview moveto [expr {([lindex [$w.c coords selection] 0]-[winfo width $w.c]/2.0)/$x1}]
  } else {
    set a [lindex $abi_info($w,basecall_loc,edited) [expr {$to_c-1}]]
    set b [lindex $abi_info($w,basecall_loc,edited) $to_c]
    set c [lindex $abi_info($w,basecall_loc,edited) [expr {$to_c+1}]]

    if {$to_c == 0} {
      set loc_list [concat $loc_list 0 [expr {($b+$c)/2}]]
    } elseif {[expr {$to_c+1}] == [llength $abi_info($w,basecall_loc,edited)]} {
      set loc_list [concat $loc_list [expr {($a+$b)/2}] $x1]
    } else {
      set loc_list [concat $loc_list [expr {($a+$b)/2}] [expr {($b+$c)/2}]]
    }
    set loc_list [lsort -integer $loc_list]
    $w.c coords selection [expr {$abi_info($w,rev_com)?($x1-$abi_info($w,horiz_scale)*([lindex $loc_list 0]+3)):($abi_info($w,horiz_scale)*[lindex $loc_list 0])}] -2 [expr {$abi_info($w,rev_com)?($x1-$abi_info($w,horiz_scale)*([lindex $loc_list 3]+3)):($abi_info($w,horiz_scale)*[lindex $loc_list 3])}] [winfo height $w.c]
    $w.c xview moveto [expr {([lindex [$w.c coords selection] 0]-[winfo width $w.c]/2.0)/$x1}]
  }
}

###########
## get sequence and indicies can use pixel values, "start", "end", "sel.first", "sel.last"
###########
proc abi_get_seq {w from to} {
  global abi_info

  foreach {x0 i x1 j} [$w.c bbox all] {break}
  foreach {s0 i s1 j} [$w.c coords selection] {break}

  set from [string map "start $x0 end $x1 sel.first $s0 sel.last $s1" $from]
  set to [string map "start $x0 end $x1 sel.first $s0 sel.last $s1" $to]

  if {$abi_info($w,rev_com)} {
    set from [expr {($x1-$from) / $abi_info($w,horiz_scale)}]
    set to [expr {($x1-$to) / $abi_info($w,horiz_scale)}]
  } else {
    set from [expr {$from /$abi_info($w,horiz_scale)}]
    set to [expr {$to /$abi_info($w,horiz_scale)}]
  }

  if {$from > $to} {
    set i $from
    set from $to
    set to $i   
  }

  set i 0
  set j -1
  foreach index $abi_info($w,basecall_loc,edited) {
    if {$index < $from} {incr i}
    if {$index > $to} break
    incr j
  }

  if {$j < $i} {
    return [list]
  }
  if {$abi_info($w,rev_com)} {
    return [list [expr {1+$j}] [expr {1+$i}] [string range $abi_info($w,sequence) end-$j end-$i]]
  } else {
    return [list [expr {1+$i}] [expr {1+$j}] [string range $abi_info($w,sequence) $i $j]]
  } 
}

###########
## reverse complement an abi window
###########
proc abi_revcom {w} {
  global abi_info
  set c $w.c

  set x0 [lindex [$c bbox all] 0]
  set x1 [lindex [$c bbox all] 2]
  $c scale all [expr {($x1+$x0)/2}] 0 -1 1
  foreach call [$c find withtag basecall] {
    $c itemconfigure $call -text [string map {A T B V C G D H G C H D K M M K N N * * R Y S S T A V B W W Y R a t b v c g d h g c h d k m m k n n r y s s t a v b w w y r} [$c itemcget $call -text]]
  }
  set abi_info($w,sequence) [revcom $abi_info($w,sequence)]

  $w.c xview moveto [expr {1.0-[lindex [$w.xs get] 1]}]
}

###########
## change y scaling of the traces in an abi window
###########
proc abi_scale {w new_scale} {
  global abi_info
    set baseline [lindex [$w.c bbox trace] 3]
    set ratio [expr {$new_scale/$abi_info($w,vert_scale)}]
    $w.c scale trace  0 $baseline 1 $ratio
    set abi_info($w,vert_scale) $new_scale
}

###########
## change x scaling of the traces in an abi window
###########
proc abi_x_scale {w new_scale} {
  global abi_info
    set left [lindex [$w.c cget -scrollregion] 0] 
    set center [expr {([lindex [$w.xs get] 1] + [lindex [$w.xs get] 0]) / 2.0}]
    set ratio [expr {$new_scale/$abi_info($w,horiz_scale)}]
    $w.c scale all  0 0 $ratio 1
    $w.c configure -scrollregion [lreplace [$w.c cget -scrollregion] 2 2  [expr {$ratio* [lindex [$w.c cget -scrollregion] 2]}]]
    set abi_info($w,horiz_scale) $new_scale
    set half_width [expr {([lindex [$w.xs get] 1] - [lindex [$w.xs get] 0]) / 2.0/$ratio}]
    $w.c xview moveto [expr {$center-$half_width}]
}

###########
## abi file info dialog
###########
proc abi_file_info {w} {
  global abi_info
  global info

  if {[winfo exists $w.info] == 1} {
    wm deiconify $w.info
    raise $w.info
  } else {
    toplevel $w.info
    wm withdraw $w.info
    wm title $w.info "[mc {File Info:}]$abi_info($w,filename)"
    wm protocol $w WM_DELETE_WINDOW "closewindow $w"
    grid [text $w.info.text -height 5 -width 50] -row 0 -column 0 -sticky nsew
    grid rowconfigure $w.info 0 -weight 1
    grid columnconfigure $w.info 0 -weight 1
    $w.info.text insert end "[mc Sample:]$abi_info($w,sample_name) \n"
    $w.info.text insert end "[mc Scaling:][string index $abi_info($w,filter_wheel) 0]:$abi_info($w,scale_track,0),[string index $abi_info($w,filter_wheel) 1]:$abi_info($w,scale_track,1),[string index $abi_info($w,filter_wheel) 2]:$abi_info($w,scale_track,2),[string index $abi_info($w,filter_wheel) 3]:$abi_info($w,scale_track,3)\n"
    $w.info.text insert end "[mc {Run began:}][clock format [clock scan "$abi_info($w,run_date,1) $abi_info($w,run_time,1)"] -format "%l:%M%p %a,%b %d, %Y"]\n"
    $w.info.text insert end "[mc {Run ended:}][clock format [clock scan "$abi_info($w,run_date,2) $abi_info($w,run_time,2)"] -format "%l:%M%p %a,%b %d, %Y"]\n"
    wm deiconify $w.info 
  }
}

###########
## returns new canvas with copy of abi trace canvas c from x1 to x2.  
###########
proc abi_copy_region {fromc fromx1 fromx2 toc tox1 toy1 xscale yscale {clip_y 0}} {
  regexp {(-?[0-9]*) (-?[0-9]*) (-?[0-9]*) (-?[0-9]*)} [$fromc bbox all] l x0 fromy1 x3 fromy2


#copy the non-trace objects (should be all text)
  set objects [$fromc find enclosed [expr {$fromx1-15.0}] $fromy1 [expr {$fromx2+15.0}] $fromy2]
  foreach object $objects {
    switch [$fromc type $object] {
      text {
        regexp {([0-9]*.[0-9]) ([0-9]*.[0-9])} [$fromc coords $object] l textx texty
        $toc create text [expr {($textx-$fromx1)*$xscale+$tox1}] [expr {$texty*$yscale+$toy1}] -text [$fromc itemcget $object -text] -fill [$fromc itemcget $object -fill] -anchor [$fromc itemcget $object -anchor] -font [$fromc itemcget $object -font]
        #need to adjust scale if yscale<1
      }
      default {#sputs [$fromc type $object]}
    }
  }

 ##do all traces at once so you can do only the x region between fromx1 to fromx2

  set Aline [$fromc find withtag {trace && A}] 
  set Cline [$fromc find withtag {trace && C}] 
  set Gline [$fromc find withtag {trace && G}] 
  set Tline [$fromc find withtag {trace && T}] 
  set newAcoords [list]
  set newCcoords [list]
  set newGcoords [list]
  set newTcoords [list]
  foreach {xA yA} [$fromc coords $Aline] {xC yC} [$fromc coords $Cline] {xG yG} [$fromc coords $Gline] {xT yT} [$fromc coords $Tline] {
    if {$xA > $fromx2} break
    if {$xA > $fromx1} {
       if {$yA < $clip_y} {set yA $clip_y}
       lappend newAcoords [expr {($xA-$fromx1)*$xscale+$tox1}] [expr {$yA*$yscale+$toy1}]
       if {$yC < $clip_y} {set yC $clip_y}
       lappend newCcoords [expr {($xC-$fromx1)*$xscale+$tox1}] [expr {$yC*$yscale+$toy1}] 
       if {$yG < $clip_y} {set yG $clip_y}
       lappend newGcoords [expr {($xG-$fromx1)*$xscale+$tox1}] [expr {$yG*$yscale+$toy1}] 
       if {$yT < $clip_y} {set yT $clip_y}
       lappend newTcoords [expr {($xT-$fromx1)*$xscale+$tox1}] [expr {$yT*$yscale+$toy1}]     
    }
  }
  $toc create line $newAcoords -fill [$fromc itemcget $Aline -fill]
  $toc create line $newCcoords -fill [$fromc itemcget $Cline -fill]
  $toc create line $newGcoords -fill [$fromc itemcget $Gline -fill]
  $toc create line $newTcoords -fill [$fromc itemcget $Tline -fill]
}


###########
## save current view as eps file
###########
proc abi_save {c} {
  set regionx1 [$c canvasx [expr {[winfo x $c]+2}]]
  set regionx2 [$c canvasx [expr {[winfo x $c]+[winfo width $c]-4}]]
  set c2 [canvas .print_canvas -width [expr {$regionx2-$regionx1}] -height [winfo height $c]]
  abi_copy_region $c $regionx1 $regionx2 $c2 0 0 1.0 1.0
  canvas_save $c2
  destroy $c2
}

###########
## copy current view to clipboard as metafile
###########
proc abi_copy_metafile {c {mode both}} {
  set regionx1 [$c canvasx [expr {[winfo x $c]+2}]]
  set regionx2 [$c canvasx [expr {[winfo x $c]+[winfo width $c]-4}]]
  set c2 [canvas .print_canvas -width [expr {$regionx2-$regionx1}] -height [winfo height $c]]
  abi_copy_region $c $regionx1 $regionx2 $c2 0 0 1.0 1.0
  canvas_to_metafile copy $c2 both
  destroy $c2
}

###########
## print current view
###########
proc abi_print {c} {
  set regionx1 [$c canvasx [expr {[winfo x $c]+2}]]
  set regionx2 [$c canvasx [expr {[winfo x $c]+[winfo width $c]-4}]]
  set c2 [canvas .print_canvas -width [expr {$regionx2-$regionx1}] -height [winfo height $c]]
  abi_copy_region $c $regionx1 $regionx2 $c2 0 0 1.0 1.0
  print_canvas $c2 $c
  destroy $c2
}

###########
## find dialog
###########
proc abi_find_dialog {w} {
  global abi_info
  global info
  global ok dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  set ok 0
  set s [toplevel .find_dialog]
  wm title $s "[mc Find]..."
  wm protocol $s WM_DELETE_WINDOW {set ok -1}

  raise $s

  set wgeomlist [split [wm geometry $w] "x+"]
  set sgeom "+[expr [lindex $wgeomlist 2]+3]+[expr [lindex $wgeomlist 3]+150]"
  wm geometry $s $sgeom

  #label $s.parent_info -text "[mc Find]:"
  entry $s.find -width 50
  $s.find insert 0 $abi_info(last_find)
  frame $s.mod_frame 
    pack [checkbutton  $s.mod_frame.revcom -text [mc "also find rev-com of string"] -onvalue 1 -offvalue 0 -variable abi_info(find_revcom)] -side left
    pack [checkbutton  $s.mod_frame.exact -text [mc "as literal"] -onvalue 1 -offvalue 0 -variable abi_info(find_literal) ] -side left 
  frame $s.action_frame
    pack [button $s.action_frame.next -text [mc "find next"] -command "set ok 1" -default active] -side left 
    pack [button $s.action_frame.cancel -text [mc "cancel"] -command "set ok -1"] -side left 

  #pack $s.parent_info -side top -fill x -expand true
  pack $s.find -side top -fill x -expand true
  pack $s.mod_frame -side top -fill x -expand true
  pack $s.action_frame -side top -fill x -expand true
 
  grab set $s
  bind $s <Key-Return> "$s.action_frame.next invoke"
  focus $s
  focus $s.find

  vwait ok
  set dialogblock 0
  if {$ok >0} {
    set abi_info(last_find) [$s.find get]
    do_abi_find $w $abi_info(last_find) $abi_info(find_revcom) $abi_info(find_literal)
  }
  grab release $s
  destroy $s
}

############
## Find text in a trace window
############
proc do_abi_find {w search_text revcom find_literal} {
  global abi_info

  set left_x [$w.c canvasx [winfo x $w.c]]
  set i 0
  foreach x $abi_info($w,basecall_loc,edited) {
    if {$x>= $left_x} break
    incr i
  }
  set text [string range $abi_info($w,basecall,edited) $i end]
  set search_text [string toupper $search_text]
  if {$find_literal == 0} {
    set search_text [string map {N .} $search_text]
  }
  if {$revcom == 1} {
    set search_text "$search_text|[revcom_regexp $search_text]"
  }
#sputs $search_text
  if {[regexp -indices $search_text $text index] == 1} {
    set new_x [lindex $abi_info($w,basecall_loc,edited) [expr {[lindex $index 0]+$i}]]
    set trace_len [llength $abi_info($w,processed_data,0)]
    set half_width [expr {[winfo width $w.c]/2}]
    if {$new_x < $half_width} {set new_x $halfwidth}
    if {$new_x > [expr {$trace_len-$half_width}]} {set new_x [expr {$trace_len-$half_width}]}
    $w.c xview moveto [expr {($new_x-$half_width+0.0)/$trace_len}]
    #$w.c delete crosshair
    #$w.c create line $new_x 0 $new_x [winfo height $w.c] -tags crosshair
    $w.c move crosshair [expr {$new_x-[lindex [$w.c bbox crosshair] 0]}] 0
  } else {
    bell
  }
}

##############
## show all sites that can be generated keeping the same translation
##############
proc silent_sites {w selected_enzymes} {
  global info
  global enzinfo
  global enzymes
  global modstring modifier

  if {[$w.textarea tag ranges sel] == ""} {
    set start 0
    regexp "\\.(.*)" [$w.textarea index end-1chars] fill end
  } else {
    regexp "\\.(.*)" [$w.textarea index sel.first] fill start 
    regexp "\\.(.*)" [$w.textarea index sel.last] fill end
  }

  if {$selected_enzymes == [list]} {
    tk_messageBox -message [mc "No enzymes selected, searching all enzymes."] -type ok -icon info -default ok
    set selected_enzymes $enzymes
  }

  if {[array names enzinfo 3ftrxl*] == [list]} {
    foreach enzyme $enzymes {
      set enzinfo(3ftrxl,$enzyme) [degen_transl $enzinfo(flatpat,$enzyme)]
    }
  }

  findenzymes $w

  set DNA [$w.textarea get 1.$start 1.$end]
  incr start
  incr end
  set AA [translate $DNA]
  set templist [list]

  foreach enzyme $selected_enzymes {
    foreach trxl $enzinfo(3ftrxl,$enzyme) frame [list 2 1 0] {
      foreach pair [regexp -indices -all -inline $trxl $AA] {
        if {[lsearch $info($w,$enzyme) [expr {$start+3*[lindex $pair 0]+$frame}]] == -1} {
          lappend templist [list $enzyme $frame [expr {$start+3*[lindex $pair 0]+$frame}] [string range $DNA [expr {3*[lindex $pair 0]}] [expr {2+3*[lindex $pair 1]}]]]
        }
      }
    }
  }

  set a [new_analysis_window $w "Silent Sites"]
  create_textframe $a 40
  $a.textframe.text configure -tabs 2c

  $a.textframe.text insert end "[clock format [clock seconds] -format $info(clock_format)]\n"
  $a.textframe.text insert end "$info($w,win_title) \n"
  $a.textframe.text insert end "[mc From] $start [mc to] [expr {$end-1}]\. \n"
  $a.textframe.text insert end "Silent Sites\n\n"

  set templist [lsort -integer -index 2 $templist]
  foreach site $templist {
    $a.textframe.text insert end "[expr {[lindex $site 2]-[lindex $site 1]}]\t" {} "[regexp -inline -all {(?:...)} [lindex $site 3]]" dnatag "\n"
    $a.textframe.text insert end "[lindex $site 0]\t[join [regexp -inline -all {(?:...)} "[string repeat " " [lindex $site 1]]$enzinfo(flatpat,[lindex $site 0])   "] " "]\n"
    $a.textframe.text insert end "\t [translate [lindex $site 3] 1 1 3]\n\n"
  }
  if {$templist == ""} {
    $a.textframe.text insert end "No sites Found"
  }
  $a.textframe.text tag bind dnatag <Double-Button> "alignlink $a.textframe.text $w \[$a.textframe.text index @%x,%y\] 1 goto 1; break"
  $a.textframe.text tag bind dnatag <$modifier-Button> "alignlink $a.textframe.text $w \[$a.textframe.text index @%x,%y\] 1 goto 0; break"
  $a.textframe.text tag bind dnatag <Shift-$modifier-Button> "alignlink $a.textframe.text $w \[$a.textframe.text index @%x,%y\] 1 select 0; break"
  $a.textframe.text tag bind dnatag <Shift-Double-Button> "alignlink $a.textframe.text $w \[$a.textframe.text index @%x,%y\] 1 select 1; break"
  lappend info($w,linked_events) [list $a.textframe.text dnatag <Double-Button>]
  lappend info($w,linked_events) [list $a.textframe.text dnatag <$modifier-Button>]
  lappend info($w,linked_events) [list $a.textframe.text dnatag <Shift-$modifier-Button>]
  lappend info($w,linked_events) [list $a.textframe.text dnatag <Shift-Double-Button>]
  $a.textframe.text tag bind dnatag <Any-Leave> {%W tag remove highlite 1.0 end}
  $a.textframe.text tag bind dnatag <Any-Motion> {%W tag remove highlite 1.0 end; %W tag add highlite current}
  $a.textframe.text tag configure highlite -background red

  bind $a <MouseWheel> {[winfo toplevel %W].text yview scroll [expr {%D/-$info(mousewheelunits)}] units}
  set outputmenu [add_output_menu $a.textframe.text "Text" text $w]
  bind $a <Button-1> "focus $a.textframe.text" 
  bindtags $a.textframe.text [list $a.textframe.text AnalysisText . all]
  wm deiconify $a
  window_on_screen $a
  focus $a.textframe.text
} 

##############
## Make an orf map
##############
proc orf_map {w region} {
  global enzymes
  global info
  global modstring
  global modifier

  if {[$w.textarea tag ranges tempsel] == ""} {set region "all"}
  if {$region == "all"} {
    set start 0
    regexp "\\.(.*)" [$w.textarea index end-1chars] fill total_length
    set end $total_length
  } else {
    regexp "\\.(.*)" [$w.textarea index tempsel.first] fill start
    regexp "\\.(.*)" [$w.textarea index tempsel.last] fill end
    set total_length [expr {$end - $start+1}]
  }
  if {$total_length == 0} return
  set text [$w.textarea get 1.$start 1.$end]

  incr start

  foreach frame [list 0 1 2] {
    set metlist($frame) [list]
    set stoplist($frame) [list]
    set temlist($frame) [list]
    set potslist($frame) [list]
  }

  foreach met [regexp -inline -all -nocase -indices {ATG} $text] {
    lappend metlist([expr {[lindex $met 0] % 3}]) [list met [lindex $met 0]]
  }
  foreach stop [regexp -inline -all -nocase -indices {TAG|TAA|TGA} $text] {
    lappend stoplist([expr {[lindex $stop 0] %3}]) [list stop [lindex $stop 0]]
  }
  foreach tem [regexp -inline -all -nocase -indices {CAT} $text] {
    lappend temlist([expr {[lindex $tem 0] % 3}]) [list tem [lindex $tem 0]]
  }
  foreach pots [regexp -inline -all -nocase -indices {CTA|TTA|TCA} $text] {
    lappend potslist([expr {[lindex $pots 0] % 3}]) [list pots [lindex $pots 0]]
  }

  set a [new_analysis_window $w "ORF Map"]  

  grid [canvas $a.c -yscrollcommand "optionscrollbar $a.sy" -xscrollcommand "optionscrollbar $a.sx" -width 700 -height 450 -background white] -row 0 -column 0 -sticky nswe
  grid [scrollbar $a.sy -orient vertical -command "$a.c yview"] -row 0 -column 1 -sticky nswe
  grid [scrollbar $a.sx -orient horizontal -command "$a.c xview"] -row 1 -column 0 -sticky nswe
  grid rowconfigure $a 0 -weight 1
  grid columnconfigure $a 0 -weight 1

  set scale [expr {1.0/5}]
  set tick_incr 100
  if {$total_length <= 1000} {
    set scale [expr {1.0/2}]
    set tick_incr 10
    if {$total_length <= 100} {
      set scale [expr {2.0}]
      set tick_incr 1
    }
  }

  ##make the outer box and horizontal lines
  set x_right [expr {50+$total_length*$scale}]
  $a.c create rectangle 50 100 $x_right 400
  foreach y [list 150 200 250 300 350] {
    $a.c create line 50 $y $x_right $y
  }
  $a.c create text 50 405 -text "$info($w,filename)" -anchor nw 
  $a.c create text 50 417 -text "ORF Map from $start to $end" -anchor nw 

  ##make the top row tick marks
  set x [expr {$start-(($start-1) % $tick_incr)-1+$tick_incr}]
  while {$x <= $end} {
    set x_pos [expr {50+($x-$start)*$scale}]
    if {[expr {$x %(5*$tick_incr)}] == 0} {
      $a.c create line $x_pos 90 $x_pos 100
      if {[expr {$x % (10*$tick_incr)}] == 0} {
        $a.c create text $x_pos 87 -text $x -anchor s
      }
    } else {
      $a.c create line $x_pos 95 $x_pos 100
    }
    incr x $tick_incr
  }

  ##now put in the met and stops
  foreach frame [list 0 1 2] {
    set fwd [lsort -integer -index 1 [concat $metlist($frame) $stoplist($frame)]]
    foreach item $fwd {
      set x_pos [expr {50+[lindex $item 1]*$scale}]
      if {[lindex $item 0] == "met"} {
        $a.c create line $x_pos [expr {125+50*$frame}] $x_pos [expr {150+50*$frame}]
      } else {
        $a.c create line $x_pos [expr {100+50*$frame}] $x_pos [expr {150+50*$frame}] 
      }
    }
    set rev [lsort -integer -index 1 [concat $temlist($frame) $potslist($frame)]]
    foreach item $rev {
      set x_pos [expr {50+[lindex $item 1]*$scale}]
      if {[lindex $item 0] == "tem"} {
        $a.c create line $x_pos [expr {250+50*$frame}] $x_pos [expr {275+50*$frame}]
      } else {
        $a.c create line $x_pos [expr {250+50*$frame}] $x_pos [expr {300+50*$frame}] 
      }
    }
  }
  $a.c configure -scrollregion [$a.c bbox all]

  set outputmenu [add_output_menu $a.c "Image" image $w]
  bind $a <Button-1> "focus $a.c" 
  wm deiconify $a
  window_on_screen $a
  focus $a.c
}

##############
## Translation analysis window
##############
proc translation_window {w} {
  global info
  global modifier
  global modstring
  global ok dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1


  toplevel .dialog
  wm title .dialog [mc "Translate..."]
  wm protocol .dialog WM_DELETE_WINDOW "set ok -1"
  bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry .dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .dialog $w
  } else {
    ::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
  }
  set ok 0
  grid [frame .dialog.f1 -relief ridge -borderwidth 3] -row 1 -column 0 -sticky nwe
  grid [label .dialog.f1.label -text [mc "Code:"]] -row 0 -column 0 -sticky w
  grid [radiobutton .dialog.f1.1let -text [mc "1 Letter"] -variable info(translation_window_code) -value 1  -selectcolor white -activebackground $info(bg_color)] -row 0 -column 1 -sticky w
  grid [radiobutton .dialog.f1.3let -text [mc "3 Letter"] -variable info(translation_window_code) -value 3 -selectcolor white -activebackground $info(bg_color)] -row 0 -column 2 -sticky w
  grid [checkbutton .dialog.f1.spacing -text [mc "Codon Spacing"] -variable info(translation_window_spacing) -onvalue 1 -offvalue 0 -selectcolor white -activebackground $info(bg_color)] -row 1 -column 0 -columnspan 2 -sticky w
  grid [checkbutton .dialog.f1.revcom -text [mc "Reverse Complement"] -variable info(translation_window_revcom) -onvalue 1 -offvalue 0 -selectcolor white -activebackground $info(bg_color)] -row 1 -column 2 -sticky w

  grid [frame .dialog.f2 -relief ridge -borderwidth 3] -row 2 -column 0  -sticky nwe
  grid [label .dialog.f2.label1 -text [mc "Line Width:"]] -row 0 -column 0 -sticky w
  grid [entry .dialog.f2.width -width 4 -textvariable info(translation_window_width) -validate key -vcmd "check_char %S"] -row 0 -column 1 -columnspan 3 -sticky w
  grid [label .dialog.f2.label2 -text [mc "Line Numbers:"]] -row 1 -column 0 -sticky w
  grid [radiobutton .dialog.f2.lnone -text [mc "None"] -variable info(translation_window_line_numbers) -value none -selectcolor white -activebackground $info(bg_color)] -row 1 -column 1 -sticky w
  grid [radiobutton .dialog.f2.lleft -text [mc "Left"] -variable info(translation_window_line_numbers) -value left -selectcolor white -activebackground $info(bg_color)] -row 1 -column 2 -sticky w
  grid [radiobutton .dialog.f2.lright -text [mc "Right"] -variable info(translation_window_line_numbers) -value right -selectcolor white -activebackground $info(bg_color)] -row 1 -column 3 -sticky w
  grid [radiobutton .dialog.f2.lboth -text [mc "Both"] -variable info(translation_window_line_numbers) -value both -selectcolor white -activebackground $info(bg_color)] -row 1 -column 4 -sticky w

  grid [frame .dialog.f3 -relief ridge -borderwidth 3] -row 3 -column 0  -sticky nwe
  #grid [checkbutton .dialog.f3.revcom -text [mc "Index Line"] -variable info(translation_window_index_line) -onvalue 1 -offvalue 0 -selectcolor white -activebackground $info(bg_color)] -row 1 -column 0 -columnspan 2 -sticky w
  grid [label .dialog.f3.label1 -text [mc "DNA:"]] -row 2 -column 0 -sticky w
  grid [radiobutton .dialog.f3.dnone -text [mc "None"] -variable info(translation_window_dna_line) -value none -selectcolor white -activebackground $info(bg_color)] -row 2 -column 1 -sticky w
  grid [radiobutton .dialog.f3.dabove -text [mc "Above"] -variable info(translation_window_dna_line) -value above -selectcolor white -activebackground $info(bg_color)] -row 2 -column 2 -sticky w
  grid [radiobutton .dialog.f3.dbelow -text [mc "Below"] -variable info(translation_window_dna_line) -value below -selectcolor white -activebackground $info(bg_color)] -row 2 -column 3 -sticky w
  grid [checkbutton .dialog.f3.dcopyhigh -text [mc "Copy Highlight"] -variable info(translation_window_copy_highlight) -onvalue 1 -offvalue 0 -selectcolor white -activebackground $info(bg_color)] -row 3 -column 0 -sticky w 

  grid [button .dialog.ok -command {set ok 1} -text [mc "OK"] -default active] -row 4 -column 0 -sticky n

  update idletasks
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .dialog] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry .dialog "+$winx+$winy"

  vwait ok
  set dialogblock 0
  destroy .dialog
  bind . <<RaiseDialogs>> ""
  if {$ok != 1} {return}
  unset ok
  if {$info(translation_window_width) == ""} {set info(translation_window_width) 20}
  if {[$w.textarea tag ranges sel] == ""} {
    set start 0
    regexp "\\.(.*)" [$w.textarea index end-1chars] fill end
  } else {
    regexp "\\.(.*)" [$w.textarea index sel.first] fill start 
    regexp "\\.(.*)" [$w.textarea index sel.last] fill end
  }

  set a [new_analysis_window $w "Translation"]

  create_textframe $a 120

  set dna [$w.textarea get 1.$start 1.$end]
  incr start
  if {$info(translation_window_revcom)} {
    set dna [revcom $dna]
    set start [expr {$start+([string length $dna] % 3)}]
  } else {
    set end [expr {$end-([string length $dna] % 3)}]
  }
  regexp {(^[^*]*)\*?(.*)} [translate $dna 1 1 0] temp mwaa mwxs
  set dna [regexp -inline {(?:...)*} $dna]
  set spacing 0

  if {$info(translation_window_spacing)} {
    if {($info(translation_window_code) == 1) && ($info(translation_window_dna_line) != "none")} {
      set spacing 3
    } else {
      set spacing 1
    }
  } elseif {($info(translation_window_code) == 1) && ($info(translation_window_dna_line) != "none")} {
    set spacing 2
  }

  set protein [translate $dna $info(translation_window_code) 1 $spacing]

  if {$info(translation_window_spacing)} {
    set dna [regexp -all -inline {(?:...)} $dna]
  }
  if {($info(translation_window_code) == 1) && ($info(translation_window_dna_line) != "none")} {
    set protein " $protein"
  }
  if {($info(translation_window_code) == 1) && ($info(translation_window_dna_line) != "none")} {
    set protein "$protein "
  }

  if {($info(translation_window_code) == 1)} {
    incr spacing 1
  } else {
    incr spacing 3
  }
    
  $a.textframe.text insert end "[clock format [clock seconds] -format $info(clock_format)]\n"
  $a.textframe.text insert end "$info($w,win_title) \n"
  $a.textframe.text insert end "$info($w,filename) \n"
  if {$info(translation_window_revcom)} {
    $a.textframe.text insert end "From $end to $start\. \n"
  } else {
    $a.textframe.text insert end "From $start to $end\. \n"
  }
  $a.textframe.text insert end "Translation"

  set mw [expr "[string map -nocase {X +0.0 G +57.05 A +71.09 S +87.08 T +101.11 C +103.15 V +99.14 L +113.16 I +113.16 M +131.19 P +97.12 F +147.18 Y +163.18 W +186.21 D +115.09 E +129.12 N +114.11 Q +128.14 H +137.14 K +128.17 R +156.19 * +0.0} $mwaa]+18.02"]
  $a.textframe.text insert end "[string length $mwaa] a.a. MW=$mw\n"
  if {$mwxs!=""} {
    $a.textframe.text insert end "???([string length $mwxs] extra codons after stop)\n"
  } else {
    $a.textframe.text insert end "\n"
  }

  for {set x 0} {[expr {$x*($spacing)}]< [string length $protein]} {incr x $info(translation_window_width)} {
    set width $info(translation_window_width)
    set pline [string range $protein [expr {$x*($spacing)}] [expr {($width+$x)*($spacing)-1}]]
    set dline [string range $dna [expr {$x*($spacing)}] [expr {($width+$x)*($spacing)-1}]]
    set linestart [expr {$info(translation_window_revcom)?($end-3*$x):($start+3*$x)}]
    if {[string length $dline] == [expr {$width*$spacing}]} {
      set lineend [expr {$info(translation_window_revcom)?($end-3*($x+$width)+1):($start+3*($x+$width)-1)}]
    } else {
      set lineend [expr {$info(translation_window_revcom)?$start:$end}]
    }
    if {$info(translation_window_dna_line) == "above"} { 
      if {($info(translation_window_line_numbers) == "left") ||($info(translation_window_line_numbers) == "both")} {
        $a.textframe.text insert end "[format {%7u } $linestart]"
      }
      set dlinestartindex [$a.textframe.text index "end -1char"]
      $a.textframe.text insert end "$dline" [list linktag "ls$linestart\_[expr {$info(translation_window_revcom)?-1:1}]"]
      if {$info(translation_window_copy_highlight)} {
        copy_tags $w.textarea [list 1.[expr {$linestart-1}] 1.[expr {$lineend-1}]] $a.textframe.text $dlinestartindex [expr {$info(translation_window_revcom)?"reverse":"forward"}] ""
      }
      if {($info(translation_window_line_numbers) == "right") ||($info(translation_window_line_numbers) == "both")} {
        $a.textframe.text insert end "[format { %-7u} $lineend]\n"
      } else {
        $a.textframe.text insert end "\n"
      }
    }

    if {($info(translation_window_line_numbers) == "left") ||($info(translation_window_line_numbers) == "both")} {
      $a.textframe.text insert end "[format {%7u } [expr {$x+1}]]"
    }
    $a.textframe.text insert end "$pline" [list linktag "ls$linestart\_[expr {$info(translation_window_revcom)?(($info(translation_window_code)==1)?-3:-1):(($info(translation_window_code)==1)?3:1)}]"]
    if {($info(translation_window_line_numbers) == "right") ||($info(translation_window_line_numbers) == "both")} {
      $a.textframe.text insert end "[format { %-7u} [expr {$x+[string length $pline]/$spacing}]]\n"
    } else {
      $a.textframe.text insert end "\n"
    }

    if {$info(translation_window_dna_line) == "below"} { 
      if {($info(translation_window_line_numbers) == "left") ||($info(translation_window_line_numbers) == "both")} {
        $a.textframe.text insert end "[format {%7u } $linestart]"
      }
      set dlinestartindex [$a.textframe.text index "end -1char"]
      $a.textframe.text insert end "$dline" [list linktag "ls$linestart\_[expr {$info(translation_window_revcom)?-1:1}]"]
      if {$info(translation_window_copy_highlight)} {
        copy_tags $w.textarea [list 1.[expr {$linestart-1}] 1.[expr {$lineend-1}]] $a.textframe.text $dlinestartindex [expr {$info(translation_window_revcom)?"reverse":"forward"}] ""
      }
      if {($info(translation_window_line_numbers) == "right") ||($info(translation_window_line_numbers) == "both")} {
        $a.textframe.text insert end "[format { %-7u} $lineend]\n"
      } else {
        $a.textframe.text insert end "\n"
      }
    }
    if {($info(translation_window_dna_line) != "none") || ($info(translation_window_index_line))} {
      $a.textframe.text insert end "\n"
    }
  }

  foreach tag [$w.textarea tag names] {
     if {[lsearch -exact [$a.textframe.text tag names] $tag] > -1} {
       $a.textframe.text tag raise $tag
     }
  }
  $a.textframe.text tag bind linktag <Double-Button> "alignlink2 $a.textframe.text $w \[$a.textframe.text index @%x,%y\] goto 1; break"
  $a.textframe.text tag bind linktag <Shift-Double-Button> "alignlink2 $a.textframe.text $w \[$a.textframe.text index @%x,%y\] select 1; break"
  $a.textframe.text tag bind linktag <$modifier-Button> "alignlink2 $a.textframe.text $w \[$a.textframe.text index @%x,%y\] goto 0; break"
  $a.textframe.text tag bind linktag <Shift-$modifier-Button> "alignlink2 $a.textframe.text $w \[$a.textframe.text index @%x,%y\] select 0; break"
  $a.textframe.text tag bind linktag <Any-Leave> {%W tag remove highlite 1.0 end}
  $a.textframe.text tag bind linktag <Any-Motion> {%W tag remove highlite 1.0 end; %W tag add highlite current}
  $a.textframe.text tag configure highlite -background red
  lappend info($w,linked_events) [list $a.textframe.text linktag <Double-Button>]
  lappend info($w,linked_events) [list $a.textframe.text linktag <$modifier-Button>]
  lappend info($w,linked_events) [list $a.textframe.text linktag <Shift-$modifier-Button>]
  lappend info($w,linked_events) [list $a.textframe.text linktag <Shift-Double-Button>]

  bind $a <MouseWheel> {[winfo toplevel %W].textframe.text yview scroll [expr {%D/-$info(mousewheelunits)}] units}
  set outputmenu [add_output_menu $a.textframe.text "Text" text $w]
  bind $a <Button-1> "focus $a.textframe.text" 
  bindtags $a.textframe.text [list $a.textframe.text AnalysisText . all]
  wm deiconify $a
  window_on_screen $a
  focus $a.textframe.text
}


##############
## save a canvas as encapsulated postscript or enhanced metafile
##############
proc canvas_save {c} {
  global tcl_platform ps_fontmap

  if {[set coords [$c bbox all]] != [list]} {
    set filetypes {{"Encapsulated Postscript" {.eps}} {"Scalable Vector Graphics" {.svg}}} 
    if {$tcl_platform(platform) == "windows"} {
      lappend filetypes {"Enhanced Metafile" {.emf}}
    }

    set filename [tk_getSaveFile -defaultextension ".eps" -title [mc "Save Image"] -filetypes $filetypes -parent [winfo toplevel $c]]

    if {$filename == ""} {return}
    if {[file extension $filename] == ""} {set filename "$filename\.eps"}

    switch [file extension $filename] {
      .eps {
        set x1 [lindex $coords 0]
        set y1 [lindex $coords 1]
        set x2 [lindex $coords 2]
        set y2 [lindex $coords 3]
        array unset ps_fontmap *
        get_canv_fontmap $c ps_fontmap
        $c postscript -file $filename -fontmap ps_fontmap -height [expr ($y2-$y1)] -width [expr ($x2-$x1)] -x $x1 -y $y1
      } 
      .emf {      
        if {([lsearch -glob [info loaded] "*wmf.dll*"] > -1)} {
          set hDCname [wmf open]
          set hDC [hdc addr $hDCname]
          canvas_to_hdc $c $hDC
          set emf_file [wmf close $hDCname]
          wmf copy $emf_file -file $filename
          wmf delete $emf_file
        }
      }
      .svg {
        set file [open $filename w]
        puts $file [canvas_to_svg $c]
        close $file
      }
    } 
  }
}

##############
## get a postscript fontmap for a canvas
##############
proc get_canv_fontmap {c fontmap} {
  global $fontmap tk_fonts ps_fonts
  set fontlist [list]
  set scale [tk scaling]
  array unset $fontmap *
  foreach i [$c find all] {
    if {[$c type $i] == "text"} {
      set font [$c itemcget $i -font]
      if {[lsearch -exact $fontlist $font] == -1} {
        set family [string tolower [font actual "$font" -family]]
        set weight [font actual "$font" -weight]
        set size [font actual "$font" -size]
        if {[lsearch -exact $tk_fonts "$family\_$weight"] > -1} {
          set ps_font [lindex $ps_fonts [lsearch -exact $tk_fonts "$family\_$weight"]]
        } elseif {[font metrics $font -fixed]} {
          set ps_font Courier  
        } else {
          set ps_font Helvetica
        }
        set "$fontmap\($font)" [list $ps_font [expr {1.0*$size*$scale}]]
      }
    }
  }
}

##############
## print a canvas- windows uses printer extension, aqua uses maccarbonprint extension, UNIX uses postscript
##############
proc print_canvas {c {dialog_parent ""}} {
  global tcl_platform ps_fontmap
  if {[set sc [$c bbox all]] != [list]} {
    foreach {x1 y1 x2 y2} $sc {}
  } else {
    return
  }
  if {$dialog_parent eq ""} {
    if {[winfo ismapped $c]} {
      set dialog_parent $c
    } else {
      tk_messageBox -message "bad window parent: $c" -icon error 
      return
    }
  }
  if {($tcl_platform(platform) == "windows")} {
    if {[set print_parameters [windows_print_dialog]] != ""} {
      foreach {hDC page_width page_height pm_left pm_top pm_right pm_bottom ppi_x ppi_y first_page last_page copies spool_copies print_flag} $print_parameters {}
      set pages [expr {($copies<$spool_copies)?$spool_copies:$copies}]
      ###need to scale the output to fit on page
      set window_x [expr {$x2-$x1}]
      set window_y [expr {$y2-$y1}]
      set printer_x [expr {($page_width - $pm_left - $pm_right) * $ppi_x / 1000.0}]
      set printer_y [expr {($page_height - $pm_top - $pm_bottom) * $ppi_y / 1000.0}]
      set factor_x [ expr $window_x / $printer_x ]
      set factor_y [ expr $window_y / $printer_y ]
      if { $factor_x < $factor_y } {
        set lo $window_y
        set ph $printer_y
      } else {
        set lo $window_x
        set ph $printer_x
      }
      printer job start
      for {set i 0} {$i < $pages} {incr i} {
        printer page start
        gdi map $hDC -logical $lo -physical $ph -offset [list [expr {int(($pm_left* $ppi_x / 1000)-(1.0*$x1/$lo*$ph))}] [expr {int(($pm_top* $ppi_x / 1000)-(1.0*$y1/$lo*$ph))}]]
        canvas_to_hdc $c $hDC
        printer page end
      }
      printer job end
      printer close
    }
  } elseif {$tcl_platform(platform) == "unix"} {
    if  {([lsearch -glob [info loaded] "*MacCarbonPrint.dylib*"] > -1)} {
    ########Aqua native printing
      if {[set print_parameters [aqua_print_dialog $dialog_parent]] != ""} {
        foreach {printObj printtype page_width page_height ppi_x ppi_y copies firstpage lastpage} $print_parameters {}

        set window_x [expr {$x2-$x1}]
        set window_y [expr {$y2-$y1}]
sputs "x1:$x1 y1:$y1"
        set scale_x [expr {($page_width/($window_x))<1?($page_width/($window_x)):1}]
        set scale_y [expr {($page_height/($window_y))<1?($page_height/($window_y)):1}]
        set scale [expr {($scale_x<$scale_y)?$scale_x:$scale_y}]
sputs $scale_x $scale_y $scale
        if {![catch {::maccarbonprint::opendoc $printObj 1 1} errMsg]} {
          for {set i 0} {$i<$copies} {incr i} {
            if {![catch {set win [::maccarbonprint::openpage $printObj -scale $scale -offsetx [expr {int($scale*$x1)}] -offsety [expr {int($scale*$y1)}]]}]} {
              canvas_to_carbonprint $c $win $scale
              # Close down.
              catch {::maccarbonprint::closepage $printObj}
            }
          }
          catch {::maccarbonprint::closedoc $printObj}
        }
      }
    } else {
      #####non-aqua unix printing
      if {[llength [set print_parameters [unix_print_dialog [winfo toplevel $c]]]] > 3} {
        puts $print_parameters
        foreach {printer page_width page_height pm_left pm_top pm_right pm_bottom ppi_x ppi_y copies orient scale} $print_parameters {}
        array unset ps_fontmap *
        get_canv_fontmap $c ps_fontmap
        #exec lpr -P $printer << [$c postscript -height [expr ($y2-$y1)] -width [expr ($x2-$x1)] -x $x1 -y $y1]
        if {$orient=="landscape"} {
          exec lpr -P $printer << [$c postscript -height [expr ($y2-$y1)] -width [expr ($x2-$x1)] -x $x1 -y $y1 -pagex "$pm_top\i" -pagey "[expr {$pm_left}]\i" -pageanchor nw -rotate 1]
        } else {
          exec lpr -P $printer << [$c postscript -height [expr ($y2-$y1)] -width [expr ($x2-$x1)] -fontmap ps_fontmap -x $x1 -y $y1 -pagex "$pm_left\i" -pagey "[expr {$page_height-$pm_top}]\i" -pageanchor nw]
        }
        #if {[tk_messageBox -title [mc "Printing..."] -message "[mc {Printing will occur on your default printer:\n %1$s} [exec lpstat -d]]" -type okcancel -default ok -parent [winfo toplevel $c]] == "ok"} {
        #  exec lpr << [$c postscript -height [expr ($y2-$y1)] -width [expr ($x2-$x1)] -x $x1 -y $y1]
        #}
      }
    }
  } else {
    warning_dialog "Printing isn't available from non-windows or -unix"
    canvas_save $c
  }
}

##############
## print a text window- windows uses printer extension, aqua uses maccarbonprint extension, UNIX uses postscript
##############
proc print_text {t} {
  global ps_fontmap info

  if {[tk windowingsystem] == "win32"} {
    set page_width 8500
    set page_height 11000
    set pm_left 1000
    set pm_right 1000
    set pm_top 1000
    set pm_bottom 1000
    set ppi_x 1000
    set ppi_y 1000
    set attr_list [printer attr]
    foreach attr $attr_list {
      switch -exact [lindex $attr 0] {
        "page dimensions" {
          set page_width [lindex [lindex $attr 1] 0]
          set page_height [lindex [lindex $attr 1] 1]
        }
        "page margins" {
          set pm_left [lindex [lindex $attr 1] 0]
          set pm_top [lindex [lindex $attr 1] 1]
          set pm_right [lindex [lindex $attr 1] 2]
          set pm_bottom [lindex [lindex $attr 1] 3]
        }
        "pixels per inch" {
          set ppi_x [lindex [lindex $attr 1] 0]
          set ppi_y [lindex [lindex $attr 1] 1]
        }
      }
    }
    set print_font [list [font actual courier -family] $info(print_font_size)]
    #set lines_per_inch [expr {[tk scaling]*72/[font metrics $print_font -linespace]}]
    set lines_per_inch [expr {11.0*6/$info(print_font_size)}] ; #11.25??
    set lines_per_page [expr {int(($page_height-$pm_top-$pm_bottom)/1000*$lines_per_inch)}]
    set total_pages [expr {int(ceil(([$t index end]-1)/$lines_per_page))}]
    if {[set print_parameters [windows_print_dialog [list 1 $total_pages]]] != ""} {
#sputs $print_parameters
      foreach {hDC page_width page_height pm_left pm_top pm_right pm_bottom ppi_x ppi_y first_page last_page copies spool_copies print_flag} $print_parameters {}
      switch $print_flag {
        all {
          set first_line 1.0
          set last_line [expr {[$t index end]-1}]
        }
        selection {
          if {[$t tag ranges sel] != ""} {
            set first_line [$t index sel.first\linestart]
            set last_line [$t index sel.last\linestart]
          } else {
            set first_line 1.0
            set last_line [expr {[$t index end]-1}]
          }
        }
        pagenums {
            set first_line [expr {(($first_page-1)*$lines_per_page)+1.0}]
            set last_line [expr {(($last_page)*$lines_per_page*1.0)}]
        }
      } 
#sputs first_line $first_line last_line $last_line lines_per_page $lines_per_page lines_per_inch $lines_per_inch page_height $page_height ppi_y $ppi_y
      set copies_spooled 0
      printer job start
      while {$copies_spooled < $spool_copies} {
        set next_line $first_line
        while {$next_line <= $last_line} {
          set copies_printed 0
          if {[expr {$next_line+$lines_per_page-1}] > $last_line} {
            set last_page_line $last_line
          } else {
            set last_page_line [expr {$next_line+$lines_per_page-1}]
          }
          while {$copies_printed < $copies} {
            printer page start
            gdi map $hDC -offset [list [expr {int(($pm_left* $ppi_x / 1000))}] [expr {int(($pm_top* $ppi_x / 1000))}]]
            gdi text $hDC 0 0 -text [$t get $next_line [$t index $last_page_line\lineend]] -anchor nw -font $print_font -justify left
            #sputs [$t get $next_line [$t index $last_page_line\lineend]]
            printer page end
            incr copies_printed
          }
          set next_line [expr {$next_line+$lines_per_page}]
        }
        incr copies_spooled
      }
      printer job end
      printer close
    }
  } elseif {[tk windowingsystem] == "aqua"} {

    if  {([lsearch -glob [info loaded] "*MacCarbonPrint.dylib*"] > -1)} {
    ########Aqua native printing
      global info
      set pf [eval {::maccarbonprint::pagesetup -dialog 0} $info(aqua_page_setup)]
      array set optsArr [maccarbonprint::pageformatconfigure $pf]
      sputs :pageformatconfigure:
      foreach o [array names optsArr] {
        sputs $o $optsArr($o)
      }
      foreach {pageX pageY page_width page_height} $optsArr(-adjustedpagerect) {}
      if {[llength [set resol $optsArr(-resolution)]]==1} {
        set ppi_x $resol
        set ppi_y $resol
      } else {
        foreach {ppi_x ppi_y} $resol {}
      }

      set print_font [list [font actual courier -family] $info(print_font_size)]
      set lines_per_inch [expr {11.0*6/$info(print_font_size)}] ; #11.25??
      set lines_per_page [expr {int(($page_height)/$ppi_y*$lines_per_inch)}]
      set total_pages [expr {int(ceil(([$t index end]-1)/$lines_per_page))}]

      if {[set print_parameters [aqua_print_dialog [winfo toplevel $t] [list 1 $total_pages]]] != ""} {

        foreach {printObj printtype page_width page_height ppi_x ppi_y copies first_page last_page} $print_parameters {}
        set first_line [expr {(($first_page-1)*$lines_per_page)+1.0}]
        set last_line [expr {(($last_page)*$lines_per_page*1.0)}]
        set scale 1

        if {![catch {::maccarbonprint::opendoc $printObj 1 [expr {($last_page-$first_page+1)*$copies}]} errMsg]} {
          for {set i 0} {$i<$copies} {incr i} {
            #Print a document copy
            set next_line $first_line
            while {$next_line <= $last_line} {
              if {[expr {$next_line+$lines_per_page-1}] > $last_line} {
                set last_page_line $last_line
              } else {
                set last_page_line [expr {$next_line+$lines_per_page-1}]
              }
              #Print page
              if {![catch {set win [::maccarbonprint::openpage $printObj -scale $scale]}]} {
                $win create text 0 0 -text [$t get $next_line [$t index $last_page_line\lineend]] -anchor nw -font $print_font -justify left
                catch {::maccarbonprint::closepage $printObj}
              }
              set next_line [expr {$next_line+$lines_per_page}]
            }
          }
          catch {::maccarbonprint::closedoc $printObj}
        }
      }
    } else {
      #####non-aqua unix printing
    }
  } else {
    warning_dialog "Printing isn't available from non-windows or -unix"
  }
}

##############
## display printer dialog in windows
## return $hDC $page_width $page_height $pm_left $pm_top $pm_right $pm_bottom $ppi_x $ppi_y $copies
## return {} and close printer if no print connection is established
##############
proc windows_print_dialog {{pagerange {1 1}}} {
  if  {([lsearch -glob [info loaded] "*printer.dll*"] > -1)} {
    set minimum_page [lindex $pagerange 0]
    set maximum_page [lindex $pagerange 1]
    printer attr -set [list [list "first page" $minimum_page] [list "last page" $maximum_page] [list "maximum page" $maximum_page] [list "minimum page" $minimum_page]]
    set hDC [printer dialog select]
    if { [lindex $hDC 1] == 0 } {
    # User has canceled printing
      catch {printer close}
	return {}
    }
    set hDC [lindex $hDC 0]
    if {[string match "?" $hDC] || [string match 0x0 $hDC]} {
      catch {printer close}
      tk_messageBox -title [mc "Printer problem"] -message [mc "Problem opening printer: printer context cannot be established"] -type ok
      return {}
    }
    set page_width 8500
    set page_height 1100
    set pm_left 1000
    set pm_top 1000
    set pm_right 1000
    set pm_bottom 1000
    set ppi_x 1000
    set ppi_y 1000
    set first_page 1
    set last_page 1
    set spool_copies 1
    set copies 1
    set print_flag "all"
    set attr_list [printer attr]
    foreach attr $attr_list {
      switch -exact [lindex $attr 0] {
        "page dimensions" {
          set page_width [lindex [lindex $attr 1] 0]
          set page_height [lindex [lindex $attr 1] 1]
        }
        "page margins" {
          set pm_left [lindex [lindex $attr 1] 0]
          set pm_top [lindex [lindex $attr 1] 1]
          set pm_right [lindex [lindex $attr 1] 2]
          set pm_bottom [lindex [lindex $attr 1] 3]
        }
        "pixels per inch" {
          set ppi_x [lindex [lindex $attr 1] 0]
          set ppi_y [lindex [lindex $attr 1] 1]
        }
        "first page" {
          set first_page [lindex $attr 1]
        }
        "last page" {
          set last_page [lindex $attr 1]
          if {$last_page > $maximum_page} {
            set last_page $maximum_page
          }
        }
        "copiesToSpool" {
          set spool_copies [lindex $attr 1]
        }
        "copies" {
          set copies [lindex $attr 1]
        }
        "print flag" {
          set print_flag [lindex $attr 1]
        }
      }
    }
    return [list $hDC $page_width $page_height $pm_left $pm_top $pm_right $pm_bottom $ppi_x $ppi_y $first_page $last_page $copies $spool_copies $print_flag]
  } else {
    tk_messageBox -title [mc "Printer problem"] -message [mc "Problem opening printer: printer dll is not loaded"] -type ok
    return {}
  }
}

##############
## display printer dialog in MacCarbon (Aqua)
## return $printerObject $type $page_width $page_height $ppi_x $ppi_y $copies
## return {} and close printer if no print connection is established
##############
proc aqua_print_dialog {w {pagerange {1 1}}} {
  global info

sputs $w $pagerange $info(aqua_page_setup)
  if {[set printlist [eval {::maccarbonprint::print -parent $w -pagerange $pagerange -copies 1} $info(aqua_page_setup)]] != {}} {
    set type [lindex $printlist 0]
    set printObj [lindex $printlist 1]
    array set optsArr [::maccarbonprint::printconfigure $printObj]
    sputs dialog-printconfigure
    foreach o [array names optsArr] {
      sputs $o $optsArr($o)
    }

    foreach {pageX pageY pagewidth pageheight} $optsArr(-adjustedpagerect) {}

    if {[llength [set resol $optsArr(-resolution)]]==1} {
      set ppi_x $resol
      set ppi_y $resol
    } else {
      foreach {ppi_x ppi_y} $resol {}
    }

    if {[set last_page $optsArr(-lastpage)] > [lindex $pagerange 1]} {
      set last_page [lindex $pagerange 1]
    }

    return [list $printObj $type $pagewidth $pageheight $ppi_x $ppi_y $optsArr(-copies) $optsArr(-firstpage) $last_page]
  } else {
    return ""
  }
}

##############
## display printer setup dialog in Aqua
##############
proc page_setup_dialog {w} {
  global info
  if  {([lsearch -glob [info loaded] "*MacCarbonPrint.dylib*"] > -1)} {
    set pf [eval {::maccarbonprint::pagesetup -parent $w} $info(aqua_page_setup)]
    if {$pf != ""} {
      set info(aqua_page_setup) [list -pageformat $pf]
    }
  } elseif {([lsearch -glob [info loaded] "*printer.dll*"] > -1)} {
    printer dialog page_setup
  }
}

##############
## display printer dialog in unix
## return $printer $page_width $page_height $pm_left $pm_top $pm_right $pm_bottom $ppi_x $ppi_y $copies
## return {} and close printer if no print connection is established
##############
proc unix_print_dialog {w} {
  global ok dialogblock info tcl_platform
  global printer orient scale copies
  if {$tcl_platform(platform) != "unix"} {
    set prlist [list printer1 printer2 printer3]
    set defaultprinter printer2
  } else {
    if {([catch {set lplist [exec lpstat -a]}]) || ($lplist == "")} {return "no_lplist"}
    set lplist [split $lplist \n]
    foreach lp $lplist {
      puts $lp
      if {![regexp "not accepting" $lp]} {
        lappend prlist [regexp -inline {[^ ]*} $lp]
      }
    }
    if {([catch {set defaultprinter [exec lpstat -d]}]) || ([regexp "no system default" $defaultprinter])}  {return "no_default"}
    set defaultprinter [regexp -inline {[^ ]*$} $defaultprinter]
  }

  if {$dialogblock == 1} {return "dialogblock"}
  set dialogblock 1
  set a [toplevel .dialog]
  wm title $a [mc "Printer..."]
  wm protocol $a WM_DELETE_WINDOW "set ok -1"
  set oldbind [bind . <<RaiseDialogs>>]
  bind . <<RaiseDialogs>> "wm deiconify $a; raise $a"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry $a "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient $a $w
  } else {
    ::tk::unsupported::MacWindowStyle style $a document {closeBox}
  }
  set ok 0
  set page_width 8.500
  set page_height 11.00
  set pm_left 1.0
  set pm_top 1.0
  set pm_right 1.0
  set pm_bottom 1.0
  set ppi_x 1000
  set ppi_y 1000
  set copies 1
  set orient portrait
  set scale 100
  set printer $defaultprinter


  grid [frame $a.frame1 -relief flat] -row 0 -column 0 -sticky nswe
  grid [label $a.frame1.label1 -text [mc "Printer:"]] -row 0 -column 0 -sticky e
  set printer_menu [eval "tk_optionMenu $a.frame1.printer printer $prlist"]
  grid $a.frame1.printer -row 0 -column 1 -sticky nw -columnspan 2
  grid [label $a.frame1.label2 -text [mc "Orientation:"]] -row 1 -column 0 -sticky e
  set orientmenu [tk_optionMenu $a.frame1.orient orient portrait landscape]
  grid $a.frame1.orient -row 1 -column 1 -sticky nw -columnspan 2
  #grid [label $a.frame1.label3 -text [mc "Scale:"]] -row 2 -column 0 -sticky e
  #set scalemenu [tk_optionMenu $a.frame1.scale scale 10 20 30 40 50 60 70 80 90 100]
  #grid $a.frame1.scale -row 2 -column 1 -sticky nw
  #grid [label $a.frame1.label4 -text "%"] -row 2 -column 2 -sticky w
  #grid [label $a.frame1.label5 -text [mc "Copies:"]] -row 3 -column 0 -sticky e
  #grid [entry $a.frame1.copies -textvariable copies -width 2 -validate key -validatecommand "string is integer %P"] -row 3 -column 1 -sticky nw -columnspan 2


  grid [frame $a.frame2 -relief flat] -row 1 -column 0 -sticky nswe
  grid [button .dialog.frame2.ok -text [mc "OK"] -command "set ok 1" -default active] -row 0 -column 0 -padx 5 -sticky we
  grid [button .dialog.frame2.cancel -text [mc "Cancel"] -command "set ok -1"] -row 0 -column 1 -padx 5 -sticky we
  grid rowconfigure .dialog 0 -weight 1
  grid rowconfigure .dialog 1 -weight 0
  grid rowconfigure .dialog 2 -weight 0
  grid columnconfigure .dialog 0 -weight 1

  if {[tk windowingsystem] == "aqua"} {
    foreach button [list .dialog.frame2.ok .dialog.frame2.cancel] {
      grid configure $button -sticky n
      update
      grid configure $button -sticky ne
    }
  }

  vwait ok
  set dialogblock 0
  destroy .dialog
  bind . <<RaiseDialogs>> $oldbind
  if {$ok == 1} {
    if {[catch {set optionlist [exec lpoptions -p $printer -l]}]} {
      set optionlist ""
    } else {
      foreach attr [split $optionlist \n] {
        set attr [split $attr :]
        switch -regexp [lindex $attr 0] {
          "Resolution" {
            regexp {\*([0-9]*)x([0-9]*)} [lindex $attr 1] i ppi_x ppi_y
           }
        }
      }
    }
    set scale [expr {$scale/100.0}]
    puts "printer $printer optionlist:$optionlist"
    set result [list $printer $page_width $page_height $pm_left $pm_top $pm_right $pm_bottom $ppi_x $ppi_y $copies $orient $scale]
    unset printer orient scale copies
    return $result
  } else {
    return "cancel"
  }
}

##############
## save or copy to clipboard a canvas to wmf or emf format
##############
proc canvas_to_metafile {operation c {type both}} {
  set hDCname [wmf open]
  set hDC [hdc addr $hDCname]
  canvas_to_hdc $c $hDC
  set emf_file [wmf close $hDCname ]
  if {$operation == "copy"} {
    wmf copy $emf_file -type $type
  } else {
    set filename [tk_getSaveFile -defaultextension ".emf" -title [mc "Save Enhanced Metafile..."] -parent [winfo toplevel $c]]
    if {$filename != ""} {
      wmf copy $emf_file -file $filename
    }
  }
  wmf delete $emf_file
}

##############
## print a canvas to an hDC (device context handle) for windows printing and metafile output
##############
proc canvas_to_hdc {c hdc} {
  foreach item [$c find all] {
    set coords [$c coords $item]

    switch [$c type $item] {
      arc {
        #set cmd  "gdi arc $hdc $coords -extent [$c itemcget $item -extent] -outline [$c itemcget $item -outline] -start [$c itemcget $item -start] -style [$c itemcget $item -style]"
        #if {[set width  [$c itemcget $item -width]] > 1} {set cmd "$cmd -width $width"}
        set x0 [expr {([lindex $coords 0]+[lindex $coords 2])/2.0}]
        set y0 [expr {([lindex $coords 1]+[lindex $coords 3])/2.0}]
        set xr [expr {[lindex $coords 2]-$x0}]
        set yr [expr {[lindex $coords 3]-$y0}]
        set start_angle [$c itemcget $item -start]
        set stop_angle [expr {[$c itemcget $item -extent]+$start_angle}]
        if {$stop_angle < $start_angle} {
          set a $start_angle
          set start_angle $stop_angle
          set stop_angle $a
        }
        set deg2rad [expr {3.14159/180}]
        set start_angle [expr {$deg2rad*$start_angle}]
        set stop_angle [expr {$deg2rad*$stop_angle}]
        set incr [expr {$deg2rad*2.5}]
        set coords [list [expr {$x0+cos($start_angle)*$xr}] [expr {$y0-sin($start_angle)*$yr}]]
        for {set a $start_angle} {$a <$stop_angle} {set a [expr {$a+$incr}]} {
          lappend coords [expr {$x0+cos($a)*$xr}] [expr {$y0-sin($a)*$yr}]
        }
        lappend coords [expr {$x0+cos($stop_angle)*$xr}] [expr {$y0-sin($stop_angle)*$yr}]
        set cmd  "gdi line $hdc $coords -fill [$c itemcget $item -outline]"
        if {[set width  [$c itemcget $item -width]] > 1} {set cmd "$cmd -width $width"}
      } 
      bitmap {
        #sputs "No handler for item type: Bitmap"
      }
      image {
        #sputs "No handler for item type: Image"
      }
      line {
        set cmd  "gdi line $hdc $coords -fill [$c itemcget $item -fill]"
        if {[set width  [$c itemcget $item -width]] > 1} {set cmd "$cmd -width $width"}
      }
      oval {
        set cmd "gdi oval $hdc $coords -outline [$c itemcget $item -outline]"
        if {[set width  [$c itemcget $item -width]] > 1} {set cmd "$cmd -width $width"}
      }
      polygon {
        set cmd "gdi polygon $hdc $coords -width [$c itemcget $item -width]"
        if {[set outline [$c itemcget $item -outline]] != ""} {set cmd "$cmd  -outline $outline"}
      }
      rectangle {
        set cmd "gdi rectangle $hdc $coords -outline [$c itemcget $item -outline]"
        if {[set width  [$c itemcget $item -width]] > 1} {set cmd "$cmd -width $width"}
      }
      text {
        if {[$c itemcget $item -text] != ""} {
          set font [$c itemcget $item -font]
          #set font [list [font configure $font -family]  -[font configure $font -size]]
          set font [list [font actual $font -family] [expr {int([font actual $font -size]*[tk scaling])}] [font actual $font -weight] [font actual $font -slant]]
          set bbox [$c bbox $item]
          #set width [expr {[lindex $bbox 2] - [lindex $bbox 0]+2}]
          set width [font measure $font [$c itemcget $item -text]]
          set cmd "gdi text $hdc $coords -text [list [$c itemcget $item -text]] \
		-anchor [$c itemcget $item -anchor] -font [list $font] \
		-width $width -justify [$c itemcget $item -justify]"
        } else {
          set cmd ""
        }
      }
      window {
        #sputs "No handler for item type: Window"
      }
      default {
        #sputs "No handler for item type:[$c type $item]"
      }
    }

    if {([set fill [$c itemcget $item -fill]] != "") && ($cmd !="")} {set cmd "$cmd -fill $fill"}
    eval $cmd
  }
}

##############
## print a canvas to a printer window for Aqua printing
##############
proc canvas_to_carbonprint {c win scale} {
  foreach item [$c find all] {
    switch [$c type $item] {
      text {
        if {[$c itemcget $item -text] != ""} {
          set font [$c itemcget $item -font]
          set font [list [font actual $font -family] [expr {int([font actual $font -size]*[tk scaling]*$scale)}] [font actual $font -weight] [font actual $font -slant]]
          set bbox [$c bbox $item]
          set width [font measure $font [$c itemcget $item -text]]
          eval {$win create text} [$c coords $item] -text [list [$c itemcget $item -text]] -anchor [$c itemcget $item -anchor] -font [list $font] -justify [$c itemcget $item -justify] -fill [$c itemcget $item -fill]
        }
      }
      arc {
        set opts {}
        foreach option [$c itemconfigure $item] {
          foreach {key blank blank2 def val} $option {}
          if {![string equal $def $val]} {
            lappend opts $key $val
          }
        }
        eval {$win create [$c type $item]} [$c coords $item] $opts
      }
      window {
        #sputs "No handler for item type: Window"
      }
      default {
        set opts {}
        foreach option [$c itemconfigure $item] {
          foreach {key blank blank2 def val} $option {}
          if {![string equal $def $val]} {
            lappend opts $key $val
          }
        }
        eval {$win create [$c type $item]} [$c coords $item] $opts
      }
    }
  }
}

##############
## generate a canvas from a simple text box
##############
proc text_to_canvas {t c {x0 2} {y0 2}} {
  set i 1
  set x $x0
  set y $y0
  set font [$t cget -font]
  set linespace [font metrics [$t cget -font] -linespace]
  while {$i < [$t index end]} {
    $c create text $x $y -text [$t get $i.0 $i.0lineend] -anchor nw -font $font
    incr i
    incr y $linespace 
  }
}

##############
## save text analysis as text
##############
proc text_save {t {filename {}}} {
global info
  while {($info(default_dnadir) != "") && (![file isdirectory $info(default_dnadir)])} {
    set info(default_dnadir) [file dirname $info(default_dnadir)]
    if {[file dirname $info(default_dnadir)] == $info(default_dnadir)} {
      set info(default_dnadir) ""
    }
  }
  if {$filename == ""} {
    set filename [tk_getSaveFile -title [mc "Save As Text File"] -initialdir $info(default_dnadir) -defaultextension ".txt"]
  }
  while {([file exists $filename]) && (![file writable $filename])} {
    tk_messageBox -title "Text File not Writable" -message "The file \"$filename\" is not writable. It may be locked." -type ok -icon warning -default ok
    set filename [tk_getSaveFile -title "Save As Text File" -initialdir $info(default_dnadir) -defaultextension ".txt"]
  }
  if {$filename != ""} {
    ###save the text file
    set text_file [open $filename w] 
    set i 1
    while {$i < [$t index end]} {
      puts $text_file [$t get $i.0 $i.0lineend]
      incr i
    }
    close $text_file
  }
}

##############
## convert named color to svg rgb value
##############
proc color_to_svgrgb {color} {
    if {$color == ""} {return none}
    foreach {r g b} [winfo rgb . $color] {}
    format #%02x%02x%02x [expr {$r/256}] [expr {$g/256}] [expr {$b/256}]
}


##############
## print a canvas to an svg string
##############
proc canvas_to_svg {c} {
  
  set result "<svg version=\"1.1\" preserveAspectRatio=\"xMidYMid meet\" viewBox=\"[lindex [$c bbox all] 0] [lindex [$c bbox all] 1] [lindex [$c bbox all] 2] [lindex [$c bbox all] 3]\">\n"
  foreach item [$c find all] {
    set coords [$c coords $item]

    switch [$c type $item] {
      arc {
        set x0 [expr {([lindex $coords 0]+[lindex $coords 2])/2.0}]
        set y0 [expr {([lindex $coords 1]+[lindex $coords 3])/2.0}]
        set xr [expr {[lindex $coords 2]-$x0}]
        set yr [expr {[lindex $coords 3]-$y0}]
        set start_angle [$c itemcget $item -start]
        set stop_angle [expr {[$c itemcget $item -extent]+$start_angle}]
        if {$stop_angle < $start_angle} {
          set a $start_angle
          set start_angle $stop_angle
          set stop_angle $a
        }
        set deg2rad [expr {3.14159/180}]
        set start_angle [expr {$deg2rad*$start_angle}]
        set stop_angle [expr {$deg2rad*$stop_angle}]

        set mcoords "[expr {$x0+cos($start_angle)*$xr}],[expr {$y0-sin($start_angle)*$yr}]"
        set ecoords "[expr {$x0+cos($stop_angle)*$xr}],[expr {$y0-sin($stop_angle)*$yr}]"

        set cmd "<path d=\"M $mcoords A$xr,$yr 0 [expr {(abs([$c itemcget $item -extent])>180)?1:0}] 0 $ecoords\" stroke-width=\"[$c itemcget $item -width]\" stroke=\"[color_to_svgrgb [$c itemcget $item -outline]]\" fill=\"[color_to_svgrgb [$c itemcget $item -fill]]\"/>\n"
      } 
      bitmap {
        #sputs "No handler for item type: Bitmap"
        set cmd ""
      }
      image {
        #sputs "No handler for item type: Image"
        set cmd ""
      }
      line {
        if {[llength $coords] == 4} {
          set cmd "<line x1=\"[lindex $coords 0]\" y1=\"[lindex $coords 1]\" x2=\"[lindex $coords 2]\" y2=\"[lindex $coords 3]\" stroke-width=\"[$c itemcget $item -width]\" stroke=\"[color_to_svgrgb [$c itemcget $item -fill]]\"/>\n"
        } else {
          set cmd "<polyline points=\"$coords\" stroke-width=\"[$c itemcget $item -width]\" stroke=\"[color_to_svgrgb [$c itemcget $item -fill]]\" fill=\"none\"/>\n"
        }
      }
      oval {
          set cmd "<ellipse cx=\"[expr {([lindex $coords 2]+[lindex $coords 0])/2.0}]\" cy=\"[expr {([lindex $coords 3]+[lindex $coords 1])/2.0}]\" rx=\"[expr {([lindex $coords 2]-[lindex $coords 0])/2.0}]\" ry=\"[expr {([lindex $coords 3]-[lindex $coords 1])/2.0}]\" stroke-width=\"[$c itemcget $item -width]\" stroke=\"[color_to_svgrgb [$c itemcget $item -outline]]\" fill=\"[color_to_svgrgb [$c itemcget $item -fill]]\"/>\n"
      }
      polygon {
          set cmd "<polygon points=\"$coords\" stroke-width=\"[$c itemcget $item -width]\" stroke=\"[color_to_svgrgb [$c itemcget $item -outline]]\" fill=\"[color_to_svgrgb [$c itemcget $item -fill]]\"/>\n"

      }
      rectangle {
          set cmd "<rect x=\"[lindex $coords 0]\" y=\"[lindex $coords 1]\" width=\"[expr {[lindex $coords 2]-[lindex $coords 0]}]\" height=\"[expr {[lindex $coords 3]-[lindex $coords 1]}]\" stroke-width=\"[$c itemcget $item -width]\" stroke=\"[color_to_svgrgb [$c itemcget $item -outline]]\" fill=\"[color_to_svgrgb [$c itemcget $item -fill]]\"/>\n"
      }
      text {
        set font [$c itemcget $item -font]
        if {[$c itemcget $item -text] != ""} {
          set bbox [$c bbox $item]
          set cmd "<text x=\"[lindex $bbox 0]\" y=\"[lindex $bbox 3]\" font-weight=\"[font actual $font -weight]\" font-style=\"[font actual $font -slant]\" font-family=\"[font actual $font -family]\" font-size=\"[expr {[tk scaling]*[font actual $font -size]}]\" fill=\"[color_to_svgrgb [$c itemcget $item -fill]]\" >[$c itemcget $item -text]</text>\n"
        } else {
          set cmd ""
        }
      }
      window {
        #sputs "No handler for item type: Window"
        set cmd ""
      }
      default {
        #sputs "No handler for item type:[$c type $item]"
        set cmd ""
      }
    }
    append result $cmd
  }
  append result "</svg>\n"
  return $result
}

##############
## scans a dna window textbox and adds feature tags and metainfo 
##############
#<<Metadata>>:(0)name, (1)type, (2)plain_text, (3)fwd/rev, (4)annotations_string
#<<Revcolors>>: binding used for the color if the sequence is reversed
#library list: (0)name (1)regexp_fwd(flanked by paren) (2)regexp_rev (3)preseq (4)postseq (5)plain_text_def (6)type (7)fwd_color (8)rev_color (9)bold (10)ital (11)annotations_string

proc feature_scan {w} {
  global fea_info
  global info dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  $w configure -cursor watch
  $w.textarea configure -cursor watch
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy


  #toplevel .wait
  #wm overrideredirect .wait 1
  #wm geometry .wait "+[expr (($winx<0)?0:$winx)+50]+[expr (($winy<0)?0:$winy)+130]"
  #wm title .wait "Searching for Features, Please wait..." 
  #bind . <<RaiseDialogs>> {wm deiconify .wait; raise .wait}
  #grab set .wait
  #.wait configure -cursor watch
  #pack [label .wait.label -text "Searching for Features, Please wait..." -bg yellow -font boldlabelfont]

  #update
  $w.infobox configure -bg yellow -text "Searching for Features, Please wait..."


  set temp_text [$w.textarea get 1.0 1.end]
  set text_end [expr {[string length $temp_text]-1}]
  if {$info($w,circular)== "circular"} {
    set temp_text "$temp_text$temp_text"
  }

  #feature_cleanup $w.textarea

  foreach feature $fea_info(library) {
    foreach direction [list 1 2]  {
      set start 0
      if {[lindex $feature $direction] == ""} continue
      while {[set found_list [regexp -inline -indices -nocase -start $start [lindex $feature $direction] $temp_text]] != ""} {
        ##when adding features, set saved to 0 (keep links)
        set info($w,saved) 0

        set start [lindex [lindex $found_list 0] 0]
        if {$start > $text_end} {break}
        incr start
        set tagname "f$fea_info($w,count)#"

        ##end extension scan
        if {$direction == 1} {
          set preseq [lindex $feature 3]
          set postseq [lindex $feature 4]
        } else {
          set preseq [revcom_regexp [lindex $feature 4]]
          set postseq [revcom_regexp [lindex $feature 3]]
        }
        if {$preseq != ""} {
          set i [expr {[lindex [lindex $found_list 0] 0]}]
          if {($i <= 0) && ($info($w,circular)== "circular")} {
            set i [expr {1+$text_end}]
          }
          set j [expr {[string length $preseq]}]
          set k 0
          while {([string match -nocase [string index $preseq [expr {$j-1}]] [string index $temp_text [expr {$i-1}]]]) || ([string index $preseq $j] == "N")} {
            incr i -1
            incr j -1
            incr k
            if {($i <= 0)} {
              if {($info($w,circular)== "circular")} {
                set i [expr {1+$text_end}]
              } else {
                break
              }
            }
          }
          ##add the extension
          if {$i > $start} {
            #wrapped around
            set temp_list [list]
            foreach r $found_list {
              lappend temp_list [list [expr {[lindex $r 0]+$text_end+1}] [expr {[lindex $r 1]+$text_end+1}]]
            }
            set found_list $temp_list
          }
          set found_list [lreplace $found_list 0 1 [list $i [lindex [lindex $found_list 0] 1]] [list $i [lindex [lindex $found_list 1] 1]]] 
        }

        if {$postseq != ""} {
          set last [lindex [lindex $found_list 0] 1]
          set postf [string toupper [string range $temp_text [expr {$last + 1}] [expr {$last + [string length $postseq]}]]]
          set i 0
          while {($i < [string length $postseq]) && (([string index $postseq $i] == [string index $postf $i]) || ([string index $postseq $i] == "N"))} {incr i} 
          ##add the extension
          set found_list [lreplace $found_list 0 1 [list [lindex [lindex $found_list 0] 0] [expr {[lindex [lindex $found_list 0] 1]+$i}]] [list [lindex [lindex $found_list end] 0] [expr {[lindex [lindex $found_list end] 1]+$i}]]]
        }
        ##add feature fn(feature_number)# to indicate a split feature
        if {[llength $found_list] > 2} {
          set first [lindex [lindex $found_list 0] 0]
          set last [lindex [lindex $found_list 0] 1]
          if {$last <= $text_end} {
            $w.textarea tag add "fn$fea_info($w,count)#" "1.$first" "1.[expr {1 + $last}]"
          } else {
            $w.textarea tag add "fn$fea_info($w,count)#" "1.$first" 1.end
            $w.textarea tag add "fn$fea_info($w,count)#" 1.0 "1.[expr {$last-$text_end}]"
          }
        }
        incr fea_info($w,count)

        foreach region [lrange $found_list 1 end] {
#sputs "$tagname 1.[lindex $region 0] 1.[expr {1+[lindex $region 1]}]"
update
          if {[lindex $region 1] <= $text_end} {
            $w.textarea tag add $tagname "1.[lindex $region 0]" "1.[expr {1+[lindex $region 1]}]"
          } else {
            if {[lindex $region 0] <= $text_end} {
              $w.textarea tag add $tagname "1.[lindex $region 0]" 1.end
              $w.textarea tag add $tagname 1.0 "1.[expr {[lindex $region 1]-$text_end}]"
            } else {
              $w.textarea tag add $tagname "1.[expr {[lindex $region 0]-$text_end-1}]" "1.[expr {[lindex $region 1]-$text_end}]"
            }
          }
        }
        #<<Metadata>>:(0)name, (1)special_info, (2)type, (3)fwd/rev, (4)annotations_string
        set metadata [list [lindex $feature 0] [list 0 0 0] [lindex $feature 6] [expr {$direction-1}] [lindex $feature 11]]
        $w.textarea tag bind $tagname <<Metadata>> $metadata
        set color [lindex $feature 7]
        set revcolor [lindex $feature 8]
        $w.textarea tag bind $tagname <<Revcolors>> [list $color $revcolor]
        $w.textarea tag configure $tagname -background [lindex $feature [expr {$direction + 6}]]

        ###check for dup tags and delete the old tag
        set tag_start_index [lindex [$w.textarea tag ranges $tagname] 0]
        set tag_ranges [$w.textarea tag ranges $tagname]
        foreach pos_dup_tag [$w.textarea tag names $tag_start_index] {
          if {$pos_dup_tag == $tagname} continue
          if {([$w.textarea tag ranges $pos_dup_tag] == $tag_ranges) && ([lindex [$w.textarea tag bind $pos_dup_tag <<Metadata>>] 0] == [lindex $metadata 0])} {
            $w.textarea tag delete $pos_dup_tag
          }
        }
      }
    }
  }

  #destroy .wait 
  #bind . <<RaiseDialogs>> ""
  set dialogblock 0
  $w.infobox configure -bg $info(bg_color) -text ""
  $w configure -cursor arrow
  $w.textarea configure -cursor xterm
  update
}

##############
## remove unused feature tags, split features that are broken
##############
proc feature_cleanup {text} {
  set deleted [list]
  foreach text_tag [$text tag names] {
    if {([regexp {fn?[0-9]+#} $text_tag]) && ([$text tag ranges $text_tag] == {})} {
      $text tag delete $text_tag 
      lappend deleted $text_tag
    }
  }
  return 
}

##############
## returns all features with elements between c1 and c2 as a list of feature names and tag names 
##############
proc feature_find {w {c1 -1} {c2 -1}} {
  if {$c1 == -1} {
    set c1 1.0
  } else {
    set c1 "1.$c1"
  }
  if {$c2 == -1} {
    set c2 [$w.textarea index 1.end]
  } else {
    set c2 "1.$c2"
  }
  #feature_cleanup $w.textarea
  set result [list]
  foreach text_tag [$w.textarea tag names] {
    if {([regexp {f[0-9]+#} $text_tag]) && (([lsearch -exact [$w.textarea tag names $c1] $text_tag] != -1) || ([$w.textarea tag nextrange $text_tag $c1 $c2+1char] != {}))} {
      set metadata [$w.textarea tag bind $text_tag <<Metadata>>]
      set name [lindex $metadata 0]
      regsub "f" $text_tag "fn" tt2
      if {[$w.textarea tag ranges $tt2] != {}} {
        lappend result $name $tt2
      } else {
        lappend result $name $text_tag
      }
    }
  }
  return $result
}

##############
## returns coordinates of fea between c1 and c2
##############
proc feature_coords {w fea {c1 -1} {c2 -1}} {
  if {$c1 == -1} {
    set c1 0
  } 
  if {$c2 == -1} {
    set c2 [string range [$w.textarea index 1.end] 2 end]
  }
  set result [list]
  foreach {x1 x2} [$w.textarea tag ranges $fea] {
    set x1 [string range $x1 2 end]
    set x2 [string range $x2 2 end]
    incr x2 -1
    if {$x1 >= $c1} {
      if {$x1 <= $c2} {
        if {$x2 <= $c2} {
          lappend result [list $x1 $x2]
        } else {
          lappend result [list $x1 $c2]
        }
      }
    } else {
      if {$x2 >= $c1} {
        if {$x2 <= $c2} {
          lappend result [list $c1 $x2]
        } else {
          lappend result [list $c1 $c2]
        }
      } 
    }
  }
  set fntag_ranges [$w.textarea tag ranges [regsub f $fea fn]]
  if {$fntag_ranges == ""} {set fntag_ranges [$w.textarea tag ranges $fea]}
  if {[llength $fntag_ranges]> 2} {
    set start [string range [lindex $fntag_ranges 2] 2 end]
    set pos 0
    while {([lindex [lindex $result $pos] 0] < $start) && ($pos < [llength $result])} {incr pos}
    set result [concat [lrange $result $pos end] [lrange $result 0 [expr {$pos-1}]]]
  }
  return $result
}

##############
## deletes all feature tags in a textbox
##############
proc feature_clear {w} {
  global info
  if {[tk_messageBox -message [mc "Clear all features from the file (NO UNDO)?"] -type okcancel -icon warning -default cancel] == "cancel"} return
  foreach text_tag [$w.textarea tag names] {
    if {[regexp {fn?[0-9]+#} $text_tag]} {
        $w.textarea tag delete $text_tag 
    }
  }
  set info($w,saved) 0
}

##############
## Utility for dealing with selections in listboxes
##############
proc listbox_multiple {w el} {
  if {[$w selection includes $el]} {
    $w selection clear $el
  } else {
    $w selection set $el
  }
}

##############
## Configure display properties of features in a window
##############
proc configure_features {w} {
  global info 
  global ok modifier
  global mem deleted dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1
  toplevel .dialog
  wm title .dialog [mc "Edit Features..."]
  wm protocol .dialog WM_DELETE_WINDOW "set ok -1"
  bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry .dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .dialog $w
  } else {
    ::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
    #.dialog configure -menu [menu .dialog.menubar]
  }
  set ok 0

  grid [frame .dialog.frame1 -relief flat] -row 0 -column 0 -sticky nswe
  grid [listbox .dialog.frame1.list -width 60 -yscrollcommand "optionscrollbar .dialog.frame1.scroll" -selectmode extended -activestyle none -background white -exportselection 0] -row 0 -column 0 -sticky nswe
  grid [scrollbar .dialog.frame1.scroll -command ".dialog.frame1.list yview"] -row 0 -column 1 -sticky nse
  grid rowconfigure .dialog.frame1 0 -weight 1
  grid columnconfigure .dialog.frame1 0 -weight 1
  grid columnconfigure .dialog.frame1 1 -weight 0
  grid [frame .dialog.frame2 -relief ridge -borderwidth 2] -row 1 -column 0 -sticky swe
  grid [button .dialog.frame2.raise -text [mc "Raise"] -command "event generate .dialog <<raise>>"] -row 0 -column 0 -sticky we
  grid [button .dialog.frame2.lower -text [mc "Lower"] -command "event generate .dialog <<lower>>"] -row 0 -column 1 -sticky we
  grid [button .dialog.frame2.delete -text [mc "Delete"] -command "event generate .dialog <<delete>>"] -row 0 -column 2 -sticky we
  grid [button .dialog.frame2.rename -text [mc "Edit"] -command "event generate .dialog.frame1.list <<edit>> -x -1 -y -1"] -row 0 -column 3 -sticky we
  #grid [button .dialog.frame2.recolor -text [mc "Recolor"] -command "event generate .dialog <<recolor>>"] -row 1 -column 1 -sticky we


  grid [frame .dialog.frame3 -relief raised] -row 2 -column 0 -sticky swe
  grid [button .dialog.frame3.ok -text [mc "OK"] -command "set ok 1" -default active] -row 0 -column 0 -padx 5 -sticky we
  grid [button .dialog.frame3.cancel -text [mc "Cancel"] -command "set ok -1"] -row 0 -column 1 -padx 5 -sticky we
  grid rowconfigure .dialog 0 -weight 1
  grid rowconfigure .dialog 1 -weight 0
  grid rowconfigure .dialog 2 -weight 0
  grid columnconfigure .dialog 0 -weight 1

  #bind .dialog.frame1.list <$modifier-Button-1> {listbox_multiple %W [%W index @%x,%y];break}
  bind .dialog.frame1.list <Double-Button-1> "event generate .dialog.frame1.list <<edit>> -x %x -y %y"
  #bind .dialog.frame1.list <Button-1> {bell; .dialog.frame1.list selection clear 0 end; .dialog.frame1.list selection set [%W index @%x,%y]}

  set i [expr {[llength [set features [feature_find $w]]]/2-1}]
  foreach {name tag} $features {
    .dialog.frame1.list insert 0 $name
    regsub "fn" $tag "f" tag
    .dialog.frame1.list itemconfigure 0 -background [$w.textarea tag cget $tag -background]
    set mem($i) $tag
    incr i -1
  }
  set deleted [list]
  .dialog.frame1.list selection set 0

  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .dialog] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry .dialog "+$winx+$winy"
  update
  focus .dialog.frame1.list

  bind .dialog.frame1.list <<edit>> "
    if {%y == -1} {
      set sel_index \[lindex \[%W curselection\] 0\]
    } else {
      set sel_index \[%W index @%x,%y\]
    }
    if {\$sel_index != {}} {
      if {\[rename_recolor_window_feature .dialog $w \$sel_index\] eq \"ok\"} {
        .dialog.frame1.list delete \$sel_index
        .dialog.frame1.list insert \$sel_index \[lindex \[$w.textarea tag bind \$mem(\$sel_index) <<Metadata>>\] 0\]
        if {\[lindex \[$w.textarea tag bind \$mem(\$sel_index) <<Metadata>>\] 3\] == 0} {
          .dialog.frame1.list itemconfigure \$sel_index -background \[lindex \[$w.textarea tag bind \$mem(\$sel_index) <<Revcolors>>\] 0\]
        } else {
          .dialog.frame1.list itemconfigure \$sel_index -background \[lindex \[$w.textarea tag bind \$mem(\$sel_index) <<Revcolors>>\] 0\]
        }
        .dialog.frame1.list selection set \$sel_index \$sel_index
      }
    }
  "
  bind .dialog <<raise>> {
    foreach i [.dialog.frame1.list curselection] {
      if {$i > 0} {
        set nstore [.dialog.frame1.list get $i]
        set cstore [.dialog.frame1.list itemcget $i -background]
        .dialog.frame1.list delete $i
        set tstore $mem($i)
        set mem($i) $mem([expr {$i-1}])
        incr i -1
        .dialog.frame1.list insert $i $nstore
        .dialog.frame1.list itemconfigure $i -background $cstore       
        set mem($i) $tstore
        .dialog.frame1.list selection set $i
      }
    }
    if {[.dialog.frame1.list curselection] != {}} {
      .dialog.frame1.list see [lindex [.dialog.frame1.list curselection] 0]
    }
  } 
  bind .dialog <<lower>> {
    for {set j [expr {[llength [.dialog.frame1.list curselection]]-1}]} {$j > -1} {incr j -1} {
      set i [lindex [.dialog.frame1.list curselection] $j]
      if {$i < [expr {[.dialog.frame1.list index end] -1}]} {
        set nstore [.dialog.frame1.list get $i]
        set cstore [.dialog.frame1.list itemcget $i -background]
        .dialog.frame1.list delete $i
        set tstore $mem($i)
        set mem($i) $mem([expr {$i+1}])
        incr i 1
        .dialog.frame1.list insert $i $nstore
        .dialog.frame1.list itemconfigure $i -background $cstore       
        set mem($i) $tstore
        .dialog.frame1.list selection set $i
      }
    }
    if {[.dialog.frame1.list curselection] != {}} {
      .dialog.frame1.list see [lindex [.dialog.frame1.list curselection] end]
    }
  }
  bind .dialog <<delete>> {
    for {set j [expr {[llength [.dialog.frame1.list curselection]]-1}]} {$j > -1} {incr j -1} {
      set i [lindex [.dialog.frame1.list curselection] $j]
      .dialog.frame1.list delete $i
      .dialog.frame1.list selection set $i
      lappend deleted $mem($i)
      while {$i <= [expr {[.dialog.frame1.list index end] - 1}]} {
        set mem($i) $mem([expr {$i+1}])
        incr i
      }
      array unset mem $i 
    }
    .dialog.frame1.list selection clear 0 end
  } 

  vwait ok
  set dialogblock 0
  destroy .dialog

  bind . <<RaiseDialogs>> ""

  if {$ok == 1} {
    foreach i [lsort -integer [array names mem]] {
      $w.textarea tag lower $mem($i)
    }
    $w.textarea tag lower sel
    $w.textarea tag lower tempsel
    set tags [$w.textarea tag names]
    foreach tag $deleted {
      set info($w,saved) 0
      $w.textarea tag delete $tag
      if {[lsearch $tags [regsub "f" $tag "fn"]] > -1} {
        $w.textarea tag delete [regsub "f" $tag "fn"]
      }
    }
  }
  unset ok
  unset deleted
  array unset mem
}

##############
## sub-procedure of configure_features to rename
##############
proc rename_recolor_window_feature {parent w i} {
  global mem ok2 temp_info info
  if {$i == [list]} return
  set a [toplevel .edit_feature_dialog2]
  wm title $a [mc "Edit Feature..."]
  wm protocol $a WM_DELETE_WINDOW "set ok2 -1"
  set oldbind [bind . <<RaiseDialogs>>]
  bind . <<RaiseDialogs>> "wm deiconify $a; raise $a"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $parent] wgeom winx winy
  wm geometry $a "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient $a $parent
  } else {
    ::tk::unsupported::MacWindowStyle style $a document {closeBox}
  }
  set ok2 0

  set temp_info(name) [lindex [$w.textarea tag bind $mem($i) <<Metadata>>] 0]
  set temp_info(other) [lindex [$w.textarea tag bind $mem($i) <<Metadata>>] 1]
  set temp_info(type) [lindex [$w.textarea tag bind $mem($i) <<Metadata>>] 2]
  set temp_info(dir) [lindex [$w.textarea tag bind $mem($i) <<Metadata>>] 3]
  set temp_info(comment) [lindex [$w.textarea tag bind $mem($i) <<Metadata>>] 4]

  set temp_info(fwd_color) [lindex [$w.textarea tag bind $mem($i) <<Revcolors>>] 0]
  set temp_info(rev_color) [lindex [$w.textarea tag bind $mem($i) <<Revcolors>>] 1]

  grid [frame $a.textframe -relief ridge -borderwidth 2] -row 0 -column 1 -sticky nswe
  grid [frame $a.infoframe -relief ridge -borderwidth 2] -row 1 -column 1 -sticky nswe
  grid [frame $a.okframe -relief ridge -borderwidth 2] -row 2 -column 1 -columnspan 2 -sticky nswe
  grid rowconfigure $a 0 -weight 1
  grid rowconfigure $a 1 -weight 0
  grid rowconfigure $a 2 -weight 0
  grid columnconfigure $a 1 -weight 1

  grid [entry $a.textframe.name -textvariable temp_info(name) -font dnafont -width 50] -row 0 -column 0 -columnspan 2 -sticky nwe
  grid rowconfigure $a.textframe 1 -weight 1
  grid columnconfigure $a.textframe 0 -weight 1


  grid [label $a.infoframe.label1 -text "Feature type:"] -row 0 -column 1
  grid [menubutton $a.infoframe.type -menu $a.infoframe.type.menu -textvariable temp_info(type) -font labelfont] -row 0 -column 2
  menu $a.infoframe.type.menu
  set typelist [list source primer_bind CDS exon intron 5'UTR 3'UTR mRNA prim_transcript precursor_RNA 5'clip 3'clip misc_RNA rep_origin gene promoter enhancer polyA_site polyA_signal terminator CAAT_signal TATA_signal -35_signal -10_signal GC_signal RBS attenuator misc_signal sig_peptide transit_peptide mat_peptide variation STS unsure conflict modified_base misc_difference old_sequence misc_feature LTR repeat_region repeat_unit satellite rRNA tRNA scRNA snRNA snoRNA protein_bind misc_binding misc_recomb misc_structure iDNA stem_loop D-loop C_region D_segment J_segment N_region S_region V_region V_segment] 
  set divided_typelist [list Genes [list CDS exon intron gene 5'UTR 3'UTR mRNA prim_transcript precursor_RNA 5'clip 3'clip] Signals [list rep_origin promoter enhancer polyA_site polyA_signal terminator CAAT_signal TATA_signal -35_signal -10_signal GC_signal RBS attenuator misc_signal sig_peptide transit_peptide mat_peptide] Binding [list primer_bind protein_bind misc_binding] Variation [list variation STS unsure conflict modified_base misc_difference old_sequence] Repeats [list LTR repeat_region repeat_unit satellite] Misc_RNA [list misc_RNA rRNA tRNA scRNA snRNA snoRNA] Misc [list source misc_feature misc_binding misc_recomb misc_structure iDNA stem_loop D-loop] Ig [list C_region D_segment J_segment N_region S_region V_region V_segment]] 
  foreach {header types} $divided_typelist {
    $a.infoframe.type.menu add cascade -label $header -menu [menu $a.infoframe.type.menu.[string tolower $header]]
    foreach type $types {
      $a.infoframe.type.menu.[string tolower $header] add radiobutton -label $type -variable temp_info(type) -value $type
    }
  }
  if {[lsearch -exact $typelist $temp_info(type)] == -1} {
    $a.infoframe.type.menu add radiobutton -label $temp_info(type) -variable temp_info(type) -value $temp_info(type)
  }

  grid [label $a.infoframe.label2 -text "Forward color:"] -row 1 -column 1
  grid [color_button $a.infoframe.fwdcolor temp_info(fwd_color) $temp_info(fwd_color) 12] -sticky w -row 1 -column 2
  grid [label $a.infoframe.label3 -text "Reverse color:"] -row 1 -column 3
  grid [color_button $a.infoframe.revcolor temp_info(rev_color) $temp_info(rev_color) 12] -sticky w -row 1 -column 4
  grid [button $a.okframe.ok -text [mc "OK"] -command "set ok2 1"] -row 0 -column 0

  focus $a
  vwait ok2
  set return "cancel"
  ###execute changes
  if {$ok2 == 1} {
    set info($w,saved) 0
    if {[winfo rgb . $temp_info(fwd_color)] == [winfo rgb . $info(transp_feature_color)]} {
      set fwd_color {}
    } else {
      set fwd_color $temp_info(fwd_color)
    }
    if {[winfo rgb . $temp_info(rev_color)] == [winfo rgb . $info(transp_feature_color)]} {
      set rev_color {}
    } else {
      set rev_color $temp_info(rev_color)
    }

    $w.textarea tag bind $mem($i) <<Revcolors>> [list $fwd_color $rev_color]
    $w.textarea tag bind $mem($i) <<Metadata>> [list $temp_info(name) $temp_info(other) $temp_info(type) $temp_info(dir) $temp_info(comment)]
    if {$temp_info(dir) == 0} {
      $w.textarea tag configure $mem($i) -background $fwd_color
    } else {
      $w.textarea tag configure $mem($i) -background $rev_color
    }
    set return "ok"
  }
  destroy $a
  bind . <<RaiseDialogs>> $oldbind
  array unset temp_info
  unset ok2
  return $return
}



##############
## converts a list of lines from a genbank feature table to a tag list for put_tags
##############
proc feature_table_to_taglist {table} {
  global info
  set result [list]
  set subtype ""
  set subdata ""
  set type ""
  set range ""
  set source_tag ""
  set i 0
  set c 0
  set genbank_type_list [list]
  set genbank_data_list [list]
  set Apeinfo_label ""
  set Apeinfo_fwd ""
  set Apeinfo_rev ""
  lappend table "     end   end"
if {1} {
  foreach line $table {
    if {[regexp { {21,23}/(.*)=(.*)} $line line next_subtype next_subdata]} {
      ##store the sub data
      if {$subtype == "ApEinfo_label"} {
        set Apeinfo_label $subdata
      } elseif {$subtype == "ApEinfo_fwdcolor"} {
        set Apeinfo_fwd $subdata
      } elseif {$subtype == "ApEinfo_revcolor"} {
        set Apeinfo_rev $subdata
      } elseif {$subtype != ""} {
        lappend genbank_data_list $subdata
      }
      set subtype $next_subtype
      if {[lsearch [list ApEinfo_label ApEinfo_fwdcolor ApEinfo_revcolor] $subtype] == -1} {
        lappend genbank_type_list $subtype
      }
      set subdata $next_subdata
      set c 1
    } elseif {[regexp { {21,23}(.*)} $line line continue]} {
      if {$c == 0} {
        set range "$range$continue"
      } else {
        set subdata "$subdata $continue"
      }
    } elseif {[regexp { {5,7}([^ ]*) *(.*)} $line line next_type next_range]} {
      ##next feature heading line found
      ##finish the previous feature
      if {$subtype == "ApEinfo_label"} {
        set Apeinfo_label $subdata
      } elseif {$subtype == "ApEinfo_fwdcolor"} {
        set Apeinfo_fwd $subdata
      } elseif {$subtype == "ApEinfo_revcolor"} {
        set Apeinfo_rev $subdata
      } elseif {$subtype != ""} {
        lappend genbank_data_list $subdata
      }
      if {($type != "")} {
        ##interpret the range data
        if {[string match "*complement*" $range]} {
          set dir 1
        } else {
          set dir 0
        }
        regsub -all {(([0-9]+)\.([0-9]+))} $range {\2} new_range
        regsub -all {(([0-9]+)\^([0-9]+))} $new_range {\2} new_range
        regsub -all {join|complement|\(|\)|\.|\,} $new_range " " new_range
        regsub -all {\ +} $new_range " " new_range
        set new_range [string trim $new_range]
        set new_range [split $new_range " "]
        set temp_range [list]
        foreach {x y} $new_range {
          if {[string is integer -strict $x] && [string is integer -strict $y]} {
            lappend temp_range [expr {$x-1}] $y
          }
        }
        set new_range [lsort -integer -increasing $temp_range]
        if {[expr {[llength $new_range] % 2}] == 1} {lappend new_range [lindex $new_range end]}

        ##need to assign the name based on the sub-data
        #label gene product notes
        if {$Apeinfo_label != ""} {
          set name $Apeinfo_label
        } elseif {[set index [lsearch $genbank_type_list "label"]] > -1} {
          set name [lindex $genbank_data_list $index]
        } elseif {[set index [lsearch $genbank_type_list "product"]] > -1} {
          set name [lindex $genbank_data_list $index]
        } elseif {[set index [lsearch $genbank_type_list "gene"]] > -1} {
          set name [lindex $genbank_data_list $index]
        } elseif {[set index [lsearch $genbank_type_list "note"]] > -1} {
          set name [lindex $genbank_data_list $index]
        } elseif {$type == "source"} {
          if {[set index [lsearch $genbank_type_list "organism"]] > -1} {
            set name "source:[lindex $genbank_data_list $index]"
          } else {
            set name "source"
          }
        } else {
          set name $type
        }
        set name [string range $name 0 50]
        regsub -all {\"} $name "" name

        ##need to assign the colors based on the type
        if {$Apeinfo_fwd != ""} {
          set fwd_color $Apeinfo_fwd
          set rev_color $Apeinfo_rev
        } else {
          switch -- $type {
            "source" {
              set fwd_color {}
              set rev_color {}
            }
            exon -
            "CDS" {
              set fwd_color DarkOrange1
              set rev_color DarkOrange1
            }
            "primer_bind" {
              set fwd_color cyan
              set rev_color green
            }
            "intron" -
            "gene" {
              set fwd_color gray45
              set rev_color gray45
            }
            "rep_origin"  {
              set fwd_color SlateGray3
              set rev_color SlateGray3
            }
            "promoter" -
            "enhancer" - 
            "polyA_site" -
            "polyA_signal" - 
            "terminator" - 
            "CAAT_signal" -
            "TATA_signal" -
            "-35_signal" -
            "-10_signal" -
            "GC_signal" -
            "RBS" {
              set fwd_color seashell4
              set rev_color seashell4
            }
            "variation" -
            "STS" -  
            "unsure" - 
            "conflict" - 
            "modified_base" - 
            "misc_difference" - 
            "old_sequence" {
              set fwd_color goldenrod2
              set rev_color goldenrod2
            }
            "misc_feature" {
              set fwd_color maroon4
              set rev_color maroon4
            }
            default {
              set fwd_color pink
              set rev_color pink
            }
          }
        }

        set genbank_list [list]
        foreach subtype $genbank_type_list subdata $genbank_data_list {
            lappend genbank_list $subtype $subdata
        }
        if {1} {
          if {$dir == 0} {
            set formatlist [list $fwd_color {} {}]
          } else {
            set formatlist [list $rev_color {} {}]
          }
          set bindlist [list <<Revcolors>> <<Metadata>>]
          set actionlist [list [list $fwd_color $rev_color] [list $name {} $type $dir $genbank_list]]
          lappend result "f$i#" [list $new_range $formatlist $bindlist $actionlist]
          if {[llength $new_range] > 2} {
            lappend result "fn$i#" [list [list [lindex $new_range 0] [lindex $new_range end]] [list {} {} {}] [list] [list]]
          }
          incr i
        } else {
         set source_tag $genbank_list
        }
      }
      set type $next_type
      set range $next_range
      set subtype ""
      set subdata ""
      set genbank_type_list [list]
      set genbank_data_list [list]
      set Apeinfo_label ""
      set Apeinfo_fwd ""
      set Apeinfo_rev ""
      set c 0
    }
  }
}

##also need to return $source_tag
  return $result
}

##############
## reads a new feature library and stores the feature list in fea_info(library) 
##############
#file format(tab sep): (0)name (1)plain_text_def (2)type (3)fwd_color (4)rev_color (5)bold (6)ital (7)annotations_string
#library list: (0)name (1)regexp_fwd(flanked by paren) (2)regexp_rev (3)preseq (4)postseq (5)plain_text_def (6)type (7)fwd_color (8)rev_color (9)bold (10)ital (11)annotations_string

proc read_featurelib {{filename ""} {mode "new"}} {

  global info
  global fea_info dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1
sputs "reading feature file:$filename"
  set error 1
  while {$error == 1} {
    if {($filename == "") || (![file exists $filename])} {
      set focus_win [focus]
      if {![file isdirectory $info(default_featuredir)]} {
        global argv0
        set info(default_featuredir) [file dirname $argv0]
      }
      set filename [tk_getOpenFile -initialdir $info(default_featuredir) -title "open ApE Features file"]
      if {([winfo exists $focus_win]) && ($focus_win != ".")} {
        wm deiconify [winfo toplevel $focus_win]
        focus $focus_win
      }
    }
    if {[file exists $filename] == 1} {
      set fileid [open $filename r]
      set error 0
      set templibrary ""
      while {([eof $fileid] != 1) && $error == 0} {
        set nextline [gets $fileid]
        if {([string index $nextline 0] != "#") && ($nextline != "")} {
          set nextfea [split $nextline \t]
          if {([llength $nextfea] < 2) ||([llength $nextfea] > 8) || ([regexp -nocase {[^ACGTNBDHKMRSVWY\s\+\#<>]} [lindex $nextfea 1]] == 1)} {
            if {[tk_messageBox -message "$filename is an improperly formatted oligo library. \n $nextfea\nTry another?" -type yesno -icon error -default yes] == "yes"} {
              set error 1; #try again
              set filename ""
            } else {
              set error -1; #cancel
            }
          } else {
            set nextfea [concat $nextfea [lrange {{} {} {} {} {}} 0 [expr {6-[llength $nextfea]}]]]

            ######set defaults here
            if {[lindex $nextfea 2] == ""} {set nextfea [lreplace $nextfea 2 2 "misc_feature"]}
            if {[catch {winfo rgb . [lindex $nextfea 3]}]} {set nextfea [lreplace $nextfea 3 3 $info(feature_default_fcolor)]}
            if {[catch {winfo rgb . [lindex $nextfea 4]}]} {set nextfea [lreplace $nextfea 4 4 $info(feature_default_rcolor)]}
            if {[lindex $nextfea 5] != 1} {set nextfea [lreplace $nextfea 5 5 0]}
            if {[lindex $nextfea 6] != 1} {set nextfea [lreplace $nextfea 6 6 0]}


            ######set the regexp strings
            set regexpfwd [lindex $nextfea 1]
            #for all lower case, set to all upper case
            if {![regexp {[A-Z]} $regexpfwd]} {set regexpfwd [string toupper $regexpfwd]}   
            if {![regexp {(.*)<(.*)} $regexpfwd a preseq regexpfwd]} {
              set preseq ""
              ######for primer sequences, set the preseq to all leading lower case chars
              if {[lindex $nextfea 2] == "primer_bind"} {
                if {![regexp {([a-z]+)([^a-z]+.*)} $regexpfwd a preseq regexpfwd]} {
                  set preseq ""
                }
                set regexpfwd [string toupper $regexpfwd]
              }
            }
            if {![regexp {(.*)>(.*)} $regexpfwd a regexpfwd postseq]} {set postseq ""}
            set regexpfwd "($regexpfwd)"
            set preseq [string toupper $preseq]
            set postseq [string toupper $postseq]
            regsub -all {\s} $regexpfwd "" regexpfwd
            regsub -all {[a-z]+} $regexpfwd ")&(" regexpfwd
            regsub -all {[nN]} $regexpfwd "." regexpfwd
            #remove duplicate + and #
            regsub -all {\++} $regexpfwd "+" regexpfwd
            regsub -all {[\+\#][\+\#]+} $regexpfwd "#" regexpfwd
            #remove leading and trailing # and +
            regsub -all {^\+} $regexpfwd "" regexpfwd
            regsub -all {^\#} $regexpfwd "" regexpfwd
            regsub -all {\+$} $regexpfwd "" regexpfwd
            regsub -all {\#$} $regexpfwd "" regexpfwd
            #convert # and + to .*?
            regsub -all {\+} $regexpfwd ").*?(" regexpfwd
            regsub -all {\#} $regexpfwd ".*?" regexpfwd
            #remove empty parens
            regsub -all {\(\)} $regexpfwd "" regexpfwd

            #substitute degenerate codes
            set regexpfwd [regsubpattern $regexpfwd]

            #make the reverse pattern
            set regexprev [revcom_regexp $regexpfwd]
            regsub -all {\?\*\.} $regexprev .*? regexprev


            if {[string length [regsub -all -nocase {[^a-z]} $regexpfwd ""]] > 5} {
              #featurelib element:
              #(0)name (1)regexp_fwd(flanked by paren) (2)regexp_rev (3)preseq (4)postseq (5)plain_text_def (6)type (7)fwd_color (8)rev_color (9)bold (10)ital (11)annotations_string
              if {$regexpfwd != $regexprev} {
                lappend templibrary [linsert $nextfea 1 $regexpfwd $regexprev $preseq $postseq]
              } else {
                lappend templibrary [linsert $nextfea 1 $regexpfwd "" $preseq $postseq]
              }
            } else {
              sputs "Feature too short:[lindex $nextfea 0] $regexpfwd"
            }
          }
        }
      }; #while not end of file
      close $fileid
    } elseif {$filename == ""} {
      #open cancelled, don't try again
      set error -1 
    } else {
      #file doesn't exist, try again
      set error 1
    }
  } ; #while error reading file
  if {$error == 0} {
    if {$mode == "new"} {
      set fea_info(library) $templibrary
      regsub -all {\\} $filename {/} filename
      set fea_info(library_file) $filename
    } else {
      lappend fea_info(library) $templibrary
      set fea_info(library_file) ""
    }
    if {$info(featuredir_follows_open)} {
      set info(default_featuredir) [file dirname $filename]
    }
  } 
  set dialogblock 0
}

##############
## lists feature tags found in a DNA window 
##############
proc list_features {w} {
  global info toolbar_images
  global modstring modifier dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

#set t $a.textframe.text
#$t image create 7.0 -image $toolbar_images(info_button)
#$t tag add 7.0 7.0 button_tag
#$t tag bind button_tag <Enter> "$t configure -cursor arrow"
#$t tag bind button_tag <Leave> "$t configure -cursor xterm"

  set a [new_analysis_window $w "Features List"]
  create_textframe $a 80
  $a.textframe.text configure -tabs {6c 9c}

  $a.textframe.text insert end "[clock format [clock seconds] -format $info(clock_format)]\n"
  $a.textframe.text insert end "$info($w,win_title) \n"
  $a.textframe.text insert end "$info($w,filename) \n"
  $a.textframe.text insert end "Features List\n\n"

  set temp_fea_list [list]
  foreach tag [$w.textarea tag names] {
    if {[regexp {f[0-9]+#} $tag]} {
      set x1 [lindex [$w.textarea tag ranges $tag] 0]
      regsub "1." $x1 "" x1
      if {$x1 != "" } {
        lappend temp_fea_list [list $x1 $tag]
      } else {
        sputs "feature_list- empty feature $x1 $tag"
      }
    }
  }
  #set temp_fea_list [lsort -index 0 -integer $temp_fea_list]

  $a.textframe.text insert end "Name\ttype\tdirection\tlocation\n" heading
  $a.textframe.text tag configure heading -underline true

  foreach fea $temp_fea_list {
    set tag [lindex $fea 1]
    set metadata [$w.textarea tag bind $tag <<Metadata>>]

 
    set loc [list]
    foreach {x1 x2} [$w.textarea tag ranges $tag] {
      regsub "1." $x1 "" x1
      incr x1
      regsub "1." $x2 "" x2
      if {[lindex $metadata 3] == 1} {
        set loc [concat "$x2..$x1" $loc]
      } else {
        lappend loc "$x1..$x2"
      }
    }
    set loc [join $loc ,]
    $a.textframe.text insert end "[lindex $metadata 0]\t[lindex $metadata 2]\t[expr {[lindex $metadata 3]?"rev":"fwd"}]\t$loc\n" $tag

    $a.textframe.text tag bind $tag <Any-Button> "select_region $w {1.[lindex $fea 0]} {1.$x2}"
    lappend info($w,linked_events) [list $a.textframe.text $tag <Any-Button>]
    $a.textframe.text tag bind $tag <Any-Enter> "%W tag configure $tag -foreground red; %W tag raise $tag sel"
    $a.textframe.text tag bind $tag <Any-Leave> "%W tag configure $tag -foreground black; %W tag lower $tag sel"

  }

  set outputmenu [add_output_menu $a.textframe.text "Text" text $w]
  bind $a <Button-1> "focus $a.textframe.text" 
  bindtags $a.textframe.text [list $a.textframe.text AnalysisText . all]
  wm deiconify $a
  window_on_screen $a
  focus $a.textframe.text
  set dialogblock 0
}

##############
## edit a feature element
##############
#file format(tab sep): (0)name (1)plain_text_def (2)type (3)fwd_color (4)rev_color (5)bold (6)ital (7)annotations_string
#library list: (0)name (1)regexp_fwd(flanked by paren) (2)regexp_rev (3)preseq (4)postseq (5)plain_text_def (6)type (7)fwd_color (8)rev_color (9)bold (10)ital (11)annotations_string
proc edit_feature {w feature_number} {
  global info temp_info fea_info ok2 dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  bind $w <FocusIn> "event generate . <<RaiseDialogs>>"
  bind $w <Map> "event generate . <<RaiseDialogs>>"
  bind $w <Activate> "event generate . <<RaiseDialogs>>"

  set temp_info(name) [lindex [lindex $fea_info(library) $feature_number] 0]
  set temp_info(regexp) [lindex [lindex $fea_info(library) $feature_number] 5]
  set temp_info(fwd_color) [lindex [lindex $fea_info(library) $feature_number] 7]
  set temp_info(rev_color) [lindex [lindex $fea_info(library) $feature_number] 8]
  set temp_info(type) [lindex [lindex $fea_info(library) $feature_number] 6]
  set temp_info(notes) [lindex [lindex $fea_info(library) $feature_number] 11]

  set a [toplevel .edit_feature_dialog]
  wm title $a [mc "Edit Feature..."]
  wm protocol $a WM_DELETE_WINDOW "set ok2 -1"
  set raisedialogstore [bind . <<RaiseDialogs>>]
  bind . <<RaiseDialogs>> "wm deiconify $a; raise $a"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry $a "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient $a $w
  } else {
    ::tk::unsupported::MacWindowStyle style $a document {closeBox}
  }
  set ok2 0

  grid [frame $a.textframe -relief ridge -borderwidth 2] -row 0 -column 1 -sticky nswe
  grid [frame $a.infoframe -relief ridge -borderwidth 2] -row 1 -column 1 -sticky nswe
  grid [frame $a.okframe -relief ridge -borderwidth 2] -row 2 -column 1 -columnspan 2 -sticky nswe
  grid rowconfigure $a 0 -weight 1
  grid rowconfigure $a 1 -weight 0
  grid rowconfigure $a 2 -weight 0
  grid columnconfigure $a 1 -weight 1

  grid [entry $a.textframe.name -textvariable temp_info(name) -font dnafont -width 50] -row 0 -column 0 -columnspan 2 -sticky nwe
  grid [text $a.textframe.regexp -height 10 -width 50 -wrap char -yscrollcommand "optionscrollbar $a.textframe.scroll" -font dnafont] -row 1 -column 0 -sticky nswe
  grid [scrollbar $a.textframe.scroll -command "$a.textframe.regexp yview"] -row 1 -column 1 -sticky nse
  grid [frame $a.textframe.buttons] -row 2 -column 0 -columnspan 2 -sticky nswe
  grid [button $a.textframe.buttons.upper -text "Uppercase" -command "event generate $a.textframe.regexp <<toupper>>"] -row 0 -column 0 -sticky ns
  grid [button $a.textframe.buttons.lower -text "Lowercase" -command "event generate $a.textframe.regexp <<tolower>>"] -row 0 -column 1 -sticky ns

  grid rowconfigure $a.textframe 1 -weight 1
  grid columnconfigure $a.textframe 0 -weight 1

  $a.textframe.regexp insert 1.0 $temp_info(regexp)

  grid [label $a.infoframe.label1 -text "Feature type:"] -row 0 -column 1
  grid [menubutton $a.infoframe.type -menu $a.infoframe.type.menu -textvariable temp_info(type) -font labelfont] -row 0 -column 2
  grid [label $a.infoframe.label2 -text "Forward color:"] -row 1 -column 1
  grid [color_button $a.infoframe.fwdcolor temp_info(fwd_color) $temp_info(fwd_color) 12] -sticky w -row 1 -column 2
  grid [label $a.infoframe.label3 -text "Reverse color:"] -row 1 -column 3
  grid [color_button $a.infoframe.revcolor temp_info(rev_color) $temp_info(rev_color) 12] -sticky w -row 1 -column 4
  grid [button $a.okframe.ok -text [mc "OK"] -command "set ok2 1"] -row 0 -column 0

  menu $a.infoframe.type.menu
  set typelist [list primer_bind CDS exon intron 5'UTR 3'UTR mRNA prim_transcript precursor_RNA 5'clip 3'clip misc_RNA rep_origin gene promoter enhancer polyA_site polyA_signal terminator CAAT_signal TATA_signal -35_signal -10_signal GC_signal RBS attenuator misc_signal sig_peptide transit_peptide mat_peptide variation STS unsure conflict modified_base misc_difference old_sequence misc_feature LTR repeat_region repeat_unit satellite rRNA tRNA scRNA snRNA snoRNA protein_bind misc_binding misc_recomb misc_structure iDNA stem_loop D-loop C_region D_segment J_segment N_region S_region V_region V_segment] 
  set divided_typelist [list Genes [list CDS exon intron gene 5'UTR 3'UTR mRNA prim_transcript precursor_RNA 5'clip 3'clip] Signals [list rep_origin promoter enhancer polyA_site polyA_signal terminator CAAT_signal TATA_signal -35_signal -10_signal GC_signal RBS attenuator misc_signal sig_peptide transit_peptide mat_peptide] Binding [list primer_bind protein_bind misc_binding] Variation [list variation STS unsure conflict modified_base misc_difference old_sequence] Repeats [list LTR repeat_region repeat_unit satellite] Misc_RNA [list misc_RNA rRNA tRNA scRNA snRNA snoRNA] Misc [list misc_feature misc_binding misc_recomb misc_structure iDNA stem_loop D-loop] Ig [list C_region D_segment J_segment N_region S_region V_region V_segment]] 
  foreach {header types} $divided_typelist {
    $a.infoframe.type.menu add cascade -label $header -menu [menu $a.infoframe.type.menu.[string tolower $header]]
    foreach type $types {
      $a.infoframe.type.menu.[string tolower $header] add radiobutton -label $type -variable temp_info(type) -value $type
    }
  }
  if {[lsearch -exact $typelist $temp_info(type)] == -1} {
    $a.infoframe.type.menu add radiobutton -label $temp_info(type) -variable temp_info(type) -value $temp_info(type)
  }
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $a] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry $a "+$winx+$winy"
  update

  bind $a.textframe.regexp <<toupper>> "
    if {\[$a.textframe.regexp tag ranges sel\] != {}} {
      set temptext \[$a.textframe.regexp get sel.first sel.last\] 
      $a.textframe.regexp delete sel.first sel.last
      $a.textframe.regexp insert insert \[string toupper \$temptext\]
      $a.textframe.regexp tag add sel insert-\[string length \$temptext\]char insert
    }
  "
  bind $a.textframe.regexp <<tolower>> "
    if {\[$a.textframe.regexp tag ranges sel\] != {}} {
      set temptext \[$a.textframe.regexp get sel.first sel.last\] 
      $a.textframe.regexp delete sel.first sel.last
      $a.textframe.regexp insert insert \[string tolower \$temptext\]
      $a.textframe.regexp tag add sel insert-\[string length \$temptext\]char insert
    }
  "

  focus $a
  set continue 1

  while {$continue} {

    vwait ok2
    ###execute changes
    if {$ok2 == 1} {
      set featuretext [$a.textframe.regexp get 1.0 end]
      regsub -all "\ +" $featuretext "" featuretext 
      regsub -all "\n+" $featuretext "" featuretext 
      set regexpfwd $featuretext
      #for all lower case, set to all upper case
      if {![regexp {[A-Z]} $regexpfwd]} {set regexpfwd [string toupper $regexpfwd]}   
      if {![regexp {(.*)<(.*)} $regexpfwd fill preseq regexpfwd]} {
        set preseq ""
        ######for primer sequences, set the preseq to all leading lower case chars
        if {$temp_info(type) == "primer_bind"} {
          if {![regexp {([a-z]+)([^a-z]+.*)} $regexpfwd fill preseq regexpfwd]} {
            set preseq ""
          }
          set regexpfwd [string toupper $regexpfwd]
        }
      }
      if {![regexp {(.*)>(.*)} $regexpfwd blank regexpfwd postseq]} {set postseq ""}
      set regexpfwd "($regexpfwd)"
      set preseq [string toupper $preseq]
      set postseq [string toupper $postseq]
      regsub -all {\s} $regexpfwd "" regexpfwd
      regsub -all {[a-z]+} $regexpfwd ")&(" regexpfwd
      regsub -all {[nN]} $regexpfwd "." regexpfwd
      #remove duplicate + and #
      regsub -all {\++} $regexpfwd "+" regexpfwd
      regsub -all {[\+\#][\+\#]+} $regexpfwd "#" regexpfwd
      #remove leading and trailing # and +
      regsub -all {^\+} $regexpfwd "" regexpfwd
      regsub -all {^\#} $regexpfwd "" regexpfwd
      regsub -all {\+$} $regexpfwd "" regexpfwd
      regsub -all {\#$} $regexpfwd "" regexpfwd
      #convert # and + to .*?
      regsub -all {\+} $regexpfwd ").*?(" regexpfwd
      regsub -all {\#} $regexpfwd ".*?" regexpfwd
      #remove empty parens
      regsub -all {\(\)} $regexpfwd "" regexpfwd

      #substitute degenerate codes
      set regexpfwd [regsubpattern $regexpfwd]

      #make the reverse pattern
      set regexprev [revcom_regexp $regexpfwd]
      regsub -all {\?\*\.} $regexprev .*? regexprev


      if {[string length [regsub -all -nocase {[^a-z]} $regexpfwd ""]] > 5} {
        #featurelib element:
        #(0)name (1)regexp_fwd(flanked by paren) (2)regexp_rev (3)preseq (4)postseq (5)plain_text_def (6)type (7)fwd_color (8)rev_color (9)bold (10)ital (11)annotations_string
        if {$regexpfwd != $regexprev} {
          set newfea [list $temp_info(name) $regexpfwd $regexprev $preseq $postseq $featuretext $temp_info(type) $temp_info(fwd_color) $temp_info(rev_color) 0 0 ""]
        } else {
          set newfea [list $temp_info(name) $regexpfwd "" $preseq $postseq $featuretext $temp_info(type) $temp_info(fwd_color) $temp_info(rev_color) 0 0 ""]
        }
        #sputs $newfea
        set fea_info(library) [lreplace [K $fea_info(library) [set fea_info(library) {}]] $feature_number $feature_number $newfea]
        set return "ok"
        set continue 0

      } else {
        bell
        tk_messageBox -message "$preseq<$regexpfwd>$postseq isn't valid, try again." -icon warning -type ok 
        set ok2 0
        set continue 1

      }
    } else {
      set return "cancel"
      set continue 0

    }
  }


  set dialogblock 0
  destroy $a
  bind . <<RaiseDialogs>> $raisedialogstore
  bind $w <FocusIn> ""
  bind $w <Map> ""
  bind $w <Activate> ""

  unset ok2
  array unset temp_info
  set return $return
}

##############
## adds selected DNA to curent feature library via edit_feature
##############
#library list: (0)name (1)regexp_fwd(flanked by paren) (2)regexp_rev (3)preseq (4)postseq (5)plain_text_def (6)type (7)fwd_color (8)rev_color (9)bold (10)ital (11)annotations_string

proc add_sel_feature_library {w} {
  global info fea_info dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  if {[$w.textarea tag ranges sel] != {}} {
    set text [$w.textarea get sel.first sel.last]
    set regexpfwd $text
    regsub -all {[a-z]+} $regexpfwd ")&(" regexpfwd
    regsub -all {[nN]} $regexpfwd "." regexpfwd
    regsub -all {\(\)} $regexpfwd "" regexpfwd
    set regexpfwd [regsubpattern $regexpfwd]

    set regexprev [revcom_regexp $regexpfwd]
    regsub -all {\?\*\.} $regexprev .*? regexprev
  } else {
    set text ""
    set regexpfwd ""
    set regexprev ""
  }
  set new_fea [list "New Feature" "($regexpfwd\)" "($regexprev\)" "" "" $text "misc_feature" $info(feature_default_fcolor) $info(feature_default_rcolor)]
  lappend fea_info(library) $new_fea
  set dialogblock 0
  edit_feature_library $w [expr {[llength $fea_info(library)]-1}]
}

##############
## edit, import and save curent feature library
##############
#library list: (0)name (1)regexp_fwd(flanked by paren) (2)regexp_rev (3)preseq (4)postseq (5)plain_text_def (6)type (7)fwd_color (8)rev_color (9)bold (10)ital (11)annotations_string

proc edit_feature_library {w {select -1}} {
#new
#import
#save as
  global info fea_info
  global ok
  global mem deleted dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  toplevel .dialog
  wm title .dialog [mc "Edit Feature LIbrary..."]
  wm protocol .dialog WM_DELETE_WINDOW "set ok -1"
  bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry .dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .dialog $w
  } else {
    #::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
    #.dialog configure -menu [menu .dialog.menubar]
  }
  set ok 0

  menu .dialog.menubar
  menu .dialog.menubar.librarymenu
    .dialog.menubar add cascade -menu .dialog.menubar.librarymenu -label [mc "Library"]
    .dialog.menubar.librarymenu add command -label [mc "Open"] -state disabled   -command "bell"
    .dialog.menubar.librarymenu add command -label [mc "Import Features"]  -state disabled  -command "bell"
    .dialog.menubar.librarymenu add command -label [mc "Save As..."] -command "set dialogblock 0; save_feature_library .dialog; set dialogblock 1"

  .dialog configure -menu .dialog.menubar

  if {[tk windowingsystem] == "aqua"} {
    #define the apple menu
    menu .dialog.menubar.apple
    .dialog.menubar add cascade -menu .dialog.menubar.apple
    .dialog.menubar.apple add command -label [mc "About ApE..."] -command "about_dialog $w" -state disabled
    .dialog.menubar.apple add separator
    .dialog.menubar.apple add command -label [mc "Preferences..."]  -command "configure_preferences $w" -state disabled 
  }


  grid [frame .dialog.frame1 -relief flat] -row 0 -column 0 -sticky nswe
  grid [listbox .dialog.frame1.list -yscrollcommand "optionscrollbar .dialog.frame1.scroll" -selectmode extended -activestyle none -background white -exportselection 0] -row 0 -column 0 -sticky nswe
  grid [scrollbar .dialog.frame1.scroll -command ".dialog.frame1.list yview"] -row 0 -column 1 -sticky nse
  grid rowconfigure .dialog.frame1 0 -weight 1
  grid columnconfigure .dialog.frame1 0 -weight 1
  grid columnconfigure .dialog.frame1 1 -weight 0
  grid [frame .dialog.frame2 -relief ridge -borderwidth 2] -row 1 -column 0 -sticky swe
  grid [button .dialog.frame2.selall -text [mc "Select All"] -command "event generate .dialog <<sel_all>>"] -row 0 -column 0 -sticky we
  grid [button .dialog.frame2.raise -text [mc "Raise"] -command "event generate .dialog <<raise>>"] -row 0 -column 1 -sticky we
  grid [button .dialog.frame2.lower -text [mc "Lower"] -command "event generate .dialog <<lower>>"] -row 0 -column 2 -sticky we
  grid [button .dialog.frame2.delete -text [mc "Delete"] -command "event generate .dialog <<delete>>"] -row 0 -column 3 -sticky we
  grid [button .dialog.frame2.new -text [mc "New"] -command "event generate .dialog <<new>>"] -row 0 -column 4 -sticky we


  grid [frame .dialog.frame3 -relief raised] -row 2 -column 0 -sticky swe
  grid [button .dialog.frame3.ok -text [mc "Save Changes"] -command "set ok 1" -default active] -row 0 -column 0 -padx 5 -sticky we
  grid [button .dialog.frame3.ok_scan -text [mc "Save Changes As..."] -command "set ok 2"] -row 0 -column 1 -padx 5 -sticky we
  grid [button .dialog.frame3.cancel -text [mc "Cancel"] -command "set ok -1"] -row 0 -column 2 -padx 5 -sticky we
  grid [checkbutton .dialog.frame3.scan -text [mc "Scan on Close"] -variable info(feature_scan_on_close) -onvalue 1 -offvalue 0 -selectcolor white] -row 0 -column 3 -padx 5 -sticky we
  grid [label .dialog.filelabel -textvariable fea_info(library_file) -justify right -width 50] -row 3 -column 0 -sticky new
  grid rowconfigure .dialog 0 -weight 1
  grid rowconfigure .dialog 1 -weight 0
  grid rowconfigure .dialog 2 -weight 0
  grid columnconfigure .dialog 0 -weight 1

  #bind .dialog.frame1.list <Shift-Button-1> {listbox_multiple %W [%W index @%x,%y];break}
  #bind .dialog.frame1.list <Button-1> {bell; .dialog.frame1.list selection clear 0 end; .dialog.frame1.list selection set [%W index @%x,%y]}

  bind .dialog.frame1.list <Double-Button-1> "
    set dialogblock 0
    set pointer \[%W index @%x,%y\]
    edit_feature .dialog \$mem(\$pointer)
    set dialogblock 1
    .dialog.frame1.list delete \$pointer
    .dialog.frame1.list insert \$pointer \[lindex \[lindex \$fea_info(library) \$mem(\$pointer)] 0\]
    .dialog.frame1.list itemconfigure \$pointer -background \[lindex \[lindex \$fea_info(library) \$mem(\$pointer)] 7\]
    unset pointer
  "



  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .dialog] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry .dialog "+$winx+$winy"
  update
  focus .dialog.frame1.list


  bind .dialog <<fill_list>> {
    array unset mem
    set i [expr {[llength $fea_info(library)]-1}]
    set n 0
    foreach feature $fea_info(library) {
      .dialog.frame1.list insert 0 [lindex $feature 0]
      .dialog.frame1.list itemconfigure 0 -background [lindex $feature 7]
      set mem($i) $n
      incr i -1
      incr n 1
    }
    set feature ""
    unset n i feature
    set deleted [list]
    .dialog.frame1.list selection set 0
  }

  bind .dialog <<sel_all>> {
    .dialog.frame1.list selection set 0 end
  }

  bind .dialog <<raise>> {
    foreach i [.dialog.frame1.list curselection] {
      if {$i > 0} {
        set nstore [.dialog.frame1.list get $i]
        set cstore [.dialog.frame1.list itemcget $i -background]
        .dialog.frame1.list delete $i
        set tstore $mem($i)
        set mem($i) $mem([expr {$i-1}])
        incr i -1
        .dialog.frame1.list insert $i $nstore
        .dialog.frame1.list itemconfigure $i -background $cstore       
        set mem($i) $tstore
        .dialog.frame1.list selection set $i
        unset nstore cstore tstore
      }

    }
    if {[.dialog.frame1.list curselection] != {}} {
      .dialog.frame1.list see [lindex [.dialog.frame1.list curselection] 0]
      unset i
    }

  } 

  bind .dialog <<lower>> {
    for {set j [expr {[llength [.dialog.frame1.list curselection]]-1}]} {$j > -1} {incr j -1} {
      set i [lindex [.dialog.frame1.list curselection] $j]
      if {$i < [expr {[.dialog.frame1.list index end] -1}]} {
        set nstore [.dialog.frame1.list get $i]
        set cstore [.dialog.frame1.list itemcget $i -background]
        .dialog.frame1.list delete $i
        set tstore $mem($i)
        set mem($i) $mem([expr {$i+1}])
        incr i 1
        .dialog.frame1.list insert $i $nstore
        .dialog.frame1.list itemconfigure $i -background $cstore       
        set mem($i) $tstore
        .dialog.frame1.list selection set $i
        unset nstore cstore tstore
      }
    }
    if {[.dialog.frame1.list curselection] != {}} {
      .dialog.frame1.list see [lindex [.dialog.frame1.list curselection] end]
      unset i j
    }
  }
  bind .dialog <<delete>> {
    for {set j [expr {[llength [.dialog.frame1.list curselection]]-1}]} {$j > -1} {incr j -1} {
      set i [lindex [.dialog.frame1.list curselection] $j]
      .dialog.frame1.list delete $i
      .dialog.frame1.list selection set $i
      lappend deleted $mem($i)
      while {$i <= [expr {[.dialog.frame1.list index end] - 1}]} {
        set mem($i) $mem([expr {$i+1}])
        incr i
      }
      array unset mem $i 
    }
    unset j i
  } 

  bind .dialog <<new>> {
    set new_fea [list "New Feature" "" "" "" "" "" "misc_feature" $info(feature_default_fcolor) $info(feature_default_rcolor)]
    lappend fea_info(library) $new_fea
    set dialogblock 0
    set result [edit_feature .dialog [expr {[llength $fea_info(library)]-1}]]
    sputs "result=$result"
    if {$result != "ok"} {
      set fea_info(library) [lrange $fea_info(library) 0 end-1]
    } else {
      for {set i [.dialog.frame1.list index end]} {$i > 0} {incr i -1} {
        set mem($i) $mem([expr {$i-1}])
      }
      set mem(0) [expr {[llength $fea_info(library)]-1}]
      .dialog.frame1.list insert 0 [lindex [lindex $fea_info(library) end] 0]
      .dialog.frame1.list itemconfigure 0 -background [lindex [lindex $fea_info(library) end] 7]
      .dialog.frame1.list selection set 0
      unset i new_fea
    }
  }


  event generate .dialog <<fill_list>>

 ##if select !=-1, then we are editing a new element
  if {$select != -1} {
    .dialog.frame1.list selection clear 0 end
    .dialog.frame1.list selection set $mem($select)
    .dialog.frame1.list see $mem($select)
    set dialogblock 0
    edit_feature .dialog $select
    set dialogblock 1
    #.dialog.frame1.list itemconfigure $mem($select) -background [lindex [lindex $fea_info(library) $select] 7]
    .dialog.frame1.list delete $mem($select)
    .dialog.frame1.list insert $mem($select) [lindex [lindex $fea_info(library) $select] 0] 
    .dialog.frame1.list itemconfigure $mem($select) -background [lindex [lindex $fea_info(library) $select] 7]
  }
  vwait ok
  set dialogblock 0
  destroy .dialog

  bind . <<RaiseDialogs>> ""

  if {$ok > 0} {
    ##put the changes into the feature library
    foreach i [lsort -decreasing -integer [array names mem]] {
      lappend new_feature_list [lindex $fea_info(library) $mem($i)]
    }
    set fea_info(library) $new_feature_list

    if {$ok == 1} { 
      save_feature_library $w $fea_info(library_file)
    } elseif {$ok ==2} {
      save_feature_library $w
    }

    if {$info(feature_scan_on_close) == 1} {
      feature_scan $w
    }
  } else {
    ###revert to the saved version of the file
    if {[file readable $fea_info(library_file)]} {
      read_featurelib $fea_info(library_file) new
    } else {
     set fea_info(library_file) ""
    }
  }
  unset ok
  unset deleted
  array unset mem

}

##############
## save a feature library
##############
#file format(tab sep): (0)name (1)plain_text_def (2)type (3)fwd_color (4)rev_color (5)bold (6)ital (7)annotations_string
#library list: (0)name (1)regexp_fwd(flanked by paren) (2)regexp_rev (3)preseq (4)postseq (5)plain_text_def (6)type (7)fwd_color (8)rev_color (9)bold (10)ital (11)annotations_string
proc save_feature_library {w {filename ""}} {
  global info fea_info
  global mem deleted dialogblock

  if {$dialogblock == 1} {return}
  set dialogblock 1
  if {![file isdirectory $info(default_featuredir)]} {
    global argv0
    set info(default_featuredir) [file dirname $argv0]
  }
  if {![file exists $filename]} {
    set filename [tk_getSaveFile -title "Save Feature File" -parent $w -initialdir $info(default_featuredir) -defaultextension ".txt"]
  }

  while {([file exists $filename]) && (![file writable $filename])} {
    tk_messageBox -title "Feature File not Writable" -message "The file \"$filename\" is not writable. It may be locked." -type ok -icon warning -default ok
    set filename [tk_getSaveFile -title "Save Feature File" -initialdir $info(default_featuredir) -defaultextension ".txt"]
  }

  if {$filename != ""} {
    set feature_file [open $filename w]
  
    foreach feature $fea_info(library) {
      puts $feature_file "[lindex $feature 0]\t[lindex $feature 5]\t[lindex $feature 6]\t[lindex $feature 7]\t[lindex $feature 8]\t[lindex $feature 9]\t[lindex $feature 10]\t[lindex $feature 11]"
    }
    close $feature_file
    set fea_info(library_file) $filename

  } else {
      set fea_info(library_file) ""
  }
  set dialogblock 0
}


##############
## scans a dna window textbox and stores the cutsites in the info array as info($w,$enzyme) 
##############
##this can scan 34kb in about a second
##----need to add way to deal with degenerate bases (regexp??)

proc findenzymes {w} {
 
  #set atime [clock clicks]

  global info

  if {$info($w,scanned) == 1} {return}

  global longnamelist 
  global hexlist 
  global enzymes
  global enzinfo


  foreach enzyme $enzymes {
    set info($w,$enzyme) [list]
  }
  set info($w,MDam) [list]
  set info($w,MDcm) [list]

  set str [$w.textarea get 1.0 "end-1 chars"]
  set last [lindex [split [$w.textarea index end-1chars] "."] 1]

  if {[string length $str] < 4} {return}

  if {$info($w,circular) == "circular"} {
    set str "$str[$w.textarea get 1.0 1.63]"
  }

  ## scan the longnames using the regexp search function
  set ctime [clock clicks]

  foreach enzymeelement $longnamelist {
    set enzname [lindex $enzymeelement 0]
    set pattern [lindex $enzymeelement 1]

    foreach found [regexp -nocase -indices -all -inline -- $pattern $str] {
      if {[lindex $found 0] < $last} {
        lappend info($w,$enzname) [expr {[lindex $found 0]+1}]
      }
    }
  }
  set dtime [clock clicks]

  ## scan using the hexarray for all sites <6bp long
  array set a {a 0 A 0 c 1 C 1 g 2 G 2 t 3 T 3 n 0 N 0 * 0 b 1 B 1 d 0 D 0 h 0 H 0 k 2 K 2 m 0 M 0 r 0 R 0 s 1 S 1 v 0 V 0 w 0 W 0 y 1 Y 1}
  array set b {a 8 A 8 c 4 C 4 g 2 G 2 t 1 T 1 n 15 N 15 * 15 b 7 B 7 d 11 D 11 h 13 H 13 k 3 K 3 m 12 M 12 r 10 R 10 s 6 S 6 v 14 V 14 w 9 W 9 y 5 Y 5}

  if {$info($w,circular) == "circular"} {
    set len [expr ($last+5)]
  } else {
    set len $last
  }

  set h 0

  for {set i 0} {$i < 5} {incr i} {
    set h [expr {($h << 2) & 0xfff | $a([string index $str $i])}]
  }

  set v "0x[string map {A 8 B 7 C 4 D b G 2 H d K 3 M c N f * 15 R a S 6 T 1 V e W 9 Y 5} [string toupper [string range $str 0 4]]]"

  for {set i 5} {$i < $len} {incr i} {
    set h [expr {($h << 2) & 0xfff | $a([string index $str $i])}]
    set v [expr {($v << 4) & 0xffffff | $b([string index $str $i])}]
    if {[lindex $hexlist $h] != {}} {
      foreach enzname [lindex $hexlist $h] {
        if {[expr {$enzinfo(enz_vcode,$enzname) & $v}] == 0} {
          lappend info($w,$enzname) [expr {$i-4}]
        }
      }
    }
  }
  #set btime [clock clicks]
  initialize_methylation $w
  set info($w,scanned) 1
  if {$info($w,Dam_Dcm_methylated) == 1} {toggle_methylation $w}

#warning_dialog "Total time [expr $btime-$atime]ms."
#warning_dialog "Longnames time [expr $dtime-$ctime]ms."
}

################
## for Dam/Dcm blocked enzymes, add a list of unblocked sites into info($w,mstore,$enzyme)
################
proc initialize_methylation {w} {
  global info
  global enzinfo

  if {($info($w,circular) == "circular")} {
    set last [expr {[lindex [split [$w.textarea index 1.end] .] 1]}]
    if {[llength $info($w,MDcm)] > 1 } {
      set info($w,MDcm) [concat [expr {[lindex $info($w,MDcm) end-1]-$last}] [expr {[lindex $info($w,MDcm) end]-$last}] $info($w,MDcm) [expr {[lindex $info($w,MDcm) 0]+$last}] [expr {[lindex $info($w,MDcm) 1]+$last}]]
    } elseif {[llength $info($w,MDcm)] > 0} {
      set info($w,MDcm) [concat [expr {[lindex $info($w,MDcm) end]-$last}] $info($w,MDcm) [expr {[lindex $info($w,MDcm) 0]+$last}]]
    }
    if {[llength $info($w,MDam)] > 1 } {
      set info($w,MDam) [concat [expr {[lindex $info($w,MDam) end-1]-$last}] [expr {[lindex $info($w,MDam) end]-$last}] $info($w,MDam) [expr {[lindex $info($w,MDam) 0]+$last}] [expr {[lindex $info($w,MDam) 1]+$last}]]
    } elseif {[llength $info($w,MDam)] > 0} {
      set info($w,MDam) [concat [expr {[lindex $info($w,MDam) end]-$last}] $info($w,MDam) [expr {[lindex $info($w,MDam) 0]+$last}]]
    }
  }
#example: {MDam {{BcgI 2 9} {BclI 2 3} {BsaBI 1 8} {BspHI 5 0} {ClaI 0 5} {HphI 4 0} {Hpy188I 4 0} {MboI 1 2} {MboII 4 0} {NruI 5 0} {TaqI 3 0} {XbaI 5 0}}} {MDcm {{AlwNI 0 8} {ApaI 5 0} {AvaII 4 0} {PflMI 1 9} {PpuMI 6 0} {ScrFI 1 3} {SexAI 2 4} {SfiI 12 0} {StuI 4 1}}}
  foreach enzyme $info(blocked_enzymes) {
    set info($w,mstore,$enzyme) $info($w,$enzyme)
  }
  foreach methylase $enzinfo(enz_methylaselist) {
    set metsiteslist $info($w,[lindex $methylase 0])
    foreach blocked_enzyme [lindex $methylase 1] { 
      set i 0
      set enzyme [lindex $blocked_enzyme 0]
      while {$i < [llength $info($w,mstore,$enzyme)]} {
        set site [lindex $info($w,mstore,$enzyme) $i]
        if {([lsearch $metsiteslist [expr {$site+[lindex $blocked_enzyme 1]}] ]> -1) || ([lsearch $metsiteslist [expr {$site+[lindex $blocked_enzyme 2]}]] > -1)} {
          set info($w,mstore,$enzyme) [lreplace $info($w,mstore,$enzyme) $i $i]
        } else { 
          incr i
        }
      }      
    }
  }
  if {[array names info "$w,DpnI"] != ""} {
    set info($w,DpnI) [list]
  }
}

################
## Turn Dam/Dcm methylation on/off by swapping info($w,$enzyme) and info($w,mstore,$enzyme)
################
proc toggle_methylation {w} {
  global info

  if {$info($w,scanned) == 0} {
    findenzymes $w
    return
  }

  foreach enzyme $info(blocked_enzymes) {
    set templist $info($w,$enzyme) 
    set info($w,$enzyme) $info($w,mstore,$enzyme)
    set info($w,mstore,$enzyme) $templist
  }

}
################
## Primer scan conditions dialog
################
proc primer_window {w} {
  global info
  global ok dialogblock modifier
  if {$dialogblock == 1} {return}
  set dialogblock 1

  if {[winfo exists .primer_conditions]} {destroy .primer_conditions}
  set a [toplevel .primer_conditions]
  wm protocol .primer_conditions WM_DELETE_WINDOW "set ok -1"
  bind . <<RaiseDialogs>> "wm deiconify $a; raise $a"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry $a "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient $a $w
  } else {
    ::tk::unsupported::MacWindowStyle style $a document {closeBox}
  }
  grid [label $a.lmin -text [mc "Minimum"]] -row 0 -column 2
  grid [label $a.lmax -text [mc "Maximum"]] -row 0 -column 3
  grid [label $a.llen -text [mc "Length: "]] -row 1 -column 1 -sticky e
  grid [entry $a.elenmin -width 10 -textvariable info(primer_len_min) -validate all -vcmd "entry_integer %P %W %V 1 100"] -row 1 -column 2
  grid [entry $a.elenmax -width 10 -textvariable info(primer_len_max) -validate all -vcmd "entry_integer %P %W %V 1 100"] -row 1 -column 3
  grid [label $a.lTm -text "Tm: "] -row 2 -column 1 -sticky e
  grid [entry $a.eTmmin -width 10 -textvariable info(primer_Tm_min) -validate all -vcmd "entry_integer %P %W %V 1 100"] -row 2 -column 2
  grid [entry $a.eTmmax -width 10 -textvariable info(primer_Tm_max) -validate all -vcmd "entry_integer %P %W %V 1 100"] -row 2 -column 3
  grid [label $a.l%gc -text "%GC: " -justify right] -row 3 -column 1 -sticky e
  grid [entry $a.e%gcmin -width 10 -textvariable info(primer_gc_min) -validate all -vcmd "entry_integer %P %W %V 0 100"] -row 3 -column 2
  grid [entry $a.e%gcmax -width 10 -textvariable info(primer_gc_max) -validate all -vcmd "entry_integer %P %W %V 0 100"] -row 3 -column 3
  grid [label $a.lclamp -text [mc "GC clamp: "]] -row 4 -column 1 -sticky e
  grid [entry $a.eclampmin -width 10 -textvariable info(primer_clamp_min) -validate all -vcmd "entry_integer %P %W %V 0 100"] -row 4 -column 2
  grid [entry $a.eclampmax -width 10 -textvariable info(primer_clamp_max) -validate all -vcmd "entry_integer %P %W %V 0 100"] -row 4 -column 3

  grid [frame $a.scomp -relief ridge -borderwidth 3] -row 6 -column 1 -columnspan 3 -pady 1
    grid [label $a.scomp.ltot -text [mc "total"]] -row 0 -column 1
    grid [label $a.scomp.ladj -text [mc "adjacent"]] -row 0 -column 2
    grid [label $a.scomp.ltp -text "3'"] -row 0 -column 3
    grid [checkbutton $a.scomp.lscomp -text [mc "Self complement: "] -variable info(primer_check_sc)] -row 1 -column 0 -sticky e
    grid [entry $a.scomp.scomptot -width 10 -textvariable info(primer_scomp_tot_max) -validate all -vcmd "entry_integer %P %W %V 1 1000" -width 3] -row 1 -column 1
    grid [entry $a.scomp.scompadj -width 10 -textvariable info(primer_scomp_adj_max) -validate all -vcmd "entry_integer %P %W %V 1 1000" -width 3] -row 1 -column 2
    grid [entry $a.scomp.scompthree -width 10 -textvariable info(primer_scomp_3_max) -validate all -vcmd "entry_integer %P %W %V 1 1000" -width 3] -row 1 -column 3
    grid [label $a.scomp.lhcomp -text [mc "Heterodimer: "]] -row 2 -column 0 -sticky e
    grid [entry $a.scomp.hcomptot -width 10 -textvariable info(primer_hcomp_tot_max) -validate all -vcmd "entry_integer %P %W %V 1 1000" -width 3] -row 2 -column 1
    grid [entry $a.scomp.hcompadj -width 10 -textvariable info(primer_hcomp_adj_max) -validate all -vcmd "entry_integer %P %W %V 1 1000" -width 3] -row 2 -column 2
    grid [entry $a.scomp.hcompthree -width 10 -textvariable info(primer_hcomp_3_max) -validate all -vcmd "entry_integer %P %W %V 1 1000" -width 3] -row 2 -column 3
    #grid [label $a.scomp.lholigo -text [mc "Oligo: "]] -row 3 -column 0 -sticky e
    grid [entry $a.scomp.scompoligo -width 10 -textvariable info(primer_hcomp_oligo) -validate key -vcmd {if {[regexp {[^atcgATGC]} %P] || ([string length %P] > 40)} {bell; return 0} else {return 1}} -width 35 -selectforeground $info(text_select_fg_color) -selectbackground $info(text_select_bg_color) -foreground $info(text_fg_color) -background $info(text_bg_color)] -row 3 -column 0 -columnspan 4

  grid [frame $a.concframe -relief ridge -borderwidth 3] -row 5 -column 1 -columnspan 3 -pady 3
    grid [label $a.concframe.lsalt -text [mc "Salt (mM):"]] -row 0 -column 1
    grid [entry $a.concframe.esalt -width 10 -textvariable info(primer_salt) -validate all -vcmd "entry_integer %P %W %V 1 1000"] -row 0 -column 2
    grid [label $a.concframe.ldna -text [mc "Primer (nM):"]] -row 1 -column 1
    grid [entry $a.concframe.edna -width 10 -textvariable info(primer_dna) -validate all -vcmd "entry_integer %P %W %V 1 1000"] -row 1 -column 2
  grid [label $a.lscan -text "Orientation:"] -row 7 -column 1 -sticky e
  tk_optionMenu $a.direction info(primer_scan) "5'-->3'" "3'<--5'"
  grid $a.direction -row 7 -column 2
  grid [label $a.lsort -text "Sort by:"] -row 8 -column 1 -sticky e
  tk_optionMenu $a.sort info(primer_sort) " 5'  " "  3'  " "length" "  %GC " "  Tm  "
  grid $a.sort -row 8 -column 2

  set ok 0
  grid [button $a.ok -width 9 -text [mc "OK"] -command "set ok 1" -default active] -row 9 -column 1 -columnspan 3 -pady 5
  #grab $a
  focus $a
  update idletasks
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $a] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry $a "+$winx+$winy"

  vwait ok
  set dialogblock 0
  foreach entry [list concframe.edna concframe.esalt scomp.scompoligo scomp.hcompthree scomp.hcompadj scomp.hcomptot scomp.scompthree scomp.scompadj scomp.scomptot eclampmax eclampmin e%gcmax e%gcmin eTmmax eTmmin elenmin elenmax] {
    $a.$entry validate
  }
  destroy $a
  bind . <<RaiseDialogs>> ""
  if {$ok == 1} {primer_scan $w}
  unset ok
}

################
## Primer scan
################
proc primer_scan {w} {
  global info
  global modifier
  global modstring

  $w configure -cursor watch
  $w.textarea configure -cursor watch
  update
  set log_dna [expr {1.987 * log( $info(primer_dna) / 1000000000.0)}] 
  set log_salt [expr {12 * log10($info(primer_salt) / 1000.0)}]

  if {[$w.textarea tag ranges sel] == ""} {
    if {[tk_messageBox -type okcancel -message "No DNA is selected. Do you want to search the entire selection?" -icon question -default ok] == "cancel"} {
      $w configure -cursor arrow
      $w.textarea configure -cursor xterm
      return
    }
    set first 1.0
    set last [$w.textarea index end-1chars]
  } else {
    set first [$w.textarea index sel.first]
    set last  [$w.textarea index sel.last]
  }
  if {$info(primer_scan) == "3'<--5'"} {
    set text [$w.textarea get $first $last]
  } else {
    set text [revcom [$w.textarea get $first $last]]
  }
  
  set hc_oligo [string tolower $info(primer_hcomp_oligo)]

  regsub -nocase -all {[^acgt]} $text "t" text
  if {[string length $text] < $info(primer_len_min)} {
    tk_messageBox -type ok -message "Selection is smaller than the minimum primer size" -icon warning -default ok
    $w configure -cursor arrow
    $w.textarea configure -cursor xterm
    raise $w
    focus $w
    return
  }
  set text [string tolower $text]
  array set sadd {aa 222 tt 222 at 204 ta 213 ca 158 tg 296 gt 155 ac 293 ct 141 ag 279 ga 153 tc 291 cg 272 gc 244 gg 199 cc 199}
  array set hadd {aa 79 tt 79 at 72 ta 72 ca 63 tg 107 gt 62 ac 106 ct 56 ag 100 ga 60 tc 104 cg 106 gc 98 gg 80 cc 80}
  array set srem {aa -222 tt -222 at -204 ta -213 ca -296 tg -158 gt -293 ac -155 ct -279 ag -141 ga -291 tc -153 cg -272 gc -244 gg -199 cc -199}
  array set hrem {aa -79 tt -79 at -72 ta -72 ca -107 tg -63 gt -106 ac -62 ct -100 ag -56 ga -104 tc -60 cg -106 gc -98 gg -80 cc -80}
  
  if {[string first [string index $text 0] "at"] > -1} {
    set htot -46
    set stot -82 
    set gc 0
  } else {
    set htot -2
    set stot 56
    set gc 1
  }

  for {set i 0} {$i < [expr {$info(primer_len_min)-1}]} {incr i} {
    incr htot $hadd([string range $text $i [expr $i+1]])
    incr stot $sadd([string range $text $i [expr $i+1]])
    if {[string first [string index $text [expr $i+1]] "gc"] > -1} {incr gc}
  }
  set hmem $htot
  set smem $stot
  set gcmem $gc
 
  #set min_sc $info(primer_scomp_3_max)
  #if {$info(primer_scomp_adj_max) < $min_sc} {
  #  set min_sc $info(primer_scomp_adj_max)
  #}
  #if {$info(primer_scomp_tot_max) < $min_sc} {
  #  set min_sc $info(primer_scomp_tot_max)
  #}

  set j $i
  set k 0

  set primerlist [list]
  while {$k < [expr {[string length $text]-$info(primer_len_min)-1}]} {
    regexp "^(g|c)*" [string range $text $k $j] clamp
    if {([string length $clamp] >= $info(primer_clamp_min)) && ([string length $clamp] <= $info(primer_clamp_max))} {
      while {[expr {$j-$k}] < $info(primer_len_max)} {
        set Tm [expr {$htot * 100.0 / ($stot * 0.1 - $log_dna) - 273.15 + $log_salt}]
        if {($Tm > $info(primer_Tm_max))} {
          break
        }
        if {($Tm > $info(primer_Tm_min) && ([expr {$gc * 100 / ($j-$k+1)}] >= $info(primer_gc_min)) && ([expr {$gc * 100 / ($j-$k+1)}] <= $info(primer_gc_max)))} {
         #######add primer to list of acceptable primers here
          if {$info(primer_check_sc)} {
            set potential [revcom [string range $text $k $j]]
            set sc_list [oligo_self_compl $potential $potential]
            if {([lindex $sc_list 0] < $info(primer_scomp_tot_max)) && ([lindex $sc_list 1] < $info(primer_scomp_adj_max)) && ([lindex $sc_list 2] < $info(primer_scomp_3_max))} {
              if {$hc_oligo != ""} {
                set hc_list [oligo_self_compl $potential $hc_oligo]
                if {([lindex $hc_list 0] < $info(primer_hcomp_tot_max)) && ([lindex $hc_list 1] < $info(primer_hcomp_adj_max)) && ([lindex $hc_list 2] < $info(primer_hcomp_3_max))} {
                  lappend primerlist [list [expr $k+1] [expr $j+1] [string range $text $k $j] [expr {$gc * 100 / ($j-$k+1)}] $Tm [expr {$j-$k+1}] $sc_list $hc_list]
                }
              } else {
                lappend primerlist [list [expr $k+1] [expr $j+1] [string range $text $k $j] [expr {$gc * 100 / ($j-$k+1)}] $Tm [expr {$j-$k+1}] $sc_list "-----"]
              }
            }
          } elseif {$hc_oligo != ""} {
            set potential [revcom [string range $text $k $j]]
            set hc_list [oligo_self_compl $potential $hc_oligo]
            if {([lindex $hc_list 0] < $info(primer_hcomp_tot_max)) && ([lindex $hc_list 1] < $info(primer_hcomp_adj_max)) && ([lindex $hc_list 2] < $info(primer_hcomp_3_max))} {
              lappend primerlist [list [expr $k+1] [expr $j+1] [string range $text $k $j] [expr {$gc * 100 / ($j-$k+1)}] $Tm [expr {$j-$k+1}] "-----" $hc_list]
            }
          } else {
            lappend primerlist [list [expr $k+1] [expr $j+1] [string range $text $k $j] [expr {$gc * 100 / ($j-$k+1)}] $Tm [expr {$j-$k+1}] "-----" "-----"]
          }
        }
        if {($j == [expr {[string length $text]-1}])} {
          break
        }
        incr htot $hadd([string range $text $j [expr $j+1]])
        incr stot $sadd([string range $text $j [expr $j+1]])
        if {[string first [string index $text [expr $j+1]] "gc"] > -1} {incr gc}
        incr j
      }
    }
    incr hmem $hrem([string range $text $k [expr $k+1]])
    incr smem $srem([string range $text $k [expr $k+1]])
    if {[string first [string index $text $k] "gc"] > -1} {incr gcmem -1}
    incr k
    if {$i < [expr {[string length $text]-1}]} {
      incr hmem $hadd([string range $text $i [expr $i+1]])
      incr smem $sadd([string range $text $i [expr $i+1]])
      if {[string first [string index $text [expr $i+1]] "gc"] > -1} {incr gcmem}
      set j [incr i]
    } else {
      set j $i
    }
    set htot $hmem
    set stot $smem
    set gc $gcmem  
  }
  if {$primerlist != [list]} {
    set a [new_analysis_window $w "Find Primers"]

    set first_tab [expr {[font measure dnafont "1000 AAAAAAAAAAAAAAAAAAAAAAAAAAAA 1028"]/[winfo pixels $a "1c"]*1.01}]
    set tab_spacing [expr {[font measure dnafont "length "]/[winfo pixels $a "1c"]*1.0}]
    grid [text $a.heading -background $info(bg_color) -foreground $info(label_fg_color) -height 1 -relief flat -font labelfont -tabs "[expr {$first_tab}]c center [expr {$first_tab+$tab_spacing}]c center [expr {$first_tab+2*$tab_spacing}]c center [expr {$first_tab+3*$tab_spacing}]c left"] -row 0 -column 0 -columnspan 2 -sticky nswe
    $a.heading insert end "Primer (5'-->3') \tlength\t%GC\tTm(C)\tself/other(max adj 3') compl.\n" 
    $a.heading configure -state disabled 
    grid [text $a.text -yscrollcommand "$a.scroll set" -width 80 -font dnafont -tabs "[expr {$first_tab}]c center [expr {$first_tab+$tab_spacing}]c center [expr {$first_tab+2*$tab_spacing}]c center [expr {$first_tab+3*$tab_spacing}]c left"] -row 1 -column 0 -rowspan 2 -sticky nswe
    grid [scrollbar $a.scroll -orient vertical -command "$a.text yview"] -row 1 -column 1 -sticky nswe
    grid [frame $a.sp -height 10] -row 2 -column 1 -sticky swe
    grid columnconfigure $a 0 -weight 1
    grid columnconfigure $a 1 -weight 0
    grid rowconfigure $a 1 -weight 1

    $a.text insert end "[clock format [clock seconds] -format $info(clock_format)]\n"
    $a.text insert end "$info($w,filename) \n"
    $a.text insert end "From [expr {[string range $first 2 end]+1}] to [string range $last 2 end] ($info(primer_scan)) \n\n"
    $a.text insert end "Find Primers" 
    $a.text insert end "Primer (5'-->3') \tlength\t%GC\tTm(C)\tself/other(max adj 3') compl.\n" 

    set first [lindex [regexp -inline 1.(.*) $first] 1]
    set last [lindex [regexp -inline 1.(.*) $last] 1]
    set t 0

    switch $info(primer_sort) {
      " 5'  " {set primerlist [lsort -decreasing -integer -index 1 $primerlist]}
      "  3'  " {set primerlist [lsort -decreasing -integer -index 0 $primerlist]}
      "length" {set primerlist [lsort -integer -index 5 $primerlist]}
      "  %GC " {set primerlist [lsort -integer -index 3 $primerlist]}
      "  Tm  " {set primerlist [lsort -index 4 $primerlist]}
    }
    foreach primer $primerlist {
      if {$info(primer_scan) == "3'<--5'"} {
        $a.text insert end "[expr {$first+[lindex $primer 1]}] "
        $a.text insert end "[string toupper [revcom [lindex $primer 2]]] " "tag$t"
        $a.text insert end "[expr {$first+[lindex $primer 0]}]\t"
        $a.text tag bind "tag$t" <$modifier-Button-1> "$a.text tag remove sel 1.0 end
          $a.text tag add sel tag$t.first tag$t.last
          $a.text mark set insert tag$t.last
          $a.text mark set anchor tag$t.first
          select_region $w [$w.textarea index 1.[expr {$first+[lindex $primer 0]-1}]] [$w.textarea index 1.[expr {$first+[lindex $primer 1]}]]"
      } else {
        $a.text insert end "[expr {1+$last-[lindex $primer 1]}] "
        $a.text insert end "[string toupper [revcom [lindex $primer 2]]] " "tag$t"
        $a.text insert end "[expr {1+$last-[lindex $primer 0]}]\t"
        $a.text tag bind "tag$t" <$modifier-Button-1> "$a.text tag remove sel 1.0 end
          $a.text tag add sel tag$t.first tag$t.last
          $a.text mark set insert tag$t.last
          $a.text mark set anchor tag$t.first
          select_region $w [$w.textarea index 1.[expr {$last-[lindex $primer 1]}]] [$w.textarea index 1.[expr {$last-[lindex $primer 0]+1}]]"
      }
      $a.text tag bind "tag$t" <Double-Button-1> [$a.text tag bind "tag$t" <$modifier-Button-1>]
      $a.text tag bind "tag$t" <Any-Leave> "%W tag remove highlite 1.0 end"
      $a.text tag bind "tag$t" <Any-Motion> "%W tag remove highlite 1.0 end; %W tag add highlite tag$t.first tag$t.last"
      $a.text tag bind "tag$t" <Enter> "%W tag remove highlite 1.0 end; %W tag add highlite tag$t.first tag$t.last"


      lappend info($w,linked_events) [list $a.text "tag$t" <$modifier-Button-1>]

      $a.text insert end "[lindex $primer 5]\t"
      $a.text insert end "[lindex $primer 3]\t"
      $a.text insert end "[format %2.0f [lindex $primer 4]]\t"
      $a.text insert end "[lindex $primer 6] /"
      $a.text insert end "[lindex $primer 7]\n"
      incr t
    }
    $a.text tag configure highlite -background red -foreground white
    $a.text tag configure mmhighlite -background red -foreground white
    bind $a.text <$modifier-Button-1> "if \{\[lsearch -exact \[$a.text tag names @%x,%y\] sel\] != -1\} break"
    set outputmenu [add_output_menu $a.text "Text" text $w]
    bind $a <Button-1> "focus $a.text" 
    bindtags $a.text [list $a.text AnalysisText . all]
    wm deiconify $a
    window_on_screen $a
    focus $a.text
  } else {
    tk_messageBox -type ok -message "No primers found" -icon info -default ok
  }
  $w configure -cursor arrow
  $w.textarea configure -cursor xterm
}

################
## find the complementarity between two oligos
################
#returns list of total, adjacent, and 3'

proc oligo_self_compl {o1 o2} {
  if {[string length $o2] < [string length $o1]} {
    set o3 $o1
    set o1 $o2
    set o2 $o3
  }
  set o2 [revcom $o2]
  set len_dif [expr {[string length $o2] - [string length $o1]}]

  set max_match 0
  set max_adj 0
  set max_three 0
  for {set i [expr {-[string length $o1]}]} {$i < [expr {[string length $o2]}]} {incr i} {
    if {$i <= 0} {
      set top [string range $o1 [expr {-$i}] end]
      set bottom [string range $o2 0 [expr {[string length $top]-1}]]
    } else {
      set bottom [string range $o2 $i [expr {$i+[string length $o1]-1}]]
      set top [string range $o1 0 [expr {[string length $bottom]-1}]]
    }
    set this_max_match 0
    set this_adj 0
    set this_adj_working 0
    set bottom_flag 1
    foreach x [split $top ""] y [split $bottom ""] {
      if {$x == $y} {
        incr this_max_match
        incr this_adj_working
        if {$this_adj_working > $this_adj} {set this_adj $this_adj_working}
      } else {
        if {($this_adj_working > $max_three) && ($i <= 0) && ($bottom_flag)} {set max_three $this_adj_working}
        set this_adj_working 0
        set bottom_flag 0
      }
    }
    #sputs "$top\n$bottom\n$this_max_match $this_adj $this_adj_working $max_three"
    if {$this_max_match > $max_match} {set max_match $this_max_match}
    if {$this_adj > $max_adj} {set max_adj $this_adj}
    if {($this_adj_working > $max_three) && ($i <= $len_dif)} {set max_three $this_adj_working}
  }
  return [list $max_match $max_adj $max_three]
} 

################
## DNA alignment dialog
################
proc align_dialog {w} {
  global info  
  global align_select_window1
  global align_select_window2
  global ok dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1


  if {[winfo exists .aligndialog]} {destroy .aligndialog}
  set s [toplevel .aligndialog]
  wm title $s [mc "Align DNA"]
  wm protocol $s WM_DELETE_WINDOW "set ok -1"
  bind . <<RaiseDialogs>> "wm deiconify $s; raise $s"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry $s "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient $s $w
  } else {
    ::tk::unsupported::MacWindowStyle style $s document {closeBox}
  }
  grid [label $s.label1 -text [mc "Window"]] -row 1 -column 1 -sticky ne
  menubutton $s.win_button1
    menu $s.win_button1.menu
    foreach window $info(dnawindows_list) {
      $s.win_button1.menu add radiobutton -command "$s.win_button1 configure -text \"[wm title $window]\"" -label [wm title $window] -variable align_select_window1 -value $window 
    }
    set abi_list [lsort -unique [regexp -inline -all {\.abi_window[0-9]*} "[winfo children .] "]]
    if {[llength $abi_list] > 0} {
      $s.win_button1.menu add separator
      foreach window $abi_list {
        $s.win_button1.menu add radiobutton -command "$s.win_button1 configure -text \"[wm title $window]\"" -label [wm title $window] -variable align_select_window1 -value $window 
      }
    }
  $s.win_button1 configure -text [wm title $w] -menu $s.win_button1.menu -indicatoron 1 -relief raised -width 60 
  bind $s.win_button1.menu <<MenuSelect>> "focus $s"
  set align_select_window1 $w
  grid $s.win_button1 -row 1 -column 2 -columnspan 3

  grid [checkbutton $s.local1 -selectcolor $info(bg_color) -width 10 -variable info(align_select_region1) -textvariable info(align_select_region1) -onvalue "selection" -offvalue "all" -indicator off] -row 1 -column 5
  grid [checkbutton $s.rc1 -selectcolor $info(bg_color) -width 8 -variable info(align_select_rc1) -textvariable info(align_select_rc1) -offvalue "forward" -onvalue "rev/com" -indicator off] -row 1 -column 6

  grid [label $s.label2 -text [mc "Window"]] -row 2 -column 1 -sticky ne
  menubutton $s.win_button2
    menu $s.win_button2.menu
    foreach window $info(dnawindows_list) {
      $s.win_button2.menu add radiobutton -command "$s.win_button2 configure -text \"[wm title $window]\"" -label [wm title $window] -variable align_select_window2 -value $window 
    }
    if {[llength $abi_list] > 0} {
      $s.win_button2.menu add separator
      foreach window $abi_list {
        $s.win_button2.menu add radiobutton -command "$s.win_button2 configure -text \"[wm title $window]\"" -label [wm title $window] -variable align_select_window2 -value $window 
      }
    }
  bind $s.win_button2.menu <<MenuSelect>> "focus $s"
  if {[winfo exists $info(align_select_window2)]} {
    set align_select_window2 $info(align_select_window2)
  } else {
    set align_select_window2 $w
  }
  $s.win_button2 configure -text [wm title $align_select_window2] -menu $s.win_button2.menu -indicatoron 1 -relief raised -width 60 
  grid $s.win_button2 -row 2 -column 2 -columnspan 3

  grid [checkbutton $s.local2 -selectcolor $info(bg_color) -width 10 -variable info(align_select_region2) -textvariable info(align_select_region2) -onvalue "selection" -offvalue "all" -indicator off] -row 2 -column 5
  grid [checkbutton $s.rc2 -selectcolor $info(bg_color) -width 8 -variable info(align_select_rc2) -textvariable info(align_select_rc2) -offvalue "forward" -onvalue "rev/com" -indicator off] -row 2 -column 6

  grid [label $s.blockslabel -text [mc "Blocks:"]] -row 4 -column 1 -sticky ne
  menubutton $s.blocksbut -indicatoron 1 -relief raised -width 10 -textvariable info(align_blocksize)
  menu $s.blocksbut.blocks
    $s.blocksbut.blocks add radiobutton -label "20" -value 20 -variable info(align_blocksize)
    $s.blocksbut.blocks add radiobutton -label "15" -value 15 -variable info(align_blocksize)
    $s.blocksbut.blocks add radiobutton -label "10" -value 10 -variable info(align_blocksize)
    $s.blocksbut.blocks add radiobutton -label "N-W align all" -value "N-W align all" -variable info(align_blocksize)
  $s.blocksbut configure -menu $s.blocksbut.blocks
  bind $s.blocksbut <<MenuSelect>> "focus $s"

  grid $s.blocksbut -row 4 -column 2 -sticky nw

  grid [label $s.nwmaxlabel -text "N-W max:"] -row 4 -column 3 -sticky ne
  grid [entry $s.nwmaxentry -validate key -width 4 -vcmd "check_char %S" -textvariable info(NWmax)] -row 4 -column 4 -sticky nw

  grid [label $s.mmplabel -text [mc "Mismatch Penalty:"]] -row 5 -column 1 -sticky nw
  menubutton $s.mmpbut -indicatoron 1 -relief raised -width 10 -textvariable info(mm_penalty)
  menu $s.mmpbut.mmpmenu
    $s.mmpbut.mmpmenu add radiobutton -label "0" -value 0 -variable info(mm_penalty)
    $s.mmpbut.mmpmenu add radiobutton -label " -1" -value -1 -variable info(mm_penalty)
    $s.mmpbut.mmpmenu add radiobutton -label " -2" -value -2 -variable info(mm_penalty)
  $s.mmpbut configure -menu $s.mmpbut.mmpmenu
  bind $s.mmpbut.mmpmenu <<MenuSelect>> "focus $s"
  grid $s.mmpbut -row 5 -column 2 -sticky nw

  grid [label $s.gapplabel -text [mc "Gap Penalty:"]] -row 5 -column 3 -sticky ne
  menubutton $s.gappbut -indicatoron 1 -relief raised -width 10 -textvariable info(gap_penalty)
  menu $s.gappbut.gapp
    $s.gappbut.gapp add radiobutton -label "0" -value 0 -variable info(gap_penalty)
    $s.gappbut.gapp add radiobutton -label " -1" -value -1 -variable info(gap_penalty)
    $s.gappbut.gapp add radiobutton -label " -2" -value -2 -variable info(gap_penalty)
  $s.gappbut configure -menu $s.gappbut.gapp
  bind $s.gappbut.gapp <<MenuSelect>> "focus $s"
  grid $s.gappbut -row 5 -column 4 -sticky nw

  grid [label $s.widthentrylabel -text [mc "Line width:"]] -row 6 -column 1 -sticky ne
  grid [entry $s.widthentry -validate key -vcmd "check_char %S" -textvariable info(align_linewidth)] -row 6 -column 2 -sticky nw
  grid [checkbutton $s.highlight -selectcolor white -text [mc "Copy Highlighting from Sequence"] -onvalue 1 -offvalue 0 -variable info(align_copy_highlight)] -row 6 -column 4

  grid [button $s.ok -text [mc "OK"] -command "set ok 1" -default active] -row 7 -column 1 -columnspan 2
  grid [button $s.cancel -text [mc "Cancel"] -command "set ok -1"] -row 7 -column 2 -columnspan 2
  #grab $s
  focus $s

  update idletasks
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $s] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry $s "+$winx+$winy"

  vwait ok
  set dialogblock 0
  if {$info(NWmax) == ""} {set info(NWmax) 300}
  if {$info(align_linewidth) == ""} {set info(align_linewidth) 100}
  set info(align_select_window2) $align_select_window2
  #grab release $s
  destroy $s
  bind . <<RaiseDialogs>> ""
  if {$ok == 1} {blockalign}
  unset ok 
}

################
## align DNA sequences and make a window with the output
################
proc blockalign {} {
  global align_select_window1
  global align_select_window2
  global info abi_info
  global modifier
  global modstring

  $align_select_window1 configure -cursor watch
  $align_select_window2 configure -cursor watch
  set xdir 1; set ydir 1

  if {[string first .abi_window $align_select_window1] == -1} {
    ###window1 is a sequence window
    #$align_select_window1.textarea configure -cursor watch
    set window_type1 "sequence"
    if {($info(align_select_region1) == "selection") && ([$align_select_window1.textarea tag nextrange tempsel 1.0] != "")} {
      set xdna [$align_select_window1.textarea get tempsel.first tempsel.last]
      if {$info(align_select_rc1) == "rev/com"} {
        set xfirst [expr {1+[lindex [split [$align_select_window1.textarea index tempsel.last] .] 1]}]
        set xdna [revcom $xdna]
        set xdir -1
      } else {
        set xfirst [lindex [split [$align_select_window1.textarea index tempsel.first] .] 1]
      }
    } else {
      set xdna [$align_select_window1.textarea get 1.0 end-1chars]
      if {$info(align_select_rc1) == "rev/com"} {
        set xfirst [expr {1+[lindex [split [$align_select_window1.textarea index end-1chars] .] 1]}]
        set xdna [revcom $xdna]
        set xdir -1
      } else {
        set xfirst 0
      }
    }
  } else {
    ###window1 is an abi window
    #$align_select_window1.c configure -cursor watch
    set window_type1 "abi"
    if {($info(align_select_region1) == "selection")} {
      set abi_seq [abi_get_seq $align_select_window1 sel.first sel.last]
      if {$abi_seq == ""} {
        set abi_seq [abi_get_seq $align_select_window1 start end]
      }
    } else {
      set abi_seq [abi_get_seq $align_select_window1 start end]
    }
    if {$info(align_select_rc1) == "rev/com"} {
      set xfirst [expr {[lindex $abi_seq 1]-1}]
      set xdna [revcom [lindex $abi_seq 2]]
      if {!$abi_info($align_select_window1,rev_com)} {
        set xdir -1
        incr xfirst 2
      }
    } else {
      set xfirst [expr {[lindex $abi_seq 0]-1}]
      set xdna [lindex $abi_seq 2]
      if {$abi_info($align_select_window1,rev_com)} {
        set xdir -1
        incr xfirst 2
      }
    }
  }

  if {[string first .abi_window $align_select_window2] == -1} {
    ###window2 is a sequence window
    #$align_select_window2.textarea configure -cursor watch
    set window_type2 "sequence"
    if {($info(align_select_region2) == "selection") && ([$align_select_window2.textarea tag nextrange tempsel 1.0] != "")} {
      set ydna [$align_select_window2.textarea get tempsel.first tempsel.last]
      if {$info(align_select_rc2) == "rev/com"} {
        set yfirst [expr {1+[lindex [split [$align_select_window2.textarea index tempsel.last] .] 1]}]
        set ydna [revcom $ydna]
        set ydir -1
      } else {
        set yfirst [lindex [split [$align_select_window2.textarea index tempsel.first] .] 1]
      }
    } else {
      set ydna [$align_select_window2.textarea get 1.0 end-1chars]
      if {$info(align_select_rc2) == "rev/com"} {
        set yfirst [expr {1+[lindex [split [$align_select_window2.textarea index end-1chars] .] 1]}]
        set ydna [revcom $ydna]
        set ydir -1
      } else {
        set yfirst 0
      }
    }
  } else {
    ###window2 is an abi window
    #$align_select_window2.c configure -cursor watch
    set window_type2 "abi"
    if {($info(align_select_region2) == "selection")} {
      set abi_seq [abi_get_seq $align_select_window2 sel.first sel.last]
      if {$abi_seq == ""} {
        set abi_seq [abi_get_seq $align_select_window2 start end]
      }
    } else {
      set abi_seq [abi_get_seq $align_select_window2 start end]
    }
    if {$info(align_select_rc2) == "rev/com"} {
      set yfirst [expr {[lindex $abi_seq 1]-1}]
      set ydna [revcom [lindex $abi_seq 2]]
      if {!$abi_info($align_select_window2,rev_com)} {
        set ydir -1
        incr yfirst 2
      }
    } else {
      set yfirst [expr {[lindex $abi_seq 0]-1}]
      set ydna [lindex $abi_seq 2]
      if {$abi_info($align_select_window2,rev_com)} {
        set ydir -1
        incr yfirst 2
      }
    }
  }
  
  update

  if {$info(align_blocksize) != "N-W align all"} {
    set xseq [string toupper $xdna]
    set yseq [string toupper $ydna]
    set i 0
    set j 0
    set xout ""
    set yout ""
    set xend [expr {[string length $xseq]-$info(align_blocksize)+1}]
    while {$i < [string length $xseq]} {
      set xun ""
      set yun ""
    #find a match
      while {($i < $xend) && ([set found [string first [string range $xseq $i [expr {$i+$info(align_blocksize)-1}]] $yseq $j]] == -1)} {
        append xun [string range $xdna $i [expr {$i+$info(align_blocksize)-1}]]
        incr i $info(align_blocksize)
      }
      if {($i >= $xend)} {
        set found [string length $ydna]
        append xun [string range $xdna $i end]
      }
      set yun [string range $ydna $j [expr {$found-1}]]
 ###align the unaligned regions between matches: xun and yun here
      if {($xun != "") || ($yun != "")} {
        set alignlist [finealign $xun $yun]
        append xout [lindex $alignlist 0]
        append yout [lindex $alignlist 1]
        append lineout [lindex $alignlist 2]
      }
##now add in the match
      if {($i < $xend)} {
        append xout [string range $xdna $i [expr {$i+$info(align_blocksize)-1}]]
        append yout [string range $ydna $found [expr {$found+$info(align_blocksize)-1}]]
        append lineout [string repeat "|" $info(align_blocksize)]
        set j [expr {$found+$info(align_blocksize)}]
      }
      incr i $info(align_blocksize)
    }
    if {$i == [string length $xseq]} {
        set yextra [string range $ydna $j end]
        append xout [string repeat "-" [string length $yextra]]
        append yout $yextra
        append lineout [string repeat " " [string length $yextra]]
    }
  } else {
    set resultlist [NWalign $xdna $ydna 1]
    set xout [lindex $resultlist 0]
    set yout [lindex $resultlist 1]
    set lineout [lindex $resultlist 2]
  }

  ###make the top index line
  set spacing [expr {($xdir==-1)? (1 + ($xfirst-1) % 10) : (10 - ($xfirst % 10))}]
  #sputs "$xfirst $spacing"
  set dot 0
  set x 0
  set xdots ""
  while {[set dot [string length [regexp -inline "(?:-*\[^-\]){$spacing}" [string range $xout $x end]]]] != 0} {
    set xdots "$xdots[string repeat " " [expr {$dot-1}]]*"
    set spacing 10
    incr x $dot
  }
  set spacing [expr {($ydir == -1) ? (1 + ($yfirst-1) % 10) : (10 - $yfirst%10)}]
  set dot 0
  set x 0
  set ydots ""
  while {[set dot [string length [regexp -inline "(?:-*\[^-\]){$spacing}" [string range $yout $x end]]]] != 0} {
    set ydots "$ydots[string repeat " " [expr {$dot-1}]]*"
    set spacing 10
    incr x $dot
  }
  if {$align_select_window1 != $align_select_window2} {
    set winlist [list $align_select_window1 $align_select_window2]
  } else {
    set winlist [list $align_select_window1]
  }
  set a [new_analysis_window $winlist "[mc {Alignment to}] [wm title $align_select_window2]"]

  ##make output
  create_textframe $a [expr {$info(align_linewidth)+15}]
  #summary info
  $a.textframe.text insert end "[clock format [clock seconds] -format $info(clock_format)]\n"
  $a.textframe.text insert end "[wm title $align_select_window1] from [expr {$xfirst + $xdir}] to [expr {$xfirst + $xdir*[string length $xdna]}]\n"
  $a.textframe.text insert end "Alignment to\n"
  $a.textframe.text insert end "[wm title $align_select_window2] from [expr {$yfirst + $ydir}] to [expr {$yfirst + $ydir*[string length $ydna]}]\n\n"
  $a.textframe.text insert end "[mc Matches](|):[string length [string map {" " "" # "" . ""} $lineout]] \n"
  $a.textframe.text insert end "[mc Mismatches](#):[string length [string map {" " "" | "" . ""} $lineout]] \n"
  $a.textframe.text insert end "[mc Gaps]( ):[string length [string map {"|" "" # "" . ""} $lineout]] \n"
  $a.textframe.text insert end "[mc Unattempted](.):[string length [string map {" " "" # "" | ""} $lineout]] \n\n"

  ##replace leading and trailing -'s
  regexp {(^-*)(.*?)(-*$)} $xout fill leading body trailing
  set xout "[string repeat "~" [string length $leading]]$body[string repeat "~" [string length $trailing]]"
  regexp {(^-*)(.*?)(-*$)} $yout fill leading body trailing
  set yout "[string repeat "~" [string length $leading]]$body[string repeat "~" [string length $trailing]]"

  for {set x 0} {$x < [expr {[string length $xout]}]} {incr x $info(align_linewidth)} {
    set linelength [expr {[string length $xout]-$x-1}]
    if {$linelength > [expr {$info(align_linewidth)-1}]} {set linelength [expr {$info(align_linewidth)-1}]}
    set xchars [regexp -all  {[^- ~]} [string range $xout $x [expr {$x+$linelength}]]]
    set ychars [regexp -all  {[^- ~]} [string range $yout $x [expr {$x+$linelength}]]]
    if {$xchars > 0} {
      $a.textframe.text insert end "[string repeat { } 7][string range $xdots $x [expr {$x+$linelength}]]\n"
      $a.textframe.text insert end "[format {%6u } [incr xfirst $xdir]]"
      set dnastarts [$a.textframe.text index end-1char]
      $a.textframe.text insert end "[string range $xout $x [expr {$x+$linelength}]]" w1tag
      if {($info(align_copy_highlight)) && ($window_type1 eq "sequence")} {
        copy_tags $align_select_window1.textarea [list 1.[expr {$xfirst-1}] 1.[expr {$xfirst-1+$xdir*($xchars-1)}]] $a.textframe.text $dnastarts [expr {($xdir==1)?"forward":"reverse"}] "w1"
      }
      $a.textframe.text insert end "[format { %-6u} [incr xfirst [expr {$xdir*($xchars-1)}]]]\n"
    } else {
      $a.textframe.text insert end "[format {%6u } [expr {$xfirst+1}]]" {} "[string range $xout $x [expr {$x+$linelength}]]" w1tag "[format { %-6u} [expr {$xfirst+1}]]\n" {}
    }
    set lineout_sub [string range $lineout $x [expr {$x+$linelength}]]
    $a.textframe.text insert end "[string repeat { } 7]$lineout_sub[string repeat { } 7]\n"
    ##add highlight tag to mismatches and gaps
    if {[regexp {[^\.\ ]} $lineout_sub]} {
      set ls [$a.textframe.text index "end-2char linestart"]
      foreach highliterange [regexp -inline -indices -all {[^|]+} $lineout_sub] {
        $a.textframe.text tag add mmhighlite [$a.textframe.text index $ls+[expr {7+[lindex $highliterange 0]}]chars] [$a.textframe.text index $ls+[expr {[lindex $highliterange 1]+8}]chars]
      }
    }
    if {$ychars > 0} {
      $a.textframe.text insert end "[format {%6u } [incr yfirst $ydir]]"
      set dnastarts [$a.textframe.text index end-1char]
      $a.textframe.text insert end "[string range $yout $x [expr {$x+$linelength}]]" w2tag
      if {($info(align_copy_highlight)) && ($window_type2 eq "sequence")} {
        copy_tags $align_select_window2.textarea [list 1.[expr {$yfirst-1}] 1.[expr {$yfirst-1+$ydir*($ychars-1)}]] $a.textframe.text $dnastarts [expr {($ydir==1)?"forward":"reverse"}] "w2"
      }
      $a.textframe.text insert end "[format { %-6u} [incr yfirst [expr {$ydir*($ychars-1)}]]]\n"
      $a.textframe.text insert end "[string repeat { } 7][string range $ydots $x [expr {$x+$linelength}]]\n\n"
    } else {
      $a.textframe.text insert end "[format {%6u } [expr {$yfirst+1}]]" {} "[string range $yout $x [expr {$x+$linelength}]]" w2tag "[format { %-6u} [expr {$yfirst+1}]]\n\n"
    }
  }

  if {$window_type2 eq "sequence"} {
    foreach tag [$align_select_window2.textarea tag names] {
     if {[lsearch -exact [$a.textframe.text tag names] $tag] > -1} {
       $a.textframe.text tag raise $tag
     }
    }
  }

  if {$window_type1 eq "sequence"} {
    foreach tag [$align_select_window1.textarea tag names] {
      if {[lsearch -exact [$a.textframe.text tag names] $tag] > -1} {
        $a.textframe.text tag raise $tag
      }
    }
  }

  $a.textframe.text tag bind w1tag <Double-Button> "$a.textframe.text tag add sel @%x,%y; alignlink $a.textframe.text $align_select_window1 \[$a.textframe.text index @%x,%y\] $xdir goto 1; break"
  $a.textframe.text tag bind w2tag <Double-Button> "$a.textframe.text tag add sel @%x,%y; alignlink $a.textframe.text $align_select_window2 \[$a.textframe.text index @%x,%y\] $ydir goto 1; break"
  $a.textframe.text tag bind w1tag <Shift-Double-Button> "$a.textframe.text tag add sel @%x,%y; alignlink $a.textframe.text $align_select_window1 \[$a.textframe.text index @%x,%y\] $xdir select 1; break"
  $a.textframe.text tag bind w2tag <Shift-Double-Button> "$a.textframe.text tag add sel @%x,%y; alignlink $a.textframe.text $align_select_window2 \[$a.textframe.text index @%x,%y\] $ydir select 1; break"
  $a.textframe.text tag bind w1tag <$modifier-Button> "alignlink $a.textframe.text $align_select_window1 \[$a.textframe.text index @%x,%y\] $xdir goto 0; break"
  $a.textframe.text tag bind w2tag <$modifier-Button> "alignlink $a.textframe.text $align_select_window2 \[$a.textframe.text index @%x,%y\] $ydir goto 0; break"
  $a.textframe.text tag bind w1tag <Shift-$modifier-Button> "alignlink $a.textframe.text $align_select_window1 \[$a.textframe.text index @%x,%y\] $xdir select 0; break"
  $a.textframe.text tag bind w2tag <Shift-$modifier-Button> "alignlink $a.textframe.text $align_select_window2 \[$a.textframe.text index @%x,%y\] $ydir select 0; break"
  lappend info($align_select_window1,linked_events) [list $a.textframe.text w1tag <Double-Button>]
  lappend info($align_select_window2,linked_events) [list $a.textframe.text w2tag <Double-Button>]
  lappend info($align_select_window1,linked_events) [list $a.textframe.text w1tag <Shift-Double-Button>]
  lappend info($align_select_window2,linked_events) [list $a.textframe.text w2tag <Shift-Double-Button>]
  lappend info($align_select_window1,linked_events) [list $a.textframe.text w1tag <$modifier-Button>]
  lappend info($align_select_window2,linked_events) [list $a.textframe.text w2tag <$modifier-Button>]
  lappend info($align_select_window1,linked_events) [list $a.textframe.text w1tag <Shift-$modifier-Button>]
  lappend info($align_select_window2,linked_events) [list $a.textframe.text w2tag <Shift-$modifier-Button>]
  $a.textframe.text tag bind w1tag <Any-Leave> {%W tag remove highlite 1.0 end}
  $a.textframe.text tag bind w1tag <Any-Motion> {%W tag remove highlite 1.0 end; %W tag add highlite current}
  $a.textframe.text tag bind w2tag <Any-Leave> {%W tag remove highlite 1.0 end}
  $a.textframe.text tag bind w2tag <Any-Motion> {%W tag remove highlite 1.0 end; %W tag add highlite current}
  $a.textframe.text tag configure highlite -background red -foreground white
  $a.textframe.text tag configure mmhighlite -background red -foreground white

  bind $a.textframe.text <Double-Button> break
  bind $a.textframe.text <Shift-Double-Button> break

  bind $a <Button-1> "focus $a.textframe.text"
  set outputmenu [add_output_menu $a.textframe.text "Text" text $align_select_window1]
  bind $a.textframe.text <Double-Button-1> ""
  $outputmenu entryconfigure [mc "Raise Parent"] -label [mc "Raise Window 1"]
  $outputmenu insert [expr {1+[$outputmenu index [mc "Raise Window 1"]]}] command -label [mc "Raise Window 2"] -accelerator "$modstring+Shift+Space" -command "event generate $a.textframe.text <<Return2>>"
  bind $a.textframe.text <Shift-$modifier-KeyPress-space> "event generate $a.textframe.text <<Return2>>"
  if {$window_type1 eq "abi"} {
    bind $a.textframe.text <<Return>> "wm deiconify $align_select_window1; raise $align_select_window1; focus $align_select_window1.c"
  }
  if {$window_type2 eq "sequence"} {
    bind $a.textframe.text <<Return2>> "wm deiconify $align_select_window2; raise $align_select_window2; focus $align_select_window2.textarea"
  } else {
    bind $a.textframe.text <<Return2>> "wm deiconify $align_select_window2; raise $align_select_window2; focus $align_select_window2.c"
  }
  lappend info($align_select_window2,linked_events) [list $a.textframe.text <<Return2>>]
  bindtags $a.textframe.text [list $a.textframe.text AnalysisText . all]
  wm deiconify $a
  window_on_screen $a
  focus $a.textframe.text

  $align_select_window1 configure -cursor arrow
  #$align_select_window1.textarea configure -cursor xterm
  $align_select_window2 configure -cursor arrow
  #$align_select_window2.textarea configure -cursor xterm
}


################
## sub-procedure of blockalign- expand the blocks and then N-W align the unaligned
################
proc finealign {xdna ydna} {
  set xseq [string toupper $xdna]
  set yseq [string toupper $ydna]
  set i 0
  set j 0
  if {[string length $xseq] > [string length $yseq]} {
    set max [string length $yseq]
  } else {
    set max [string length $xseq]
  }
  while {([string index $xseq $i] == [string index $yseq $i]) && ($i < $max)} {incr i}
  incr i -1
  set maxj [expr $max - $i -2]
  while {([string index $xseq end-$j] == [string index $yseq end-$j]) && ($j <= $maxj)} {incr j}
  incr j -1
  set NWlist [NWalign [string range $xdna [expr $i+1] end-[expr $j +1]] [string range $ydna [expr $i+1] end-[expr $j +1]]]
  append xout [string range $xdna 0 $i] [lindex $NWlist 0] [string range $xdna end-$j end]
  append yout [string range $ydna 0 $i] [lindex $NWlist 1] [string range $ydna end-$j end]
  append lineout [string repeat | [expr $i+1]] [lindex $NWlist 2] [string repeat | [expr $j+1]]
  return [list $xout $yout $lineout]
}

################
## sub-procedure of finealing or blockalign - Needleman-Wunsch Alignment (affine gap penalty)
################
#Needleman-Wunsch Alignment
# for veclist: 0=left 1=diag 2=up 3=diag,mismatch
#time= 56usec * len(x) * len(y)
proc NW_affine_align {xdna ydna {ignoremax 0}} {
  set atime [clock clicks]
  global info

  if {(!$ignoremax) && ([expr {sqrt([string length $xdna]* [string length $ydna])}] > $info(NWmax))} {
    if {[set diff [expr {[string length $xdna] - [string length $ydna]}]] > 0} {
      return [list $xdna "[string repeat - $diff]$ydna" [string repeat . [string length $xdna]]]
    } else {
      return [list "[string repeat - [expr {-$diff}]]$xdna" $ydna [string repeat . [string length $ydna]]]
    }
  }

  set gop $info(gap_penalty)
  set gep $info(gap_extension_penalty)
  set mp $info(mm_penalty)


  set allvecs [list]
  set veclist [list]
  set ylist [split [string toupper $ydna] ""]
  set xlist [split [string toupper $xdna] ""]
  set mlist [list]
  set ilist [list]

 #initialize the first row of mlist and ilist
  set ychar [lindex $ylist 0]
  set xchar [lindex $xlist 0]
  if {$xchar == $ychar} {
    set dm 1
    lappend veclist 1
  } else {
    set dm $mm
    lappend veclist 3
  }
  set di $gop

  set i $gop
  foreach xchar [lrange $xlist 1 end] {
    if {$xchar == $ychar} {
      lappend mlist [expr {$i+1}]
      lappend veclist 1
    } else {
      lappend mlist [expr {$i+$mm}]
      lappend veclist 3
    }
    lappend ilist [incr i $gep]
  }
  lappend allvecs $veclist

  #now do all of the rest of the rows
  set ygp $gop
  foreach ychar $ylist {
    set veclist [list]
    set newmlist [list]
    set newilist [list]
    set xchar [lindex $xlist 0]
      if {$xchar == $ychar} {
        set lm [expr {$ygp+1}]
      } else {
        set lm [expr {$ygp+$mm}]
      }
      set li [incr ygp $gep]
      lappend veclist 2
      set lmstore $lm
      set listore $li

    foreach xchar [lrange $xlist 1 end] um $mlist ui $ilist {
      if {$xchar == $ychar} {
      } else {
      }
    } 
    set mlist $newmlist
    set ilist $newilist
    lappend allvecs $veclist
  }
  set btime [clock clicks]
  set x [expr {[string length $xdna]-1}] 
  set y [expr {[string length $ydna]-1}] 
  set yout ""
  set xout ""
  set lineout ""
  while {($x >-1) && ($y > -1)} {
    switch [lindex [lindex $allvecs $y] $x] {
      0 {
        set yout "-$yout"
        set xout "[string index $xdna $x]$xout"
        set lineout " $lineout"
        incr x -1 
      }
      1 {
        set yout "[string index $ydna $y]$yout"
        set xout "[string index $xdna $x]$xout"
        set lineout "|$lineout"
        incr x -1 
        incr y -1
      }
      3 {
        set yout "[string index $ydna $y]$yout"
        set xout "[string index $xdna $x]$xout"
        set lineout "#$lineout"
        incr x -1 
        incr y -1
      }
      2 {
        set yout "[string index $ydna $y]$yout"
        set xout "-$xout" 
        set lineout " $lineout"
        incr y -1
      }
    }
  }
  if {$x > -1} {
    set yout "[string repeat - [expr {1+$x}]]$yout"
    set xout "[string range $xdna 0 $x]$xout"
    set lineout "[string repeat " " [expr {1+$x}]]$lineout"
  } else {
    set xout "[string repeat - [expr {1+$y}]]$xout"
    set yout "[string range $ydna 0 $y]$yout"
    set lineout "[string repeat " " [expr {1+$y}]]$lineout"
  }
  set ctime [clock clicks]
  return [list $xout $yout $lineout]
}

################
## sub-procedure of finealing or blockalign - Needleman-Wunsch Alignment
################
#Needleman-Wunsch Alignment
#for comparisons:
#g f
#m ?
# for veclist: 0=left 1=diag 2=up 3=diag,mismatch
#time= 56usec * len(x) * len(y)
proc NWalign {xdna ydna {ignoremax 0}} {
  set atime [clock clicks]
  global info

  if {(!$ignoremax) && ([expr {sqrt([string length $xdna]* [string length $ydna])}] > $info(NWmax))} {
    if {[set diff [expr {[string length $xdna] - [string length $ydna]}]] > 0} {
      return [list $xdna "[string repeat - $diff]$ydna" [string repeat . [string length $xdna]]]
    } else {
      return [list "[string repeat - [expr {-$diff}]]$xdna" $ydna [string repeat . [string length $ydna]]]
    }
  }

  set gp $info(gap_penalty)
  set mp $info(mm_penalty)
  set scorelist1 [list]

  for {set i 1} {$i <= [string length $xdna]} {incr i} {lappend scorelist1 [expr {$i*$gp}]}
  set i 0
  set allvecs [list]
  set ylist [split [string toupper $ydna] ""]
  set xlist [split [string toupper $xdna] ""]
  foreach ychar $ylist {
    set veclist [list]
    set scorelist [list]
    set g $i
    set m [incr i $gp]
    foreach xchar $xlist f $scorelist1 {
      if {$xchar == $ychar} {
        set s [expr {$g+1}]
        set v 1
      } else {
        set s [expr {$g+$mp}]
        set v 3
      }
      if {[expr {$f+$gp}] > $s} {
        set s [expr {$f+$gp}]
        set v 2
      }
      if {[expr {$m+$gp}] > $s} {
        set m [expr {$m+$gp}]
        lappend veclist 0
      } else {
        set m $s
        lappend veclist $v
      }
      lappend scorelist $m
      set g $f

    } 
    set scorelist1 $scorelist
    lappend allvecs $veclist
  }
  set btime [clock clicks]
  set x [expr {[string length $xdna]-1}] 
  set y [expr {[string length $ydna]-1}] 
  set yout ""
  set xout ""
  set lineout ""
  while {($x >-1) && ($y > -1)} {
    switch [lindex [lindex $allvecs $y] $x] {
      0 {
        set yout "-$yout"
        set xout "[string index $xdna $x]$xout"
        set lineout " $lineout"
        incr x -1 
      }
      1 {
        set yout "[string index $ydna $y]$yout"
        set xout "[string index $xdna $x]$xout"
        set lineout "|$lineout"
        incr x -1 
        incr y -1
      }
      3 {
        set yout "[string index $ydna $y]$yout"
        set xout "[string index $xdna $x]$xout"
        set lineout "#$lineout"
        incr x -1 
        incr y -1
      }
      2 {
        set yout "[string index $ydna $y]$yout"
        set xout "-$xout" 
        set lineout " $lineout"
        incr y -1
      }
    }
  }
  if {$x > -1} {
    set yout "[string repeat - [expr {1+$x}]]$yout"
    set xout "[string range $xdna 0 $x]$xout"
    set lineout "[string repeat " " [expr {1+$x}]]$lineout"
  } else {
    set xout "[string repeat - [expr {1+$y}]]$xout"
    set yout "[string range $ydna 0 $y]$yout"
    set lineout "[string repeat " " [expr {1+$y}]]$lineout"
  }
  set ctime [clock clicks]
  return [list $xout $yout $lineout]
}


################
## link a double click in an analysis window to the original wndow
################
proc alignlink {textbox window index dir mode {raise 1}} {
  #set text [split [$textbox get [$textbox index "$index linestart"] [$textbox index $index]] "\t"]
  #set linestart [lindex $text 0]
  #set chars [regexp -all {[^-]} [lindex $text 1]]

  if {[regexp {([0-9]+)(?:[[:blank:]~]*)(.*)} [$textbox get [$textbox index "$index linestart"] [$textbox index $index]] line linestart text]} {
    set chars [regexp -all {[^-\~\ ]} $text]

    if {[string first ".abi_window" $window] == -1} {
      #####sequence window
      if {([$textbox get $index] == "-") || ([$textbox get $index] == " ")} {
        if {$mode == "select"} {
          select_region $window [$window.textarea index insert] 1.[expr {$linestart-1+$dir*$chars}]
        } else {
          select_region $window 1.[expr {$linestart-1+$dir*$chars}] 1.[expr {$linestart-1+$dir*$chars}]
        }
      } else {
        if {$mode == "select"} {
          if {[$window.textarea compare [$window.textarea index insert] > 1.[expr {$linestart+$dir*$chars}]]} {
            select_region $window [$window.textarea index insert] 1.[expr {$linestart-1+$dir*$chars}]
          } else {
            select_region $window [$window.textarea index insert] 1.[expr {$linestart+$dir*$chars}]
          }
        } else {
          select_region $window 1.[expr {$linestart-1+$dir*$chars}] 1.[expr {$linestart-1+$dir*$chars}]
        }
      }
      if {$raise} {
        wm deiconify $window
        raise $window
        focus $window.textarea
      }
    } else {
      ####abi window
      if {$mode == "select"} {
        abi_select_region $window [expr {$linestart-1+$dir*$chars}] select
      } else {
        abi_select_region $window [expr {$linestart-1+$dir*$chars}] [expr {$linestart-1+$dir*$chars}]
      }
      if {$raise} {
        wm deiconify $window
        raise $window
        focus $window.c
      }
    }
  } else {
    bell
  }
}

#############
##align ling that uses lsxx_yy tag to find line start index(xx) and spacing(yy) for linking translated sequences, or DNA without indexes 
#############
proc alignlink2 {textbox window index mode {raise 1}} {

  if {([set lstag [lsearch -inline  [$textbox tag names $index] ls*]] != "") && [regexp {ls([0-9]+)_(-?[0-9]+)} $lstag blank linestart spacing]} {
    set text [$textbox get [$textbox index "$index linestart"] [$textbox index $index]]
    regsub -all {[0-9]|[[:blank:]~]} $text "" text
    set chars [regexp -all {[^-]} $text]

    if {[$textbox get $index] == "-"} {
      if {$mode == "select"} {
        select_region $window [$window.textarea index insert] 1.[expr {$linestart-1+$spacing*$chars}]
      } else {
        select_region $window 1.[expr {$linestart-1+$spacing*$chars}] 1.[expr {$linestart-1+$spacing*$chars}]
      }
    } else {
      if {$mode == "select"} {
        if {[$window.textarea compare [$window.textarea index insert] > 1.[expr {$linestart+$spacing*$chars}]]} {
          select_region $window [$window.textarea index insert] 1.[expr {$linestart-1+$spacing*$chars}]
        } else {
          select_region $window [$window.textarea index insert] 1.[expr {$linestart-1+$spacing*($chars+1)}]
        }
      } else {
          select_region $window 1.[expr {$linestart+(($spacing>0)?-1:0)+$spacing*($chars+1)}] 1.[expr {$linestart+(($spacing>0)?-1:0)+$spacing*$chars}]
      }
    }
    if {$raise} {
      wm deiconify $window
      raise $window
      focus $window.textarea
    }
  } else {
    bell
  }
}

################
## Send DNA to the NCBI BLAST server and open results in a browser
################
proc NCBI_Blast {w} {
  global ok
  global info dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  if {[$w.textarea tag ranges sel] == ""} {
    set first 1.0
    set last [$w.textarea index end-1chars]
  } else {
    set first [$w.textarea index sel.first]
    set last [$w.textarea index sel.last]
  }
  set text [$w.textarea get $first $last]

  if {[string length $text] > 15} {
    toplevel .dialog 
    wm protocol .dialog WM_DELETE_WINDOW "set ok -1"
    wm title .dialog [mc "BLAST DNA at NCBI..."]
    set raisedialog_store [bind . <<RaiseDialogs>>]
    bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
    regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
    wm geometry .dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
    if {[tk windowingsystem] != "aqua"} {
      wm transient .dialog $w
    } else {
      ::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
    }
    set ok 0


    grid [frame .dialog.frame1] -row 0 -column 0
    grid [label .dialog.frame1.proglabel -text [mc "BLAST Program:"]] -row 0 -column 0 -sticky nse
    grid [menubutton .dialog.frame1.prog -textvariable info(ncbi_blast,program) -menu .dialog.frame1.prog.menu -indicatoron 1 -relief raised] -row 0 -column 1 -sticky w
    menu .dialog.frame1.prog.menu
     foreach bprog [list blastn blastp blastx tblastn tblastx] {
      .dialog.frame1.prog.menu add radiobutton -label $bprog -indicatoron 1 -value $bprog -variable info(ncbi_blast,program)
    }
    bind .dialog.frame1.prog.menu <<MenuSelect>> "focus .dialog"
    grid [label .dialog.frame1.expectlabel -text [mc "Expect:"]] -row 1 -column 0 -sticky nse
    grid [entry .dialog.frame1.expect -textvariable info(ncbi_blast,expect)] -row 1 -column 1 -sticky w
    grid [checkbutton .dialog.frame1.filter -text [mc "Low Complexity Filter"] -onvalue 1 -offvalue 0 -variable info(ncbi_blast,low_filter)] -row 2 -column 0 -columnspan 2

    grid [frame .dialog.frame2] -row 1 -column 0
    grid [button .dialog.frame2.ok -text [mc "OK"] -command "set ok 1" -default active] -row 1 -column 0 -sticky ns
    grid [button .dialog.frame2.cancel -text [mc "Cancel"] -command "set ok -1"] -row 1 -column 1 -sticky ns

    grid columnconfigure .dialog 0 -weight 1
    grid rowconfigure .dialog 0 -weight 1
    grid rowconfigure .dialog 1 -weight 0

    regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .dialog] wgeom winw winh winx winy
    if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
    if {$winx < 0} {set winx 0}
    if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
    if {$winy < 0} {set winy 0}
    wm geometry .dialog "+$winx+$winy"
    update

    focus .dialog
    vwait ok 
    set dialogblock 0
    destroy .dialog
    bind . <<RaiseDialogs>> ""
    if {$ok == 1} {
      if {($info(ncbi_blast,program) == "blastp") || ($info(ncbi_blast,program) == "tblastn")} {
        set text [translate $text]
      }
      set url "http://www.ncbi.nlm.nih.gov/blast/Blast.cgi?QUERY=$text&DATABASE=$info(ncbi_blast,database)&EXPECT=$info(ncbi_blast,expect)&PROGRAM=$info(ncbi_blast,program)&SERVICE=plain&CMD=Put"
      if {$info(ncbi_blast,low_filter)} {append url "&FILTER=L"}

      set x [::http::geturl $url -blocksize 4096 -timeout 4000 -command blast_callback]
    }
  } else {
    tk_messageBox -message [mc "Not enough sequence selected for BLASTing"] -type ok -icon warning -default ok
    set dialogblock 0
  }
}


################
## called by blast to open the browser to display results
################
proc blast_callback {token} {
  global info

  if {[::http::status $token] == "timeout"} {
    warning_dialog "Command timed out- possibly a firewall issue?"
  } else {
    set html  [::http::data $token]

    #get the RID:
    if {[regexp {.*RID = (.*?)\n.*} $html fill RID]} {
     #set the web page URL
      set url "http://www.ncbi.nlm.nih.gov/blast/Blast.cgi?CMD=Get&RID=$RID&FORMAT_OBJECT=Alignment&FORMAT_TYPE=HTML&DESCRIPTIONS=100&ALIGNMENTS=200&ALIGNMENT_TYPE=Pairwise&SHOW_OVERVIEW=yes"
      sputs $RID
      sputs $url
      #open browser and display
      if {[tk windowingsystem] == "win32" } {
        if {($info(browser) != "")} {
          if {[catch {dde request $info(browser) WWW_OpenURL $url,,0x0}]} {
            sputs "No dde, using exec"
            regsub -all {&} $url {\\\&} url
            if {[string first %1 $info(appCmd)] > -1} {
              eval exec [regsub {"%1"} $info(appCmd) \"$url\"] &
            } else {
              eval exec $info(appCmd) \"$url\" &
            }
          }
        } else {
          if {[string first %1 $info(appCmd)] > -1} {
            eval exec [regsub {"%1"} $info(appCmd) \"$url\"] &
          } else {
            eval exec $info(appCmd) \"$url\" &
          }
          #warning_dialog "Sorry, I don't know how to open your browser. Your RID is:$RID"
        }
      } elseif {[tk windowingsystem] == "aqua"} {
        exec open $url
      } elseif {[tk windowingsystem] == "x11"} {
        foreach executable {mozilla netscape iexplorer opera lynx w3m links galeon konquerer mosaic firefox amaya browsex elinks} {
          set executable [auto_execok $executable]
          if {[string length $executable]} {
            set command [list $executable $url &]
            break
          }
        }
        if {[info exists command]} {
          if {[catch {eval exec $command} err]} {
            tk_messageBox -icon error -message "error '$err' with '$command'"
          }
        } else {
          warning_dialog "Sorry, I don't know how to open a browser in your OS."
        }
      } else {
        warning_dialog "Sorry, I don't know how to open a browser in your OS. Your RID is:$RID"
      }
    } else {
      warning_dialog "Error: $html"
    }
  }
  ::http::cleanup $token
}

################
## Send DNA to the wormbase BLAST server and open results in a browser
################
proc Wormbase_Blast {w} {
  global ok
  global info dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  if {1} {
    toplevel .dialog
    wm protocol .dialog WM_DELETE_WINDOW "set ok -1"
    wm title .dialog [mc "BLAST DNA at Wormbase..."]
    set raisedialog_store [bind . <<RaiseDialogs>>]
    bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
    regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
    wm geometry .dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
    if {[tk windowingsystem] != "aqua"} {
      wm transient .dialog $w
    } else {
      ::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
    }
    set ok 0


    grid [frame .dialog.frame1] -row 0 -column 0
    grid [label .dialog.frame1.proglabel -text [mc "BLAST Program:"]] -row 0 -column 0 -sticky nse
    grid [menubutton .dialog.frame1.prog -textvariable info(wb_blast,program) -menu .dialog.frame1.prog.menu -indicatoron 1 -relief raised] -row 0 -column 1 -sticky w
    menu .dialog.frame1.prog.menu
     foreach label [list blastn blastp blastx tblastn tblastx BLAT] {
      .dialog.frame1.prog.menu add radiobutton -label $label -indicatoron 1 -value $label -variable info(wb_blast,program)
    }
    bind .dialog.frame1.prog.menu <<MenuSelect>> "focus .dialog"

    grid [label .dialog.frame1.databaselabel -text [mc "Database:"]] -row 1 -column 0 -sticky nse
    grid [menubutton .dialog.frame1.database -textvariable info(wb_blast,database) -menu .dialog.frame1.database.menu -indicatoron 1 -relief raised] -row 1 -column 1 -sticky w
    menu .dialog.frame1.database.menu
     foreach bprog [list {C. elegans genomic} {C. elegans proteins (WormPep)} {C. briggsae genomic} {C. elegans ESTs} {C. briggsae proteins (Brigpep)}] {
      .dialog.frame1.database.menu add radiobutton -label $bprog -indicatoron 1 -value $bprog -variable info(wb_blast,database)
    }

    grid [label .dialog.frame1.expectlabel -text [mc "Expect Threshold:"]] -row 2 -column 0 -sticky nse
    grid [menubutton .dialog.frame1.expect -textvariable info(wb_blast,expect) -menu .dialog.frame1.expect.menu -indicatoron 1 -relief raised] -row 2 -column 1 -sticky w
    menu .dialog.frame1.expect.menu
     foreach bprog [list .0001 .01 1 10 100 1000 BLAT] {
      .dialog.frame1.expect.menu add radiobutton -label $bprog -indicatoron 1 -value $bprog -variable info(wb_blast,expect)
    }

    set info(wb_blast,maxhits) 100

    grid [frame .dialog.frame2] -row 1 -column 0
    grid [button .dialog.frame2.ok -text [mc "OK"] -command "set ok 1" -default active] -row 1 -column 0 -sticky ns
    grid [button .dialog.frame2.cancel -text [mc "Cancel"] -command "set ok -1"] -row 1 -column 1 -sticky ns

    grid columnconfigure .dialog 0 -weight 1
    grid rowconfigure .dialog 0 -weight 1
    grid rowconfigure .dialog 1 -weight 0

    regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .dialog] wgeom winw winh winx winy
    if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
    if {$winx < 0} {set winx 0}
    if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
    if {$winy < 0} {set winy 0}
    wm geometry .dialog "+$winx+$winy"
    update

    focus .dialog
    vwait ok
    set dialogblock 0
    destroy .dialog
    bind . <<RaiseDialogs>> ""
    if {$ok == 1} {

      array set dblookup [list {C. elegans genomic} Elegans {C. elegans proteins (WormPep)} WormPep {C. briggsae genomic} Briggsae_genomic {C. elegans ESTs} EST_Elegans {C. briggsae proteins (Brigpep)} Briggpep]
      if {[$w.textarea tag ranges sel] == ""} {
        set first 1.0
        set last [$w.textarea index end-1chars]
      } else {
        set first [$w.textarea index sel.first]
        set last [$w.textarea index sel.last]
      }
      set text [$w.textarea get $first $last]

      if {($info(wb_blast,program) == "blastp") || ($info(wb_blast,program) == "tblastn")} {
        set text [translate $text]
      }

      set url "http://blast.wormbase.org/db/searches/blat?sequence=$text&db=$dblookup($info(wb_blast,database))&cutoff=$info(wb_blast,expect)&program=$info(wb_blast,program)&max_hits=$info(wb_blast,maxhits)&name=&action=Run%20BLAST%20or%20BLAT"

      if {[tk windowingsystem] == "win32" } {
        set htmlfile [open [file join $info(Accdir) "last_WB_BLAST.html"] w]
        puts $htmlfile "<form name=\"myform\" method=\"post\" action=\"http://blast.wormbase.org/db/searches/blat\" enctype=\"application/x-www-form-urlencoded\">"
        puts $htmlfile "<input type=\"hidden\" name=\"sequence\" value=\"$text\">"
        puts $htmlfile "<input type=\"hidden\" name=\"db\" value=\"$dblookup($info(wb_blast,database))\">"
        puts $htmlfile "<input type=\"hidden\" name=\"cutoff\" value=\"$info(wb_blast,expect)\">"
        puts $htmlfile "<input type=\"hidden\" name=\"max_hits\" value=\"$info(wb_blast,maxhits)\">"
        puts $htmlfile "<input type=\"hidden\" name=\"program\" value=\"$info(wb_blast,program)\">"
        puts $htmlfile "<input type=\"hidden\" name=\"name\" value=\"\">"
        puts $htmlfile "<input type=\"hidden\" name=\"action\" value=\"Run BLAST or BLAT\">"
        puts $htmlfile "<h1>If Javascript is disabled, click here</h1>"
        puts $htmlfile "<INPUT TYPE=SUBMIT VALUE=\"Get results\"> "
        puts $htmlfile "</form>" 
        puts $htmlfile "<SCRIPT language=\"JavaScript1.2\">"
        puts $htmlfile " document.myform.submit();"
        puts $htmlfile "</SCRIPT>"

        close $htmlfile
        after 500
        set htmlfile "[file join $info(Accdir) "last_WB_BLAST.html"]"
sputs $htmlfile

        if {($info(browser) != "")} {
          if {[catch {dde request $info(browser) WWW_OpenURL \"file://$htmlfile\",,0x0}]} {
            sputs "No dde, using exec"
            if {[string first %1 $info(appCmd)] > -1} {
              eval exec [regsub {"%1"} $info(appCmd) \"file://$htmlfile\"] &
            } else {
              eval exec $info(appCmd) \"file://$htmlfile\" &
            }
            #eval exec [regsub {"%1"} $info(appCmd) ""] &
            #set i 0
            #while {([catch {dde request $info(browser) WWW_OpenURL \"file://$htmlfile\",,-1}]) && ($i < 10)} {
            #  after 500
            #  incr i
            #  sputs "trying dde $i"
            #}
          }
        } else {
          warning_dialog "Sorry, I don't know how to open your browser."
        }

      } elseif {[tk windowingsystem] == "aqua"} {
        exec open $url
      } elseif {[tk windowingsystem] == "x11"} {
        foreach executable {mozilla netscape iexplorer opera lynx w3m links galeon konquerer mosaic firefox amaya browsex elinks} {
          set executable [auto_execok $executable]
          if [string length $executable] {
            set command [list $executable $url &]
            break
          }
        }
        if [info exists command] {
          if [catch {eval exec $command} err] {
            tk_messageBox -icon error -message "error '$err' with '$command'"
          }
        } else {
          warning_dialog "Sorry, I don't know how to open a browser in your OS."
        }
      } else {
        warning_dialog "Sorry, I don't know how to open a browser in your OS."
      }
    }
  } else {
    tk_messageBox -message [mc "Not enough sequence selected for BLASTing"] -type ok -icon warning -default ok
  }
}

################
## Generic function to get info from a web post operation
################
proc URL_Post {base_url function vars_list} {
  global info

      if {[tk windowingsystem] == "win32" } {
        set htmlfile [open [file join $info(Accdir) "last_$function.html"] w]
        puts $htmlfile "<form name=\"myform\" method=\"post\" action=\"$base_url\" enctype=\"application/x-www-form-urlencoded\">"
        foreach name $vars_list {
          puts $htmlfile "<input type=\"hidden\" name=\"$name\" value=\"$info($function,$name)\">"
        }
        puts $htmlfile "<h1>If Javascript is disabled, click here</h1>"
        puts $htmlfile "<INPUT TYPE=SUBMIT VALUE=\"Get results\"> "
        puts $htmlfile "</form>" 
        puts $htmlfile "<SCRIPT language=\"JavaScript1.2\">"
        puts $htmlfile " document.myform.submit();"
        puts $htmlfile "</SCRIPT>"

        close $htmlfile
        after 500
        set htmlfile "[file join $info(Accdir) "last_$function.html"]"
sputs $htmlfile

        if {($info(browser) != "")} {
          if {[catch {dde request $info(browser) WWW_OpenURL \"file://$htmlfile\",,0x0}]} {
            sputs "No dde, using exec"
            if {[string first %1 $info(appCmd)] > -1} {
              eval exec [regsub {"%1"} $info(appCmd) \"file://$htmlfile\"] &
            } else {
              eval exec $info(appCmd) \"file://$htmlfile\" &
            }
          }
        } else {
          warning_dialog "Sorry, I don't know how to open your browser."
        }
      } elseif {[tk windowingsystem] == "aqua"} {
        set url "$base_url"
        set div "?"
        foreach name $vars_list {
          set url "$url$div$name=$info($function,$name)"
          set div "&"
        }
        exec open $url
      } elseif {[tk windowingsystem] == "x11"} {
        foreach name $vars_list {
          set url "$url$div$name=$info($function,$name)"
          set div "&"
        }
        foreach executable {mozilla netscape iexplorer opera lynx w3m links galeon konquerer mosaic firefox amaya browsex elinks} {
          set executable [auto_execok $executable]
          if [string length $executable] {
            set command [list $executable $url &]
            break
          }
        }
        if [info exists command] {
          if [catch {eval exec $command} err] {
            tk_messageBox -icon error -message "error '$err' with '$command'"
          }
        } else {
          warning_dialog "Sorry, I don't know how to open a browser in your OS."
        }
      } else {
        warning_dialog "Sorry, I don't know how to open a browser in your OS."
      }

}

################
## initialize some info for MS Windows
################
proc initialize_win32 {} {
  package require registry
  global info possible_extensions tcl_platform env

  ##set browser info
  set root HKEY_CLASSES_ROOT
  if {![catch {set appKey [registry get $root\\.html ""]}]} {
    set appCmd [registry get $root\\$appKey\\shell\\open\\command ""]
    regsub -all {\\} $appCmd  {\\\\} info(appCmd)
    if {[string match -nocase *iexplore.exe* $info(appCmd)]} {
      set info(browser) iexplore
    } elseif {[string match -nocase *netscape.exe* $info(appCmd)]} {
      set info(browser) NETSCAPE
    } elseif {[string match -nocase *mozilla.exe* $info(appCmd)]} {
      set info(browser) Mozilla
    } elseif {[string match -nocase *opera.exe* $info(appCmd)]} {
      set info(browser) Opera
    } elseif {[string match -nocase *firefox.exe* $info(appCmd)]} {
      set info(browser) Firefox
    } else {
      sputs "Sorry, I don't recognize the browser: [file tail $info(appCmd)]"
      set info(browser) ""
    }
  } else {
      sputs "Sorry, no application registered for .html files"
      set info(browser) ""
  }
sputs "command:$info(appCmd)"
sputs "browser:$info(browser)"
  ##set system colors info
  set info(bg_color) SystemButtonFace
  set info(menu_bg_color) SystemMenu
  set info(menu_fg_color) SystemMenuText
  set info(menu_select_bg_color) SystemHighlight
  set info(menu_select_fg_color) SystemHighlightText
  set info(text_bg_color) SystemWindow
  set info(text_fg_color) SystemWindowText
  set info(text_select_bg_color) SystemHighlight
  set info(text_select_fg_color) SystemHighlightText
  set info(tk_ps_fonts) [list helvetica Helvetica arial Helvetica times TimesNewRomanPSMT {news gothic mt} TimesNewRomanPSMT {times new roman} TimesNewRomanPSMT courier Courier {courier new} Courier]

  ##set user defaults directory
  if {$tcl_platform(os) == "Windows 95"} {
   ##Windows 95/98
    if {[catch {set info(user_defaults_dir) [registry get {HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders} AppData]}]} {
      sputs "Can't get the user defaults dir from registry, using internal"
      set info(user_defaults_dir) $info(Accdir)
    } else {
      set info(user_defaults_dir) [file join $info(user_defaults_dir) Ape]
    }

  } else {
   ##NT and XP
    if {[catch {set info(user_defaults_dir) $env(APPDATA)}]} {
      sputs "Can't get the user defaults dir from environment variables, using internal"
      set info(user_defaults_dir) $info(Accdir)
    } else {
      set info(user_defaults_dir) [file join $info(user_defaults_dir) Ape]
    }
  }

  ##set window icon
  if {[file exists [file join $info(Accdir) "Icons and images" "APE_icon.ico"]]} {
    set window_icon [file join $info(Accdir) "Icons and images" "APE_icon.ico"]
    catch {wm iconbitmap . -default $window_icon}
  }
  
  ##load the wmf, gdi, hdc and printer dlls
  if {[file exists [file join $info(Accdir) "lib" "printer.dll"]]} {
    if {[file exists [file join $info(Accdir) "lib" "hdc.dll"]]} {
      if {[file exists [file join $info(Accdir) "lib" "gdi.dll"]]} {
        if {[file exists [file join $info(Accdir) "lib" "wmf.dll"]]} {
          load [file join $info(Accdir) "lib" "hdc.dll"]
          load [file join $info(Accdir) "lib" "printer.dll"]
          load [file join $info(Accdir) "lib" "gdi.dll"]
          load [file join $info(Accdir) "lib" "wmf.dll"]
        }
      }
    }
  }

  ##load the drag and drop dll
  if {[file exists [file join $info(Accdir) "lib" "libtkdnd.dll"]]} {
    load [file join $info(Accdir) "lib" "libtkdnd.dll"]
    proc HandleFileDrop {data type} {
      sputs "data:$data type:$type"
      if {$type == "Files"} {
        foreach filename $data {
          if {[file extension $filename] == ".lnk"} {
            if {[file exists [set filename [find_link_target $filename]]]} {
              open_file $filename
            }
          } else {
            open_file $filename
          }
        }
      }
    }
    proc find_link_target {filename} {
      set fileid [open $filename r]
 
      if {[read $fileid 1] != "L"} {close $fileid; return ""}
      #read flags
      seek $fileid 0x14
      binary scan [read $fileid 4] b7 flags
      set has_id_list [string index $flags 0]
      set is_file [string index $flags 1]
      set has_desc [string index $flags 2]
      set has_rel_path [string index $flags 3]
      set has_working_dir [string index $flags 4]
      set has_command_line [string index $flags 5]
      set has_custom_icon [string index $flags 6]


      if {!$is_file} {close $fileid; return ""}

      #if item id list, skip it
      if {$has_id_list} {
        #read the item ID list size
        seek $fileid 0x4C
        binary scan [read $fileid 2] s id_size
        set file_loc_off [expr (0x4E+$id_size)]
      } else {
        set file_loc_off 0x4E
      }

      #read the location info_header
      seek $fileid $file_loc_off
      set x [read $fileid 28]
      binary scan $x iiiiiii structure_len offset_last_item location_flags local_v_off   local_path_off net_v_off final_path_off
      set has_local [expr {$location_flags % 2}]
      set has_net [expr {$location_flags >> 1}]


      if {$has_local} {
        #read local_volume_table_header
        seek $fileid [expr {$file_loc_off+$local_v_off}]
        binary scan [read $fileid 16] iiii len_loc_v_table fixed_disk vol_ser vol_label_off
        set vol_label ""
        while {(![string equal [set c [read $fileid 1]] \0]) && (![eof $fileid])} {
          set vol_label "$vol_label$c"
        }

        seek $fileid [expr {$file_loc_off+$local_path_off}]
        set local_path ""
        while {(![string equal [set c [read $fileid 1]] \0]) && (![eof $fileid])} {
          set local_path "$local_path$c"
        }
      }
 
      if {$has_net} {
        #read net_volume_table_header
        seek $fileid [expr {$file_loc_off+$net_v_off}]
        binary scan [read $fileid 20] iiiii len_net_v_table x share_name_off x x

        set share_name ""
        while {(![string equal [set c [read $fileid 1]] \0]) && (![eof $fileid])} {
          set share_name "$share_name$c"
        }
      }

      seek $fileid [expr {$file_loc_off+$final_path_off}]
      set final_path ""
      while {(![string equal [set c [read $fileid 1]] \0]) && (![eof $fileid])} {
        set final_path "$final_path$c"
      }

      close $fileid
      if {$has_local} {
        set leader $local_path
      } elseif {$has_net} {
        set leader $share_name
      }

      if {$final_path != ""} {
        return "$leader[file separator]$final_path"
      } else {
        return $leader
      }
    }
  }

  set possible_extensions [list .ape .dna .str .seq .abi .ab1 .gb .embl]

}
################
## initialize some info for Aqua (OS X)
################
proc initialize_aqua {} {
  global info modstring

  if {[file exists [file join $info(Accdir) lib MacCarbonPrint.dylib]]} {
    load [file join $info(Accdir) lib MacCarbonPrint.dylib]
  }

  #add menubar to the window
. configure -menu [menu .menubar]

  menu .menubar.filemenu
  .menubar add cascade -menu .menubar.filemenu -label "File"
  .menubar.filemenu add command -label "New" -accelerator "$modstring+N"  -command "create_window" -state disabled

  .menubar.filemenu add command -label "Open..." -accelerator "$modstring+O" -command "open_file_dialog" -state disabled
  .menubar.filemenu add command -label "Duplicate"  -accelerator "$modstring+D"  -state disabled
  .menubar.filemenu add command -label "Close"  -accelerator "$modstring+W" -state disabled
  .menubar.filemenu add separator
  .menubar.filemenu add command -label "Save"  -accelerator "$modstring+S" -state disabled
  .menubar.filemenu add command -label "Save As..." -state disabled


#define edit menu
  menu .menubar.edit
  .menubar add cascade -menu .menubar.edit -label "Edit"
  .menubar.edit add command -label "Can't Undo" -state disabled -accelerator "$modstring+Z"
  .menubar.edit add command -label "Cut"  -accelerator "$modstring+X" -state disabled
  .menubar.edit add command -label "Copy"  -accelerator "$modstring+C" -state disabled
  .menubar.edit add command -label "Paste"  -accelerator "$modstring+V" -state disabled
  .menubar.edit add separator
  .menubar.edit add command -label "Cut Rev-Com"  -accelerator "$modstring+Shift+X" -state disabled
  .menubar.edit add command -label "Copy Rev-Com"  -accelerator "$modstring+Shift+C" -state disabled
  .menubar.edit add command -label "Paste Rev-Com"  -accelerator "$modstring+Shift+V" -state disabled
  .menubar.edit add separator
  .menubar.edit add command -label "Preferences..." -state disabled
  .menubar.edit add separator
  .menubar.edit add command -label "Select All"  -accelerator "$modstring+A" -state disabled
  .menubar.edit add command -label "Select From-To..."  -accelerator "$modstring+H"  -state disabled
  .menubar.edit add command -label "Jump To..."  -accelerator "$modstring+J" -state disabled
  .menubar.edit add separator
  .menubar.edit add command -label "Find..."  -accelerator "$modstring+F" -state disabled
  .menubar.edit add command -label "Find Again"  -accelerator "$modstring+G" -state disabled
  .menubar.edit add command -label "Clear Find"  -state disabled
  .menubar.edit add separator
  .menubar.edit add command -label "Convert to UPPERCASE"  -accelerator "$modstring+Shift+U" -state disabled
  .menubar.edit add command -label "Convert to lowercase"  -accelerator "$modstring+L" -state disabled
  .menubar.edit add command -label "UPPER<->lower" -accelerator "$modstring+="  -state disabled
  .menubar.edit add separator
  .menubar.edit add command -label "Reverse-Complement"  -accelerator "$modstring+/" -state disabled
  .menubar.edit add separator
  .menubar.edit add command -label "Circular<->Linear" -accelerator "$modstring+U" -state disabled
  .menubar.edit add command -label "Set Origin"  -accelerator "$modstring+I" -state disabled
  .menubar.edit add command -label "Linearize @ insert site" -state disabled

#add enzymes menu
menu .menubar.enzymes
  .menubar add cascade -menu .menubar.enzymes -label "Enzymes"
  .menubar.enzymes add command -label "Enzyme Selector..." -state disabled
  .menubar.enzymes add separator
  .menubar.enzymes add checkbutton -label "Selection Only" -onvalue "selection" -offvalue "all" -variable info(analyze_selection) -state disabled
  .menubar.enzymes add separator
  .menubar.enzymes add command -label "Graphic Map" -state disabled
  .menubar.enzymes add command -label "Graphic Map +U" -state disabled
  .menubar.enzymes add separator
  .menubar.enzymes add command -label "Highlight" -state disabled
  .menubar.enzymes add command -label "Clear Highlighting" -state disabled
  .menubar.enzymes add separator
  .menubar.enzymes add command -label "Digestion" -state disabled
  .menubar.enzymes add cascade -label "Ladder"  -state disabled
  .menubar.enzymes add separator
  .menubar.enzymes add command -label "List Enzymes..."  -state disabled
  .menubar.enzymes add cascade -label "Quick Lists"  -state disabled
  .menubar.enzymes add separator
  .menubar.enzymes add command -label "Silent Sites" -state disabled
  .menubar.enzymes add separator
  .menubar.enzymes add command -label "Text Map..." -state disabled

#add ORFs menu
menu .menubar.orf
  .menubar add cascade -menu .menubar.orf -label "ORFs"
    .menubar.orf add command -label "Find Next" -accelerator "$modstring+>" -state disabled
    .menubar.orf add command -label "Find Previous" -accelerator "$modstring+<"  -state disabled

    .menubar.orf add separator
    .menubar.orf add command -label "Translate..." -state disabled
    .menubar.orf add separator
    .menubar.orf add command -label "ORF map"  -state disabled
    .menubar.orf add separator
    .menubar.orf add radiobutton -label "Selection Translate" -variable info(selection_shows) -value "Translate" -state disabled
    #.menubar.orf add radiobutton -label "Selection None" -variable info(selection_shows) -value "None"  -state disabled

#add oligos menu
#menu .menubar.oligos
    #.menubar add cascade -menu .menubar.oligos -label "Oligos"
    #.menubar.oligos add command -label "Find Primers..."  -state disabled
    #.menubar.oligos add separator
    #.menubar.oligos add radiobutton -label "Selection Tm, %GC" -variable info(selection_shows) -value "Tm" -state disabled
    #.menubar.oligos add radiobutton -label "Selection None" -variable info(selection_shows) -value "None" -state disabled

#add features menu
menu .menubar.features
  .menubar add cascade -menu .menubar.features -label "Features"
  $w.menubar.features add command -label [mc "New Feature..."] -state disabled
  $w.menubar.features add command -label [mc "List Features"] -state disabled
  $w.menubar.features add command -label [mc "Edit Features..."]  -state disabled
  $w.menubar.features add command -label [mc "Clear Features"] -accelerator "$modstring+Shift+H"  -state disabled
  $w.menubar.features add separator
  $w.menubar.features add command -label [mc "Edit Feature Library..."]  -state disabled
  $w.menubar.features add command -label [mc "New Feature in Library..."] -state disabled
  $w.menubar.features add command -label [mc "Open Feature Library..."]  -state disabled
  $w.menubar.features add command -label [mc "Annotate Features using Library"] -accelerator "$modstring+H"  -state disabled



#add Tools menu
menu .menubar.tools
  .menubar add cascade -menu .menubar.tools -label "Tools"
    .menubar.tools add command -label "Find Primers..."  -state disabled
    .menubar.tools add separator
    .menubar.tools add command -label "Align Sequences..."  -state disabled
    .menubar.tools add separator
    .menubar.tools add command -label "BLAST Sequences at NCBI..."  -state disabled
    .menubar.tools add command -label [mc "BLAST Sequences at Wormbase..."] -state disabled

#add Windows menu stub (menu items are maintained by update_windows_menu procedure)
menu .menubar.windows
    .menubar add cascade -menu .menubar.windows -label "Windows"

#add Help menu
menu .menubar.help
  .menubar add cascade -menu .menubar.help -label "Help"
    .menubar.help add command -label "Standard Genetic Code"  -state disabled 
    .menubar.help add command -label "AA Info"  -state disabled


   #define the apple menu
    menu .menubar.apple
      .menubar add cascade -menu .menubar.apple
      .menubar.apple add command -label [mc "About ApE..."] -command "about_dialog ." 
      .menubar.apple add separator
      .menubar.apple add command -label [mc "Preferences..."]  -command "configure_preferences ." -state disabled
  set info(newaquax) 20
  set info(newaquay) 20
  set info(aqua_page_setup) {}
  set info(user_defaults_dir) "~/Library/Preferences/ApE/"
  bind Listbox <Command-Button-1> [bind Listbox <Control-Button-1>]
  bind Text
}

################
## set up file type associations on Windows
################
proc file_typer {{on_extensions {}} {off_extensions {}} {clear_type 0}} {
  global possible_extensions
  global info argv0

  if {($clear_type) && ([registry keys "HKEY_CLASSES_ROOT\\" ApE.DNA.file] != "")} {
    #remove all extensions
    registry delete HKEY_CLASSES_ROOT\\ApE.DNA.file
  } elseif {($on_extensions != [list]) && ([registry keys "HKEY_CLASSES_ROOT\\" "ApE.DNA.file"] == "")} {
      #set up file description
      set classpath "HKEY_CLASSES_ROOT\\ApE.DNA.file"
        registry set $classpath {} "ApE DNA File" sz
        registry set "$classpath\\DefaultIcon" {} "[file nativename [file join $info(Accdir) "Icons and images" "APE_icon.ico"]],0" sz
        registry set "$classpath\\Shell\\open\\command" {} "\"[file nativename [info nameofexecutable]]\" \"[file nativename $argv0]\" \"%1\"" sz
        registry set $classpath\\Shell\\open\\ddeexec {} "open_file {%1}" sz
      registry set $classpath\\Shell\\open\\ddeexec\\ifexec {} "#" sz
        registry set $classpath\\Shell\\open\\ddeexec\\Application {} "TclEval" sz
        registry set $classpath\\Shell\\open\\ddeexec\\topic {} "readDNAfiles" sz
  }
  foreach extension $on_extensions {
    if {([registry keys "HKEY_CLASSES_ROOT\\" $extension] == "")} {
      registry set "HKEY_CLASSES_ROOT\\$extension" {} "ApE.DNA.file" sz
    } else {
      set type [registry get "HKEY_CLASSES_ROOT\\[registry get "HKEY_CLASSES_ROOT\\$extension" {}]" {}]
      if {[tk_messageBox -message "The extension $extension is already registered as type: $type.\n Are you sure you want to replace it?" -type okcancel -default cancel] == "ok"} {
        registry set "HKEY_CLASSES_ROOT\\$extension" {} "ApE.DNA.file" sz
      }
    }
  }
  foreach extension $off_extensions {
    registry delete "HKEY_CLASSES_ROOT\\$extension"
  }
}

################
## returns filetypes that are registered on windows as ApE.DNA.file and files that are registered as other types
################
proc get_registered_types {{file_descriptor ApE.DNA.file}} {
  global possible_extensions

  set r [list]
  set b [list]
  set a [list]
  foreach extension $possible_extensions {
    if {([registry keys "HKEY_CLASSES_ROOT\\" $extension] != [list]) && (![catch {registry get  "HKEY_CLASSES_ROOT\\$extension" {}}]) && ([registry get  "HKEY_CLASSES_ROOT\\$extension" {}] != [list])} {
     if {[set descript [registry get "HKEY_CLASSES_ROOT\\$extension" {}]] == $file_descriptor} {
       lappend r $extension
     } else {
       lappend a $extension 
       if {[catch {set temp [registry get "HKEY_CLASSES_ROOT\\$descript" {}]}] == 0} {
         lappend b $temp
       } else {
         lappend b "Unknown Program:$descript"
       }
     }
    }
  }
  return [list $r $a $b]
}

################
## maintain the windows menu
################
proc old_update_windows_menu {} {
  global info
  global toolbar_images

  foreach window [winfo children .] {
    if {[winfo exists $window.menubar.windows]} {
      $window.menubar.windows delete 0 end

      set analysiswindows [regexp -all -inline "$window\_analysis\[0-9\]+" [winfo children .]]
      foreach awindow $analysiswindows {
        $window.menubar.windows add command -label " [wm title $awindow]" -command "wm deiconify $awindow; raise $awindow; focus $awindow; event generate $awindow <Button-1> ";#  -compound left -image $toolbar_images(ana_win)
      }
      if {$analysiswindows != [list]} {
        $window.menubar.windows add separator
      }

      foreach windowelement [regexp -all -inline ".dna_window\[0-9\]* " "[winfo children .] "] {
        set windowelement [string range $windowelement 0 end-1]
        if {($windowelement != $window)} {
          #$window.menubar.windows add command -label $info($windowelement,win_title) -command "wm deiconify $windowelement; raise $windowelement"
          $window.menubar.windows add command -label [wm title $windowelement] -command "wm deiconify $windowelement; raise $windowelement; focus $windowelement.textarea";# -compound left -image $toolbar_images(dna_win)  
        } 
      }
      set abiwindows [regexp -all -inline ".abi_window\[0-9\]* " "[winfo children .] "]
      if {$abiwindows != [list]} {
        $window.menubar.windows add separator
      }
      foreach abwindow $abiwindows {
        set abwindow [string range $abwindow 0 end-1]
        if {($abwindow != $window) || ([tk windowingsystem] == "aqua")} {
          $window.menubar.windows add command -label "[wm title $abwindow]" -command "wm deiconify $abwindow; raise $abwindow; focus $abwindow";#-compound left -image $toolbar_images(abi_win)
        } 
      }
    }
  }
}

################
## maintain the windows menu (2nd version)
################
proc update_windows_menu {} {
  global info
  global toolbar_images global_true

  set global_true 1
  foreach window [winfo children .] {
    if {[winfo exists $window.menubar.windows]} {
      $window.menubar.windows delete 0 end

      ####add in the sequence windows
      foreach windowelement [regexp -all -inline ".dna_window\[0-9\]* " "[winfo children .] "] {
        set windowelement [string range $windowelement 0 end-1]
        if {($windowelement != $window)} {
          $window.menubar.windows add command -label [wm title $windowelement] -command "wm deiconify $windowelement; raise $windowelement; focus $windowelement.textarea";# -compound left -image $toolbar_images(dna_win)  
        } else {
          $window.menubar.windows add checkbutton -variable global_true -label [wm title $windowelement] -command "set global_true 1;wm deiconify $windowelement; raise $windowelement; focus $windowelement.textarea";# -compound left -image $toolbar_images(dna_win)  
        }
        ##add the analysis children of the window
        set analysiswindows [regexp -all -inline "$windowelement\_analysis\[0-9\]+" [winfo children .]]
        foreach awindow $analysiswindows {
          if {$window != $awindow} {
            $window.menubar.windows add command -label "     [wm title $awindow]" -command "wm deiconify $awindow; raise $awindow; focus $awindow; event generate $awindow <Button-1> ";#  -compound left -image $toolbar_images(ana_win)
          } else {
            $window.menubar.windows add checkbutton -variable global_true -label "     [wm title $awindow]" -command "set global_true 1;wm deiconify $awindow; raise $awindow; event generate $awindow <Button-1> ";#  -compound left -image $toolbar_images(ana_win)
          }
        }
      }

      ####add in the abi windows
      set abiwindows [regexp -all -inline ".abi_window\[0-9\]* " "[winfo children .] "]
      if {$abiwindows != [list]} {
        $window.menubar.windows add separator
      }
      foreach abwindow $abiwindows {
        set abwindow [string range $abwindow 0 end-1]
        if {($abwindow != $window)} {
          $window.menubar.windows add command -label "[wm title $abwindow]" -command "wm deiconify $abwindow; raise $abwindow; focus $abwindow";#-compound left -image $toolbar_images(abi_win)
        }  else {
          $window.menubar.windows add checkbutton -variable global_true -label [wm title $abwindow] -command "set global_true 1;wm deiconify $abwindow; raise $abwindow; focus $abwindow.c";# -compound left -image $toolbar_images(abi_win)  
        }
        ##add the analysis children of the window
        set analysiswindows [regexp -all -inline "$abwindow\_analysis\[0-9\]+" [winfo children .]]
        foreach awindow $analysiswindows {
          if {$window != $awindow} {
            $window.menubar.windows add command -label "     [wm title $awindow]" -command "wm deiconify $awindow; raise $awindow; focus $awindow; event generate $awindow <Button-1> ";#  -compound left -image $toolbar_images(ana_win)
          } else {
            $window.menubar.windows add checkbutton -variable global_true -label "     [wm title $awindow]" -command "set global_true 1;wm deiconify $awindow; raise $awindow; event generate $awindow <Button-1> ";#  -compound left -image $toolbar_images(ana_win)
          }
        }
      }
      
      ###add orphaned analysis windows
      set ping 0
      set awindows [regexp -all -inline {([^[:space:]]+)_analysis[0-9]+} [winfo children .]]
      foreach {awindow parent} $awindows {
        if {![winfo exists $parent]} {
          if {!$ping} {
            $window.menubar.windows add separator
            set ping 1
          }
          if {$window != $awindow} {
            $window.menubar.windows add command -label "     [wm title $awindow]" -command "wm deiconify $awindow; raise $awindow; focus $awindow; event generate $awindow <Button-1> ";#  -compound left -image $toolbar_images(ana_win)
          } else {
            $window.menubar.windows add checkbutton -variable global_true -label "     [wm title $awindow]" -command "set global_true 1;wm deiconify $awindow; raise $awindow; event generate $awindow <Button-1> ";#  -compound left -image $toolbar_images(ana_win)
          }
        }
      }
    }
  }
}


###########
##About ApE dialog
###########
proc about_dialog {w} {
  global ok
  global info 
  global version dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  toplevel .about_dialog 
  wm withdraw .about_dialog
  wm title .about_dialog [mc "About ApE"]
  wm protocol .about_dialog WM_DELETE_WINDOW "set ok -1"
  bind . <<RaiseDialogs>> "wm deiconify .about_dialog; raise .about_dialog"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry .about_dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .about_dialog $w
  } else {
    ::tk::unsupported::MacWindowStyle style .about_dialog document {closeBox}
  }
  #wm deiconify .about_dialog
  #raise .about_dialog
  set ok 0

  set filename [file join $info(Accdir) "Icons and images" "gibbons.gif"]
  if {[file exists $filename]} {
    image create photo "gibbons_image" -file $filename
  } else {
    image create photo "gibbons_image" -height 1 -width 1
  }
  set filename [file join $info(Accdir) "Icons and images" "pwrdLogo75.gif"]
  if {[file exists $filename]} {
    image create photo "Tcl_image" -file $filename
  } else {
    image create photo "Tcl_image" -height 1 -width 1
  }

  grid [canvas .about_dialog.t -relief flat -bg {light steel blue} -width 550 -height 250 -highlightthickness 0] -row 1 -column 0  -sticky nwe

  grid [frame .about_dialog.f3 -relief ridge -bd 0 -bg {light steel blue}] -row 2 -column 0 -sticky nswe
  grid [button .about_dialog.f3.ok -command "set ok 1" -text [mc "OK"] -highlightbackground {light steel blue} -width 10 -default active] -pady 10 -row 0 -column 0 -sticky nswe

  set y 10
  set title [.about_dialog.t create text 250 $y -text "ApE- A plasmid Editor v$version" -anchor n -font "Helvetica 24 bold" -justify center]
  incr y 36
  .about_dialog.t create image [lindex [.about_dialog.t bbox $title] 2] [lindex [.about_dialog.t bbox $title] 1] -image "Tcl_image" -anchor nw
  .about_dialog.t create text 250 $y -text "Copyright  2003-2005 by M. Wayne Davis" -anchor n  -font boldlabelfont -justify center
  incr y [expr {[font metrics labelfont -linespace]+5}]

  if {[mc "Translation by:"] != "Translation by:"} {
    .about_dialog.t create text 250 $y -text "[mc Locale:][mclocale].  [mc "Translation by:"]" -anchor n  -font labelfont -justify center
  } else {
    .about_dialog.t create text 250 $y -text "[mc Locale:][mclocale]." -anchor n  -font labelfont -justify center
  }
  incr y [font metrics labelfont -linespace]

  .about_dialog.t create text 250 $y -text "Running with Tclkit by Jean-Claude Wippler (equi4 software) \nusing Tcl/Tk, IncrTcl, TclVFS, Zlib, and Metakit software." -anchor n  -font labelfont -justify center
  incr y [expr {2*[font metrics labelfont -linespace]}]
  .about_dialog.t create text 250 $y -text "Tcl/Tk  Regents of the University of California, Sun Microsystems, Inc., \nScriptics Corporation, ActiveState Corporation and other parties." -anchor n  -font labelfont -justify center
  incr y [expr {2*[font metrics labelfont -linespace]}]
  .about_dialog.t create text 250 $y -text "IncrTcl  Cadence Design Systems, Inc., and other parties." -anchor n  -font labelfont -justify center
  incr y [font metrics labelfont -linespace]
  .about_dialog.t create text 250 $y -text "TclVFS  Vince Darley, and other parties." -anchor n  -font labelfont -justify center
  incr y [font metrics labelfont -linespace]
  .about_dialog.t create text 250 $y -text "Zlib  1995-2003 Jean-loup Gailly and Mark Adler" -anchor n  -font labelfont -justify center
  incr y [font metrics labelfont -linespace]
  .about_dialog.t create text 250 $y -text "Metakit  1996-2004 Jean-Claude Wippler " -anchor n  -font labelfont -justify center
  incr y [font metrics labelfont -linespace]
  if {[tk windowingsystem] == "aqua"} {
    .about_dialog.t create text 250 $y -text "Tcl/Tk Aqua port by Jim Ingham & Ian Reid  2001-2002 Apple Computer" -anchor n  -font labelfont -justify center
    incr y [font metrics labelfont -linespace]
    .about_dialog.t create text 250 $y -text "Wiskit for Aqua  2003-2004, Daniel A. Steffen" -anchor n  -font labelfont -justify center
    incr y [font metrics labelfont -linespace]
    if  {([lsearch -glob [info loaded] "*MacCarbonPrint.dylib*"] > -1)} {
      .about_dialog.t create text 250 $y -text "OSX carbon printing extensions  Mats Bengtsson and other parties." -anchor n -font labelfont -justify center
      incr y [font metrics labelfont -linespace]
    }
  } else {
    if  {([lsearch -glob [info loaded] "*printer.dll*"] > -1)} {
      .about_dialog.t create text 250 $y -text "Windows printing and metafile extensions  Michael I. Schwartz" -anchor n -font labelfont -justify center
      incr y [font metrics labelfont -linespace]
    }
    if  {([lsearch -glob [info loaded] "*libtkdnd.dll*"] > -1)} {
      .about_dialog.t create text 250 $y -text "Windows drag and drop extensions  George Petasis and Laurent Riesterer" -anchor n -font labelfont -justify center
      incr y [font metrics labelfont -linespace]
    }
  }
  .about_dialog.t create image 250 $y -image "gibbons_image" -anchor n

  .about_dialog.t configure -width [expr {10+[lindex [.about_dialog.t bbox all] 2]}] -height [expr {10+[lindex [.about_dialog.t bbox all] 3]}]
  wm resizable .about_dialog 0 0
  update idletasks
  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .about_dialog] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry .about_dialog "+$winx+$winy"

  wm deiconify .about_dialog
  raise .about_dialog
  focus .about_dialog.f3.ok

  bind .about_dialog <Key-Return> ".about_dialog.f3.ok invoke"
  vwait ok
  set dialogblock 0
  #grab release .about_dialog
  destroy .about_dialog
  image delete "gibbons_image"
  image delete "Tcl_image"
  bind . <<RaiseDialogs>> ""
}

###########
##Codon Table Help Window
###########
proc genetic_code_dialog {w} {
  if {[winfo exists .gen_code]} {
    wm deiconify .gen_code
    raise .gen_code
  } else {
    toplevel .gen_code 
    wm title .gen_code [mc "Standard Genetic Code..."]
    wm protocol .gen_code WM_DELETE_WINDOW ".gen_code.f3.ok invoke"
    regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
    wm geometry .gen_code "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"

    grid [text .gen_code.t -bg white -wrap none -width 80 -height 30 ] -row 1 -column 0  -sticky nwe

    grid [frame .gen_code.f3 -relief flat] -row 2 -column 0 -sticky nswe
    grid [button .gen_code.f3.ok -command "destroy .gen_code" -text "OK"] -row 0 -column 0 -sticky nswe

    .gen_code.t insert end "               Second Position\n" topline
    .gen_code.t insert end "        T         C         A          G\n" midlines
    .gen_code.t insert end "   T | TTT Phe | TCT Ser | TAT Tyr  | TGT Cys  | T\n" midlines
    .gen_code.t insert end "F  T | TTC Phe | TCC Ser | TAC Tyr  | TGC Cys  | C  T\n" midlines
    .gen_code.t insert end "i  T | TTA Leu | TCA Ser | TAA Stop | TGA Stop | A  h\n" midlines
    .gen_code.t insert end "r  T | TTG Leu | TCG Ser | TAG Stop | TGG Trp  | G  i\n" midlines
    .gen_code.t insert end "s    +---------+---------+----------+----------+--- r\n" midlines

    .gen_code.t insert end "t  C | CTT Leu | CCT Pro | CAT His  | CGT Arg  | T  d\n" midlines
    .gen_code.t insert end "   C | CTC Leu | CCC Pro | CAC His  | CGC Arg  | C   \n" midlines
    .gen_code.t insert end "P  C | CTA Leu | CCA Pro | CAA Gln  | CGA Arg  | A  P\n" midlines
    .gen_code.t insert end "o  C | CTG Leu | CCG Pro | CAG Gln  | CGG Arg  | G  o\n" midlines
    .gen_code.t insert end "s    +---------+---------+----------+----------+--- s\n" midlines

    .gen_code.t insert end "i  A | ATT Ile | ACT Thr | AAT Asn  | AGT Ser  | T  i\n" midlines
    .gen_code.t insert end "t  A | ATC Ile | ACC Thr | AAC Asn  | AGC Ser  | C  t\n" midlines
    .gen_code.t insert end "i  A | ATA Ile | ACA Thr | AAA Lys  | AGA Arg  | A  i\n" midlines
    .gen_code.t insert end "o  A | ATG Met | ACG Thr | AAG Lys  | AGG Arg  | G  o\n" midlines
    .gen_code.t insert end "n    +---------+---------+----------+----------+--- n\n" midlines

    .gen_code.t insert end "   G | GTT Val | GCT Ala | GAT Asp  | GGT Gly  | T   \n" midlines
    .gen_code.t insert end "   G | GTC Val | GCC Ala | GAC Asp  | GGC Gly  | C   \n" midlines
    .gen_code.t insert end "   G | GTA Val | GCA Ala | GAA Glu  | GGA Gly  | A   \n" midlines
    .gen_code.t insert end "   G | GTG Val | GCG Ala | GAG Glu  | GGG Gly  | G   \n" midlines

    .gen_code.t insert end "\nTAA=Ochre, TAG=Amber, TGA=Opal"

    .gen_code.t tag configure topline -font "Courier 10 bold" -tabs {5c center}
    .gen_code.t tag configure midlines -font "Courier 10" -tabs {1c 2c center 3c 4c center 5c 6c center 7c 8c center 9c 10c}

    .gen_code.t configure -state disabled

    wm resizable .gen_code 0 0
    bind .gen_code <Key-Return> ".gen_code.f3.ok invoke"
    update idletasks
    regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .gen_code] wgeom winw winh winx winy
    if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
    if {$winx < 0} {set winx 0}
    if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
    if {$winy < 0} {set winy 0}
    wm geometry .gen_code "+$winx+$winy"
  }
}

###########
##AA Info help window
###########
proc aa_info_dialog {w} {
  global info

  if {[winfo exists .aa_data]} {
    wm deiconify .aa_data
    raise .aa_data
  } else {
    set aa_names [list Glycine Alanine Serine Threonine Cysteine Valine Leucine Isoleucine Methionine Proline Phenylalanine Tyrosine Tryptophan "Aspartic Acid" "Glutamic Acid" Asparagine Glutamine Histidine Lysine Arginine]
    set aa_3let [list Gly Ala Ser Thr Cys Val Leu Ile Met Pro Phe Tyr Trp Asp Glu Asn Gln His Lys Arg]
    set aa_1let [list G A S T C V L I M P F Y W D E N Q H K R]
    set aa_MW [list 57.05 71.09 87.08 101.11 103.15 99.14 113.16 113.16 131.19 97.12 147.18 163.18 186.21 115.09 129.12 114.11 128.14 137.14 128.17 156.19]

   #read in gif images if available. If not, make blank images
    if {[lsearch [image names] Glycine_image] == -1} {
      set dirname [file join $info(Accdir) "Icons and images" "Amino Acids"]
      foreach aa $aa_names {
        set filename [file join $dirname "$aa.gif"]
        if {[file exists $filename] == 1} {
          image create photo "$aa\_image" -file $filename
        } else {
          image create photo "$aa\_image" -height 1 -width 1
        }
      }
    }
    toplevel .aa_data 
    wm title .aa_data [mc "Amino Acids Data..."]
    wm protocol .aa_data WM_DELETE_WINDOW ".aa_data.f3.ok invoke"
    regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
    wm geometry .aa_data "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"

    grid [canvas .aa_data.c -bg wheat -bd 0 -highlightthickness 0 -width 200 -height 300 -yscrollcommand ".aa_data.ys set" ] -row 1 -column 0  -sticky nswe
    grid [scrollbar .aa_data.ys -command ".aa_data.c yview"] -row 1 -column 1 -rowspan 1 -sticky nsw

    grid [frame .aa_data.f3 -relief flat -bg wheat -bd 0] -row 2 -column 0 -columnspan 2 -sticky nswe
    grid [button .aa_data.f3.ok -bg {light steel blue} -command "destroy .aa_data" -text "OK"] -row 0 -column 0 -sticky nswe

    grid rowconfigure .aa_data 1 -weight 1
    grid columnconfigure .aa_data 0 -weight 1

    set y 10
    foreach aa $aa_names 3let $aa_3let 1let $aa_1let MW $aa_MW {
    set img [.aa_data.c create image 10 $y -image "$aa\_image" -anchor nw]
    set imgy [lindex [.aa_data.c bbox $img] 3]
    set imgx [lindex [.aa_data.c bbox $img] 0]
    .aa_data.c create text $imgx $imgy -text "$aa ($3let, $1let)\nMW: $MW" -anchor nw -font "Helvetica 10 bold" 
    set y [expr {$imgy+50}]
    }

    .aa_data.c configure -scrollregion [.aa_data.c bbox all]
    bindtags .aa_data.c .aa_data.c
    bind .aa_data.c <MouseWheel> {.aa_data.c yview scroll [expr {- (%D / $info(mousewheelunits)) * 4}] units}
    bind .aa_data <MouseWheel> {.aa_data.c yview scroll [expr {- (%D / $info(mousewheelunits)) * 4}] units}

   # wm resizable .aa_data 0 1
    bind .aa_data <Key-Return> ".aa_data.f3.ok invoke"
    update idletasks
    regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .aa_data] wgeom winw winh winx winy
    if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
    if {$winx < 0} {set winx 0}
    if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
    if {$winy < 0} {set winy 0}
    wm geometry .aa_data "+$winx+$winy"
    wm deiconify .aa_data
  }
}

#################
## initialize bindings for AnalysisText type
#################
proc setup_AlalysisText_bindings {} {
  global modifier
  #set default binding for Button-Mouse wheel events for all text boxes
  if {[tk windowingsystem] != "aqua"} {
    bind Text <B1-MouseWheel> "[bind Text <MouseWheel>]
      [bind Text <Shift-Button-1>]"
  } else {
    bind Text <B1-MouseWheel> [bind Text <MouseWheel>]
  }
  bind AnalysisText <$modifier-Key-a> "%W tag add sel 1.0 end"
  bind AnalysisText <Double-Button-1> "%W tag add sel @%x,%y"
  foreach action [list <B1-MouseWheel> <MouseWheel> <<Copy>> <B1-Motion> <Button-1> <B1-Leave> <B1-Enter> <Shift-Button-1> <ButtonRelease-1>] {
    bind AnalysisText $action [bind Text $action]
  }
}
#################
## if a process is already running, pass off to it (only on windows so far)
#################
proc check_for_running_process {} {
  global argv tcl_platform

  ##if on windows, check for running version and pass the file off, else register dde to catch new instances
  if {$tcl_platform(platform) == "windows"} {
    package require dde 1.1
    if {[dde services TclEval readDNAfiles] != ""} {
      if {[llength $argv] == 0} {
        dde execute TclEval readDNAfiles "create_window"
      } else {
        dde execute TclEval readDNAfiles "uplevel #0 \{open_file \{[lindex $argv 0]\}\}"
      }
      final_exit
    } else {
      dde servername "readDNAfiles"
    }
  } elseif {($tcl_platform(platform) == "unix") && ([tk windowingsystem] != "aqua")} {
    if {[tk appname "dna_ape"] != "dna_ape"} {
      if {[llength $argv] == 0} {
        send -async dna_ape "create_window"
      } else {
        send -async dna_ape "uplevel #0 \{open_file \{[lindex $argv 0]\}\}"
      }
      final_exit
    }
  }
}

#################
##returns a 3 byte hex number (24 bits), each base has 4 bits a=0111, c=1011, g=1101, t=1110 , n=0000
#################
proc validation_code {p} {
  set p [string map {^ "" ( "" - "" ) "" / ""} $p]
  regsub -all {([0-9])*} $p "" p
  set p [string trim $p "N"] 
  if {[string length $p] < 7} {
    return "0x[string map "A 7 B 8 C b D 4 G d H 2 K c M 3 N 0 R 5 S 9 T e V 1 W 6 Y a" [string range [append $p "NNNNN"] 0 5]]"
  } else {
    return {}
  }
}

#################
##adds compatible enzymes to a list at enzinfo(compat,$enzyme)
#################
proc add_compatibles {} {
  global enzymes
  global enzinfo

  foreach enzyme $enzymes {
    set enzinfo(compat,$enzyme) ""
    set len [string length $enzinfo(flatpat,$enzyme)]
    set caret [string first ^ $enzinfo(enz_pattern,$enzyme)]
    if {($caret != -1) && ([set oh [expr {$len/2.0-$caret}]] != 0)} {
      if {$oh > 0} {
        set overhang [string range $enzinfo(flatpat,$enzyme) $caret [expr {$len-$caret-1}]]
        if {([regexp {[^ACGT]} $overhang] == 0)} {
          lappend h(+$overhang) $enzyme
        }
      } else {
        set overhang [string range $enzinfo(flatpat,$enzyme) [expr {$len-$caret}] [expr {$caret-1}]]
        if {([regexp {[^ACGT]} $overhang] == 0)} {
          lappend h(-$overhang) $enzyme 
        }  
      }
    } 
  }

  foreach ov [array names h] {
    if {[llength $h($ov)] > 1} {
      set i 0
      foreach enz $h($ov) {
        #sputs "$enz:[lreplace $h($ov) $i $i]"
        set enzinfo(compat,$enz) [lreplace $h($ov) $i $i]
        incr i
      }
    }
  }
}

#############
## return all aa encoded by a potentially degenerate 3 base sequence
#############
proc triplet_transl {t} {
  set s [string map {b [cgt] d [agt] h [act] k [gt] m [ac] n ? r [ag] s [gc] v [acg] w [at] y [ct]} $t] 
  array set aa {ttt F ttc F tta L ttg L tct S tcc S tca S tcg S tat Y tac Y taa * tag * tgt C tgc C tga * tgg W ctt L ctc L cta L ctg L cct P ccc P cca P ccg P cat H cac H caa Q cag Q cgt R cgc R cga R cgg R att I atc I ata I atg M act T acc T aca T acg T aat N aac N aaa K aag K agt S agc S aga R agg R gtt V gtc V gta V gtg V gct A gcc A gca A gcg A gat D gac D gaa E gag E ggt G ggc G gga G ggg G}
  set result [list]
  foreach i [array names aa $s] {
    lappend result $aa($i)
  }
  set result [lsort -unique $result]
  if {[llength $result] < 21} {
    return [join $result ""]
  } else {
    return .
  }
}

#############
## return all possible peptides encoded in all three frames of a sequence
#############
proc degen_transl {p} {
  set p [string tolower $p]
  regsub -all -- {[^abcdghkmnrstvwy]} $p {} p
  set p "nn$p"
  append p "nn"
  set r [list]
  for {set i 0} {$i < 3} {incr i} {
    set k ""
    for {set j $i} {$j < [expr [string length $p]-2]} {incr j 3} {
      append k "\[[triplet_transl [string range $p $j [expr $j+2]]]\]"
    }
  lappend r $k
  }
return $r
}

#############
## read the enzymes file
#############
proc read_enzymes {enzfilename} {
  global longnamelist
  global hexlist
  global enzymes
  global enzinfo
  global info

  set error 0
  if {($enzfilename == "") || ([file exists $enzfilename] == 0)} {
    warning_dialog "Can't find the enzymes file"
    set error 1
  } else {
    set fileid [open $enzfilename r]
    set header [gets $fileid]
    if {$header == "tkDNAview enzymes file type 2"} {
      array unset enzinfo
      set enzinfo(enz_groups_list) {All {Mem Recall} {Difference...}}
      set enzinfo(enz_groups,All) [list]
      set enzinfo(enz_groups,Mem\ Recall) [list]
      set enzinfo(enz_groups,Difference...) "Difference..."

      set longnamelist [gets $fileid]
      set hexlist [gets $fileid]
      set enzymes [gets $fileid]
      set pats [gets $fileid]
      set enzcomments [gets $fileid]

      set enzinfo(enz_groups_list) {All {Mem Recall} {Difference...}}
      set enzinfo(enz_groups,All) $enzymes
      set enzinfo(enz_groups,Mem\ Recall) [list]
      set enzinfo(enz_groups,Difference...) "Difference..."
      set newgroups [gets $fileid]
      set enzinfo(enz_groups_list) [concat $enzinfo(enz_groups_list) $newgroups]
      foreach group $newgroups {
        set enzinfo(enz_groups,$group) [gets $fileid] 
      }
      unset newgroups

      #read list of Dam and Dcm blocked enzymes
      set enzinfo(enz_methylaselist) [gets $fileid]

      #make list of all enzymes blocked by any methylase
      set info(blocked_enzymes) [list]
      foreach methylase $enzinfo(enz_methylaselist) {
        foreach blocked_enzyme [lindex $methylase 1] {
          lappend info(blocked_enzymes) [lindex $blocked_enzyme 0]
        }
      }
      #add DpnI specifically
      if {[lsearch -exact $enzymes DpnI] > -1} {
        lappend info(blocked_enzymes) "DpnI"
      }

      set info(blocked_enzymes) [lsort -unique $info(blocked_enzymes)]

      set info(enz_currently_selected) [list]

      foreach enzyme $enzymes pattern $pats comment $enzcomments {
        set enzinfo(enz_vcode,$enzyme) [validation_code $pattern]
        set enzinfo(enz_pattern,$enzyme) $pattern
        set enzinfo(enz_comment,$enzyme) $comment
        set p [string map {^ "" ( "" - "" ) "" / ""} $pattern]
        regsub -all {([0-9])*} $p "" p
        set p [string trim $p "N"] 
        set enzinfo(flatpat,$enzyme) $p
      }
      add_compatibles
      set enzinfo(enz_vcode,MDam) [validation_code GATC]
      set enzinfo(enz_vcode,MDcm) [validation_code CCWGG]
      set info(enzymes_saved) 1
    } else {
      tk_messageBox -message "Wrong enzyme file type- no enzymes read" -type ok -icon warning -default ok
      set error 1
    }
    close $fileid
  }
  return $error
}

#############
## reload the enzymes file
#############
proc reload_enzymes {} {
  global enz_select_window
  global info
  global enz_select_window

  if {$info(enzymes_saved) != 1} {
    if {[tk_messageBox -message "Do you want to save changes to the enzyme set?" -type yesno] == "yes"} {
      save_enzymes "" 0
    }
  }
  if {![file isdirectory $info(default_enzymedir)]} {
    global argv0
    set info(default_enzymedir) [file dirname $argv0]
  }
  set filename [tk_getOpenFile -initialdir $info(default_enzymedir) -title "Open An Enzyme File" -defaultextension ".txt"]
  if {($filename != "") && ([file exists $filename] == 1)} {
    set error [read_enzymes $filename]
    if {($error == 0)} {
      foreach window $info(dnawindows_list) {
        set info($window,scanned) 0
        findenzymes $window
      }

      if {[winfo exists .enz_selector.textframe.textbox] == 1} {
        fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment $enz_select_window
        wm deiconify .enz_selector
      }
    }
  }
}

#############
## save a copy of the enzymes file
#############
proc save_enzymes {{filename ""} {default 0}} {
  global longnamelist
  global hexlist
  global enzymes
  global enzinfo
  global info

  if {[winfo exists .enz_selector]} {
    set parent_win .enz_selector
  } else {
    set parent_win .
  }
  if {![file isdirectory $info(default_enzymedir)]} {
    global argv0
    set info(default_enzymedir) [file dirname $argv0]
  }
  if {$default} {set filename [file join $info(default_enzymedir) Default_Enzymes.txt]}
  if {($filename == "")} {
    set filename [tk_getSaveFile -title "Save Enzymes File" -parent $parent_win -initialdir $info(default_enzymedir) -defaultextension ".txt"]
  }

  while {([file exists $filename]) && (![file writable $filename])} {
    warning_dialog "$filename\nFile isn't writable"
    set filename [tk_getSaveFile -title "Save Enzymes File" -parent $parent_win -initialdir $info(default_enzymedir) -defaultextension ".txt"]
  }

  if {$filename == ""} return

  set fileid [open $filename w]
  puts $fileid "tkDNAview enzymes file type 2"
  puts $fileid $longnamelist 
  puts $fileid $hexlist
  puts $fileid $enzymes

  foreach enzyme $enzymes {
    lappend pats $enzinfo(enz_pattern,$enzyme)
    lappend enzcomments $enzinfo(enz_comment,$enzyme)
  }
  puts $fileid $pats
  puts $fileid $enzcomments

  set groupslist [list]
  foreach group $enzinfo(enz_groups_list) {
    if {($group != "All") && ($group != "Mem Recall") && ($group != "Difference...")} {
      lappend groupslist $group
    }
  }
  puts $fileid $groupslist
  foreach group $groupslist {
    puts $fileid $enzinfo(enz_groups,$group)
      }

  puts $fileid $enzinfo(enz_methylaselist)

  close $fileid
}

#################
##return the hexlist address for a pattern
#################
proc hexmult {pattern} {
set n [list 0]
foreach c [split $pattern ""] {
  set newlist [list]
  switch $c {
    "A" {foreach m $n {set newlist [concat $newlist [expr $m*4]]}}
    "B" {foreach m $n {set newlist [concat $newlist  [expr $m*4+1] [expr $m*4+2] [expr $m*4+3]]}}
    "C" {foreach m $n {set newlist [concat $newlist  [expr $m*4+1]]}}
    "D" {foreach m $n {set newlist [concat $newlist [expr $m*4]  [expr $m*4+2] [expr $m*4+3]]}}
    "G" {foreach m $n {set newlist [concat $newlist  [expr $m*4+2]]}}
    "H" {foreach m $n {set newlist [concat $newlist [expr $m*4] [expr $m*4+1] [expr $m*4+3]]}}
    "K" {foreach m $n {set newlist [concat $newlist  [expr $m*4+2] [expr $m*4+3]]}}
    "M" {foreach m $n {set newlist [concat $newlist [expr $m*4] [expr $m*4+1]] }}
    "N" {foreach m $n {set newlist [concat $newlist [expr $m*4] [expr $m*4+1] [expr $m*4+2] [expr $m*4+3]]}}
    "R" {foreach m $n {set newlist [concat $newlist [expr $m*4]  [expr $m*4+2]]}}
    "S" {foreach m $n {set newlist [concat $newlist [expr $m*4+1]  [expr $m*4+2]]}}
    "T" {foreach m $n {set newlist [concat $newlist [expr $m*4+3]]}}
    "V" {foreach m $n {set newlist [concat $newlist [expr $m*4] [expr $m*4+1] [expr $m*4+2]]}}
    "W" {foreach m $n {set newlist [concat $newlist [expr $m*4] [expr $m*4+3]]}}
    "Y" {foreach m $n {set newlist [concat $newlist [expr $m*4+1] [expr $m*4+3]]}}
   }
  set n $newlist
}
return $n
}

#################
##return a pattern with degenerate code substituted with regexp equivalents
#################
proc regsubpattern {pattern} {
  regsub -all " " $pattern "" pattern
  regsub -all {[\,\;\|]} $pattern "|" pattern
  return [string map {b [cgtsky] d [agtkrw] h [actryw] k [gt] m [ac] n . r [ag] s [cg] v [acgmrs] w [at] y [ct] B [CGTSKY] D [AGTKRW] H [ACTRYW] K [GT] M [AC] N . R [AG] S [CG] V [ACGMRS] W [AT] Y [CT]} $pattern]
}

#################
##proc K returns a variable and allows it to be deleted at the same time
##by calling "[K $var [set var ""]]
#################
proc K { x y } { set x }


#################
##add a pattern to either the longnamelist or hexlist
#################
proc addhexarray {name pattern} {
  global longnamelist
  global hexlist


  regsub -all {\(([0-9]|[/-])*\)} $pattern "" spattern
  regsub "\\^" $spattern "" spattern
  set spattern [string trim $spattern "N"] 
  set rpattern [revcom $spattern]
  if {[string length $spattern] >6} {
    if {$rpattern == $spattern} {
      lappend longnamelist [list $name [regsubpattern $spattern]]
    } else {
      ###other option is to add the enzyme twice, same name both patterns
      lappend longnamelist [list $name "[regsubpattern $spattern]|[regsubpattern $rpattern]"]
    }
  } else {
    if {$rpattern == $spattern} {
      set spattern [string range [join "$spattern NNNNN" ""] 0 5]
      set patternlist [hexmult $spattern]
    } else {
      set spattern [string range [join "$spattern NNNNN" ""] 0 5]
      set rpattern [string range [join "$rpattern NNNNN" ""] 0 5]
      set patternlist [concat [hexmult $spattern] [hexmult $rpattern]]
    }

    foreach location $patternlist {
      set newlist [concat [lindex $hexlist $location] [list $name]]
      set hexlist [lreplace [K $hexlist [set hexlist {}]] $location $location $newlist]
    }
  }
}

###############
## delete an enzyme
###############
proc delete_enzyme {enzyme} {
  proc K { x y } { set x }
  global longnamelist
  global hexlist
  global enzymes
  global enzinfo
  global info
  global enz_select_window

  set listpos [lsearch -exact $enzymes $enzyme]
  if {$listpos == -1} {return}

  ##take the enzyme out of the enzyme list
  set enzymes [lreplace [K $enzymes [set enzymes {}]] $listpos $listpos]

  
  ##take the enzyme out of the currently selected list
  set info(enz_currently_selected) [lsearch -exact -not -inline -all $info(enz_currently_selected) $enzyme]

  ##take the enzyme out of the longname list or hexlist
  set spattern $enzinfo(flatpat,$enzyme)
  if {[string length $spattern] >6} {
    set longlistpos [lsearch -glob $longnamelist "$enzyme *"]
    set longnamelist [lreplace [K $longnamelist [set longnamelist {}]] $longlistpos $longlistpos]
  } else {
    set rpattern [revcom $spattern]
    if {$rpattern == $spattern} {
      set spattern [string range [join "$spattern NNNNN" ""] 0 5]
      set patternlist [hexmult $spattern]
    } else {
      set spattern [string range [join "$spattern NNNNN" ""] 0 5]
      set rpattern [string range [join "$rpattern NNNNN" ""] 0 5]
      set patternlist [concat [hexmult $spattern] [hexmult $rpattern]]
    }
    foreach location $patternlist {
      set newlist [lindex $hexlist $location]
      set pos [lsearch -exact $newlist $enzyme]
      set newlist [lreplace [K $newlist [set newlist {}]] $pos $pos]
      set hexlist [lreplace [K $hexlist [set hexlist {}]] $location $location $newlist] 
    }
  }

  ##take the enzyme out of the info and enzinfo arrays
  #array unset enzinfo "enz_vcode,$enzyme"
  #array unset enzinfo "enz_pattern,$enzyme"
  #array unset enzinfo "enz_comment,$enzyme"
  #array unset enzinfo "flatpat,$enzyme"
  foreach element [array names enzinfo "*,$enzyme"] {
    array unset enzinfo $element
  }
  foreach element [array names info "*,$enzyme"] {
    array unset info $element
  }

  ##take the enzyme out of the groups list
  foreach group $enzinfo(enz_groups_list) {
    set pos [lsearch -exact $enzinfo(enz_groups,$group) $enzyme]
    if {$pos > -1} {
      set enzinfo(enz_groups,$group) [lreplace [K $enzinfo(enz_groups,$group) [set enzinfo(enz_groups,$group) {}]] $pos $pos]
    }
  }

  ##take the enzyme out of the metlist
  set newmetlists [list]
  foreach metlist $enzinfo(enz_methylaselist) {
    set metlistpos [lsearch -glob [lindex $metlist 1] "$enzyme *"]
    if {$metlistpos > -1} {
      lappend newmetlists [list [lindex $metlist 0] [lreplace [lindex $metlist 1] $metlistpos $metlistpos]]
    } else {
      lappend newmetlists $metlist
    }
  }
  set enzinfo(enz_methylaselist) $newmetlists
  set info(blocked_enzymes) [list]
  foreach methylase $enzinfo(enz_methylaselist) {
    foreach blocked_enzyme [lindex $methylase 1] {
      lappend info(blocked_enzymes) [lindex $blocked_enzyme 0]
    }
  }
  #add DpnI specifically
  if {[lsearch -exact $enzymes DpnI] > -1} {
    lappend info(blocked_enzymes) "DpnI"
  }
  set info(blocked_enzymes) [lsort -unique $info(blocked_enzymes)]

  ##remake the compatibles elements 
  add_compatibles

  foreach window $info(dnawindows_list) {
    set info($window,scanned) 0
    findenzymes $window
  }

  if {[winfo exists .enz_selector.textframe.textbox] == 1} {
    .enz_selector.textframe.textbox tag delete $enzyme
    fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment $enz_select_window
  }

}

#################
## return which groups pattern belogs to
#################
proc pattern_groups {pattern} {
  set result [list All]
  regsub  {^[^ABCDGHKMNRSTVWY]*?\)} $pattern "" pattern ;#remove leading parens
  set caret [string first ^ $pattern]
  if {$caret > -1} {
    set length [expr {[string length $pattern] -1.0}]    
    set site [expr {$caret * 2 - $length}]
    if {[regexp -nocase "\[bdhkmnrsvwy\]" [string range $pattern [expr {int($length/2 - abs($caret-$length/2))}] [expr {int($length/2 + abs($caret-$length/2))}]]] == 1} {lappend result "degenerate cut"}
  } else {
    if {[regexp {(.*)\(([\-0-9]*)/([\-0-9]*)} $pattern a pattern top bottom] == 1} {
      if {$top != $bottom} {lappend result "degenerate cut"}
      set site [expr {$top -$bottom} ]
    } else {
      set site nosite
      #sputs "Can't understand pattern: $pattern, no groups added."
    }
    set length [string length $pattern]
  }

  if {[string first N $pattern] > -1} {set length -1}
  if {[regexp -nocase "\[bdhkmnrsvwy\]" $pattern] == 1} {lappend result "degenerate"}
  if {$length == 4} {lappend result "four cutter"}
  if {$length  == 5} {lappend result "five cutter"}
  if {$length  == 6} {lappend result "six cutter"}
  if {$length  > 6} {lappend result ">six cutter"}

  if {$site != "nosite"} {  
    if {$site < 0} {
      lappend result "5 prime"
    } elseif {$site == 0} {
      lappend result "blunt"
    } else {
      lappend result "3 prime"
    }
  }

  return $result
}

#################
## dialog window for adding a new enzyme
#################
proc add_enzyme_dialog {parent_win} {
  global enzymes 
  global info
  global enz_select_window
  global edit_new_enz
  global ok2

  set edit_new_enz 1
  toplevel .new_enzyme
  wm protocol .new_enzyme WM_DELETE_WINDOW "set ok2 -1"
  set raisedialog_store [bind . <<RaiseDialogs>>]
  bind . <<RaiseDialogs>> "wm deiconify .new_enzyme; raise .new_enzyme"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $parent_win] wgeom winx winy
  wm geometry .new_enzyme "+[expr (($winx<0)?0:$winx)+20]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .new_enzyme $parent_win
  } else {
    ::tk::unsupported::MacWindowStyle style .new_enzyme document {closeBox}
  }
  wm deiconify .new_enzyme
  set ok2 0

  label .new_enzyme.namelabel -text [mc "Name:"]
  entry .new_enzyme.name
  label .new_enzyme.patternlabel -text [mc "Site:"]
  entry .new_enzyme.pattern
  label .new_enzyme.commentlabel -text [mc "Comment:"]
  entry .new_enzyme.comment
  grid columnconfigure .new_enzyme 1 -weight 1
  grid .new_enzyme.namelabel -row 0 -column 0 -sticky w
  grid .new_enzyme.name -row 0 -column 1 -sticky ew
  grid .new_enzyme.patternlabel -row 1 -column 0 -sticky w
  grid .new_enzyme.pattern -row 1 -column 1 -sticky ew
  grid .new_enzyme.commentlabel -row 2 -column 0 -sticky w
  grid .new_enzyme.comment -row 2 -column 1 -sticky ew
  grid [checkbutton .new_enzyme.edit_check -text "Edit Groups and Methylation" -variable edit_new_enz] -row 3 -column 1 -columnspan 2 -sticky nw
  grid [frame .new_enzyme.okframe -relief ridge] -row 4 -column 0 -columnspan 2
  grid [button .new_enzyme.okframe.cancel -text [mc "Cancel"] -command "set ok2 -1"] -row 0 -column 1 -padx 12 -sticky nw
  grid [button .new_enzyme.okframe.ok -text [mc "OK"] -command "set ok2 1" -default active] -row 0 -column 0 -padx 12 -sticky nw
  #grab set .new_enzyme
  wm deiconify .new_enzyme
  vwait ok2
 

  if {$ok2 == 1} {
    set enzyme [.new_enzyme.name get]
    set comment [.new_enzyme.comment get]
    if {[lsearch -exact $enzymes $enzyme] != -1} {
      tk_messageBox -message "An enzyme with that name already exists" -type ok -icon warning -default ok
    } elseif {($enzyme != "" ) && ([string length [.new_enzyme.pattern get]] > 3)} {
      if {$comment == ""} {
        set comment [.new_enzyme.pattern get]
      }
      add_enzyme $enzyme [string toupper [.new_enzyme.pattern get]] $comment

      if {($edit_new_enz == 1)} {
        edit_enzyme [.new_enzyme.name get] .new_enzyme
      }

      add_compatibles

      foreach window $info(dnawindows_list) {
        set info($window,scanned) 0
        findenzymes $window
      }

      if {[winfo exists .enz_selector.textframe.textbox] == 1} {
        fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment $enz_select_window
      }
      set info(enzymes_saved) 0
    }
  }
  #grab release .new_enzyme
  destroy .new_enzyme
  bind . <<RaiseDialogs>> $raisedialog_store
  wm deiconify $parent_win
  return $ok2
}

###############
## add an enzyme 
###############
proc add_enzyme {enzyme pattern comment} {
  global enzymes
  global enzinfo
  global info


  ##add enzyme to the enzymes list
  set enzymes [lsort [lappend enzymes $enzyme]]
  set pattern [string toupper $pattern]

  ##add enzyme to the info array
  set enzinfo(enz_vcode,$enzyme) [validation_code $pattern]
  set enzinfo(enz_pattern,$enzyme) $pattern
  set enzinfo(enz_comment,$enzyme) $comment
  set p [string map {^ "" ( "" - "" ) "" / ""} $pattern]
  regsub -all {([0-9])*} $p "" p
  set p [string trim $p "N"] 
  set enzinfo(flatpat,$enzyme) $p

  ##add enzyme to the longname list or hexlist
  addhexarray $enzyme $p

  ##add enzyme to the groups list
  set grouplist [pattern_groups $pattern]
  foreach group $grouplist {
    if {[array names enzinfo "enz_groups,$group"] != {}} {
      set enzinfo(enz_groups,$group) [lsort [lappend enzinfo(enz_groups,$group) $enzyme]]
    }
  }
}

###############
## edit an enzyme (comment, groups, methylation)
###############
proc edit_enzyme {enzyme parent_win} {
  global ok2
  global enzinfo
  global info
  global g
  global dam
  global dcm
  global new_comment
  global enz_select_window

  set maxcols 4
  toplevel .edit_enzyme 
  wm protocol .edit_enzyme WM_DELETE_WINDOW "set ok2 -1"
  wm title .edit_enzyme "[mc "Edit enzyme:"] $enzyme"
  set raisedialog_store [bind . <<RaiseDialogs>>]
  bind . <<RaiseDialogs>> "wm deiconify .edit_enzyme; raise .edit_enzyme"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $parent_win] wgeom winx winy
  wm geometry .edit_enzyme "+[expr (($winx<0)?0:$winx)+20]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .edit_enzyme $parent_win
  } else {
    ::tk::unsupported::MacWindowStyle style .edit_enzyme document {closeBox}
  }
  wm deiconify .edit_enzyme
  set ok2 0

  grid [label .edit_enzyme.enzlabel -text "[mc Enzyme:] $enzyme"] -row 0 -column 0 -columnspan 2 -sticky nw
  grid [label .edit_enzyme.patlabel -text "[mc Site:] $enzinfo(enz_pattern,$enzyme)"] -row 1 -column 0 -columnspan 2 -sticky nw
  grid [label .edit_enzyme.comlabel -text "[mc Comment:]"] -row 2 -column 0 -sticky nw
  grid [entry .edit_enzyme.comentry -textvariable new_comment] -row 2 -column 1 -sticky nwe
  set new_comment $enzinfo(enz_comment,$enzyme)
  grid columnconfigure .edit_enzyme 1 -weight 1

  #groups frame
  grid [labelframe .edit_enzyme.groups -text Groups] -row 3 -column 0 -columnspan 2 -sticky nwe
  set r 0
  set c 0
  set gcb 0
  foreach group $enzinfo(enz_groups_list) {
    if {($group != "Mem Recall") && ($group != "All") && ($group != "Difference...")} {
      incr gcb
      grid [checkbutton .edit_enzyme.groups.cb$gcb -text $group -variable g($group)  -selectcolor white] -row $r -column $c -sticky w
      set g($group) [expr {([lsearch $enzinfo(enz_groups,$group) $enzyme]>-1)}]
      incr c
      if {$c >= $maxcols} {
        set c 0
        incr r
      }
    }
  }

  #Dam/Dcm metylation frame
  grid [labelframe .edit_enzyme.met -text "[mc Methylation]"] -row 4 -column 0 -columnspan 2 -sticky nwe
  set r 0
  set pattern $enzinfo(enz_pattern,$enzyme)
  regsub -all {\(([0-9]|[/-])*\)} $pattern "" pattern
  regsub "\\^" $pattern "" pattern
  set pattern [string trim $pattern "N"]
  set pattern "NN[append pattern NN]"
  set metlist [list]
  set damlist [lindex [lindex $enzinfo(enz_methylaselist) 0] 1]
  set dcmlist [lindex [lindex $enzinfo(enz_methylaselist) 1] 1]
  if {[set found [regexp -all -inline -indices {(?:[GBDKNRSV][ADHMRVW][TBDHKNWY][CBHMNSVY])|(?:[GBDKNRSV][ADHMNRVW][TBDHKWY][CBHMNSVY])} $pattern]] != "" } {
    set metlist [list]
    foreach index $found {
      lappend metlist [lsort -integer [list [expr {[lindex $index 0]-2}] [expr {[string length $pattern]-[lindex $index 0]-6}]]]
    }
    set metlist [lsort -unique $metlist]
    #sputs $metlist
    grid [label .edit_enzyme.met.damlabel -text "[mc Potential] Dam(GATC) [mc Methylation]:"] -row $r -column 0 -columnspan 2 -sticky nw
    incr r
    foreach met $metlist {
      grid [checkbutton .edit_enzyme.met.m$r -variable dam($met) -selectcolor white] -row $r -column 0 -sticky nw
      if {[lsearch $damlist "$enzyme $met"] > -1} {set dam($met) 1}
      grid [text .edit_enzyme.met.t$r -height 1 -relief flat -wrap none -background $info(bg_color)] -row $r -column 1 -sticky nw
      .edit_enzyme.met.t$r insert 1.0 $pattern
      .edit_enzyme.met.t$r tag add bold 1.[expr {[lindex $met 0]+2}] 1.[expr {[lindex $met 0]+6}] 1.[expr {[lindex $met 1]+2}] 1.[expr {[lindex $met 1]+6}]
      .edit_enzyme.met.t$r tag configure bold -background red 
      bindtags .edit_enzyme.met.t$r [list all .]
      incr r
    }

    
  }
  set pattern "N[append pattern N]"
  set metlist [list]
  if {[set found [regexp -inline -all -indices {(?:[CBHMNSVY][CBHMSVY][ATBDHKMNRVWY][GBDKNRSV][GBDKNRSV])|(?:[CBHMNSVY][CBHMNSVY][ATBDHKMNRVWY][GBDKRSV][GBDKNRSV])} $pattern]] != ""} {
    foreach index $found {
      lappend metlist [lsort -integer [list [expr {[lindex $index 0]-3}] [expr {[string length $pattern]-[lindex $index 0]-8}]]]
    }
    set metlist [lsort -unique $metlist]
    grid [label .edit_enzyme.met.dcmlabel -text "[mc Potential] Dcm(CCWGG) [mc Methylation]:"] -row $r -column 0 -columnspan 2 -sticky nw
    incr r
    foreach met $metlist {
      grid [checkbutton .edit_enzyme.met.m$r -variable dcm($met) -selectcolor white] -row $r -column 0 -sticky nw
      if {[lsearch $dcmlist "$enzyme $met"] >-1} {set dcm($met) 1}
      grid [text .edit_enzyme.met.t$r -height 1 -relief flat -wrap none -background $info(bg_color)] -row $r -column 1 -sticky nw
      .edit_enzyme.met.t$r insert 1.0 $pattern
      .edit_enzyme.met.t$r tag add bold 1.[expr {[lindex $met 0]+3}] 1.[expr {[lindex $met 0]+8}] 1.[expr {[lindex $met 1]+3}] 1.[expr {[lindex $met 1]+8}] 
      .edit_enzyme.met.t$r tag configure bold -background red 
      bindtags .edit_enzyme.met.t$r [list all .]
      incr r
    }
  }

  grid [button .edit_enzyme.cancel -text [mc "Cancel"] -command "set ok2 -1"] -row 5 -column 1 -sticky nw
  grid [button .edit_enzyme.ok -text [mc "OK"] -command "set ok2 1" -default active] -row 5 -column 0 -sticky nw
  #grab set .edit_enzyme
  wm deiconify .edit_enzyme
  vwait ok2
 
  #grab release .edit_enzyme
  destroy .edit_enzyme
  bind . <<RaiseDialogs>> $raisedialog_store

##if ok button pressed, accept changes
  if {$ok2 == 1} {
    #change comment
    set enzinfo(enz_comment,$enzyme) $new_comment

    #change groups
    foreach group $enzinfo(enz_groups_list) {
      if {($group != "Mem Recall") && ($group != "All") && ($group != "Difference...")} {
        if {$g($group) == 1} {
          if {[lsearch -exact $enzinfo(enz_groups,$group) $enzyme] == -1} {
            lappend enzinfo(enz_groups,$group) $enzyme
          }
        } else {
          set pos [lsearch -exact $enzinfo(enz_groups,$group) $enzyme]
          if {$pos > -1} {
            set enzinfo(enz_groups,$group) [lreplace $enzinfo(enz_groups,$group) $pos $pos]
          }
        }
      }
    }

    #change methylation 
    foreach dam_index [array names dam] {
      set pos [lsearch $damlist "$enzyme $dam_index"]
      if {($dam($dam_index) == 0) && ($pos > -1)} {
        set damlist [lreplace $damlist $pos $pos]
      } elseif {($dam($dam_index) == 1) && ($pos == -1)} {
        lappend damlist "$enzyme $dam_index"
      }
    }
    set damlist [lsort -index 0 $damlist]
    foreach dcm_index [array names dcm] {
      set pos [lsearch $dcmlist "$enzyme $dcm_index"]
      if {($dcm($dcm_index) == 0) && ($pos > -1)} {
        set dcmlist [lreplace $dcmlist $pos $pos]
      } elseif {($dcm($dcm_index) == 1) && ($pos == -1)} {
        lappend dcmlist "$enzyme $dcm_index"
      }
    }
    set dcmlist [lsort -index 0 $dcmlist]
    set enzinfo(enz_methylaselist) [list [list MDam $damlist] [list MDcm $dcmlist]] 
    set info(blocked_enzymes) [list]
    foreach methylase $enzinfo(enz_methylaselist) {
      foreach blocked_enzyme [lindex $methylase 1] {
        lappend info(blocked_enzymes) [lindex $blocked_enzyme 0]
      }
    }
    #add DpnI specifically
    if {[array names enzinfo "enz_pattern,DpnI"] != ""} {
      lappend info(blocked_enzymes) "DpnI"
    }
    set info(blocked_enzymes) [lsort -unique $info(blocked_enzymes)]
    #rescan
    foreach window $info(dnawindows_list) {
      set info($window,scanned) 0
      findenzymes $window
    }
    #re-fill the selector box
    if {([winfo exists .enz_selector.textframe.textbox]) && ($parent_win == ".enz_selector")} {
      fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment $enz_select_window
    }
  }

  array unset g
  array unset dam
  array unset dcm
  unset new_comment
  return $ok2
}

###############
## add an enzyme group
###############
proc add_enz_group {parent_win enzyme_list} {
  global enz_select_groups
  global info
  global enzinfo
  global ok2

  toplevel .new_group
  wm title .new_group [mc "New Enzyme Group..."]
  wm protocol .new_group WM_DELETE_WINDOW "set ok2 -1"
  set raisedialog_store [bind . <<RaiseDialogs>>]
  bind . <<RaiseDialogs>> "wm deiconify .new_group; raise .new_group"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $parent_win] wgeom winx winy
  wm geometry .new_group "+[expr (($winx<0)?0:$winx)+20]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .new_group $parent_win
  } else {
    ::tk::unsupported::MacWindowStyle style .new_group document {closeBox}
  }
  set ok2 0

  label .new_group.namelabel -text [mc "Name:"]
  entry .new_group.name
  grid columnconfigure .new_group 1 -weight 1
  grid .new_group.namelabel -row 0 -column 0 -sticky w
  grid .new_group.name -row 0 -column 1 -sticky ew
 
  grid [button .new_group.cancel -text [mc "Cancel"] -command "set ok2 -1"] -row 3 -column 1 -sticky nw
  grid [button .new_group.ok -text [mc "OK"] -command "set ok2 1" -default active] -row 3 -column 0 -sticky nw
  #grab set .new_group
  wm deiconify .new_group
  focus .new_group.name

  vwait ok2
 
  #grab release .new_group
  
  if {$ok2 == 1} {
    set newname [.new_group.name get]
    if {($newname == "All") || ($newname == "Mem Recall") || ($newname == "Difference...")} {
      tk_messageBox -default ok -title "Unmodifyable Group..." -message "The group \"$newname\" can't be modified." -type ok -icon warning -default ok
    } elseif {[lsearch -exact $enzinfo(enz_groups_list) $newname] > -1} {
      if {[tk_messageBox -default no -title "Group Exists..." -message "The group \"$newname\" already exists. Should I replace the contents of the group with the curently selected enzymes?" -type yesno -icon warning -default no] == "yes" } {
        set enzinfo(enz_groups,$newname) $enzyme_list
      }  
    } else {
      set enzinfo(enz_groups_list) [lappend enzinfo(enz_groups_list) $newname]
      set enzinfo(enz_groups,$newname) $enzyme_list
      if {[winfo exists .enz_selector.groupsframe.groups_button] == 1} {
        .enz_selector.groupsframe.groups_button.menu delete 0 end
        foreach element $enzinfo(enz_groups_list) {
          .enz_selector.groupsframe.groups_button.menu add radiobutton -label $element -variable enz_select_groups -value $element 
        }
      }
      set info(enzymes_saved) 0
    }
  }

  destroy .new_group
  unset ok2
  wm deiconify $parent_win
  bind . <<RaiseDialogs>> $raisedialog_store
} 

###############
## delete an enzyme group
###############
proc delete_enz_group {parent_win} {
  global enz_select_groups
  global info
  global enzinfo
  global ok2

  toplevel .dialog
  wm protocol .dialog WM_DELETE_WINDOW "set ok2 -1"
  wm title .dialog [mc "Delete Enzyme Groups..."]
  set raisedialog_store [bind . <<RaiseDialogs>>]
  bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $parent_win] wgeom winx winy
  wm geometry .dialog "+[expr (($winx<0)?0:$winx)+20]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .dialog $parent_win
  } else {
    ::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
  }
  set ok2 0

  label .dialog.namelabel -text [mc "Groups:"]
  set h [llength $enzinfo(enz_groups_list)]
  if {$h > 10} {set h 10}
  listbox .dialog.grouplistbox -height $h -activestyle none -exportselection false -selectmode multiple -yscrollcommand ".dialog.groupscroll set"
  scrollbar .dialog.groupscroll -command ".dialog.grouplistbox yview"
  grid columnconfigure .dialog 0 -weight 1
  grid rowconfigure .dialog 1 -weight 1
  grid .dialog.namelabel -row 0 -column 0 -sticky w
  grid .dialog.grouplistbox -row 1 -column 0 -sticky nsew -columnspan 2
  grid .dialog.groupscroll -row 1 -column 2 -sticky nsw
 
  grid [button .dialog.cancel -text [mc "Cancel"] -command "set ok2 -1"] -row 2 -column 1 -sticky nw
  grid [button .dialog.ok -text [mc "Delete"] -command "set ok2 1"] -row 2 -column 0 -sticky nw

  foreach element $enzinfo(enz_groups_list) {
    if {($element != "All") && ($element != "Mem Recall") && ($element != "Difference...")} {
      .dialog.grouplistbox insert end $element 
    }
  }
  #grab set .dialog
  wm deiconify .dialog
  focus .dialog.grouplistbox

  vwait ok2
 
  #grab release .dialog
  
  if {$ok2 == 1} {
    set selection [.dialog.grouplistbox curselection]
    if {[llength $selection] > 0} {
      if {[llength $selection] > 1} {
        if {[tk_messageBox -default no -title [mc "Multiple Group Delete"] -message [format [mc {Are you sure you want to delete these %1$s groups?}] [llength $selection]]" -type yesno -icon warning -default no] == "yes" } {
          foreach element $selection {  
            set groupname [.dialog.grouplistbox get $element]
            set index [lsearch -exact $enzinfo(enz_groups_list) $groupname]
            array unset enzinfo "enz_groups,$groupname" 
            set enzinfo(enz_groups_list) [lreplace $enzinfo(enz_groups_list) $index $index]
          }
        }  
      } else {
        set groupname [.dialog.grouplistbox get [lindex $selection 0]]
        set index [lsearch -exact $enzinfo(enz_groups_list) $groupname]
        array unset enzinfo "enz_groups,$groupname" 
        set enzinfo(enz_groups_list) [lreplace $enzinfo(enz_groups_list) $index $index]
      }
      if {[winfo exists .enz_selector.groupsframe.groups_button]} {
        .enz_selector.groupsframe.groups_button.menu delete 0 end
        foreach element $enzinfo(enz_groups_list) {
          .enz_selector.groupsframe.groups_button.menu add radiobutton -label $element -variable enz_select_groups -value $element 
        }
      }
      set info(enzymes_saved) 0
    }
  }

  destroy .dialog
  unset ok2
  wm deiconify $parent_win
  bind . <<RaiseDialogs>> $raisedialog_store
} 

###############
## load enzyme groups from file
###############
proc read_enz_groups {parent_win {filename ""}} {
  global info enzinfo enzymes

  if {$filename == ""} {
    set filename [tk_getOpenFile -title [mc "Open Group List File"] -initialdir $info(Accdir) -parent $parent_win]
    wm deiconify $parent_win; raise $parent_win
  }
  if {[file exists $filename]} {
    set f [open $filename r]
    set error 0
    while {([gets $f nextline] !=-1) && (!$error)} {
      if {([llength $nextline] == 2)} {
        set newname [lindex $nextline 0]
        set enzyme_list [lindex $nextline 1]
        
        set not_found [list]
        set found [list]
        foreach enz $enzyme_list {
          if {[lsearch -exact $enzymes $enz] == -1} {
            lappend not_found $enz
          } else {
            lappend found $enz
          }
        }
        if {[llength $not_found]} {
          tk_messageBox -title "Enzymes Not Found..." -message "Group: $newname \nThe enzymes: $not_found\nare not currently in the set of known enzymes." -type ok -icon warning
        }
        if {($newname == "All") || ($newname == "Mem Recall") || ($newname == "Difference...")} {
          tk_messageBox -default ok -title "Unmodifyable Group..." -message "The group \"$newname\" can't be modified." -type ok -icon warning -default ok
        } elseif {[lsearch -exact $enzinfo(enz_groups_list) $newname] > -1} {
          if {[tk_messageBox -title "Group Exists..." -message "The group \"$newname\" already exists. Should I replace the contents of the group with the curently selected enzymes?" -type yesno -icon warning -default no] == "yes" } {
            set enzinfo(enz_groups,$newname) $found
          }  
        } else {
          set enzinfo(enz_groups_list) [lappend enzinfo(enz_groups_list) $newname]
          set enzinfo(enz_groups,$newname) $found
          set info(enzymes_saved) 0
        }
      } elseif {[regexp {[^[:space:]]} $nextline]} {
        set errorline $nextline
        set error 1
      }
    }
    if {(!$error)} {
      if {([winfo exists .enz_selector.groupsframe.groups_button])} {
        .enz_selector.groupsframe.groups_button.menu delete 0 end
        foreach element $enzinfo(enz_groups_list) {
          .enz_selector.groupsframe.groups_button.menu add radiobutton -label $element -variable enz_select_groups -value $element 
        }
      }
    } else {
      tk_messageBox -title "Bad Group File Format" -message "The file is improperly formatted starting at line:[string range $errorline 0 15]...\nThe file should contain a group name (with quotes if the name has spaces), a space, a brace {, a list of enzymes with spaces between, a brace }, a return. No tabs, no returns, except between groups.\nFor example: \n\"test group\" {EcoRI XbaI BamHI}\n\"test group2\" {SalI XmaI}" -type ok -icon warning
    }
  }
}

###############
## import enzymes from a DNA Strider format enzyme file
###############
proc import_strider_enzymes {parent_win {filename ""}} {
  global info
  global enzymes
  global enzinfo
  global enz_select_window

  set newenz [list]
  if {$filename == ""} {
    set filename [tk_getOpenFile -title [mc "Open DNA Strider enzymes file"] -initialdir $info(Accdir) -parent $parent_win]
    wm deiconify $parent_win; raise $parent_win
  }
  if {[file exists $filename]} {
    set f [open $filename r]
    while {([gets $f nextline] !=-1) && ([string index $nextline 0] !="%")} {
      if {([string index $nextline 0]!= ";") && ([regexp  {([^,]*),([^,]*),(.*)} $nextline line enzyme pattern comment] == 1)} {
        set enzyme [string trimleft $enzyme "#"]
        set pattern [string toupper $pattern]
        set pattern [string trimleft $pattern " N"]
        set pattern [string map "/ ^" $pattern]
        if {[regexp  {([-0-9]*),([-0-9]*),(.*)} $comment fill top bottom comment]} {
          set pattern "$pattern\($top/$bottom\)"
        }
        set comment "$comment $pattern"
        if {[lsearch $enzymes $enzyme] == -1} {
          add_enzyme $enzyme $pattern $comment
          lappend newenz $enzyme
        }
       
      }
    }
    tk_messageBox -message "[llength $newenz] enzymes added." -type ok -icon info -default ok
    wm deiconify $parent_win; raise $parent_win
    if {[llength $newenz] > 0} {
      add_compatibles
      set enzinfo(enz_methylaselist) [met_list $enzymes $parent_win] 
      set info(blocked_enzymes) [list]
      foreach methylase $enzinfo(enz_methylaselist) {
        foreach blocked_enzyme [lindex $methylase 1] {
          lappend info(blocked_enzymes) [lindex $blocked_enzyme 0]
        }
      }
      #add DpnI specifically
      if {[lsearch -exact $enzymes DpnI] > -1} {
        lappend info(blocked_enzymes) "DpnI"
      }
      set info(blocked_enzymes) [lsort -unique $info(blocked_enzymes)]
      #rescan
      foreach window $info(dnawindows_list) {
        set info($window,scanned) 0
        findenzymes $window
      }
      #re-fill the selector box
      if {[winfo exists .enz_selector.textframe.textbox] == 1} {
        fill_enzymes_box .enz_selector.textframe.textbox .enz_selector.enz_comment $enz_select_window
      }
      set info(enzymes_saved) 0
    }
    close $f
  }
  return $newenz
}


#################
## take selection and list those that overlap methylation sites
#################
proc met_list {selectedenzymes parent_win} {
  global enzymes
  global info
  global enzinfo
  global ok2

  set damenz [list AccIII AlwI BcgI BclI Bsa29I BsaBI BscFI BscI BseAI BseCI Bsh1365I BsiBI BsiMI BsiQI BsiXI Bsp106I BspDI BspEI BspHI BspPI BspXi BsrBRI BstENII BstKTI Bsu15I ClaI DpnII FbaI HphI Hpy188I Hpy188III Ksp22I MamI MboI MboII MflI NdeII NruI TaqI TthHB8I XbaI]
  set dcmenz [list AatI Acc65I AccB7I AlwNI ApaI Asp718I AvaII BalI BanI Bme1390I BpmI BsaI BsaHI BseLI BshNI BslI BsmFI Bsp120I BssKI BstXI CaiI Cfr13I CfrI DraII EaeI Eco31I Eco47I EcoO109I EcoRII GsuI MlsI MscI MspR9I NlaIV PflMI PpuMI PspGI PspOMI PspPI Sau96I ScrFI SexAI SfoI StuI StyD4I VpaK11BI]

  set a [toplevel .methylation_dialog]
  wm protocol .methylation_dialog WM_DELETE_WINDOW "set ok2 1"
  wm title .methylation_dialog [mc "Verify Methylation..."]
  set raisedialog_store [bind . <<RaiseDialogs>>]
  bind . <<RaiseDialogs>> "wm deiconify .methylation_dialog; raise .methylation_dialog"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $parent_win] wgeom winx winy
  wm geometry .methylation_dialog "+[expr (($winx<0)?0:$winx)+20]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .methylation_dialog $parent_win
  } else {
    ::tk::unsupported::MacWindowStyle style .methylation_dialog document {closeBox}
  }

  set ok2 0


  grid [label .methylation_dialog.label1 -text [mc {Dam blocked:}]] -row 0 -column 1 -sticky nw
  grid [listbox $a.damlist -yscrollcommand "$a.damscroll set" -selectmode multiple -exportselection 0] -row 1 -column 1 -sticky nwe
  grid [scrollbar $a.damscroll -command "$a.damlist yview" -orient vertical] -row 1 -column 2 -sticky nswe
  bind $a <MouseWheel> "event generate $a.damlist <MouseWheel> -delta %D"
  grid [label .methylation_dialog.label2 -text [mc {Dcm blocked:}]] -row 0 -column 4 -sticky nw
  grid [listbox $a.dcmlist -yscrollcommand "$a.dcmscroll set" -selectmode multiple -exportselection 0] -row 1 -column 4 -sticky nwe
  grid [scrollbar $a.dcmscroll -command "$a.dcmlist yview" -orient vertical] -row 1 -column 5 -sticky nswe
  grid [button $a.ok -text [mc OK] -command "set ok2 1"] -row 2 -column 1 -sticky n
  grid columnconfigure $a.damlist 0 -weight 0
  grid columnconfigure $a.damlist 1 -weight 1
  grid columnconfigure $a.damlist 2 -weight 0
  grid columnconfigure $a.damlist 3 -weight 0
  grid columnconfigure $a.damlist 4 -weight 1
  grid columnconfigure $a.damlist 5 -weight 0

  foreach element $selectedenzymes {
    set pattern $enzinfo(flatpat,$element) 
    #regsub -all {\(([0-9]|[/-])*\)} $pattern "" pattern
    #regsub "\\^" $pattern "" pattern
    #set pattern [string trim $pattern "N"]

    set pattern "NN[append pattern NN]"
    set metlist [list]
    if {[set found [regexp -all -inline -indices {(?:[GBDKNRSV][ADHMRVW][TBDHKNWY][CBHMNSVY])|(?:[GBDKNRSV][ADHMNRVW][TBDHKWY][CBHMNSVY])} $pattern]] != 0 } {
      foreach index $found {
        set metlist [list]
        lappend metlist [expr {[lindex $index 0]-2}] [expr {[string length $pattern]-[lindex $index 0]-6}]
        set metlist [lsort -integer $metlist]
        if {[$a.damlist get end] != "$element $metlist" } {
          $a.damlist insert end "$element $metlist"
          if {[lsearch $damenz $element] > -1} {$a.damlist selection set end}
        }
      }
    }
    set pattern "N[append pattern N]"
    set metlist [list]
    if {[set found [regexp -inline -all -indices {(?:[CBHMNSVY][CBHMSVY][ATBDHKMNRVWY][GBDKNRSV][GBDKNRSV])|(?:[CBHMNSVY][CBHMNSVY][ATBDHKMNRVWY][GBDKRSV][GBDKNRSV])} $pattern]] !=0} {
      foreach index $found {
        set metlist [list]
        lappend metlist [expr {[lindex $index 0]-3}] [expr {[string length $pattern]-[lindex $index 0]-8}]
        set metlist [lsort -integer $metlist]
        if {[$a.dcmlist get end] != "$element $metlist" } {
          $a.dcmlist insert end "$element $metlist"
          if {[lsearch $dcmenz $element] > -1} {$a.dcmlist selection set end}
        }
      }
    }
  }

  #grab set $a
  wm deiconify $a; raise $a
  focus $a.damlist

  vwait ok2

  #grab release $a
  foreach element [$a.damlist curselection] {
    lappend dam_methylated_enzymes [$a.damlist get $element]
  }
  foreach element [$a.dcmlist curselection] {
    lappend dcm_methylated_enzymes [$a.dcmlist get $element]
  }
  set methylated_enzymes [list [list MDam $dam_methylated_enzymes] [list MDcm $dcm_methylated_enzymes]]
  destroy .methylation_dialog
  set ok2 0
  bind . <<RaiseDialogs>> $raisedialog_store
  return $methylated_enzymes
}

###############
## edit the DNA ladders file
###############
proc edit_ladders {w} {
  global info temp_info ok dialogblock
  if {$dialogblock == 1} {return}
  set dialogblock 1

  set temp_info(current_ladder) $info(current_ladder)
  set temp_info(ladderlist) $info(ladderlist)

  toplevel .dialog
  wm title .dialog [mc "Ladders..."]
  wm protocol .dialog WM_DELETE_WINDOW "set ok -1"
  bind . <<RaiseDialogs>> "wm deiconify .dialog; raise .dialog"
  regexp {[0-9]*x[0-9]*\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry $w] wgeom winx winy
  wm geometry .dialog "+[expr (($winx<0)?0:$winx)+3]+[expr (($winy<0)?0:$winy)+50]"
  if {[tk windowingsystem] != "aqua"} {
    wm transient .dialog $w
  } else {
    ::tk::unsupported::MacWindowStyle style .dialog document {closeBox}
  }
  set ok 0

  grid [frame .dialog.listframe -relief ridge -borderwidth 2] -row 0 -column 1 -sticky nswe
  grid [frame .dialog.ladderframe -relief ridge -borderwidth 2] -row 0 -column 2 -sticky nswe
  grid [frame .dialog.okframe -relief ridge -borderwidth 2] -row 1 -column 1 -columnspan 2 -sticky nswe
  grid rowconfigure .dialog 0 -weight 1
  grid columnconfigure .dialog 1 -weight 1

  grid [listbox .dialog.listframe.list -selectmode single -bg white -yscrollcommand "optionscrollbar .dialog.listframe.scroll" -exportselection 0] -row 0 -column 0 -columnspan 2 -sticky nswe
  grid [scrollbar .dialog.listframe.scroll -command ".dialog.listframe.list yview"] -row 0 -column 2 -sticky nsw
  grid [button .dialog.listframe.add -text [mc "Add New"] -command "event generate .dialog.listframe.list <<add>>"] -row 1 -column 0 -sticky we
  grid [button .dialog.listframe.delete -text [mc "Delete"] -command "event generate .dialog.listframe.list <<delete>>"] -row 1 -column 1 -sticky we
  grid [button .dialog.listframe.raise -text [mc "Raise"] -command "event generate .dialog.listframe.list <<raise>>"] -row 2 -column 0 -sticky we
  grid [button .dialog.listframe.lower -text [mc "Lower"] -command "event generate .dialog.listframe.list <<lower>>"] -row 2 -column 1 -sticky we
  grid rowconfigure .dialog.listframe 0 -weight 1
  grid columnconfigure .dialog.listframe 0 -weight 1
  grid columnconfigure .dialog.listframe 1 -weight 1

  grid [label .dialog.ladderframe.l1 -text [mc "Ladder:"]] -row 0 -column 0 -columnspan 2 -stick nwe
  grid [entry .dialog.ladderframe.ladd_name -bg white -textvariable temp_info(ladder_entry)] -row 1 -column 0 -columnspan 2 -stick nwe
  grid [canvas .dialog.ladderframe.ladder -bg white -height 300 -width 10] -row 2 -column 0 -columnspan 2 -sticky nsew
  grid [label .dialog.ladderframe.l2 -text [mc "Band:"]] -row 3 -column 0 -columnspan 2 -stick nwe
  grid [entry .dialog.ladderframe.band -bg white -validate key -validatecommand {expr {![regexp {[^0-9]} %P]}}] -row 4 -column 0 -columnspan 2 -sticky nwe
  grid [button .dialog.ladderframe.add -text [mc "Add"] -command {if {[.dialog.ladderframe.band get] != ""} {set temp_info(ladder,$temp_info(current_ladder)) [lsort -decreasing -integer [lappend temp_info(ladder,$temp_info(current_ladder)) [.dialog.ladderframe.band get]]]; event generate .dialog <<update_ladder>>}}] -row 5 -column 0 -sticky we
  grid [button .dialog.ladderframe.delete -text [mc "Delete"] -command {set temp_info(ladder,$temp_info(current_ladder)) [lsort -decreasing -integer [lsearch -inline -not -all $temp_info(ladder,$temp_info(current_ladder)) [.dialog.ladderframe.band get]]]; event generate .dialog <<update_ladder>>}] -row 5 -column 1 -sticky we
  grid rowconfigure .dialog.ladderframe 2 -weight 1

  bind .dialog.ladderframe.band <Key-Return> {if {[.dialog.ladderframe.band get] != ""} {set temp_info(ladder,$temp_info(current_ladder)) [lsort -decreasing -integer [lappend temp_info(ladder,$temp_info(current_ladder)) [.dialog.ladderframe.band get]]]; event generate .dialog <<update_ladder>>}}

  grid [button .dialog.okframe.ok -text [mc "OK"] -command "set ok 1"] -row 0 -column 0

##binding to re-draw the ladder
  bind .dialog <<update_ladder>> {
    .dialog.ladderframe.ladder delete bands
    if {([.dialog.listframe.list size] > 0)} {
      if {($temp_info(ladder,$temp_info(current_ladder)) != "")} {
        .dialog.ladderframe.delete configure -state normal
        #redraw the current ladder in the canvas
        set longfrag [lindex $temp_info(ladder,$temp_info(current_ladder)) 0]
        set shortfrag [lindex $temp_info(ladder,$temp_info(current_ladder)) end]
        while {$shortfrag < 1} {
          set temp_info(ladder,$temp_info(current_ladder)) [lrange $temp_info(ladder,$temp_info(current_ladder)) 0 end-1]
          set shortfrag [lindex $temp_info(ladder,$temp_info(current_ladder)) end]
        }
        if {$shortfrag != ""} {
          if {$shortfrag != $longfrag} {
            set gelscale [expr {210.0/(log10($longfrag)- log10($shortfrag))}]
            set gelorigin [expr {30.0 + $gelscale * log10($longfrag) }]
          } else {
            set gelscale 0
            set gelorigin 30
          }

          #draw ladder
          set lastladderband 0

          foreach band $temp_info(ladder,$temp_info(current_ladder)) {
            if {$lastladderband != $band} {
              if {[lindex $band 0] > 0} {
                .dialog.ladderframe.ladder create line [expr {5}] [expr {$gelorigin - $gelscale * log10($band)}] [expr {55}] [expr {$gelorigin - $gelscale *log10($band)}] -width 3 -capstyle round -tags [list bands ladderband$band scalable]
              }
              .dialog.ladderframe.ladder bind ladderband$band <Any-Enter> ".dialog.ladderframe.ladder itemconfigure laddertext -text [string range $band 0 end-3][expr {([string length $band]>3)?",":{}}][string range $band end-2 end]"
              .dialog.ladderframe.ladder bind ladderband$band <Any-Leave> ".dialog.ladderframe.ladder itemconfigure laddertext -text {}"
              .dialog.ladderframe.ladder bind ladderband$band <Button-1> ".dialog.ladderframe.band delete 0 end;.dialog.ladderframe.band insert 0 $band"
              set lastladderband $band
            } else {
              .dialog.ladderframe.ladder itemconfigure ladderband$band -width [expr {[.dialog.ladderframe.ladder itemcget ladderband$band -width]+1}]
            }
          }
        }
      } else {
        .dialog.ladderframe.delete configure -state disabled
      }
    }     
  } 

##binding to change to a different ladder in listbox
  bind .dialog.listframe.list <<ListboxSelect>> {
    set temp_info(ladder_entry) [.dialog.listframe.list get [.dialog.listframe.list curselection]]
    set temp_info(current_ladder) $temp_info(ladder_entry)
    event generate .dialog <<update_ladder>>
  }

##binding to raise a ladder in the list
  bind .dialog.listframe.list <<raise>> {
    if {[.dialog.listframe.list size] > 0} {
      set i [.dialog.listframe.list curselection]
      if {$i > 0} {
        set item1 [.dialog.listframe.list get $i]
        .dialog.listframe.list delete $i
        incr i -1
        set item2 [.dialog.listframe.list get $i]
        .dialog.listframe.list insert $i $item1
        .dialog.listframe.list selection set $i
        set temp_info(ladderlist) [lreplace $temp_info(ladderlist) $i [expr {$i+1}] $item1 $item2]
        event generate .dialog <<update_ladder>>
      } else {bell}
    }
  }
##binding to lower a ladder in the list
  bind .dialog.listframe.list <<lower>> {
    if {[.dialog.listframe.list size] > 0} {
      set i [.dialog.listframe.list curselection]
      if {$i < [expr {[.dialog.listframe.list size]-1}]} {
        set item1 [.dialog.listframe.list get $i]
        .dialog.listframe.list delete $i
        set item2 [.dialog.listframe.list get $i]
        incr i 1
        .dialog.listframe.list insert $i $item1
        .dialog.listframe.list selection set $i
        set temp_info(ladderlist) [lreplace $temp_info(ladderlist) [expr {$i-1}] $i $item2 $item1]
        event generate .dialog <<update_ladder>>
      } else {bell}
    }
  }
##binding to delete a ladder in the list
  bind .dialog.listframe.list <<delete>> {
    if {[.dialog.listframe.list size] > 0} {
      set i [.dialog.listframe.list curselection]
      .dialog.listframe.list delete $i
      set temp_info(ladderlist) [lreplace $temp_info(ladderlist) $i $i]
      array unset temp_info "ladder,$temp_info(current_ladder)"
      if {$i == [.dialog.listframe.list size]} {incr i -1}
      .dialog.listframe.list selection set $i
      set temp_info(ladder_entry) [.dialog.listframe.list get $i]
      set temp_info(current_ladder) $temp_info(ladder_entry)
      event generate .dialog <<update_ladder>>
    }
    if {[.dialog.listframe.list size] == 0} {
      .dialog.listframe.delete configure -state disabled
      .dialog.ladderframe.add configure -state disabled
      .dialog.ladderframe.delete configure -state disabled
      .dialog.ladderframe.ladd_name configure -state disabled
      .dialog.ladderframe.band configure -state disabled
    }
    if {[.dialog.listframe.list size] <= 1} {
      .dialog.listframe.raise configure -state disabled
      .dialog.listframe.lower configure -state disabled
    }
  }
##binding to add a ladder to the list
  bind .dialog.listframe.list <<add>> {
    if {[.dialog.listframe.list size] == 0} {
      .dialog.listframe.delete configure -state normal
      .dialog.ladderframe.add configure -state normal
      .dialog.ladderframe.delete configure -state normal
      .dialog.ladderframe.ladd_name configure -state normal
      .dialog.ladderframe.band configure -state normal
      set i 0
    } else {
      set i [.dialog.listframe.list curselection]
    }
    set new_name "New Ladder"
    set j 0
    while {[lsearch -exact $temp_info(ladderlist) $new_name] > -1} {incr j; set new_name "New Ladder($j)"}
    .dialog.listframe.list insert $i $new_name
    .dialog.listframe.list selection clear 0 end
    .dialog.listframe.list selection set $i
    set temp_info(ladder_entry) $new_name
    set temp_info(current_ladder) $new_name
    set temp_info(ladderlist) [linsert $temp_info(ladderlist) $i $new_name]
    set temp_info(ladder,$new_name) [list]
    event generate .dialog <<update_ladder>>
  }
##binding to change the name of a ladder
  bind .dialog.ladderframe.ladd_name <Key-Return> {
    if {[.dialog.listframe.list size] > 0} {
      set pos [.dialog.listframe.list curselection]
      if {[.dialog.listframe.list get $pos] != $temp_info(ladder_entry)} {
        if {[lsearch -exact $temp_info(ladderlist) $temp_info(ladder_entry)] == -1} {
          .dialog.listframe.list delete $pos
          .dialog.listframe.list insert $pos $temp_info(ladder_entry)
          .dialog.listframe.list selection set $pos
          set temp_info(ladderlist) [lreplace $temp_info(ladderlist) $pos $pos $temp_info(ladder_entry)]

          set temp_info(ladder,$temp_info(ladder_entry)) $temp_info(ladder,$temp_info(current_ladder))
          array unset temp_info "ladder,$temp_info(current_ladder)"
          set temp_info(current_ladder) $temp_info(ladder_entry)
        } else {
          bell
          set temp_info(ladder_entry) [.dialog.listframe.list get $pos]
        }
      }
    }
  }
  bind .dialog.ladderframe.ladd_name <FocusOut> [bind .dialog.ladderframe.ladd_name <Key-Return>]

  .dialog.ladderframe.ladder create text 10 10 -text "" -tags laddertext -anchor nw

  foreach ladder $temp_info(ladderlist) {
    .dialog.listframe.list insert end $ladder
    set temp_info(ladder,$ladder) $info(ladder,$ladder)
  }
  .dialog.listframe.list selection set [lsearch $temp_info(ladderlist) $temp_info(current_ladder)]
  update
  event generate .dialog.listframe.list <<ListboxSelect>>

  regexp {([0-9]*)x([0-9]*)\+(-?[0-9]*)\+(-?[0-9-]*)} [wm geometry .dialog] wgeom winw winh winx winy
  if {[expr {$winx+$winw}] > [winfo screenwidth .]} {set winx [expr {[winfo screenwidth .]-$winw}]}
  if {$winx < 0} {set winx 0}
  if {[expr {$winy+$winh}] > [winfo screenheight .]} {set winy [expr {[winfo screenheight .]-$winh}]}
  if {$winy < 0} {set winy 0}
  wm geometry .dialog "+$winx+$winy"
  update

  focus .dialog
  vwait ok 
  set dialogblock 0
  destroy .dialog
  bind . <<RaiseDialogs>> ""

  ###execute changes
  if {$ok == 1} {
    set info(current_ladder) $temp_info(current_ladder)
    set info(ladderlist) $temp_info(ladderlist)
    array unset info "ladder,*"
    foreach ladder $temp_info(ladderlist) {
      set info(ladder,$ladder) $temp_info(ladder,$ladder)
    }
    ###need to update the ladder menus on all dna windows
    foreach win [lsearch -inline -all -regexp -not [winfo children .] {abi|analysis|menubar}] {
      $win.menubar.enzymes.ladders delete 0 end
      foreach entry $info(ladderlist) {
        $win.menubar.enzymes.ladders add radiobutton -label $entry -value $entry -variable info(current_ladder)
      }
      $win.menubar.enzymes.ladders add separator
      $win.menubar.enzymes.ladders add command -label "Edit..." -command "edit_ladders $win"
    }
    ###need to save the new ladder list
    set temp_ladder_file [open [file join $info(user_defaults_dir) temp_ladderq_qqrx.txt] w]
    
    puts $temp_ladder_file "#comment lines start with #"
    puts $temp_ladder_file "#enter each new ladder on a new line"
    puts $temp_ladder_file "#start with the name of the ladder"
    puts $temp_ladder_file "#THEN A TAB"
    puts $temp_ladder_file "#then the fragment sizes (decreasing order), separated by a space"
    puts $temp_ladder_file "#no commas in the sizes."
    puts $temp_ladder_file "#enter a band twice or more to make it drawn thicker"
    foreach ladder $info(ladderlist) {
      puts $temp_ladder_file "$ladder\t$info(ladder,$ladder)"
    }
    close $temp_ladder_file
    if {[file exists [file join $info(user_defaults_dir) DNA_ladders.txt]]} {
    file delete [file join $info(user_defaults_dir) DNA_ladders.txt]
    file rename [file join $info(user_defaults_dir) temp_ladderq_qqrx.txt] [file join $info(user_defaults_dir) DNA_ladders.txt]
    }
  }
  unset ok
  array unset temp_info
}

###############
## read the DNA ladders file
###############
proc read_ladders {} {
  global info
  set info(ladderlist) [list]

  set filename [file join $info(user_defaults_dir) "DNA_ladders.txt"]
  if {[file exists $filename] == 1} {
    set fileid [open $filename r]
    while {[eof $fileid] != 1} {
      set a [gets $fileid]
      if {[string index $a 0] != "#"} {
        set name [lindex [split $a "\t"] 0]
        set sizes [lindex [split $a "\t"] 1]
        if {[llength $sizes] > 0} {
          if {[catch {set sizes [lsort -decreasing -integer $sizes]}] != 0} {
            warning_dialog "Misformatted ladder file" $name $sizes
          } else {
            lappend info(ladderlist) $name
            set info(ladder,$name) $sizes
          }
        }
      }
    }
    close $fileid
  } 
  if {[llength $info(ladderlist)] == 0} {
    set info(ladderlist) [list]
    set info(ladder,) [list]
    set info(current_ladder) ""
  } else {
    set info(current_ladder) [lindex $info(ladderlist) 0]
  }
}

###############
## read the global defaults file
###############
proc read_defaults {} {
  global info

  set filename [file join $info(user_defaults_dir) "ApE_Defaults.txt"]
  if {[file exists $filename] == 1} {
    set fileid [open $filename r]
    while {[eof $fileid] != 1} {
      set a [gets $fileid]
      if {([string index $a 0] != "#") && ([llength [set b [split $a \t]]] == 2)} {
        set info([lindex $b 0]) [lindex $b 1]
      }
    }
    close $fileid
  } 

  if ![font metrics \{$info(textfontfamily)\} -fixed] {
   set info(textfontfamily) courier
  }
  if {[expr {[llength $info(keysubstmap)] % 2}] != 0} {
    set info(keysubstmap) [list]
  }
}

###############
## shows a msg box for debugging values
###############
proc warning_dialog {args} { 
  tk_messageBox -message [list $args] -icon warning -type ok
}

#################################
##  main ########################
#################################

check_for_running_process

## initialize globals
set winum 0
set abiwinum 0
set longnamelist [list]
set hexlist [list]
set enzymes [list]
set empty_genbank_header [list {LOCUS       } {} {DEFINITION	 } {} {ACCESSION   } {} {VERSION     } {} {SOURCE      } {} {  ORGANISM  } {}]

array set info [list]
set info(enzymes_saved) 1
set info(cancel_scroll) ""
set info(mousewheelunits) 120
set info(clock_format) "%a %b %d %H:%M:%S %Z %Y"
set info(dnawindows_list) [list]
set info(dnatitles_list) [list]
set info(bg_color) gray75
set info(menu_bg_color) gray75
set info(menu_fg_color) black
set info(menu_select_bg_color) #008
set info(menu_select_fg_color) white
set info(text_bg_color) white
set info(text_fg_color) black
set info(text_select_bg_color) #008
set info(text_select_fg_color) white
set info(text_select2_bg_color) #6060d0
set info(text_select2_fg_color) white
set info(graphic_bg_color) seashell
set info(graphic_fg_color) blue
set info(enzyme_highlight_color) red
set info(find_forward_highlight) gray75
set info(find_reverse_highlight) gray50
set info(show_comment) 0
set info(comment_height) 3
set info(enz_currently_selected) [list]
set info(enz_text_2nd) 1
set info(enz_text_genes) 0
set info(enz_text_trans) "1_1"
set info(enz_text_enz) "all"
set info(enz_text_width) 100
set info(enz_text_copy_highlight) 1
set info(enz_list_enz) "all"
set info(enz_list_mode) "Name"
set info(enz_list_columns) 6
set info(find_casens) 0
set info(find_revcom) 0
set info(find_as_aa) 0
set info(find_as_literal) 0
set info(find_find_again) [list]
set info(allowedkeys) "ACGTNacgtn*"
set info(graphicfontfamily) helvetica
set info(textfontfamily) courier
set info(fixedfontlist) [list]
set info(dnafontsize) 10
set info(graphicfontsize) 10
set info(print_font_size) 8
set info(tk_ps_fonts) [list] 
set info(draw_graphic_connectors) 1
set info(strider_incompatible_warning) 1
set info(methylated_default) 1
set info(translation_window_code) 1
set info(translation_window_spacing) 0
set info(translation_window_revcom) 0
set info(translation_window_width) 20
set info(translation_window_line_numbers) none
set info(translation_window_index_line) 0
set info(translation_window_dna_line) none
set info(clipboard_tags) [list]
set info(selection_shows_Trans) 0
set info(selection_shows_Tm) 0
#setinfo(selection_shows) "None"
set info(clipboard_text) ""
set info(keysubstmap) {}
set info(find_orf_starts) "Met"
set info(feature_default_fcolor) cyan
set info(feature_default_rcolor) green
set info(transp_feature_color) "#FFFFFF"
set info(feature_scan_on_close) 1
set info(align_select_window2) ""
set info(align_select_region1) all
set info(align_select_region2) all
set info(align_select_rc1) "forward"
set info(align_select_rc2) "forward"
set info(align_linewidth) 100
set info(mm_penalty) 0
set info(gap_penalty) 0
set info(gap_extension_penalty) 0
set info(NWmax) 300
set info(align_blocksize) 15
set info(align_copy_highlight) 0
set info(default_dnadir) ""
set info(dnadir_follows_open) 1
set info(default_featuredir) ""
set info(featuredir_follows_open) 1
set info(default_file_extension) .str
set info(default_file_format) Genbank
set info(save_file_format_warning) 1

set info(ncbi_blast,program) "blastn"
set info(ncbi_blast,low_filter) 1
set info(ncbi_blast,expect) 100
set info(ncbi_blast,database) "nr"

set info(wb_blast,program) "blastn"

set info(wb_blast,expect) 0.01
set info(wb_blast,database) "C. elegans genomic"

set dialogblock 0

array set info {primer_len_min 20 primer_len_max 25 primer_Tm_min 55 primer_Tm_max 60 primer_gc_min 45 primer_gc_max 60 primer_clamp_min 1 primer_clamp_max 2 primer_scomp_tot_max 10 primer_scomp_adj_max 5 primer_scomp_3_max 3 primer_hcomp_tot_max 10 primer_hcomp_adj_max 5 primer_hcomp_3_max 3 primer_hcomp_oligo "" primer_check_sc 1 primer_salt 50 primer_dna 250 primer_scan "5'-->3'"}
set abi_info(last_find) ""
array set abi_info {colors,C blue colors,T red colors,A green colors,G black colors,N magenta colors,B magenta colors,D magenta colors,H magenta colors,K magenta colors,M magenta colors,R magenta colors,S magenta colors,V magenta colors,W magenta colors,Y magenta}


set info(Accdir) [file join [file dirname $argv0] "Accessory Files"]
while {![file isdirectory $info(Accdir)]} {
  set info(Accdir)  [tk_chooseDirectory -title "Please locate the Accessory Files Directory"]
  if {$info(Accdir) == ""} {
    final_exit
  }
}
if {[tk windowingsystem] !="aqua"} {
  set efAccdir $info(Accdir)
} else {
  set efAccdir [eval file join [linsert [lrange [file split $argv0] 0 end-5] end "Accessory Files"]]
}
set info(default_enzymedir) [file join $efAccdir "Enzymes"]
set info(default_featuredir) [file join $efAccdir "Features"]


namespace import msgcat::mc
namespace import msgcat::mclocale
proc mcload {langdir} {
  set x 0
  set encoding utf-8
  foreach p [::msgcat::mcpreferences] {
    set langfile [file join $langdir $p.msg]
    if {[file exists $langfile]} {
      incr x
      set fid [open $langfile "r"]
      gets $fid line
      if {[regexp \xFE\xFF $line]||[regexp \xFF\xFE $line]} {
        set encoding unicode
        fconfigure $fid -encoding unicode
        seek $fid 0 start ;# rewind -- real reading is still to come
        set text [read $fid [file size $langfile]]
        regsub -all "\uFEFF|\uFFFE" $text "" text
        uplevel 1 $text
      } else {
        seek $fid 0 start
        uplevel 1 [read $fid]
      }
      close $fid
    }
  }
  return $x
}
if {[string match -nocase c [mclocale]]} {mclocale en_US}
set info(locale) [mclocale]
sputs "Locale: $info(locale)"
sputs "Loaded: [mcload [file join $info(Accdir) Msgs]] locale files."

set toolbarlist [list new open save separator cut copy paste separator select find separator uppercase swap_case separator reverse_com separator enzymes separator graphic_map highlight_enzymes digest text_map separator find_primers align]
set tooltip_after "not_showing"
read_toolbar_images


package require http
::http::config -useragent "Mozilla/4.75 (X11; U; Linux 2.2.17; i586; Nav)"


if {$tcl_platform(platform) == "macintosh"} {
    set modifier Command
    set modstring "Appl"
} elseif {$tcl_platform(platform) == "windows"} {
  set modifier Control
  set modstring "Ctrl"
  initialize_win32
} elseif {$tcl_platform(platform) == "unix"} {
    set modstring "Meta"
    set modifier Meta
    if {$tcl_platform(os) == "Linux"} {
      set modifier Control
      set modstring "Ctrl"
      set info(user_defaults_dir) "~/.ApE/"
    }
    if {[tk windowingsystem] == "aqua"} {
      set modifier "Command"
      set modstring "Cmd"
      initialize_aqua
      set info(mousewheelunits) 4
    }
    bind Entry <$modifier-Key-v> "event generate %W <<Paste>>"
    bind Text <$modifier-Key-v> "event generate %W <<Paste>>"
    bind Text <Button-5> [list %W yview scroll 1 units]
    bind Text <Button-4> [list %W yview scroll -1 units]
} else {
    set modifier Control
    set modstring "Ctrl"
    #sputs "tcl_platform(os):$tcl_platform(os)"
    #sputs "tcl_platform(machine):$tcl_platform(machine)"
}
bind Entry <$modifier-Key-V> "event generate %W <<Paste>>"

##Create user defaults directory if it doesn't exist
if {![file isdirectory $info(user_defaults_dir)]} {
  file mkdir $info(user_defaults_dir)
}
if {(![file exists [file join $info(user_defaults_dir) ApE_Defaults.txt]]) && ($info(user_defaults_dir) != $info(Accdir))} {
  file copy [file join $info(Accdir) ApE_Defaults.txt] [file join $info(user_defaults_dir) ApE_Defaults.txt]
}
if {(![file exists [file join $info(user_defaults_dir) DNA_ladders.txt]]) && ($info(user_defaults_dir) != $info(Accdir))} {
  file copy [file join $info(Accdir) DNA_ladders.txt] [file join $info(user_defaults_dir) DNA_ladders.txt]
}

read_defaults
read_ladders



#check for defaults from file
#then for efAccdir/Enzymes
#then for efAccdir
#then for Accdir/Enzymes
if {(![file isdirectory $info(default_enzymedir)]) || (![file exists [file join $info(default_enzymedir) "Default_Enzymes.txt"]])} {
  if {(![file isdirectory [set info(default_enzymedir) [file join $efAccdir "Enzymes"]]])  || (![file exists [file join $info(default_enzymedir) "Default_Enzymes.txt"]])} {
    if {(![file isdirectory [set info(default_enzymedir) $efAccdir]])  || (![file exists [file join $info(default_enzymedir) "Default_Enzymes.txt"]])} {
      if {(![file isdirectory [set info(default_enzymedir) [file join $info(Accdir) "Enzymes"]]])  || (![file exists [file join $info(default_enzymedir) "Default_Enzymes.txt"]])} {

        while {![file isdirectory $info(default_enzymedir)]} {
          set info(default_enzymedir)  [tk_chooseDirectory -title "Please locate the Enzyme Files Directory"]
          if {$info(default_enzymedir) == ""} {
            final_exit
          }
        }
      }
    }
  }
}

read_enzymes [file join $info(default_enzymedir) "Default_Enzymes.txt"]

#check for defaults from file
#then for efAccdir/Features
#then for efAccdir
#then for Accdir/Features
if {![file isdirectory $info(default_featuredir)]} {
  if {![file isdirectory [set info(default_featuredir) [file join $efAccdir "Features"]]]} {
    if {(![file isdirectory [set info(default_featuredir) $efAccdir]])  || (![file exists [file join $info(default_featuredir) "Default_features.txt"]])} {
      if {![file isdirectory [set info(default_featuredir) [file join $info(Accdir) "Features"]]]} {
        while {![file isdirectory $info(default_featuredir)]} {
          set info(default_featuredir)  [tk_chooseDirectory -title "Please locate the Enzyme Files Directory"]
          if {$info(default_featuredir) == ""} {
            final_exit
          }
        }
      }
    }
  }
}
unset efAccdir

if {[file exists [file join $info(default_featuredir) "Default_Features.txt"]]} {
  read_featurelib [file join $info(default_featuredir) "Default_Features.txt"]
} else {
  set fea_info(library) [list]
  set fea_info(library_file) ""
}


foreach {index color} [list bg_color gray85 menu_bg_color gray85 menu_fg_color black menu_select_bg_color #008 menu_select_fg_color white text_bg_color white text_fg_color black text_select_bg_color #008 text_select_fg_color white text_select2_bg_color #6060d0 text_select2_fg_color white] {
  if {[catch {winfo rgb . $info($index)}]} {
    set info($index) $color
  }
}
unset index color

tk_setPalette background $info(bg_color) troughColor gray50 activeBackground $info(bg_color) activeForeground black selectForeground $info(text_select_fg_color) selectBackground $info(text_select_bg_color);#selectColor white
#bug in tk_setPalette
unset checkbutton

if {[array names info modstring] != ""} {
  set modstring $info(modstring)
}
if {[array names info modifier] != ""} {
  set modifier $info(modifier)
}

if {[array names info locale] eq "locale"} {
  mclocale $info(locale)
  sputs "Loaded: [mcload [file join $info(Accdir) Msgs]] locale files."
} 

if {(![file isdirectory $info(default_dnadir)])} {
  set info(default_dnadir) [file dirname $argv0]
}


bind . <$modifier-Key-q> "exit"
setup_AlalysisText_bindings

if {[lsearch -exact [font names] dnafont] == -1} {
  button .b
  font create labelfont -family [lindex [.b cget -font] 0] -size $info(dnafontsize)
  font create boldlabelfont -family [lindex [.b cget -font] 0] -size $info(dnafontsize) -weight bold
  destroy .b
  entry .e
  font create entryfont -family [lindex [.e cget -font] 0] -size $info(dnafontsize)
  destroy .e
  font create dnafont -family $info(textfontfamily) -size $info(dnafontsize)
}
if {[lsearch -exact [font names] graphicfont] == -1} {
  font create graphicfont -family $info(graphicfontfamily) -size $info(graphicfontsize)
}

####list of tk font names to convert to postscript font names
set tk_fonts ""
set ps_fonts ""
foreach {tk_font ps_font} $info(tk_ps_fonts) {
  lappend tk_fonts "[string tolower $tk_font]_normal" "[string tolower $tk_font]_bold"
  lappend ps_fonts "$ps_font" "$ps_font\-Bold"
}
option add *Checkbutton.font labelfont 40
option add *Radiobutton.font labelfont 40
option add *Label.font labelfont 40
option add *Labelframe.font labelfont 40
option add *Button.font labelfont 40
option add *Menu.font labelfont 40
option add *Menubutton.font labelfont 40
option add *Listbox.font labelfont 40
option add *Entry.font entryfont 40
if {[tk windowingsystem] eq "aqua"} {
option add *Button.padx 15 40
option add *Button.pady 15 40
}

option add *Text.background $info(text_bg_color)
option add *Text.foreground $info(text_fg_color)
option add *Entry.background $info(text_bg_color)
option add *Entry.foreground $info(text_fg_color)

option add *Menu*tearOff 0
option add *Menu*foreground $info(menu_fg_color)
option add *Menu*background $info(menu_bg_color)
option add *Menu*activeBackground $info(menu_select_bg_color)
option add *Menu*activeForeground $info(menu_select_fg_color)

option add *Checkbutton.activeBackground $info(bg_color)
#option add *Button*selectColor white
option add *Radiobutton.activeBackground $info(bg_color)
label .temp
set info(label_fg_color) [.temp cget -foreground]
destroy .temp

## create the first window (unless a window has been created by Apple open event)
update 
if {[lsearch [winfo children .] ".dna_window*"] == -1} {
  if {([llength $argv] == 0) || ($tcl_interactive == 1) || ([string index [lindex $argv 0] 0] == "-")} {
    create_window
  } else {
    open_file [lindex $argv 0]
  }
}
sputs "ApE version $version"
set time1 [clock clicks -milliseconds]
#warning_dialog [expr {$time1 - $time0}]
unset time0
unset time1
